/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* NEXTSTEP PPC Control Center.                                            */
/* CenterController class implementation.                                  */
/* One object of this class is instanciated at startup. It controls the    */
/* interface objects.                                                      */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 26.07.1994  Droux       Created                                         */
/*-------------------------------------------------------------------------*/
/* filename: CenterController.m                                            */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#import "CenterController.h"
#import "ProjectController.h"
#import "strings.h"
#import <syslog.h>

#define GETHOME_EXIT_CODE 123 /* value returned by gethome if execv failed */
#define PPC_HOME_NAME   "PPC_HOME"
#define PPC_SCRIPT_NAME "ppc"

@implementation CenterController

/*-------------------------------------------------------------------------*/
/* infoPanel:                                                              */
/* Shows the application info panel.                                       */
/*-------------------------------------------------------------------------*/
- infoPanel:sender
{
  if (infoPanel == nil)
  {
    if (![NXApp
          loadNibSection: "info.nib"
	  owner: self
	  withNames: NO])
    {
      return nil;
    }
  }
  [infoPanel makeKeyAndOrderFront: nil];
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* menuActionsStop:                                                        */
/* Return the above object.                                                */
/*-------------------------------------------------------------------------*/
- menuActionsStop
{
  return menuActionsStop;
}

/*-------------------------------------------------------------------------*/
/* projectOpened:                                                          */
/* Check if the specified project is already opened. Returns its id if it  */
/* is the case, nil otherwise.                                             */
/*-------------------------------------------------------------------------*/
- projectOpened: (const char *)projectName
{
  int nbProjects;                   /* number of projects in projects list */
  int actProject;                          /* actual project being checked */
  id project;                                         /* actual project Id */
  
  nbProjects = [projectsList count];
  
  for (actProject = 0; actProject < nbProjects; actProject++)
  {
    project = [projectsList objectAt: actProject];
    if (!strcmp([project projectName], projectName))
      return project;                                    /* project exists */
  }
  
  return nil;                                         /* project not found */
}

/*-------------------------------------------------------------------------*/
/* createProject:                                                          */
/* Create a new project.nib copy and initialize it.                        */
/*-------------------------------------------------------------------------*/
- createProject: (const char *)projectName
{
  newProject = [self projectOpened: projectName];
  
  if (newProject)
  {                                           /* project is already opened */
    [newProject makeActive];
    return newProject;
  }
  
  if ([NXApp loadNibSection: "project.nib" owner: self] == nil)
    return nil;
    
  if ([newProject setUp: self 
        projectName: projectName
	windowOffset: windowOffset
	: menuFilesUpdate
	: menuFilesOpenInWorkspace
	: menuFilesSelectInWorkspace
	: menuFilesAddExisting
	: menuFilesCreateNew
	: menuFilesDelete])
  {
    if ((windowOffset += 24.0) > 100.0)            /* update window offset */
      windowOffset = 0.0;
      
    [projectsList addObject: newProject];  /* add project to projects list */
    
    if ([projectsList count] == 1)
    {                         /* first project loaded, enable menu options */
      if (optionsToEnableWhenFirstProject != nil)
      {
        /* The new project has been created after appDidInit, the menu     */
	/* list is initialized and we can enable the options.              */
	[self enableMenuOptionsForFirstProject];
      }
      else
      {
        /* The new project has been created before appDidInit, the menu    */
	/* list is thus not initialized. The menu options will be          */
	/* enabled during appDidInit.                                      */
      }
    }

    return newProject;
  }

  return nil;                            /* error initializing new project */
}

/*-------------------------------------------------------------------------*/
/* openProject:                                                            */
/* Message sent when a project must be opened.                             */
/*-------------------------------------------------------------------------*/
- openProject: sender
{
  char *types[2] = {"vmc", 0};   /* allowed file extensions for open panel */

  if ([[OpenPanel new] runModalForTypes: types] == NX_OKTAG)
  {
    [self createProject: [[OpenPanel new] filename]];
    return self;
  }

  return nil;
}

/*-------------------------------------------------------------------------*/
/* saveProject:                                                            */
/* Message sent by the menu when a project must be saved.                  */
/*-------------------------------------------------------------------------*/
- saveProject: sender
{  
  if (activeProject == nil)
    return nil;
    
  [activeProject saveConfig];
  [menuProjectRevertToSaved 
    setEnabled: [activeProject canBeRevertedToSaved]];
    
  return self;
}

/*-------------------------------------------------------------------------*/
/* saveAllProjects:                                                        */
/* Message sent by the menu when all projects must be saved.               */
/*-------------------------------------------------------------------------*/
- saveAllProjects: sender
{
  int nbProjects;                   /* number of projects in projects list */
  int actProject;                          /* actual project being checked */
  
  nbProjects = [projectsList count];
  saveCanceled = NO;
  
  for (actProject = 0; 
       actProject < nbProjects && !saveCanceled; 
       actProject++)
  {
    [[projectsList objectAt: actProject] saveConfig];
  }
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* revertToSaved:                                                          */
/* Message sent by the menu when a project must be reverted to its saved   */
/* state.                                                                  */
/*-------------------------------------------------------------------------*/
- revertToSaved: sender
{
  [activeProject revertToSaved];

  return self;
}

/*-------------------------------------------------------------------------*/
/* closeProject:                                                           */
/* Message sent by the menu when a project must be closed.                 */
/*-------------------------------------------------------------------------*/
- closeProject: sender
{
  if (activeProject == nil)
    return nil;
    
  [activeProject close];
    
  return self;
}

/*-------------------------------------------------------------------------*/
/* closeAllProjects:                                                       */
/* Close all projects.                                                     */
/* Returns nil if not all projects have been closed.                       */
/*-------------------------------------------------------------------------*/
- closeAllProjects
{
  int nbProjects;                   /* number of projects in projects list */
  int actProject;                          /* actual project being checked */
  id originalList;                                /* original project list */
  BOOL editedWindows = NO;              /* YES if there are edited windows */
  id project;                               /* actual project being copied */
  
  saveCanceled = NO;
  nbProjects = [projectsList count];
  
  /* Duplicate projectsList to originalList. This is necessary since we    */
  /* cannot execute on the projectsList and delete its elements            */
  /* simultaneously.                                                       */
  originalList = [[List alloc] init];
  for (actProject = 0; actProject < nbProjects; actProject++)
  {
    project = [projectsList objectAt: actProject];
    [originalList addObject: project];
    editedWindows = editedWindows || [project isEdited];
  }
  
  if (editedWindows)
  {                                            /* there are edited windows */
    int panelResult;
    
    if (limitedMode)
      panelResult = 0;          /* force projects to be immediately closed */
    else
      panelResult = NXRunAlertPanel("Quit",
	"There are edited windows.",
	"Review Unsaved", "Quit Anyway", "Cancel");
      
    switch(panelResult)
    {
      case 1:                                            /* Review Unsaved */
        break;
      case 0:                                               /* Quit Anyway */
        for (actProject = 0; actProject < nbProjects; actProject++)
          [[projectsList objectAt: actProject] setNotEdited];
        break;
      case -1:                                                   /* Cancel */
        saveCanceled = YES;
	break;
      default:                                /* should normally not occur */
        saveCanceled = YES;
	break;
    }
  }

  for (actProject = 0; 
       actProject < nbProjects && !saveCanceled; 
       actProject++)
  {
    [[originalList objectAt: actProject] close];
  }
   
  [originalList free];                                     /* release copy */
  
  if (saveCanceled)
    return nil;                        /* not all projects have been saved */

  return self;
}

/*-------------------------------------------------------------------------*/
/* clearBuffer:                                                            */
/* Sent when the clear buffer menu option is selected.                     */
/*-------------------------------------------------------------------------*/
- clearBuffer: sender
{
  [activeProject clearBuffer: sender];
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* filesMenuSelected:                                                      */
/* Message sent by the menu when one of the Files option is selected.      */
/*-------------------------------------------------------------------------*/
- filesMenuSelected: sender
{
  [activeProject filesMenuSelected: sender];
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* actionsMenuSelected:                                                    */
/* Message sent by the menu when one of the Actions options is selected.   */
/*-------------------------------------------------------------------------*/
- actionsMenuSelected: sender
{
  [activeProject actionsMenuSelected: sender];

  return self;
}

/*-------------------------------------------------------------------------*/
/* setActiveProject:                                                       */
/* Message sent by a project when it becomes active.                       */
/*-------------------------------------------------------------------------*/
- setActiveProject: sender
{
  activeProject = sender;
  
  [menuProjectRevertToSaved setEnabled: [sender canBeRevertedToSaved]];
    
  return self;
}

/*-------------------------------------------------------------------------*/
/* removeProject:                                                          */
/* Remove the project given as parameter.                                  */
/*-------------------------------------------------------------------------*/
- removeProject: project
{
  [projectsList removeObject: project];
  
  if (![projectsList count])
  {
    /* There are no more opened project. We have to disable the necessary  */
    /* menu options.                                                       */
    [self disableMenuOptionsForLastProject];
    activeProject = nil;
  }
    
  return self;
}

/*-------------------------------------------------------------------------*/
/* saveCanceled:                                                           */
/* Message sent by a project when a save operation is canceled.            */
/*-------------------------------------------------------------------------*/
- saveCanceled
{
  saveCanceled = YES;
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* enableMenuOptionsForFirstProject:                                       */
/* Enable the menu options stored in the optionsToEnableWhenFirstProject.  */
/* This method must be called when the first project is created.           */
/*-------------------------------------------------------------------------*/
- enableMenuOptionsForFirstProject
{
  int actItem;
  int nbItems;
  
  if (optionsToEnableWhenFirstProject == nil)
    return nil;
  
  nbItems = [optionsToEnableWhenFirstProject count];
  for (actItem = 0; actItem < nbItems; actItem++)
  {
    [[optionsToEnableWhenFirstProject objectAt: actItem]
      setEnabled: YES];
  }
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* disableMenuOptionsForLastProject:                                       */
/* Disable the menu options stored in the optionsToDisableWhenLastProject  */
/* list. This method must be called when the last project is closed.       */
/*-------------------------------------------------------------------------*/
- disableMenuOptionsForLastProject
{
  int actItem;
  int nbItems;
  
  if (optionsToDisableWhenLastProject == nil)
    return nil;
  
  nbItems = [optionsToDisableWhenLastProject count];
  for (actItem = 0; actItem < nbItems; actItem++)
  {
    [[optionsToDisableWhenLastProject objectAt: actItem]
      setEnabled: NO];
  }
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* runInLimitedMode:                                                       */
/* Set the application to be executed in limited mode. In this mode, it is */
/* not possible to open projects. This method sets the instance variable   */
/* limitedMode to YES.                                                     */
/*-------------------------------------------------------------------------*/
- runInLimitedMode: (char *)title: (char *)firstPhrase
{
  
  NXRunAlertPanel(title,
    "%s%s",
    NULL, NULL, NULL,
    firstPhrase,
    strings[STR_CC_LIMITED_MODE]);
  
  [menuProjectOpen setEnabled: NO];
  limitedMode = YES;

  return self;
}

/*-------------------------------------------------------------------------*/
/* initPPCHome:                                                            */
/* Get the PPC home directory from the default environment variable        */
/* PPC Home and store it in the PPCHome instance variable.                 */
/* Returns nil if not defined.                                             */
/*-------------------------------------------------------------------------*/
- initPPCHome
{
  char *pathPtr;
  char initenvName [200];
  char *initenvArgv[] = {initenvName, NULL};

  sprintf(initenvName, "%s/%s",
    [[NXBundle mainBundle] directory], 
    "initenv");
  
  if (!fork())
  {
    execv(initenvArgv[0], initenvArgv);
    exit(1);
  }
  
  wait(NULL);
  NXUpdateDefaults();
  
  if ((pathPtr = (char *)NXReadDefault("PPC", "Home")) == NULL)
  {
    strcpy(PPCHome, "");
    return nil;
  }
  
  strcpy(PPCHome, pathPtr);   

  return self;
}

/*-------------------------------------------------------------------------*/
/* initPPCArch:                                                            */
/* Get the PPC architecture.                                               */
/* Returns nil if not defined.                                             */
/*-------------------------------------------------------------------------*/
- initPPCArch
{
  char *pathPtr;
  char initenvName [200];
  char *initenvArgv[] = {initenvName, NULL};

  sprintf(initenvName, "%s/%s",
    [[NXBundle mainBundle] directory], 
    "initarch");
  
  if (!fork())
  {
    execv(initenvArgv[0], initenvArgv);
    exit(1);
  }
  
  wait(NULL);
  NXUpdateDefaults();
  
  if ((pathPtr = (char *)NXReadDefault("PPC", "Arch")) == NULL)
  {
    strcpy(PPCArch, "UNKNOWN");
    return nil;
  }
  
  strcpy(PPCArch, pathPtr);   

  return self;
}

/*-------------------------------------------------------------------------*/
/* PPCHome:                                                                */
/* Return the PPC home directory.                                          */
/*-------------------------------------------------------------------------*/
- (char *)PPCHome
{
  return PPCHome;
}

/*-------------------------------------------------------------------------*/
/* PPCArch:                                                                */
/* Return the PPC architecture.                                            */
/*-------------------------------------------------------------------------*/
- (char *)PPCArch
{
  return PPCArch;
}

/*-------------------------------------------------------------------------*/
/* checkPPCHome:                                                           */
/* Check if the PPC directory specified by the instance variable PPCHome   */
/* exists.                                                                 */
/* Returns nil if it is not the case.                                      */
/*-------------------------------------------------------------------------*/
- checkPPCHome
{
  struct stat statBuf;

  if (stat(PPCHome, &statBuf) == -1)
    return nil;

  return self;
}

/*-------------------------------------------------------------------------*/
/* getPPCEnvironment:                                                      */
/* Get the PPC environment and initialize the corresponding instance       */
/* variables. If an error occured, an error panel is shown and the method  */
/* returns nil.                                                            */
/*-------------------------------------------------------------------------*/
- getPPCEnvironment
{
  int panelResult;
  
  if ([self initPPCHome] == nil || [self checkPPCHome] == nil)
  {
    panelResult = NXRunAlertPanel(strings[STR_CC_CONFIG_ERROR],
      strings[STR_CC_HOME_DEFINED_ERROR],
      "Run Limited...", "Quit", NULL);
    
    switch(panelResult)
    {
      case 1:                                               /* Run Limited */
	[self runInLimitedMode: "Run Limited..."
	  : "As PPC is not properly configured"];
	break;
      default:                                                     /* Quit */
	return nil;
    }
  }
  
  [self initPPCArch];
  
  return self;
}

@end

@implementation CenterController(ApplicationDelegate)

/*-------------------------------------------------------------------------*/
/* appWillInit:                                                            */
/* Called just before application initialization.                          */
/*-------------------------------------------------------------------------*/
- appWillInit: sender
{
  /* This method is necessary because the app:openFile:type: method is     */
  /* called before appDidInit: when the user starts PPCCenter from         */
  /* the workspace by double-clicking on a virtual machine configuration   */
  /* file. In this case, the new project is loaded first and then          */
  /* appDidInit is called. If we leave the initializations above in        */
  /* appDidInit, the first project is loaded, and then the initializations */
  /* are done, which cause the references to the first project loaded      */
  /* to be lost.                                                           */
  
  windowOffset = (float)0.0;
  projectsList = [[List alloc] init];
  optionsToEnableWhenFirstProject = nil;
  activeProject = nil;
  limitedMode = NO;
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* appDidInit:                                                             */
/* Called just after application initialization.                           */
/*-------------------------------------------------------------------------*/
- appDidInit: sender
{
  int actMenuOption;
  int nbEnableMenuOptions;
  
  if (init_strings((char *)[[NXBundle mainBundle] directory]) == -1)
  {
    NXRunAlertPanel("Initialization Error",
      "Impossible to read the application strings. %s",
      NULL, NULL, NULL,
      "Application is probably corrupted and will now terminate.");
    [NXApp terminate: self];
  }
  
  if ([self getPPCEnvironment] == nil)
    [NXApp terminate: self];
  
  /* The list optionsToEnableWhenFirstProject contains pointers to the     */
  /* menu options which must be enabled when the first project is loaded.  */
  optionsToEnableWhenFirstProject = [[List alloc] init];
  [optionsToEnableWhenFirstProject addObject: menuProjectSave];
  [optionsToEnableWhenFirstProject addObject: menuProjectSaveAll];
  [optionsToEnableWhenFirstProject addObject: menuProjectRevertToSaved];
  [optionsToEnableWhenFirstProject addObject: menuProjectClose];
  [optionsToEnableWhenFirstProject addObject: menuEditCut];
  [optionsToEnableWhenFirstProject addObject: menuEditCopy];
  [optionsToEnableWhenFirstProject addObject: menuEditPaste];
  [optionsToEnableWhenFirstProject addObject: menuEditClearBuffer];
  [optionsToEnableWhenFirstProject addObject: menuEditSelectAll];
  [optionsToEnableWhenFirstProject addObject: menuFilesUpdate];
  [optionsToEnableWhenFirstProject addObject: menuActionsConfigure];
  [optionsToEnableWhenFirstProject addObject: menuActionsMake];
  [optionsToEnableWhenFirstProject addObject: menuActionsInstall];
  [optionsToEnableWhenFirstProject addObject: menuActionsRun];
  [optionsToEnableWhenFirstProject addObject: menuActionsViewConfiguration];
  [optionsToEnableWhenFirstProject addObject: menuActionsClean];
  [optionsToEnableWhenFirstProject addObject: menuActionsConfigureMakeInstall];
  [optionsToEnableWhenFirstProject addObject: menuActionsConfigureMakeRun];
  
  /* The list optionsToDisableWhenLastProject contains the pointers to     */
  /* the menu options which must be disabled when the last project is      */
  /* closed.                                                               */
  optionsToDisableWhenLastProject = [[List alloc] init];
  nbEnableMenuOptions = [optionsToEnableWhenFirstProject count];
  for (actMenuOption = 0; 
       actMenuOption < nbEnableMenuOptions; 
       actMenuOption++)
  {
    [optionsToDisableWhenLastProject addObject: 
      [optionsToEnableWhenFirstProject objectAt: actMenuOption]];
  }
  [optionsToDisableWhenLastProject addObject: menuFilesOpenInWorkspace];
  [optionsToDisableWhenLastProject addObject: menuFilesSelectInWorkspace];
  [optionsToDisableWhenLastProject addObject: menuFilesAddExisting];
  [optionsToDisableWhenLastProject addObject: menuFilesCreateNew];
  [optionsToDisableWhenLastProject addObject: menuFilesDelete];
  
  if ([projectsList count])
  { 
    /* A project has been initialized before the call to this function.    */
    /* This mean that PPCCenter has been started by opening a project     */
    /* from the Workspace.                                                 */
    
    if (limitedMode)
    {
      /* A project has been loaded, but the application is running in      */
      /* limited mode. The project must be closed immediately.             */
      [self closeAllProjects];
    }
    else
    {
      /* The menu options are not yet enabled. We do this operation now.   */
      [self enableMenuOptionsForFirstProject];
    }
  }

  return self;
}

/*-------------------------------------------------------------------------*/
/* appWillTerminate:                                                       */
/* Called just before the application terminates.                          */
/*-------------------------------------------------------------------------*/
- appWillTerminate: sender
{
  if ([self closeAllProjects] == nil)
    return nil;
    
  end_strings();
  [projectsList free];
  [optionsToEnableWhenFirstProject free];
  [optionsToDisableWhenLastProject free];
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* appAcceptsAnotherFile:                                                  */
/* PPCCenter accepts files from the Workspace.                            */
/*-------------------------------------------------------------------------*/
- (BOOL)appAcceptsAnotherFile: sender
{
  return YES;
}

/*-------------------------------------------------------------------------*/
/* app:openFile:type:                                                      */
/* Message sent when the user opens a virtual machine configuration file   */
/* from the Workspace.                                                     */
/*-------------------------------------------------------------------------*/
- (int)app: sender
  openFile: (const char *)filename
  type: (const char *)aType
{
  if (limitedMode)
  {
    NXRunAlertPanel("Open from Workspace",
      strings[STR_CC_WS_LIMITED_ERROR],
      NULL, NULL, NULL);
    return NO;
  }

  if ([self createProject: filename])
    return YES;   
  
  return NO;
}

@end
