/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* NEXTSTEP PPC Control Center.                                            */
/* ProjectController class implementation.                                 */
/* One object of this class is instanciated for each opened project.       */
/* It controls the project objects.                                        */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 28.07.1994  Droux       Created                                         */
/*-------------------------------------------------------------------------*/
/* filename: ProjectController.m                                           */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#import "ProcCoordinator.h"
#import "CenterController.h"
#import "Subprocess.h"
#import "LogController.h"
#import "ProjectController.h"
#import "architectures.h"
#import <syslog.h>

typedef enum                     /* PPC command line tool keys identifiers */
{
  C_KEY,
  M_KEY,
  R_KEY,
  D_KEY,
  I_KEY,
  V_KEY,
  CMI_KEY,
  CMR_KEY,
  STOP_KEY,
  UNKNOWN_KEY
} keyId;

static char *keys[] =                        /* PPC command line tool keys */
{
  "c",
  "m",
  "r",
  "d",
  "i",
  "v",
  "cmi",
  "cmr",
  NULL
};

@implementation ProcCoordinator

/*-------------------------------------------------------------------------*/
/* init:                                                                   */
/* Instance initialization.                                                */
/*-------------------------------------------------------------------------*/
- init: theCenterController
  scrollView: scrollViewP
  procStringTable: subprocessStringTableP
  projectController: projectControllerP
  status: statusTextFieldP
{
  self = [super init];
  centerController = theCenterController;
  subprocess = [Subprocess alloc];
  subprocessStringTable = subprocessStringTableP;
  theProjectController = projectControllerP;
  statusTextField = statusTextFieldP;
  stopMenuOption = [theCenterController menuActionsStop];
  scrollView = scrollViewP;
  logController = [[LogController alloc] initWithScroll: scrollView];
  [logController setDelegate: self];
  [[scrollView docView] setEditable: NO];
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* actionsMenuSelected:                                                    */
/* Message sent by the menu when one of the Actions options is selected.   */
/*-------------------------------------------------------------------------*/
- actionsMenuSelected: sender
{
  id menuItem = [sender selectedCell];
  keyId theKeyId;
  static char cmd [1024];
  static char options [1024];
  static char cmdOutput [1024];
  static char status [100];
  archEnum arch = [theProjectController selectedArch];
  char *execPath = [theProjectController execPath];
  
  if (menuItem == nil)
    return self;

  strcpy(status, "Processing Request");                  /* default status */
  
  switch([menuItem tag])
  {
    case MENU_OPTION_TAG(MENU_ACTIONS_TAG, MENU_ACTIONS_CONFIGURE_TAG):
      theKeyId = C_KEY;
      sprintf(status, "Configuring %s for %s", 
        [theProjectController projectNameWithoutPath],
	outputArchNames[arch]);
      break;
    case MENU_OPTION_TAG(MENU_ACTIONS_TAG, MENU_ACTIONS_MAKE_TAG):
      theKeyId = M_KEY;
      sprintf(status, "Making %s for %s", 
        [theProjectController projectNameWithoutPath],
	outputArchNames[arch]);
      break;
    case MENU_OPTION_TAG(MENU_ACTIONS_TAG, MENU_ACTIONS_INSTALL_TAG):
      theKeyId = I_KEY;
      sprintf(status, "Installing %s server for %s", 
        [theProjectController projectNameWithoutPath],
	outputArchNames[arch]);
      break;
    case MENU_OPTION_TAG(MENU_ACTIONS_TAG, MENU_ACTIONS_RUN_TAG):
      theKeyId = R_KEY;
      sprintf(status, "Running %s on %s", 
        [theProjectController projectNameWithoutPath],
	outputArchNames[arch]);
      [[scrollView docView] setEditable: YES];
      break;
    case MENU_OPTION_TAG(MENU_ACTIONS_TAG, MENU_ACTIONS_STOP_TAG):
      [subprocess terminate: self];
      return self;
    case MENU_OPTION_TAG(MENU_ACTIONS_TAG, MENU_ACTIONS_VIEW_CONFIG_TAG):
      theKeyId = V_KEY;
      sprintf(status, "Visualizing Configuration of %s for %s", 
        [theProjectController projectNameWithoutPath],
	outputArchNames[arch]);
      break;
    case MENU_OPTION_TAG(MENU_ACTIONS_TAG, MENU_ACTIONS_CLEAN_TAG):
      theKeyId = D_KEY;
      sprintf(status, "Cleaning %s for %s", 
        [theProjectController projectNameWithoutPath],
	outputArchNames[arch]);
      break;
    case MENU_OPTION_TAG(MENU_ACTIONS_TAG, MENU_ACTIONS_CMI_TAG):
      theKeyId = CMI_KEY;
      sprintf(status, "Configuring, making, and installing %s for %s", 
        [theProjectController projectNameWithoutPath],
	outputArchNames[arch]);
      break;
    case MENU_OPTION_TAG(MENU_ACTIONS_TAG, MENU_ACTIONS_CMR_TAG):
      theKeyId = CMR_KEY;
      sprintf(status, "Configuring, making, and executing %s for %s", 
        [theProjectController projectNameWithoutPath],
	outputArchNames[arch]);
      [[scrollView docView] setEditable: YES];
      break;
    default:
      theKeyId = UNKNOWN_KEY;
      syslog(LOG_ERR, "INTERNAL ERROR: unknown key id in procCoordinator");
      return self;
      break;
  }
  
  [[statusTextField cell] setStringValue: status];

  sprintf(options, "-a %s",                        /* build options string */
    archNames[arch]);
    
  if (strlen([theProjectController vmcOptionalFile]))
  {
    strcat(options, " -f ");
    strcat(options, [theProjectController vmcOptionalFile]);
  }
    
  if (strlen([theProjectController hwOptionalFile: arch]))
  {
    strcat(options, " -h ");
    strcat(options, [theProjectController hwOptionalFile: arch]);
  }
    
  if((execPath != NULL) &&
     (([menuItem tag] == 
     MENU_OPTION_TAG(MENU_ACTIONS_TAG, MENU_ACTIONS_INSTALL_TAG)) ||
     ([menuItem tag] ==
     MENU_OPTION_TAG(MENU_ACTIONS_TAG, MENU_ACTIONS_CMI_TAG))))
  {
    strcat(options, " -i ");
    strcat(options, execPath);
  }
  
  if ([theProjectController verboseSelected])
    strcat(options, " -v");
  
  sprintf(cmd, "cd %s; %s/lib/%s/ppccmd %s %s",
    [theProjectController projectPath],
    [centerController PPCHome],
    [centerController PPCArch],
    keys[theKeyId],
    options);
    
  sprintf(cmdOutput, "ppc %s %s\n",
    keys[theKeyId],
    options);
  
  [logController appendString: cmdOutput user: YES];

  [subprocess init: cmd
    withDelegate: self
    andPtySupport: YES
    andStdErr: YES];

  [stopMenuOption setEnabled: YES];  

  return self;
}

/*-------------------------------------------------------------------------*/
/* clearBuffer:                                                            */
/* Sent when the clear buffer menu option is selected.                     */
/*-------------------------------------------------------------------------*/
- clearBuffer: sender
{
  [logController clearBuffer: sender];
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* subprocessOutput:                                   Subprocess Delegate */
/* Sent to the delegate when there is output data available from the       */
/* subprocess.                                                             */
/*-------------------------------------------------------------------------*/
- subprocessOutput: (char *)buffer
{
  [logController appendString: buffer];

  return self;
}

/*-------------------------------------------------------------------------*/
/* subprocessError:                                    Subprocess Delegate */
/* Sent to the delegate when a fatal error occus during the management of  */
/* the subprocess.                                                         */
/*-------------------------------------------------------------------------*/
- subprocessError: (char *)errorString
{
  const char *returnedString;
  
  if (returnedString = 
      [subprocessStringTable valueForStringKey: errorString])
    NXRunAlertPanel("PPC Control Center", returnedString, NULL, NULL, NULL);
  else
    NXRunAlertPanel("PPC Control Center", errorString, NULL, NULL, NULL);
    
  return self;
}

/*-------------------------------------------------------------------------*/
/* subprocessDone:                                     Subprocess Delegate */
/* Sent to the delegate when the subprocess is terminated.                 */
/*-------------------------------------------------------------------------*/
- subprocessDone
{
  [[statusTextField cell] setStringValue: "Idle"];
  [stopMenuOption setEnabled: NO];  
  [[scrollView docView] setEditable: NO];

  return self;
}

/*-------------------------------------------------------------------------*/
/* userEntered:                                     CommandScroll Delegate */
/*-------------------------------------------------------------------------*/
- userEntered: (char *)buffer
{
  [subprocess send: buffer withNewline: NO];
  
  return self;
}

@end
