/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Alternate construction interface.                                       */
/* Different versions of the functions below are implemented according to  */
/* the machine for which the parallel program is being compiled.           */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 02.12.1993  Droux       Created                                         */
/* 27.01.1994  Droux       Ported to PACT-C                                */
/*-------------------------------------------------------------------------*/
/* filename: alt.h                                                         */
/* Tested on :                                                             */
/*   NEXTSTEP 2.1, NeXT Computers                                          */
/*   NEXTSTEP 3.1, NeXT Computers                                          */
/*   NEXTSTEP 3.1, Intel Computers                                         */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*   SunOS Release 5.3 (Solaris 2), SPARC                                  */
/*   Linux 0.99.pl9, Intel Computers                                       */
/*   PACT-C v92.1, T805                                                    */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#ifndef altunix_h
#define altunix_h

/*-------------------------------------------------------------------------*/
/* Misc constants.                                                         */
/*-------------------------------------------------------------------------*/
#define MAX_ALT_CHANS 20           /* max # of channels in a alt construct */
#define TIMEOUT       -1                        /* special timeout channel */
#define SECONDS       1000000      /* factor to specify timeout in seconds */

typedef struct            /* records informations about the channels to be */
{                                               /* watched and the timeout */
  int nb_chans;                             /* number of channels in chans */
  channel chans [MAX_ALT_CHANS];                           /* channel id's */
  long the_timeout;                        /* timeout in microseconds or 0 */
} alt_status;

/*-------------------------------------------------------------------------*/
/* alt_init:                                                               */
/* Initializes the alternate construct.                                    */
/* This functions is not called by the user directly but it is generated   */
/* by the macro alt_begin().                                               */
/*-------------------------------------------------------------------------*/
void alt_init (alt_status *act_alt);

/*-------------------------------------------------------------------------*/
/* alt_wait:                                                               */
/* Activate the alternate construct. Waits for data comming from one of    */
/* the channels given in the specified alt status variable. Exists if some */
/* data is arrived or if the timeout is reached. Returns TIMEOUT in this   */
/* case or the winner channel id otherwise.                                */
/*-------------------------------------------------------------------------*/
void alt_wait (alt_status *act_alt, channel *winner);

/*-------------------------------------------------------------------------*/
/* alt_add_channel:                                                        */
/* Add the given channel to the list of channels to be listen.             */
/*-------------------------------------------------------------------------*/
void alt_add_channel (alt_status *act_alt, channel c);

/*-------------------------------------------------------------------------*/
/* alt_add_timeout:                                                        */
/* Specifies the timeout of the alternate construct. If this function is   */
/* not called, then the alternate constrct terminates only if something is */
/* available on one of the specified channels.                             */
/* The timeout is expressed in microseconds.                               */
/*-------------------------------------------------------------------------*/
void alt_add_timeout (alt_status *act_alt, long t);

/*-------------------------------------------------------------------------*/
/* alt_begin:                                                              */
/* Beginning of an alt construct.                                          */
/* Example:                                                                */
/*   alt_begin();                                                          */
/*-------------------------------------------------------------------------*/
#define alt_begin() { alt_status _act_alt; alt_init(&_act_alt)

/*-------------------------------------------------------------------------*/
/* alt_end:                                                                */
/* End of an alt construct.                                                */
/* Returns the winner channels, or TIMEOUT if timeout reached.             */
/* Example:                                                                */
/*   channel winner;                                                       */
/*   ...                                                                   */
/*   alt_begin();                                                          */
/*     ...                                                                 */
/*   alt_end(&winner);                                                     */
/*-------------------------------------------------------------------------*/
#define alt_end(w) alt_wait(&_act_alt, w); }

/*-------------------------------------------------------------------------*/
/* alt_channel:                                                            */
/* Declare a channel that must be listen.                                  */
/* The channel is taken in account only if the guard condition is true     */
/* (different than zero).                                                  */
/* Example:                                                                */
/*   channel winner;                                                       */
/*   channel c;                                                            */
/*   channel d;                                                            */
/*   int i;                                                                */
/*   ...                                                                   */
/*   alt_begin();                                                          */
/*     alt_channel(i > 0, c);                                              */
/*     alt_channel(1, d);                                                  */
/*   alt_end(&winner);                                                     */
/*-------------------------------------------------------------------------*/
#define alt_channel(b, c) if (b) alt_add_channel(&_act_alt, c);

/*-------------------------------------------------------------------------*/
/* alt_for:                                                                */
/* Allows the creation of a replicated alt.                                */
/* The use of alt_for is similar to the use of the well-known C for().     */
/* The alt_for() body must contain only calls to alt_channel().            */
/* Example:                                                                */
/*   #define NB_CHANS 10                                                   */
/*   channel winner;                                                       */
/*   channel c [NB_CHANS];                                                 */
/*   channel d;                                                            */
/*   int i;                                                                */
/*   int n;                                                                */
/*   ...                                                                   */
/*   alt_begin();                                                          */
/*     alt_for(n = 0; n < NB_CHANS; n++)                                   */
/*       alt_channel(1, c[n]);                                             */
/*     alt_channel(i > 0, d);                                              */
/*   alt_end(&winner);                                                     */
/*-------------------------------------------------------------------------*/
#define alt_for for

/*-------------------------------------------------------------------------*/
/* alt_timeout:                                                            */
/* Add a timeout to the alt construct. If nothing comes on one of the      */
/* specified channels after the given time (in microseconds), then the     */
/* alt construct terminates and returns the value TIMEOUT instead of the   */
/* winner.                                                                 */
/*   channel winner;                                                       */
/*   channel c;                                                            */
/*   channel d;                                                            */
/*   int i;                                                                */
/*   ...                                                                   */
/*   alt_begin();                                                          */
/*     alt_channel(i > 0, c);                                              */
/*     alt_channel(1, d);                                                  */
/*     alt_timeout(1000);                                                  */
/*   alt_end(&winner);                                                     */
/*-------------------------------------------------------------------------*/
#define alt_timeout(t) alt_add_timeout(&_act_alt, t)

#endif
