/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Virtual machine configuration implementation.                           */
/* This file contains and maintain the virtual machine configuration. It   */
/* implements two kind of functions. The first are used to define the      */
/* machine and are called from the user configuration function. The second */
/* are used to get configuration information needed to configure the       */
/* application itself. These two kind of function are accessible via two   */
/* header files, one for the user called conf.h and one for the kernel     */
/* called confpriv.h.                                                      */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 17.11.1993  Droux       Created                                         */
/* 21.01.1994  Droux       Ported to PACT-C                                */
/* 28.01.1994  Droux       Initializes channels source and destination     */
/* 17.02.1994  Droux       va_arg(char|short|float) bug fixed              */
/* 08.07.1994  Droux       New structure of source directory               */
/* 15.08.1994  Droux       bcopy defined for Solaris 2 and PACT-C          */
/*-------------------------------------------------------------------------*/
/* filename: config.c                                                      */
/* Tested on :                                                             */
/*   NEXTSTEP 2.1, NeXT Computers                                          */
/*   NEXTSTEP 3.1, NeXT Computers                                          */
/*   NEXTSTEP 3.1, Intel Computers                                         */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*   SunOS Release 5.3 (Solaris 2), SPARC                                  */
/*   Linux 0.99.pl9, Intel Computers                                       */
/*   PACT-C v92.1, T805                                                    */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define config_impl

#include <stdio.h>
#ifdef PACTC
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <ppc/config.h>
#include <ppc/config/confpriv.h>

/*-------------------------------------------------------------------------*/
/* Macros local to this module.                                            */
/*-------------------------------------------------------------------------*/
#define CHECK_INIT  if (!config_init) {init_config(); config_init=1;}
#define ERROR(msg)  {printf("ERROR: " msg "\n"); exit(1);}

/*-------------------------------------------------------------------------*/
/* Declare the processors, tasks, channels tables, and other local         */
/* variables.                                                              */
/*-------------------------------------------------------------------------*/
int config_init = 0;               /* 1 if module initialized, 0 otherwise */
processor_descr processors [MAX_PROCESSORS];
task_descr tasks [MAX_TASKS];
channel_descr channels [MAX_CHANNELS];
modules_descr modules [MAX_MODULES];

#if defined(SUN4SOL2) || defined(PACTC)
/*--------------------------------------------------------------------------*/
/* bcopy:                                                                   */
/* Equivalent of BSD bcopy() for SYS-V.                                     */
/*--------------------------------------------------------------------------*/
void bcopy (void *source, void *dest, int len)
{
  char *s = (char *)source;
  char *d = (char *)dest;

  while (len)
  {
    *d = *s;
    d++;
    s++;
    len--;
  }
}
#endif

/*-------------------------------------------------------------------------*/
/* init_config:                                                      LOCAL */
/* Initializes the module local variables. This function is called once    */
/* automatically by one of the new_xxx functions.                          */
/*-------------------------------------------------------------------------*/
void init_config (void)
{
  int n;
  
  for (n = 0; n < MAX_PROCESSORS; n++)
    processors[n].free = 1;
  for (n = 0; n < MAX_TASKS; n++)
    tasks[n].free = 1;
  for (n = 0; n < MAX_CHANNELS; n++)
  {
    /* It is possible that a declared channel is only connected to one     */
    /* task. This is correct and we must be able to test this case during  */
    /* the configuration process of some architectures (e.g. PACT-C).      */
    /* We initializes the from and to fields to -1 at the beginning.       */
    channels[n].free = 1;
    channels[n].from = -1;
    channels[n].to = -1;
  }
  for (n = 0; n < MAX_MODULES; n++)
    modules[n].free = 1;
}

/*-------------------------------------------------------------------------*/
/* place_task:                                                       LOCAL */
/* Place a task on a processor.                                            */
/*-------------------------------------------------------------------------*/
void place_task (task task_id, processor processor_id)
{
  int task_slot;
  int n;

  for (n = 0;                        /* search free task slot on processor */
    (processors[processor_id].tasks[n] != -1) && 
    (n < MAX_TASKS_PER_PROCESSOR); n++);
  if (n == MAX_TASKS_PER_PROCESSOR) ERROR ("too many tasks on processor");
  task_slot = n;
                                          /* update processor informations */
  processors[processor_id].tasks[task_slot] = task_id;
                                                       /* update task info */
  tasks[task_id].on_processor = processor_id;
}

/*-------------------------------------------------------------------------*/
/* init_channel:                                                     LOCAL */
/* Called when a task is given a channel as parameter. Update the          */
/* task channels table and the channels array. Checks the direction of the */
/* channel if it is already used by another task.                          */
/*-------------------------------------------------------------------------*/
void init_channel (task task_id, channel channel_id, int channel_type)
{
  int n;
  int chan_idx;

                             /* to_host and from_host channels are special */
  if ((channel_id == FROM_HOST_CHANNEL_ID) || 
      (channel_id == TO_HOST_CHANNEL_ID))
    return;
                                              
  if (channel_type == OUTPUT_CHANNEL)         /* update channel descriptor */
    if (channels[channel_id].from != -1)
      ERROR("channel input end-point already assigned to a task")
    else
      channels[channel_id].from = task_id;
  if (channel_type == INPUT_CHANNEL) 
    if (channels[channel_id].to != -1)
      ERROR("channel output end-point already assigned to a task")
    else
      channels[channel_id].to = task_id;
      
                            /* search free channel slot in task descriptor */
  for (n = 0; (tasks[task_id].channels[n] != -1) && 
    (n < MAX_CHANNELS_PER_TASK); n++);
  if (n == MAX_CHANNELS_PER_TASK) ERROR("too many channels for a task");
  
  chan_idx = n;
  tasks[task_id].channels[chan_idx] = channel_id;
}

/*-------------------------------------------------------------------------*/
/* new_processor:                                                          */
/* Creates a new processor.                                                */
/*-------------------------------------------------------------------------*/
processor new_processor (void)
{
  int processor_id = 0;                          /* found new processor id */
  int n;                          /* counter to find a free processor slot */
  
  CHECK_INIT;                                      /* initialization stuff */

                /* find free processor slot and change free state if found */
  for (n = 0; !processors[n].free && (n < MAX_PROCESSORS); n++);
  if (n == MAX_PROCESSORS) ERROR("no free processor slot");
  processor_id = n;
  processors[processor_id].free = 0;
                                        /* initializes processor structure */
  for (n = 0; n < MAX_TASKS_PER_PROCESSOR; n++)
    processors[processor_id].tasks[n] = -1;      /* no task for the moment */

  return processor_id;
}

/*-------------------------------------------------------------------------*/
/* new_channel:                                                            */
/* Creates a new channel.                                                  */
/*-------------------------------------------------------------------------*/
channel new_channel (void)
{
  int channel_id = 0;                              /* found new channel id */
  int n;                            /* counter to find a free channel slot */
  
  CHECK_INIT;                                      /* initialization stuff */

                  /* find free channel slot and change free state if found */
  for (n = 0; !channels[n].free && (n < MAX_CHANNELS); n++);
  if (n == MAX_CHANNELS) ERROR("no free channel slot");
  channel_id = n;
  channels[channel_id].free = 0;
                                          /* initializes channel structure */
  channels[channel_id].from = -1;                        /* not placed yet */
  channels[channel_id].to = -1;                          /* not placed yet */

  return channel_id;
}

/*-------------------------------------------------------------------------*/
/* new_implementation:                                                     */
/* Declare a new task implementation.                                      */
/* An implementation is a collection of modules which contain the source   */
/* code of the task. This function takes as parameter a serie of names     */
/* which corresponds to the name of the source files without extension.    */
/* This list must be terminated by the NULL pointer.                       */
/* For example, one can declare a task implementation as follows:          */
/*  implementation master_i = new_implementation("master", "loadb", NULL); */
/*-------------------------------------------------------------------------*/
#ifdef PACTC
implementation new_implementation (char *module_name, ...)
#else
implementation new_implementation (va_alist)
va_dcl
#endif
{
  va_list pvar;
  char *task_name;
  int n;
  char *act_module_name;
  module act_module;
  int nb_names;
  
  CHECK_INIT;                                      /* initialization stuff */
  
  /* the name of the implementation (or task linked unit) is equal to the  */
  /* first name in the list.                                               */

  #ifdef PACTC
    va_start(pvar, module_name);
    strcpy(task_name, module_name); 
  #else
    va_start(pvar);
    task_name = va_arg(pvar, char *);           /* get implementation name */
  #endif

  for (n = 0; (!modules[n].free) && (n < MAX_MODULES); n++);
  act_module = n;
  modules[act_module].free = 0;
  modules[act_module].nb_tasks = 0;
  strcpy(modules[act_module].name, task_name);

  nb_names = 1;                                     /* update module names */
  strcpy(modules[act_module].modules_names[0], task_name);
  act_module_name = va_arg(pvar, char *);
  while (act_module_name != NULL)
  {
    strcpy(modules[act_module].modules_names[nb_names], act_module_name);
    nb_names++;
    act_module_name = va_arg(pvar, char *);
  }
  modules[act_module].nb_names = nb_names;
    
  va_end(pvar);
  return act_module;
}

/*-------------------------------------------------------------------------*/
/* new_task:                                                               */
/* Declare a new task.                                                     */
/* This function returns the id of a new task. It must be used to define   */
/* the processor on which the task will be executed, the implementation    */
/* which implements it, and the parameters that must be passed to the      */
/* task at startup.                                                        */
/* The general form of this function is:                                   */
/*   t = new_task(                                                         */
/*     <processor>,                                                        */
/*     <implementation>,                                                   */
/*     <parameters>                                                        */
/*   );                                                                    */
/* The parameters are a list of tuples of the form <type>, <value> if the  */
/* value is atomic or <type>, <nitems>, <value> if the value is an array   */
/* where type is a constant defined in this file describing the type of the*/
/* parameter. The list of tuples must be terminated by the constant END.   */
/* Allowed types identifiers are defined in <ppc/types.h>.                 */
/*-------------------------------------------------------------------------*/
#ifdef PACTC
task new_task (processor p, implementation i, ...)
#else
task new_task (va_alist)
va_dcl
#endif
{
  va_list pvar;
  int task_id = 0;                                    /* found new task id */
  processor proc_id;                                 /* given processor id */
  implementation impl_id;                       /* given implementation id */
  int act_type;                                   /* actual parameter type */
  int nb_param;                                /* number of task parameter */
  int n;                               /* counter to find a free task slot */
  
  CHECK_INIT;                                      /* initialization stuff */
  
  #ifdef PACTC
    va_start(pvar, i);
    proc_id = p;
    impl_id = i;
  #else
    va_start(pvar);
    proc_id = va_arg(pvar, processor);                 /* get processor id */
    impl_id = va_arg(pvar, implementation);       /* get implementation id */
  #endif
                     /* find free task slot and change free state if found */
  for (n = 0; !tasks[n].free && (n < MAX_TASKS); n++);
  if (n == MAX_TASKS) ERROR("no free task slot");
  task_id = n;
  tasks[task_id].free = 0;
                                             /* initializes task structure */
  for (n = 0; n < MAX_CHANNELS_PER_TASK; n++)
    tasks[task_id].channels[n] = -1;          /* no channel for the moment */
  
  place_task(task_id, proc_id);               /* place task onto processor */
  tasks[task_id].implemented = impl_id;      /* update implementation info */
  
  modules[impl_id].tasks[modules[impl_id].nb_tasks] = task_id;
  modules[impl_id].nb_tasks++;
  
  nb_param = 0;
  act_type = va_arg(pvar, int);
  while(act_type != END)
  {
    nb_param++;
    tasks[task_id].parameters[nb_param-1].type = act_type;
    switch(act_type)
    {
      case OUTPUT_CHANNEL:
      {
        channel channel_id;
	channel_id = va_arg(pvar, channel);
        tasks[task_id].parameters[nb_param-1].param.value.channel_val = 
	  channel_id;
	init_channel(task_id, channel_id, OUTPUT_CHANNEL);
        break;
      }
      case INPUT_CHANNEL:
      {
        channel channel_id;
	channel_id = va_arg(pvar, channel);
        tasks[task_id].parameters[nb_param-1].param.value.channel_val = 
	  channel_id;
	init_channel(task_id, channel_id, INPUT_CHANNEL);
        break;
      }
      case INT:
      {
        int act_val;
	act_val = va_arg(pvar, int);
        tasks[task_id].parameters[nb_param-1].param.value.int_val = 
	  act_val;
        break;
      }
      case SHORT:
      {
        short act_val;
	act_val = va_arg(pvar, int);
        tasks[task_id].parameters[nb_param-1].param.value.short_val = 
	  act_val;
        break;
      }
      case UNSIGNED_SHORT:
      {
        unsigned short act_val;
	act_val = va_arg(pvar, unsigned int);
        tasks[task_id].parameters[nb_param-1].param.value.unsigned_short_val = 
	  act_val;
        break;
      }
      case LONG:
      {
        long act_val;
	act_val = va_arg(pvar, long);
        tasks[task_id].parameters[nb_param-1].param.value.long_val = 
	  act_val;
        break;
      }
      case UNSIGNED_LONG:
      {
        unsigned long act_val;
	act_val = va_arg(pvar, unsigned long);
        tasks[task_id].parameters[nb_param-1].param.value.unsigned_long_val = 
	  act_val;
        break;
      }
      case FLOAT:
      {
        float act_val;
	act_val = va_arg(pvar, double);
        tasks[task_id].parameters[nb_param-1].param.value.float_val = 
	  act_val;
        break;
      }
      case DOUBLE:
      {
        double act_val;
	act_val = va_arg(pvar, double);
        tasks[task_id].parameters[nb_param-1].param.value.double_val = 
	  act_val;
        break;
      }
      case CHAR:
      {
        char act_val;
	act_val = va_arg(pvar, int);
        tasks[task_id].parameters[nb_param-1].param.value.char_val = 
	  act_val;
        break;
      }
      case OUTPUT_CHANNEL_ARRAY:
      {
        int nitems = va_arg(pvar, int);
	channel *act_ptr = va_arg(pvar, channel *);
	int blen = sizeof(channel)*nitems;
	int m;
	channel *tmp_ptr;
	
        tasks[task_id].parameters[nb_param-1].param.ptr.nitems = nitems;
	bcopy(act_ptr, 
	  tasks[task_id].parameters[nb_param-1].param.ptr.address = 
	  (void *)malloc(blen),
	  blen);
	  
	for (tmp_ptr = act_ptr, m = 0; m < nitems; tmp_ptr++, m++)
	  init_channel(task_id, *tmp_ptr, OUTPUT_CHANNEL);
        break;
      }
      case INPUT_CHANNEL_ARRAY:
      {
        int nitems = va_arg(pvar, int);
	channel *act_ptr = va_arg(pvar, channel *);
	int blen = sizeof(channel)*nitems;
	int m;
	channel *tmp_ptr;
	
        tasks[task_id].parameters[nb_param-1].param.ptr.nitems = nitems;
	bcopy(act_ptr, 
	  tasks[task_id].parameters[nb_param-1].param.ptr.address = 
	  (void *)malloc(blen),
	  blen);
	  
	for (tmp_ptr = act_ptr, m = 0; m < nitems; tmp_ptr++, m++)
	  init_channel(task_id, *tmp_ptr, INPUT_CHANNEL);
        break;
      }
      case INT_ARRAY:
      {
	int nitems = va_arg(pvar, int);
	int *act_ptr = va_arg(pvar, int *);
	int blen = sizeof(int)*nitems;
	
	tasks[task_id].parameters[nb_param-1].param.ptr.nitems = nitems;
	bcopy(act_ptr, 
	  tasks[task_id].parameters[nb_param-1].param.ptr.address = 
	  (void *)malloc(blen),
	  blen);
	break;
      }
      case SHORT_ARRAY:
      {
	int nitems = va_arg(pvar, int);
	short *act_ptr = va_arg(pvar, short *);
	int blen = sizeof(short)*nitems;
	
	tasks[task_id].parameters[nb_param-1].param.ptr.nitems = nitems;
	bcopy(act_ptr, 
	  tasks[task_id].parameters[nb_param-1].param.ptr.address = 
	  (void *)malloc(blen),
	  blen);
	break;
      }
      case UNSIGNED_SHORT_ARRAY:
      {
	int nitems = va_arg(pvar, int);
	unsigned short *act_ptr = va_arg(pvar, unsigned short *);
	int blen = sizeof(unsigned short)*nitems;
	
	tasks[task_id].parameters[nb_param-1].param.ptr.nitems = nitems;
	bcopy(act_ptr, 
	  tasks[task_id].parameters[nb_param-1].param.ptr.address = 
	  (void *)malloc(blen),
	  blen);
	break;
      }
      case LONG_ARRAY:
      {
	int nitems = va_arg(pvar, int);
	long *act_ptr = va_arg(pvar, long *);
	int blen = sizeof(long)*nitems;
	
	tasks[task_id].parameters[nb_param-1].param.ptr.nitems = nitems;
	bcopy(act_ptr, 
	  tasks[task_id].parameters[nb_param-1].param.ptr.address = 
	  (void *)malloc(blen),
	  blen);
	break;
      }
      case UNSIGNED_LONG_ARRAY:
      {
	int nitems = va_arg(pvar, int);
	unsigned long *act_ptr = va_arg(pvar, unsigned long *);
	int blen = sizeof(unsigned long)*nitems;
	
	tasks[task_id].parameters[nb_param-1].param.ptr.nitems = nitems;
	bcopy(act_ptr, 
	  tasks[task_id].parameters[nb_param-1].param.ptr.address = 
	  (void *)malloc(blen),
	  blen);
	break;
      }
      case FLOAT_ARRAY:
      {
	int nitems = va_arg(pvar, int);
	float *act_ptr = va_arg(pvar, float *);
	int blen = sizeof(float)*nitems;
	
	tasks[task_id].parameters[nb_param-1].param.ptr.nitems = nitems;
	bcopy(act_ptr, 
	  tasks[task_id].parameters[nb_param-1].param.ptr.address = 
	  (void *)malloc(blen),
	  blen);
	break;
      }
      case DOUBLE_ARRAY:
      {
	int nitems = va_arg(pvar, int);
	double *act_ptr = va_arg(pvar, double *);
	int blen = sizeof(double)*nitems;
	
	tasks[task_id].parameters[nb_param-1].param.ptr.nitems = nitems;
	bcopy(act_ptr, 
	  tasks[task_id].parameters[nb_param-1].param.ptr.address = 
	  (void *)malloc(blen),
	  blen);
	break;
      }
      case CHAR_ARRAY:
      {
	int nitems = va_arg(pvar, int);
	char *act_ptr = va_arg(pvar, char *);
	int blen = sizeof(char)*nitems;
	
	tasks[task_id].parameters[nb_param-1].param.ptr.nitems = nitems;
	bcopy(act_ptr, 
	  tasks[task_id].parameters[nb_param-1].param.ptr.address = 
	  (void *)malloc(blen),
	  blen);
	break;
      }
      default:
      {
        ERROR("unknown type during task interface definition");
	break;
      }
    }
    act_type = va_arg(pvar, int);
  }
  
  tasks[task_id].nb_parameters = nb_param;
  va_end(pvar);
  return task_id;
}

/*-------------------------------------------------------------------------*/
/* free_config_data:                                                       */
/* Release the eventual data allocated during the virtual machine          */
/* configuration. This function is called even by the configurer main      */
/* function or from the starter at startup time.                           */
/* This function is implemented in the config.c source file.               */
/*-------------------------------------------------------------------------*/
void free_config_data(void)
{
  int n;                                                        /* counter */
  int act_param;
  
  for (n = 0; n < MAX_TASKS; n++)
  {
    if (!tasks[n].free)
    {
      for (act_param = 0; act_param < tasks[n].nb_parameters; act_param++)
      {
        switch(tasks[n].parameters[act_param].type)
	{
	  case OUTPUT_CHANNEL_ARRAY:
	  case INPUT_CHANNEL_ARRAY:
          case INT_ARRAY:
          case SHORT_ARRAY:
          case UNSIGNED_SHORT_ARRAY:
          case LONG_ARRAY:
          case UNSIGNED_LONG_ARRAY:
          case FLOAT_ARRAY:
          case DOUBLE_ARRAY:
          case CHAR_ARRAY:
	    free(tasks[n].parameters[act_param].param.ptr.address);
	    break;
	  default:
	    break;
	}
      }
    }
  }
}

