/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* PACT-C Hardware configuration implementation.                           */
/* The functions defined in this file must be called from the function     */
/* hardware_config() implemented by the user.                              */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 18.11.1993  Droux       Created                                         */
/* 15.07.1994  Droux       New structure of source directory               */
/* 17.08.1994  Droux       Support for custom hardware configuration file  */
/* 30.08.1994  Droux       PACT-C program linked with server library       */
/*-------------------------------------------------------------------------*/
/* filename: confpact.c                                                    */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   Linux 0.99.pl9, Intel Computers                                       */
/*   PACT-C v93.1, T800, T805                                              */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define confpact_serv

#ifndef PACTC
#include <time.h>
#include <pwd.h>
#endif
#include <stdio.h>
#include <ppc/csp/version.h>
#include <ppc/config/confpriv.h>
#include <ppc/config/confpact.h>
#include "confhard.h"

#define STACK_SIZE "50K"                /* default initial task stack size */

/*-------------------------------------------------------------------------*/
/* New types for Transputer configuration.                                 */
/*-------------------------------------------------------------------------*/

typedef struct                               /* properties of a transputer */
{
  tran_types type;                    /* transputer type (T414, T805, ...) */
  unsigned long memory;                        /* avaiable memory in bytes */
  int nb_procs;          /* number of virt. processors on this transputers */
  processor procs [MAX_PROCS_PER_TRANS];      /* virt. procs on transputer */
} trans_descr;

typedef char link;                 /* there are max 4 links per transputer */

typedef struct       /* properties of a connection between two transputers */
{
  transputer from_tran;                               /* connection source */
  link from_link;
  transputer to_tran;                            /* connection destination */
  link to_link;
} conn_descr;

/*-------------------------------------------------------------------------*/
/* Variables local to this configuration module.                           */
/*-------------------------------------------------------------------------*/

#define ARCH_DIR      ".pactc/"
#define MAKEFILE_NAME ARCH_DIR "makefile"

#define NB_LINKS_PER_TRANS 4              /* max # of links per transputer */
#define MAX_NB_CONN MAX_NB_TRANS*NB_LINKS_PER_TRANS   /* max # of connect. */

static int nb_trans = 0;                 /* number of declared transputers */
static trans_descr trans [MAX_NB_TRANS];         /* transputers properties */
static int nb_conn = 0;                           /* number of connections */
static conn_descr conn [MAX_NB_CONN];                       /* connections */

/* It is possible that more than one task must be linked with the same     */
/* module. In this case, it is important to avoid the duplication of this  */
/* module makefile dependency. To avoid this duplication, we record a      */
/* unique key for each encountered module.                                 */

static unsigned long modules_keys [MAX_MODULES*MAX_OBJECTS_PER_MODULE];
static int nb_keys = 0;

static char *trans_types_str[] =               /* transputer types strings */
{                           /* for example: trans_types_str[T805] = "T805" */
  "T212", "M212", "T222", "T225", 
  "T400", "T414", "T425", 
  "T800", "T801", "T805"
};

/* PPC divides the memory of the transputer into three regions: one        */
/* for the stack, one for the virtual channels routers, and the third for  */
/* the tasks workspace. We specify here how many percent of the memory are */
/* allocated to the stack and to the routers.                              */

#define STACK_PERCENT 10.0
#define ROUTERS_PERCENT 27.0
#define HEAP_PERCENT (100.0 - STACK_PERCENT - ROUTERS_PERCENT)

/*-------------------------------------------------------------------------*/
/* new_transputer:                                                         */
/* Creates a new transputer and returns its id.                            */
/*-------------------------------------------------------------------------*/
transputer new_transputer (
  tran_types type,             /* transputer types (T222, T400, T805, ...) */
  unsigned long memory                       /* transputer memory in bytes */
)
{
  if (nb_trans+1 > MAX_NB_TRANS)
  {
    fprintf(stderr, "HARDWARE CONFIGURATION ERROR:\n");
    fprintf(stderr, "  too many Transputers declared (%d allowed)\n", 
      MAX_NB_TRANS);
    exit(1);
  }
  trans[nb_trans].type = type;
  trans[nb_trans].memory = memory;
  trans[nb_trans].nb_procs = 0;
  nb_trans++;
  return nb_trans-1;
}

/*-------------------------------------------------------------------------*/
/* link_transputers:                                                       */
/* Declares the transputers physical connections.                          */
/*-------------------------------------------------------------------------*/
void link_transputers (
  transputer tran1, int link1, 
  transputer tran2, int link2)
{
  if (nb_conn+1 > MAX_NB_CONN)
  {
    fprintf(stderr, "HARDWARE CONFIGURATION ERROR:\n");
    fprintf(stderr, "  too many Connections declared (%d allowed)\n", 
      MAX_NB_CONN);
    exit(1);
  }
  if ((link1 >= NB_LINKS_PER_TRANS) || (link2 >= NB_LINKS_PER_TRANS))
  {
    fprintf(stderr, "HARDWARE CONFIGURATION ERROR:\n");
    fprintf(stderr, "  bad link identifier (allowed 0-%d)\n", 
      NB_LINKS_PER_TRANS-1);
    exit(1);
  }
  conn[nb_conn].from_tran = tran1;
  conn[nb_conn].from_link = (link)link1;
  conn[nb_conn].to_tran = tran2;
  conn[nb_conn].to_link = (link)link2;
  nb_conn++;
}

/*-------------------------------------------------------------------------*/
/* map_transputer:                                                         */
/* Maps a virtual processor on a given transputer.                         */
/*-------------------------------------------------------------------------*/
void place_processor_on_transputer (
  processor proc, 
  transputer tran
)
{
  if (trans[tran].nb_procs+1 > MAX_PROCS_PER_TRANS)
  {
    fprintf(stderr, "HARDWARE CONFIGURATION ERROR:\n");
    fprintf(stderr, "  too many virtual processors placed\n");
    fprintf(stderr, "  on transputer (%d allowed)\n", 
      MAX_PROCS_PER_TRANS);
    exit(1);
  }
  trans[tran].procs[trans[tran].nb_procs] = proc;
  trans[tran].nb_procs++;
}

/*-------------------------------------------------------------------------*/
/* generate_makefile:                                                LOCAL */
/* Generate the makefile for the current program.                          */
/*-------------------------------------------------------------------------*/
#ifndef PACTC
void generate_makefile (
  char *name, 
  char *user_name,
  char *date_str
)
{
  FILE *makefile;
  
  /*-----------------------------------------------------------------------*/
                                                                 /* header */
  printf("generating PACT-C makefile for \"%s\"\n", name);
  makefile = fopen(MAKEFILE_NAME, "w");
  fprintf(makefile, "\n");
  fprintf(makefile, " # This file has been generated by the PPC ");
  fprintf(makefile, "configurer\n");
  fprintf(makefile, " # PPC Version " PPC_VERSION "\n");
  fprintf(makefile, " # NEVER EDIT IT !\n");
  fprintf(makefile, " # Created on %s by %s\n", date_str, user_name);
  fprintf(makefile, " # PPC program name: \"%s\"\n", name);
  fprintf(makefile, " # Target architecture: PACT-C for Transputers\n\n");

  /*-----------------------------------------------------------------------*/
                                              /* compiler and linker flags */
  fprintf(makefile, "# compiler and linker flags\n");
  fprintf(makefile, "PCC=pcc\n");
  fprintf(makefile, "CFLAGS=-I$(PPC_HOME)/include ");
  fprintf(makefile, "-D_$(ARCH) \\\n");
  fprintf(makefile, "\t-D$(HOST_ARCH) -O $(OTHER_CFLAGS)\n");
  fprintf(makefile, "LFLAGS=-lconc -lconf -lconf$(ARCHLIB) -lcsp$(ARCHLIB) \\\n");
  fprintf(makefile, "\t-lserv$(ARCHLIB) -L$(PPC_HOME)/lib/PACTC\n");
  fprintf(makefile, "\n");

  /*-----------------------------------------------------------------------*/
  {                          /* list of object modules of each linked unit */
    int act_lku;
    int act_obj;
    
    fprintf(makefile, "# list of object modules of ");
    fprintf(makefile, "each linked unit\n");
  
    for (act_lku = 0; act_lku < MAX_MODULES; act_lku++)
      if (!modules[act_lku].free)
      {
	fprintf(makefile, "%sOBJ=", modules[act_lku].name);
	fprintf(makefile, ".$(ARCH)/PACTC/%s.o ", name);
	fprintf(makefile, ".$(ARCH)/PACTC/$(ARCH_FILE).o ");
	fprintf(makefile, ".$(ARCH)/PACTC/_%s.o ",
	  modules[act_lku].name);
	for (act_obj = 0; act_obj < modules[act_lku].nb_names; act_obj++)
	  fprintf(makefile, ".$(ARCH)/PACTC/%s.o ",
	    modules[act_lku].modules_names[act_obj]);
	fprintf(makefile, "\n");
      }
    fprintf(makefile, "\n");
  }

  /*-----------------------------------------------------------------------*/
  {                           /* list of all linked units, main dependency */
    int n;
    int act_obj;
    
    fprintf(makefile, "# list of all linked unit, main dependency\n");
    fprintf(makefile, ".$(ARCH)/PACTC/%s.run: .$(ARCH)/PACTC/%s.cfg ", 
      name, name);
  
    for (n = 0; n < MAX_MODULES; n++)                  
      if (!modules[n].free)
      {
	fprintf(makefile, ".$(ARCH)/PACTC/%s.lku ",
	  modules[n].name);
      }
      
    fprintf(makefile, "\n");
    fprintf(makefile, "\tcd .$(ARCH)/PACTC; $(PCC) -V %s.cfg\n", name);
    fprintf(makefile, "\n");
  }
  
  /*-----------------------------------------------------------------------*/
  {                                           /* linked units dependencies */
    int act_lku;

    fprintf(makefile, "# linked units dependencies\n");
    
    for (act_lku = 0; act_lku < MAX_MODULES; act_lku++)
      if (!modules[act_lku].free)
      {
	fprintf(makefile, ".$(ARCH)/PACTC/%s.lku: $(%sOBJ)\n",
	  modules[act_lku].name, modules[act_lku].name);
	fprintf(makefile, "\t$(PCC) -V -x $(LFLAGS) -o $@ $(%sOBJ)\n\n",
	  modules[act_lku].name);
      }
  }

  /*-----------------------------------------------------------------------*/
  {                                           /* object files dependencies */
    int act_lku;
    int act_obj;
    int duplicated;                         /* 1 if a module is duplicated */
    int key_idx;                                    /* to browse keys list */
    unsigned long act_key;                          /* actual module's key */
    
    fprintf(makefile, "# object files dependencies\n");
  
    nb_keys = 0;
    for (act_lku = 0; act_lku < MAX_MODULES; act_lku++)
      if (!modules[act_lku].free)
      {
        fprintf(makefile, ".$(ARCH)/PACTC/_%s.o: .$(ARCH)/_%s.c\n",
	  modules[act_lku].name, modules[act_lku].name);
        fprintf(makefile, "\t$(PCC) $(CFLAGS) -c $? -o _%s.o\n",
	  modules[act_lku].name);
	fprintf(makefile, "\t@mv _%s.o $@\n\n", modules[act_lku].name);
	
	for (act_obj = 0; act_obj < modules[act_lku].nb_names; act_obj++)
	{
	  act_key = confkey(modules[act_lku].modules_names[act_obj]);
	  duplicated = 0;
	  for (key_idx = 0; key_idx < nb_keys; key_idx++)
	    duplicated = duplicated || (modules_keys[key_idx] == act_key);
	  if (!duplicated)
	  {
	    fprintf(makefile, ".$(ARCH)/PACTC/%s.o: %s.c\n",
	      modules[act_lku].modules_names[act_obj],
	      modules[act_lku].modules_names[act_obj]);
	      
	    fprintf(makefile, "\t$(PCC) $(CFLAGS) -c $? -o %s.o\n",
	      modules[act_lku].modules_names[act_obj]);
	    fprintf(makefile, "\t@mv %s.o $@\n\n", 
	      modules[act_lku].modules_names[act_obj]);
	      
	    modules_keys[nb_keys] = act_key;
	    nb_keys++;
	  }
	}
      }
    fprintf(makefile, "\n");
  }

  /*-----------------------------------------------------------------------*/
                                                         /* close makefile */
  fclose(makefile);
}
#endif

/*-------------------------------------------------------------------------*/
/* cfg_found_param:                                                  LOCAL */
/* For a given task t, find the parameter that takes the channel c and put */
/* it in p. If the channel is given as the element of a channel array,     */
/* then put the index of the element in i, otherwise set i = -1.           */
/* It is assumed that the channel is in the given task, otherwise this     */
/* function may crash.                                                     */
/*-------------------------------------------------------------------------*/
#ifndef PACTC
void cfg_found_param(
  task t,                                                       /* task id */
  channel c,                                                 /* channel id */
  int *p,                          /* returned parameter index of the task */
  int *i              /* returned parameter array index if parameter array */
)
{
  int found = 0;
  
  for (*p = 0; (*p < tasks[t].nb_parameters) && (!found); (*p)++)
  {
    switch (tasks[t].parameters[*p].type)
    {
      case OUTPUT_CHANNEL:
      case INPUT_CHANNEL:
	if (tasks[t].parameters[*p].param.value.channel_val == c)
	{
	  found = 1;
	  *i = -1;
	}
	break;
      case OUTPUT_CHANNEL_ARRAY:
      case INPUT_CHANNEL_ARRAY:
	for (*i = 0; 
	     (*i < tasks[t].parameters[*p].param.ptr.nitems) && (!found);
	     (*i)++)
	  found = 
	    ((channel *)tasks[t].parameters[*p].param.ptr.address)[*i] == c;
	(*i)--;
	break;
      default:
	break;
    }
  }
  (*p)--;
}
#endif

/*-------------------------------------------------------------------------*/
/* check_main_task:                                                  LOCAL */
/* Checks if one of the tasks implemented by the given module is the main  */
/* task. The main task is the task which communicates with the user and    */
/* which starts the other tasks.                                           */
/* Returns the task if it is the case, -1 if it isn't.                     */
/* This function is used during the files generation as well as during the */
/* execution of the tasks to determine its number.                         */
/*-------------------------------------------------------------------------*/
#ifndef PACTC
int check_main_task (int impl)
{
  int act_task_idx;
  int act_chan;
  int act_task;
  int from_host_ok;
  int to_host_ok;
  int nb_param;
  int act_param_idx;
  int act_param_type;
  
  /* For each task which is implemented by the current implementation,     */
  /* checks if this task takes the two channels from_host and to_host      */
  /* as parameters. If it is the case, then the main task is implemented   */
  /* by this module.                                                       */
  
  for (act_task_idx = 0; act_task_idx < modules[impl].nb_tasks; 
       act_task_idx++)
  {
    act_task = modules[impl].tasks[act_task_idx];
    from_host_ok = 0;
    to_host_ok = 0;
    nb_param = tasks[act_task].nb_parameters;
    for (act_param_idx = 0; act_param_idx < nb_param; act_param_idx++)
    {
      act_param_type = tasks[act_task].parameters[act_param_idx].type;
      if ((act_param_type == INPUT_CHANNEL) || 
          (act_param_type == OUTPUT_CHANNEL))
      {
        act_chan =
	  tasks[act_task].parameters[act_param_idx].param.value.channel_val;
        from_host_ok = from_host_ok || (act_chan == FROM_HOST_CHANNEL_ID);
        to_host_ok = to_host_ok || (act_chan == TO_HOST_CHANNEL_ID);
      }
    }
    if (from_host_ok && to_host_ok)
      return act_task;
    if ((from_host_ok && !to_host_ok) || (!from_host_ok && to_host_ok))
    {
      printf("CONFIGURATION ERROR :\n");
      printf("  The from_host and to_host channels must be assigned");
      printf(" to the same task.\n");
      printf("  (configuration aborted)\n");
      exit(1);
    }
  }
  return -1;
}
#endif

/*-------------------------------------------------------------------------*/
/* generate_config:                                                  LOCAL */
/* Generate the PACT-C configuration file for the current program.         */
/*-------------------------------------------------------------------------*/
#ifndef PACTC
void generate_config (
  char *name, 
  char *user_name,
  char *date_str
)
{
  char cfg_name [100];
  FILE *cfg;
  
  printf("generating PACT-C config file for \"%s\"\n", name);
  sprintf(cfg_name, ARCH_DIR "/PACTC/%s.cfg", name);
  cfg = fopen(cfg_name, "w");

  /*-----------------------------------------------------------------------*/
                                                                 /* header */
  fprintf(cfg, "\n/*\n");
  fprintf(cfg, " * This file has been generated by the PPC ");
  fprintf(cfg, "configurer\n");
  fprintf(cfg, " * PPC Version " PPC_VERSION "\n");
  fprintf(cfg, " * NEVER EDIT IT !\n");
  fprintf(cfg, " * Created on %s by %s\n", date_str, user_name);
  fprintf(cfg, " * PPC program name: \"%s\"\n", name);
  fprintf(cfg, " * Configuration file\n");
  fprintf(cfg, " * Target architecture: PACT-C for Transputers\n");
  fprintf(cfg, " */\n\n");

  /*-----------------------------------------------------------------------*/
  {                                                 /* physical processors */
    int act_trans;                                    /* actual transputer */
    
    /* To each transputer transputer declared in the PACT-C hardware       */
    /* configuration corresponds exactly one physical processor            */
    /* declaration.                                                        */
    fprintf(cfg, "/* physical processors */\n");
    for (act_trans = 0; act_trans < nb_trans; act_trans++)
    {
      fprintf(cfg, "%s ", trans_types_str[trans[act_trans].type]);
      fprintf(cfg, "(memory=%lu) ", trans[act_trans].memory);
      fprintf(cfg, "p%d;\n", act_trans);
    }
    fprintf(cfg, "\n");
  }
  
  /*-----------------------------------------------------------------------*/
  {                                                /* physical connections */
    int act_conn;                                     /* actual connection */

    fprintf(cfg, "/* physical connections */\n");
    for (act_conn = 0; act_conn < nb_conn; act_conn++)
    {
      fprintf(cfg, "connect ");
      
      if (conn[act_conn].from_tran == HOST)
        fprintf(cfg, "host");
      else
        fprintf(cfg, "p%d.link[%d]", 
	  conn[act_conn].from_tran, (int)conn[act_conn].from_link);
	  
      fprintf(cfg, " to ");
      
      if (conn[act_conn].to_tran == HOST)
        fprintf(cfg, "host");
      else
        fprintf(cfg, "p%d.link[%d]", 
	  conn[act_conn].to_tran, (int)conn[act_conn].to_link);
	  
      fprintf(cfg, ";\n");
    }
    fprintf(cfg, "\n");
  }
  
  /*-----------------------------------------------------------------------*/
  {                                             /* generate task interface */
    int act_module;        /* actual implementation module being processed */
    int act_task;              /* id of the 1st task of the current module */
    int act_param;                 /* actual parameter of the current task */
    char size_str [20];                   /* array size if parameter array */
    
    for (act_module = 0; act_module < MAX_MODULES; act_module++) 
    {
      if (!modules[act_module].free)
      {
        fprintf(cfg, "/* \"%s\" task interface */\n",
	  modules[act_module].name);
        fprintf(cfg, "task\n");
	fprintf(cfg, "(\n");
	
	/* The memory of the transputer is divided into three regions:     */
	/* one region for the stack, one for the virtual channels routers, */
	/* and the third for the tasks workspace.                          */
	
	{                           /* calculate task stack and heap sizes */
	  unsigned long stacksize;
	  unsigned long heapsize;
	  int proc_id;
	  int trans_id;
	  int act_trans;
	  int act_proc;
	  
	  /* Search the transputer on which the first task of the current  */
	  /* module is placed. It is assumed that each transputer has      */
	  /* only one task running on it, and that each tasks of the       */
	  /* same implementation class require the same memory.            */
	  
	  proc_id = tasks[modules[act_module].tasks[1]].on_processor;
	  trans_id = -1;
	  for (act_trans = 0; 
	       (act_trans < nb_trans) && (trans_id == -1); 
	       act_trans++)
	  {
	    for (act_proc = 0;
	         (act_proc < trans[act_trans].nb_procs) && (trans_id == -1);
		 act_proc++)
	    {
	      if (trans[act_trans].procs[act_proc] == proc_id)
	        trans_id = act_trans;
	    }
	  }

	  stacksize = (unsigned long)(
	              (double)((double)STACK_PERCENT / (double)100.0) * 
	              (double)trans[trans_id].memory);
	  heapsize = (unsigned long)(
	              (double)((double)HEAP_PERCENT / (double)100.0) * 
	              (double)trans[trans_id].memory);
	
	  fprintf(cfg, "  stacksize=%lu,\n", stacksize);
	  fprintf(cfg, "  heapsize=%lu,\n", heapsize);
	  fprintf(cfg, "  interface\n");
	  fprintf(cfg, "  (\n");
	}
	
	/* The first parameter will be the PPC task number. It is then     */
	/* guaranteed that each task can get its number to, among other    */
	/* things, set the parameters at the correct values when calling   */
	/* the PPC task entry point.                                       */
	fprintf(cfg, "    int my_number,\n");
	
	act_task = (int)modules[act_module].tasks[0];
	for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
	      act_param++)
	{
	  char c_param_name [100];
	  int act_param_type;
	  
	  size_str[0] = '\0';
	  act_param_type = tasks[act_task].parameters[act_param].type;
	  switch(act_param_type)
	  {
	    case OUTPUT_CHANNEL:
	      strcpy(c_param_name, "output ");
	      break;
	    case INPUT_CHANNEL:
	      strcpy(c_param_name, "input ");
	      break;
	    case INT:
	      strcpy(c_param_name, "int ");
	      break;
	    case SHORT:
	      strcpy(c_param_name, "short ");
	      break;
	    case UNSIGNED_SHORT:
	      strcpy(c_param_name, "unsigned short ");
	      break;
	    case LONG:
	      strcpy(c_param_name, "long ");
	      break;
	    case UNSIGNED_LONG:
	      strcpy(c_param_name, "unsigned long ");
	      break;
	    case FLOAT:
	      strcpy(c_param_name, "float ");
	      break;
	    case DOUBLE:
	      strcpy(c_param_name, "double ");
	      break;
	    case CHAR:
	      strcpy(c_param_name, "char ");
	      break;
	    case OUTPUT_CHANNEL_ARRAY:
	      strcpy(c_param_name, "output ");
	      sprintf(size_str, "[%d]", 
	        tasks[act_task].parameters[act_param].param.ptr.nitems);
	      break;
	    case INPUT_CHANNEL_ARRAY:
	      strcpy(c_param_name, "input ");
	      sprintf(size_str, "[%d]", 
	        tasks[act_task].parameters[act_param].param.ptr.nitems);
	      break;
	    case INT_ARRAY:
	      strcpy(c_param_name, "int ");
	      sprintf(size_str, "[%d]", 
	        tasks[act_task].parameters[act_param].param.ptr.nitems);
	      break;
	    case SHORT_ARRAY:
	      strcpy(c_param_name, "short ");
	      sprintf(size_str, "[%d]", 
	        tasks[act_task].parameters[act_param].param.ptr.nitems);
	      break;
	    case UNSIGNED_SHORT_ARRAY:
	      strcpy(c_param_name, "unsigned ");
	      sprintf(size_str, "[%d]", 
	        tasks[act_task].parameters[act_param].param.ptr.nitems);
	      break;
	    case LONG_ARRAY:
	      strcpy(c_param_name, "long ");
	      sprintf(size_str, "[%d]", 
	        tasks[act_task].parameters[act_param].param.ptr.nitems);
	      break;
	    case UNSIGNED_LONG_ARRAY:
	      strcpy(c_param_name, "unsigned long ");
	      sprintf(size_str, "[%d]", 
	        tasks[act_task].parameters[act_param].param.ptr.nitems);
	      break;
	    case FLOAT_ARRAY:
	      strcpy(c_param_name, "float ");
	      sprintf(size_str, "[%d]", 
	        tasks[act_task].parameters[act_param].param.ptr.nitems);
	      break;
	    case DOUBLE_ARRAY:
	      strcpy(c_param_name, "double ");
	      sprintf(size_str, "[%d]", 
	        tasks[act_task].parameters[act_param].param.ptr.nitems);
	      break;
	    case CHAR_ARRAY:
	      strcpy(c_param_name, "char ");
	      sprintf(size_str, "[%d]", 
	        tasks[act_task].parameters[act_param].param.ptr.nitems);
	      break;
	    default:
	      fprintf(stderr, "ERROR: unknown task parameter type.\n");
	      exit(1);
	  }
	  fprintf(cfg, "    %sparam_%d", c_param_name, act_param);
	  if (strlen(size_str))
	    fprintf(cfg, size_str);
	  
	  if ((act_param_type == INPUT_CHANNEL || 
	      act_param_type == OUTPUT_CHANNEL ||
	      act_param_type == INPUT_CHANNEL_ARRAY ||
	      act_param_type == OUTPUT_CHANNEL_ARRAY) &&
	      tasks[act_task].parameters[act_param].param.value.channel_val!=
	        TO_HOST_CHANNEL_ID &&
	      tasks[act_task].parameters[act_param].param.value.channel_val!=
	        FROM_HOST_CHANNEL_ID)
	  {
	    /* the actual parameter is a channel. Each machine-wide        */
	    /* PPC channel is made by two PACT-C channels:                 */
	    /* a synchronization and a data channels. We add here the      */
	    /* synchronization channel in the parameters list.             */
	    
	    fprintf(cfg, ",\n    %sparam_%d_sync",
	      c_param_name, act_param);
	    if (strlen(size_str))
	      fprintf(cfg, size_str);
	  }
	  
	  if (act_param < (tasks[act_task].nb_parameters-1))
	    fprintf(cfg, ",");
	  fprintf(cfg, "\n");
	}
	fprintf(cfg, "  )\n");
	fprintf(cfg, ") t%d[%d];\n", 
	  act_module, modules[act_module].nb_tasks);
	fprintf(cfg, "\n");
      }
    }
  }
  
  /*-----------------------------------------------------------------------*/
  {                  /* initializes the 'my_number' parameter of each task */
    int act_module;
    int act_task;
    
    /* Note that only the my_number parameter is initialized by PACT-C.    */
    /* The other parameters are determined by the PACT-C tasks when        */
    /* calling the PPC entry-point from the virtual machine                */
    /* configuration.                                                      */
    fprintf(cfg, 
      "/* initializes the 'my_number' parameter of each task */\n");

    for (act_module = 0; act_module < MAX_MODULES; act_module++) 
    {
      if (!modules[act_module].free)
      {
        for (act_task = 0; 
	     act_task < modules[act_module].nb_tasks;
	     act_task++)
	{
	  fprintf(cfg, "t%d[%d] (my_number = %d);\n",
	    act_module, act_task, modules[act_module].tasks[act_task]);
	}
      }
    }
    fprintf(cfg, "\n");
  }

  /*-----------------------------------------------------------------------*/
  {                                  /* place tasks on physical processors */
    int act_module;
    int act_task;
    int proc_id;            /* virt. processor on which the act. task runs */
    int act_trans;                      /* actual transputer being checked */
    int act_proc;                /* actual virtual processor on transputer */
    int trans_id;         /* transputer on which the task must be executed */
    
    fprintf(cfg, "/* place tasks on physical processors */\n");

    for (act_module = 0; act_module < MAX_MODULES; act_module++) 
    {
      if (!modules[act_module].free)
      {
        for (act_task = 0; 
	     act_task < modules[act_module].nb_tasks;
	     act_task++)
	{
	  fprintf(cfg, "place t%d[%d] on ", act_module, act_task);
	  proc_id = tasks[modules[act_module].tasks[act_task]].on_processor;
	  trans_id = -1;
	  for (act_trans = 0; 
	       (act_trans < nb_trans) && (trans_id == -1); 
	       act_trans++)
	  {
	    for (act_proc = 0;
	         (act_proc < trans[act_trans].nb_procs) && (trans_id == -1);
		 act_proc++)
	    {
	      if (trans[act_trans].procs[act_proc] == proc_id)
	        trans_id = act_trans;
	    }
	  }
	  fprintf(cfg, "p%d;\n", trans_id);
	}
      }
    }
    fprintf(cfg, "\n");
  }
  
  /*-----------------------------------------------------------------------*/
  {                                              /* connect tasks together */
    int T1;                                          /* 1st task id (sink) */
    int T2;                                          /* 2nd task id (dest) */
    int i;                                   /* 1st task implementation id */
    int j;                           /* 1st task implementation task index */
    int m;                                   /* 2nd task implementation id */
    int n;                           /* 2nd task implementation task index */
    int c;                  /* channel id between the 1st and the 2nd task */
    int p1, i1;           /* 1st task parameter and evt index param_p1[i1] */
    int p2, i2;           /* 2nd task parameter and evt index param_p2[i2] */
  
    fprintf(cfg, "/* connect tasks together */\n");
    
    /* The tasks connection has the following form:                        */
    /* For each channel, find the two tasks it connects and assert:        */
    /* connect ti[j] to tm[n];                                             */
    /* where i and m are the implementation indices of the tasks and       */
    /* j and n the indices of the tasks in their respective                */
    /* implementation.                                                     */
    /* The problem here is to find i, j, m, and n, given a channel c.      */
    /* The algorithm works as follows:                                     */
    /*   For each channel c:                                               */
    /*     1. T1 <- channels[c].from                                       */
    /*     2. T2 <- channels[c].to                                         */
    /*     3. i <- tasks[T1].implemented                                   */
    /*     4. m <- tasks[T2].implemented                                   */
    /*     3. Find j such that modules[i].tasks[j] = T1                    */
    /*     4. Find n such that modules[m].tasks[n] = T2                    */
    /*  Where T1 and T2 are the tasks connected by the channel c.          */

    for (c = 0; c < MAX_CHANNELS; c++)
    {
      if ((!channels[c].free) && 
          (channels[c].from != -1) && 
	  (channels[c].to != -1))
      {
        T1 = channels[c].from;
	T2 = channels[c].to;
	i = tasks[T1].implemented;
	m = tasks[T2].implemented;
	
	for (j = 0; 
	     (j < MAX_TASKS_PER_MODULE) && (modules[i].tasks[j] != T1); 
	     j++);

	for (n = 0; 
	     (n < MAX_TASKS_PER_MODULE) && (modules[m].tasks[n] != T2); 
	     n++);
	     
	/* Once we have the implementation number and the instance         */
	/* index of the task in this implementation, we have to define     */
	/* the parameter of the task which takes the channel value. As     */
	/* it is also possible that the channel is given in a parameter    */
	/* array, its position in this array is also necessary.            */
	/* We denote here p1 and p2 the parameter indices, and i1, i2,     */
	/* the corresponding array position if the former condition is     */
	/* found to be true.                                               */
	
	cfg_found_param(T1, c, &p1, &i1);
	cfg_found_param(T2, c, &p2, &i2);
	
        fprintf(cfg, "connect ");
	
	fprintf(cfg, "t%d[%d]", i, j);                           /* source */
	fprintf(cfg, ".param_%d", p1);
	if (i1 != -1)
	  fprintf(cfg, "[%d]", i1);
	fprintf(cfg, " to ");
	
	fprintf(cfg, "t%d[%d]", m, n);                      /* destination */
	fprintf(cfg, ".param_%d", p2);
	if (i2 != -1)
	  fprintf(cfg, "[%d]", i2);
	fprintf(cfg, ";\n");

        /* connect also for each PPC channel the corresponding PACT-C      */
	/* synchronization channel.                                        */

        fprintf(cfg, "connect ");
	
	fprintf(cfg, "t%d[%d]", i, j);                           /* source */
	fprintf(cfg, ".param_%d_sync", p1);
	if (i1 != -1)
	  fprintf(cfg, "[%d]", i1);
	fprintf(cfg, " to ");
	
	fprintf(cfg, "t%d[%d]", m, n);                      /* destination */
	fprintf(cfg, ".param_%d_sync", p2);
	if (i2 != -1)
	  fprintf(cfg, "[%d]", i2);
	fprintf(cfg, ";\n");
      }
    }
  }
  
  /*-----------------------------------------------------------------------*/
  {                                /* connect the right task with the host */
    int n;
    int main_task;                /* module which implements the main task */
    int main_implementation;
    int instance;
    int from_p;
    int from_i;
    int to_p;
    int to_i;

    /* The task of the PPC program which is connected to the host is       */
    /* called the main task. The implementation which implements the main  */
    /* task is called the "main implementation". Here, we first look for   */
    /* the main task and the main implementation. Then, we look for the    */
    /* instance of the task in the implementation and the parameters which */
    /* corresponds to the to_hos and from_host channels.                   */

    main_implementation = -1;  /* search main task and main implementation */
    for (n = 0; n < MAX_MODULES; n++)
    {
      if (!modules[n].free)
      {
        main_task = check_main_task(n);
	if (main_task != -1)
	  main_implementation = n;
	break;
      }
    }

    if (main_implementation == -1)    /* verifies that a task is connected */
    {                                                       /* to the host */
      printf("CONFIGURATION ERROR :\n");
      printf("  No tasks are connected to the host by the to_host and\n");
      printf("  from_host channels.\n");
      printf("  (configuration aborted)\n");
      exit(1);
    }

    for (instance = 0;                                  /* search instance */
	 (instance < MAX_TASKS_PER_MODULE) && 
	 (modules[main_implementation].tasks[instance] != main_task); 
	  instance++);
    
    cfg_found_param(main_task, TO_HOST_CHANNEL_ID, &to_p, &to_i);
    cfg_found_param(main_task, FROM_HOST_CHANNEL_ID, &from_p, &from_i);
    
    fprintf(cfg, "connect t%d[%d]",                   /* to server channel */
      main_implementation, instance);
    fprintf(cfg, ".param_%d", to_p);
    if (to_i != -1)
      fprintf(cfg, "[%d]", to_i);
    fprintf(cfg, " to to_server;\n");

    fprintf(cfg, "connect t%d[%d]",                 /* from server channel */
      main_implementation, instance);
    fprintf(cfg, ".param_%d", from_p);
    if (from_i != -1)
      fprintf(cfg, "[%d]", from_i);
    fprintf(cfg, " to from_server;\n\n");
  }

  /*-----------------------------------------------------------------------*/
  {                                                        /* linked units */
    int act_module;
    int act_task;

    fprintf(cfg, "/* linked units */\n");
    for (act_module = 0; act_module < MAX_MODULES; act_module++) 
    {
      if (!modules[act_module].free)
      {
        fprintf(cfg, "use \"%s.lku\" for t%d;\n", 
	  modules[act_module].name, act_module);
      }
    }
  }

  /*-----------------------------------------------------------------------*/
                                                              /* close cfg */
  fclose(cfg);
}
#endif

/*-------------------------------------------------------------------------*/
/* generate_main_stub:                                               LOCAL */
/* Generate a task main stub for the given implementation.                 */
/* The main stub is a C program that is responsible to establish the       */
/* connections with other PACT-C tasks and to call the ppc_main()          */
/* of the task with the correct parameters. For each different task        */
/* implementation corresponds a main stub.                                 */
/*-------------------------------------------------------------------------*/
#ifndef PACTC
void generate_main_stub (
  char *name, 
  char *user_name,
  char *date_str,
  int impl
)
{
  FILE *stub;
  char stub_name [100];
  char ident_str [10];        /* sometimes needed to produce readable code */

  printf("generating PACT-C \"%s\" main stub file for \"%s\"\n", 
    modules[impl].name, name);
  sprintf(stub_name, ARCH_DIR "_%s.c", modules[impl].name);
  stub = fopen(stub_name, "w");

  /*-----------------------------------------------------------------------*/
                                                                 /* header */
  fprintf(stub, "\n/*\n");
  fprintf(stub, " * This file has been generated by the PPC ");
  fprintf(stub, "configurer\n");
  fprintf(stub, " * PPC Version " PPC_VERSION "\n");
  fprintf(stub, " * NEVER EDIT IT !\n");
  fprintf(stub, " * Created on %s by %s\n", date_str, user_name);
  fprintf(stub, " * PPC program name: \"%s\"\n", name);
  fprintf(stub, " * main stub for task \"%s\"\n", modules[impl].name);
  fprintf(stub, " * Target architecture: PACT-C for Transputers\n");
  fprintf(stub, " */\n\n");
  
  /*-----------------------------------------------------------------------*/
                                                           /* header files */
  fprintf(stub, "#include <ppc/config.h>\n");
  fprintf(stub, "#include <ppc/config/confpriv.h>\n");
  fprintf(stub, "#include <ppc/config/confpact.h>\n");
  fprintf(stub, "#include <ppc/csp/chanpact.h>\n");
  fprintf(stub, "\n");

  /*-----------------------------------------------------------------------*/
  {                                          /* task entry point prototype */
    int act_task;
    int act_param;
    
    fprintf(stub, "/* task entry points prototype */\n");
    fprintf(stub, "void ppc_main (\n");
    /* every tasks implemented by this module must have the same         */
    /* parameters. Here we use the first task implemented to             */
    /* determine the parameters.                                         */
    act_task = (int)modules[impl].tasks[0];
    for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
         act_param++)
    {
      char c_param_name [100];
      
      switch(tasks[act_task].parameters[act_param].type)
      {
	case OUTPUT_CHANNEL:
	  strcpy(c_param_name, "channel ");
	  break;
	case INPUT_CHANNEL:
	  strcpy(c_param_name, "channel ");
	  break;
	case INT:
	  strcpy(c_param_name, "int ");
	  break;
	case SHORT:
	  strcpy(c_param_name, "short ");
	  break;
	case UNSIGNED_SHORT:
	  strcpy(c_param_name, "unsigned short ");
	  break;
	case LONG:
	  strcpy(c_param_name, "long ");
	  break;
	case UNSIGNED_LONG:
	  strcpy(c_param_name, "unsigned long ");
	  break;
	case FLOAT:
	  strcpy(c_param_name, "float ");
	  break;
	case DOUBLE:
	  strcpy(c_param_name, "double ");
	  break;
	case CHAR:
	  strcpy(c_param_name, "char");
	  break;
	case OUTPUT_CHANNEL_ARRAY:
	  strcpy(c_param_name, "channel *");
	  break;
	case INPUT_CHANNEL_ARRAY:
	  strcpy(c_param_name, "channel *");
	  break;
	case INT_ARRAY:
	  strcpy(c_param_name, "int *");
	  break;
	case SHORT_ARRAY:
	  strcpy(c_param_name, "short *");
	  break;
	case UNSIGNED_SHORT_ARRAY:
	  strcpy(c_param_name, "unsigned *");
	  break;
	case LONG_ARRAY:
	  strcpy(c_param_name, "long *");
	  break;
	case UNSIGNED_LONG_ARRAY:
	  strcpy(c_param_name, "unsigned long *");
	  break;
	case FLOAT_ARRAY:
	  strcpy(c_param_name, "float *");
	  break;
	case DOUBLE_ARRAY:
	  strcpy(c_param_name, "double *");
	  break;
	case CHAR_ARRAY:
	  strcpy(c_param_name, "char *");
	  break;
	default:
	  fprintf(stderr, "ERROR: unknown task parameter type.\n");
	  exit(1);
      }
      fprintf(stub, "  %sparam_%d", c_param_name, act_param);
      if (act_param < (tasks[act_task].nb_parameters-1))
	fprintf(stub, ",");
      fprintf(stub, "\n");
    }
    fprintf(stub, ");\n\n");
  }

  /*-----------------------------------------------------------------------*/
                /* PACT-C task entry point and local variables declaration */
  fprintf(stub, "/* PACT-C task entry point */\n");
  fprintf(stub, "void main (void)\n");
  fprintf(stub, "{\n");
  fprintf(stub, "  int *my_nb_ptr = (int *) getparm(1); /* my task # */\n");
  fprintf(stub, "  int my_nb = *my_nb_ptr;\n");
  fprintf(stub, "  \n");

  fprintf(stub, "  printf(\"\");\n");
  fprintf(stub, "  chan_init();\n");
  fprintf(stub, "  virtual_machine(FROM_HOST_CHANNEL_ID, ");
  fprintf(stub, "TO_HOST_CHANNEL_ID);\n");
  fprintf(stub, "  \n");

  /*-----------------------------------------------------------------------*/
  {                                              /* call ppc main function */
    int act_param;                                 /* actual ppc parameter */
    int act_task;
    int act_pact_param;                    /* actual PACT-C task parameter */
    
    fprintf(stub, "  { /* call PPC main function */\n");
    
    /* channel arrays cannot be simply given by their address find in      */
    /* the task parameter because these values reference the indice of     */
    /* the channel and not its value. Thus, we have to create some         */
    /* temporary channel array that contains the correct channel           */
    /* identifiers that will be passed to the task function.               */
    
    act_task = (int)modules[impl].tasks[0];
    for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
	  act_param++)
    {
      switch(tasks[act_task].parameters[act_param].type)
      {
	case OUTPUT_CHANNEL_ARRAY:
	case INPUT_CHANNEL_ARRAY:
	  fprintf(stub, "    channel chan_array_%d [%d];\n",
	    act_param, 
	    tasks[act_task].parameters[act_param].param.ptr.nitems);
	  break;
	default:
	  break;                                           /* do nothing */
      }
    }
    fprintf(stub, "\n");
    
    /* Unconnected channels must not be opened by the PACT-C channel     */
    /* module. The open_pactc_channel() and open_pactc_channe_array()    */
    /* take as parameter a flag specifying whether the channel is        */
    /* connected to another task or not. We have here to initialize,     */
    /* for each channel of the task, a variable containing this          */
    /* information.                                                      */
    
    act_task = (int)modules[impl].tasks[0];
    for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
	  act_param++)
    {
      switch(tasks[act_task].parameters[act_param].type)
      {
        case INPUT_CHANNEL:
	case OUTPUT_CHANNEL:
	  fprintf(stub, "    connection_status_type chan_status_%d;\n",
	    act_param);
	  break;
	case OUTPUT_CHANNEL_ARRAY:
	case INPUT_CHANNEL_ARRAY:
	  fprintf(stub, "%s chan_array_status_%d [%d];\n",
	    "    connection_status_type",
	    act_param, 
	    tasks[act_task].parameters[act_param].param.ptr.nitems);
	  break;
	default:
	  break;                            /* not a channel, do nothing */
      }
    }
    
    fprintf(stub,"\n    { /* initialize connection status variables */\n");
    fprintf(stub,"      int c = 0;\n\n");
    act_task = (int)modules[impl].tasks[0];
    for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
	  act_param++)
    {
      switch(tasks[act_task].parameters[act_param].type)
      {
        case INPUT_CHANNEL:
	case OUTPUT_CHANNEL:
	  fprintf(stub, "      chan_status_%d = is_connected(my_nb, %d);\n",
	    act_param, act_param);
	  break;
	case OUTPUT_CHANNEL_ARRAY:
	case INPUT_CHANNEL_ARRAY:
	  fprintf(stub, "      for (c = 0; c < %d; c++)\n",
	    tasks[act_task].parameters[act_param].param.ptr.nitems);
	  fprintf(stub, "        chan_array_status_%d[c] = is_connected_in_array(my_nb, %d, c);\n", 
	    act_param, act_param);
	  break;
	default:
	  break;                            /* not a channel, do nothing */
      }
    }
    fprintf(stub,"    }\n\n");
    
    /* Now we can generate the call of the task entry point function.    */
    /* This call uses the variables declared and eventually initialized  */
    /* above.                                                            */
    
    fprintf(stub, "    ppc_main(\n");
    act_pact_param = 2;        /* the first parameter is used internally */
    for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
	 act_param++, act_pact_param++)
    {
      char c_param_name [100];
      
      switch(tasks[act_task].parameters[act_param].type)
      {
	case OUTPUT_CHANNEL:
	  if (tasks[act_task].parameters[act_param].param.value.channel_val 
	      == TO_HOST_CHANNEL_ID)
	    strcpy(c_param_name, "TO_HOST_CHANNEL_ID");
	  else
	  {
	    sprintf(c_param_name, 
	      "open_pactc_channel((void |*)getparm(%d), (char |*)getparm(%d), \n        CHAN_OUTPUT, chan_status_%d)",
	      act_pact_param, act_pact_param+1, act_param);
	    act_pact_param++;
	  }
	  break;
	case INPUT_CHANNEL:
	  if (tasks[act_task].parameters[act_param].param.value.channel_val 
	      == FROM_HOST_CHANNEL_ID)
	    strcpy(c_param_name, "FROM_HOST_CHANNEL_ID");
	  else
	  {
	    sprintf(c_param_name, 
	      "open_pactc_channel((void |*)getparm(%d), (char |*)getparm(%d), \n        CHAN_INPUT, chan_status_%d)",
	      act_pact_param, act_pact_param+1, act_param);
	    act_pact_param++;
	  }
	  break;
	case INT:
	  sprintf(c_param_name, 
	    "tasks[my_nb].parameters[%d].param.value.int_val",
	    act_param);
	  break;
	case SHORT:
	  sprintf(c_param_name, 
	    "tasks[my_nb].parameters[%d].param.value.short_val",
	    act_param);
	  break;
	case UNSIGNED_SHORT:
	  sprintf(c_param_name, 
	    "tasks[my_nb].parameters[%d].param.value.unsigned_short_val",
	    act_param);
	  break;
	case LONG:
	  sprintf(c_param_name, 
	    "tasks[my_nb].parameters[%d].param.value.long_val",
	    act_param);
	  break;
	case UNSIGNED_LONG:
	  sprintf(c_param_name, 
	    "tasks[my_nb].parameters[%d].param.value.unsigned_long_val",
	    act_param);
	  break;
	case FLOAT:
	  sprintf(c_param_name, 
	    "tasks[my_nb].parameters[%d].param.value.float_val",
	    act_param);
	  break;
	case DOUBLE:
	  sprintf(c_param_name, 
	    "tasks[my_nb].parameters[%d].param.value.double_val",
	    act_param);
	  break;
	case CHAR:
	  sprintf(c_param_name, 
	    "tasks[my_nb].parameters[%d].param.value.char_val",
	    act_param);
	  break;
	case OUTPUT_CHANNEL_ARRAY:
	  sprintf(c_param_name, 
	    "open_pactc_channels_array((void |**)getparm(%d), (char |**)getparm(%d),\n        chan_array_%d, tasks[my_nb].parameters[%d].param.ptr.nitems,\n        CHAN_OUTPUT, chan_array_status_%d)",
	    act_pact_param, act_pact_param+1, act_param, act_param,
	    act_param);
	    act_pact_param++;
	  break;
	case INPUT_CHANNEL_ARRAY:
	  sprintf(c_param_name, 
	    "open_pactc_channels_array((void |**)getparm(%d), (char |**)getparm(%d),\n        chan_array_%d,  tasks[my_nb].parameters[%d].param.ptr.nitems,\n        CHAN_INPUT, chan_array_status_%d)",
	    act_pact_param, act_pact_param+1, act_param, act_param,
	    act_param);
	    act_pact_param++;
	  break;
	case INT_ARRAY:
	case SHORT_ARRAY:
	case UNSIGNED_SHORT_ARRAY:
	case LONG_ARRAY:
	case UNSIGNED_LONG_ARRAY:
	case FLOAT_ARRAY:
	case DOUBLE_ARRAY:
	case CHAR_ARRAY:
	  sprintf(c_param_name, 
	    "tasks[my_nb].parameters[%d].param.ptr.address",
	    act_param);
	  break;
      }
      fprintf(stub, "      %s", c_param_name);
      if (act_param < (tasks[act_task].nb_parameters-1))
	fprintf(stub, ",");
      fprintf(stub, "\n");
    }
    fprintf(stub, "    );\n");
    fprintf(stub, "  }\n\n");
  }


  /*-----------------------------------------------------------------------*/
                                                        /* close main stub */
  fprintf(stub, "  exit(0);\n");
  fprintf(stub, "}\n");
  fclose(stub);
}
#endif

/*-------------------------------------------------------------------------*/
/* generate_files:                                                         */
/* Generate the makefile for the PPC program in order to compile it        */
/* for the target machine. One version of this function must be            */
/* implemented for each target hardware.                                   */
/*-------------------------------------------------------------------------*/
void generate_files (char *name)
{
  /* This module must be linked with the PPC program because this          */
  /* program calls the virtual machine configuration function which calls  */
  /* the hardware configuration function which calls the functions of this */
  /* module. However, the configuration files are not generated by the     */
  /* parallel program but only by the configuration program which is       */
  /* compiled on the host computer.                                        */
  
  #ifndef PACTC
  char date_str [100];
  time_t act_time;
  char user_name [100];
  struct passwd *act_passwd;
  int act_module;

  time(&act_time);                           /* get string containing date */
  strcpy(date_str, ctime(&act_time));
  date_str[strlen(date_str)-1] = '\0';
  
  act_passwd = getpwuid(getuid());                /* get user's login name */
  strcpy(user_name, act_passwd->pw_name);

  generate_makefile(name, user_name, date_str);       /* generate makefile */
  generate_config(name, user_name, date_str);      /* generate PACT-C .cfg */
  for (act_module = 0; act_module < MAX_MODULES; act_module++)
    if (!modules[act_module].free)                 /* generates main stubs */
      generate_main_stub(name, user_name, date_str, act_module);

  #else
  char *dumb = name;
  #endif
}

/*-------------------------------------------------------------------------*/
/* print_hw_report:                                                        */
/* Print a hardware configuration report. Each hardware configuration      */
/* must implement a such function. It is called by the configuration main  */
/* function (confmain.c).                                                  */
/*-------------------------------------------------------------------------*/
void print_hw_report (void)
{
  int act_trans;
  int act_proc;
  int act_conn;
  int act_link;
  int conn_idx;

  printf("\nhardware configuration for PACT-C:\n");
  printf("   # type   memory [ Link0 Link1 Link2 Link3 ] procs\n");

  for (act_trans = 0; act_trans < nb_trans; act_trans++)
  {
    printf("  %2d ", act_trans);
    printf("%s ", trans_types_str[trans[act_trans].type]);
    printf("%8lu", trans[act_trans].memory);
    printf(" [");
    for (act_link = 0; act_link < 4; act_link++)
    {
      conn_idx = -1;
      for (act_conn = 0; act_conn < nb_conn && conn_idx == -1; act_conn++)
      {
        if ((conn[act_conn].from_tran == act_trans && 
	     conn[act_conn].from_link == act_link) ||
	    (conn[act_conn].to_tran == act_trans &&
	     conn[act_conn].to_link == act_link))
	  conn_idx = act_conn;
      }
      if (conn_idx == -1)
        printf("   ...");
      else
      {
        if (conn[conn_idx].from_tran == act_trans)
	  if (conn[conn_idx].to_tran == HOST)
	    printf("  HOST");
	  else
            printf("  %2d:%d",
              conn[conn_idx].to_tran, conn[conn_idx].to_link);
        else
	  if (conn[conn_idx].from_tran == HOST)
	    printf(" HOST");
	  else
            printf("  %2d:%d",
             conn[conn_idx].from_tran, conn[conn_idx].from_link);
      }
    }
    printf(" ] ");

    for (act_proc = 0; act_proc < trans[act_trans].nb_procs; act_proc++)
      printf("%d ", trans[act_trans].procs[act_proc]);
    printf("\n");    
  }
}

/*-------------------------------------------------------------------------*/
/* is_connected:                                                           */
/* Determine if a channel is connected or not. The channel is specified    */
/* as the parameter of a task.                                             */
/* This function is called only by a PACT-C main stub before starting      */
/* the PPC task entry point.                                               */
/*-------------------------------------------------------------------------*/
connection_status_type is_connected (
  int task_id,                                          /* task identifier */
  int param_id                                        /* task parameter id */
)
{
  int c;
  
  c = tasks[task_id].parameters[param_id].param.value.channel_val;
  
  if (c == FROM_HOST_CHANNEL_ID || c == TO_HOST_CHANNEL_ID)
    return NOT_CONNECTED;
  
  if (channels[c].from == -1 || channels[c].to == -1)
    return NOT_CONNECTED;
  
  return CONNECTED;
}

/*-------------------------------------------------------------------------*/
/* is_connected_in_array:                                                  */
/* Determine if a channel is connected or not. The channel is specified    */
/* as the element of a parameter array of a specified task.                */
/* This function is called only by a PACT-C main stub before starting      */
/* the PPC task entry point.                                               */
/*-------------------------------------------------------------------------*/
connection_status_type is_connected_in_array (
  int task_id,                                          /* task identifier */
  int param_id,                                 /* task array parameter id */
  int array_idx             /* index of the element in the array parameter */
)
{
  int c;
  
  c = ((channel*)
       (tasks[task_id].parameters[param_id].param.ptr.address))[array_idx];
  
  if (c == FROM_HOST_CHANNEL_ID || c == TO_HOST_CHANNEL_ID)
    return NOT_CONNECTED;

  if (channels[c].from == -1 || channels[c].to == -1)
    return NOT_CONNECTED;
  
  return CONNECTED;
}
