/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* PVM Hardware configuration implementation.                              */
/* The functions defined in this file must be called from the function     */
/* hardware_config() implemented by the user.                              */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 07.01.1994  Droux       Created                                         */
/* 28.06.1994  Droux       UNIX programs are always linked with server lib */
/* 01.07.1994  Droux       Added print_hw_report()                         */
/* 08.07.1994  Droux       New structure of source directory               */
/* 17.08.1994  Droux       Support for custom hardware configuration file  */
/* 28.08.1994  Droux       Added "." prefix to architecture directory      */
/*-------------------------------------------------------------------------*/
/* filename: confpvm.c                                                     */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define confpvm_serv

#if defined(NEXT) || defined(NEXTINTEL)
#include <libc.h>
#endif
#include <syslog.h>
#include <time.h>
#include <stdio.h>
#include <pwd.h>
#include <pvm3.h>
#include <ppc/csp/version.h>
#include <ppc/config/confpvm.h>
#include <ppc/config/confpriv.h>
#include "confhard.h"

#define ARCH_DIR      ".pvm/"
#define MAKEFILE_NAME ARCH_DIR "makefile"

/* It is possible that more than one task must be linked with the same     */
/* module. In this case, it is important to avoid the duplication of this  */
/* module makefile dependency. To avoid this duplication, we record a      */
/* unique key for each encountered module.                                 */
static unsigned long modules_keys [MAX_MODULES*MAX_OBJECTS_PER_MODULE];
int nb_keys = 0;
                                   /* for each virtual processor, its host */
char hosts [MAX_PROCESSORS][MAX_HOST_NAME_LEN];
int max_host_idx = 0;                  /* last host defined in hosts array */
int confpvm_init = 0;       /* 0 if module not initialized, != 0 otherwise */

/*-------------------------------------------------------------------------*/
/* generate_makefile:                                                LOCAL */
/* Generate the makefile for the current program.                          */
/*-------------------------------------------------------------------------*/
void generate_makefile (
  char *name, 
  char *user_name,
  char *date_str,
  int main_implementation                   /* main task implementation id */
)
{
  FILE *makefile;
  
  /*-----------------------------------------------------------------------*/
                                                                 /* header */
  printf("generating PVM makefile for \"%s\"\n", name);
  makefile = fopen(MAKEFILE_NAME, "w");
  fprintf(makefile, "\n");
  fprintf(makefile, " # This file has been generated by the PPC ");
  fprintf(makefile, "configurer\n");
  fprintf(makefile, " # PPC Version " PPC_VERSION "\n");
  fprintf(makefile, " # NEVER EDIT IT !\n");
  fprintf(makefile, " # Created on %s by %s\n", date_str, user_name);
  fprintf(makefile, " # PPC program name: \"%s\"\n", name);
  fprintf(makefile, " # Target architecture: PVM\n\n");

  /*-----------------------------------------------------------------------*/
                                              /* compiler and linker flags */
  fprintf(makefile, "# compiler and linker flags\n");
  fprintf(makefile, "CFLAGS=-I$(PPC_HOME)/include ");
  fprintf(makefile, "-I$(PVM_ROOT)/include -D_$(ARCH) \\\n");
  fprintf(makefile, "\t-D$(HOST_ARCH) -O $(OTHER_CFLAGS)\n");
  fprintf(makefile, "LFLAGS=-lconf -lconf$(ARCHLIB) -lcsp$(ARCHLIB) \\\n");
  fprintf(makefile, "\t-lm -L$(PPC_HOME)/lib/$(HOST_ARCH) \\\n");
  fprintf(makefile, "\t-L$(PVM_ROOT)/lib/$(HOST_ARCH) -lpvm3 \\\n");
  fprintf(makefile, "\t-lserv$(ARCHLIB) $(PVMARCHLIB)\n");
  fprintf(makefile, "BDIR=$(PVM_ROOT)/bin\n");
  fprintf(makefile, "XDIR=$(BDIR)/$(PVM_ARCH)\n\n");

  /*-----------------------------------------------------------------------*/
  {                       /* list of object modules of each PVM executable */
    int act_exec;
    int act_obj;
    
    fprintf(makefile, "# list of object modules of ");
    fprintf(makefile, "each PVM executable\n");
  
    for (act_exec = 0; act_exec < MAX_MODULES; act_exec++)
      if (!modules[act_exec].free)
      {
	fprintf(makefile, "%sOBJ=", modules[act_exec].name);
	fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/%s.o ", name);
	fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/$(ARCH_FILE).o ");
	fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/_%s.o ",
	  modules[act_exec].name);
	for (act_obj = 0; act_obj < modules[act_exec].nb_names; act_obj++)
	  fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/%s.o ",
	    modules[act_exec].modules_names[act_obj]);
	fprintf(makefile, "\n");
      }
    fprintf(makefile, "\n");
  }

  /*-----------------------------------------------------------------------*/
  {                         /* list of all PVM executable, main dependency */
    int n;
    int act_obj;
    
    fprintf(makefile, "# list of all PVM executable, main dependency\n");
    fprintf(makefile, "all: $(XDIR) ");
  
    for (n = 0; n < MAX_MODULES; n++)                  
      if (!modules[n].free)
      {
	fprintf(makefile, "$(XDIR)/%s_%s ",
	  name, 
	  modules[n].name);
      }
    fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/%s.run\n\n", name);
  }
  
  /*-----------------------------------------------------------------------*/
  {                                            /* executables dependencies */
    int act_exec;

    fprintf(makefile, "# executables dependencies\n");
    
    for (act_exec = 0; act_exec < MAX_MODULES; act_exec++)
      if (!modules[act_exec].free)
      {
	fprintf(makefile, "$(XDIR)/%s_%s: $(%sOBJ)\n",
	  name, modules[act_exec].name, modules[act_exec].name);
	fprintf(makefile, "\t$(CC) $(%sOBJ) -o $@ $(LFLAGS)\n\n",
	  modules[act_exec].name);
      }
    
    fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/%s.run: $(XDIR)/%s_%s\n",
      name, name, modules[main_implementation].name);
    fprintf(makefile, "\tcp $? $@\n\n");
  }

  /*-----------------------------------------------------------------------*/
  {                                           /* object files dependencies */
    int act_exec;
    int act_obj;
    int duplicated;                         /* 1 if a module is duplicated */
    int key_idx;                                    /* to browse keys list */
    unsigned long act_key;                          /* actual module's key */
    
    fprintf(makefile, "# object files dependencies\n");
  
    nb_keys = 0;
    for (act_exec = 0; act_exec < MAX_MODULES; act_exec++)
      if (!modules[act_exec].free)
      {
        fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/_%s.o: .$(ARCH)/_%s.c\n",
	  modules[act_exec].name, modules[act_exec].name);
        fprintf(makefile, "\t$(CC) $(CFLAGS) -c $? -o $@\n\n");
	
	for (act_obj = 0; act_obj < modules[act_exec].nb_names; act_obj++)
	{
	  act_key = confkey(modules[act_exec].modules_names[act_obj]);
	  duplicated = 0;
	  for (key_idx = 0; key_idx < nb_keys; key_idx++)
	    duplicated = duplicated || (modules_keys[key_idx] == act_key);
	  if (!duplicated)
	  {
	    fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/%s.o: %s.c\n",
	      modules[act_exec].modules_names[act_obj],
	      modules[act_exec].modules_names[act_obj]);
	    fprintf(makefile, "\t$(CC) $(CFLAGS) -c $? -o $@\n\n");
	    modules_keys[nb_keys] = act_key;
	    nb_keys++;
	  }
	}
      }
                                               /* main program object file */
    fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/_%s.o : .$(ARCH)/_%s.c\n", 
      name, name);
    fprintf(makefile, "\t$(CC) $(CFLAGS) -c $? -o $@\n\n");
  }

  /*-----------------------------------------------------------------------*/
                         /* automatic creation of PVM executable directory */
  fprintf(makefile, "# automatic creation of PVM ");
  fprintf(makefile, "executable directory\n");
  fprintf(makefile, "$(XDIR):\n");
  fprintf(makefile, "\t- mkdir $(BDIR)\n");
  fprintf(makefile, "\t- mkdir $(XDIR)\n");
  fclose(makefile);
}

/*-------------------------------------------------------------------------*/
/* check_main_task:                                                        */
/* Checks if one of the tasks implemented by the given module is the main  */
/* task. The main task is the task which communicates with the user and    */
/* which starts the other tasks.                                           */
/* Returns the task number if it is the case, -1 if it isn't.              */
/* This function is used during the files generation as well as during the */
/* execution of the tasks to determine its number.                         */
/*-------------------------------------------------------------------------*/
int check_main_task (int impl)
{
  int act_task_idx;
  int act_chan;
  int act_task;
  int from_host_ok;
  int to_host_ok;
  int nb_param;
  int act_param_idx;
  int act_param_type;
  
  /* For each task which is implemented by the current implementation,     */
  /* checks if this task takes the two channels from_host and to_host      */
  /* as parameters. If it is the case, then the main task is implemented   */
  /* by this module.                                                       */
  
  for (act_task_idx = 0; act_task_idx < modules[impl].nb_tasks; 
       act_task_idx++)
  {
    act_task = modules[impl].tasks[act_task_idx];
    from_host_ok = 0;
    to_host_ok = 0;
    nb_param = tasks[act_task].nb_parameters;
    for (act_param_idx = 0; act_param_idx < nb_param; act_param_idx++)
    {
      act_param_type = tasks[act_task].parameters[act_param_idx].type;
      if ((act_param_type == INPUT_CHANNEL) || 
          (act_param_type == OUTPUT_CHANNEL))
      {
        act_chan =
	  tasks[act_task].parameters[act_param_idx].param.value.channel_val;
        from_host_ok = from_host_ok || (act_chan == FROM_HOST_CHANNEL_ID);
        to_host_ok = to_host_ok || (act_chan == TO_HOST_CHANNEL_ID);
      }
    }
    if (from_host_ok && to_host_ok)
      return act_task;
    if ((from_host_ok && !to_host_ok) || (!from_host_ok && to_host_ok))
    {
      printf("CONFIGURATION ERROR :\n");
      printf("  The from_host and to_host channels must be assigned");
      printf(" to the same task.\n");
      printf("  (configuration aborted)\n");
      exit(1);
    }
  }
  return -1;
}

/*-------------------------------------------------------------------------*/
/* generate_main_stub:                                               LOCAL */
/* Generate a task main stub for the given implementation.                 */
/* The main stub is a C program that is responsible to establish the       */
/* connections with other PVM tasks and to call the ppc_main() function    */
/* of the task with the correct parameters. For each different task        */
/* implementation corresponds a main stub.                                 */
/*-------------------------------------------------------------------------*/
void generate_main_stub (
  char *name, 
  char *user_name,
  char *date_str,
  int impl,
  int main_task            /* != 0 if this module implements the main task */
)
{
  FILE *stub;
  char stub_name [100];
  char ident_str [10];        /* sometimes needed to produce readable code */

  printf("generating PVM \"%s\" main stub file for \"%s\"\n", 
    modules[impl].name, name);
  sprintf(stub_name, ARCH_DIR "_%s.c", modules[impl].name);
  stub = fopen(stub_name, "w");

  /*-----------------------------------------------------------------------*/
                                                                 /* header */
  fprintf(stub, "\n/*\n");
  fprintf(stub, " * This file has been generated by the PPC ");
  fprintf(stub, "configurer\n");
  fprintf(stub, " * PPC Version " PPC_VERSION "\n");
  fprintf(stub, " * NEVER EDIT IT !\n");
  fprintf(stub, " * Created on %s by %s\n", date_str, user_name);
  fprintf(stub, " * PPC program name: \"%s\"\n", name);
  fprintf(stub, " * main stub for task \"%s\"\n", modules[impl].name);
  fprintf(stub, " * Target architecture: PVM\n");
  fprintf(stub, " */\n\n");
  
  /*-----------------------------------------------------------------------*/
                                                           /* header files */
  fprintf(stub, "#if defined(NEXT) || defined(NEXTINTEL)\n");
  fprintf(stub, "#include <libc.h>\n");
  fprintf(stub, "#endif\n\n");
  
  fprintf(stub, "#include <fcntl.h>\n");
  fprintf(stub, "#include <stdio.h>\n");
  fprintf(stub, "#include <sys/wait.h>\n");
  fprintf(stub, "#include <syslog.h>\n");
  fprintf(stub, "#include <pvm3.h>\n\n");
  
  fprintf(stub, "#include <ppc/config.h>\n");
  fprintf(stub, "#include <ppc/config/confpriv.h>\n");
  fprintf(stub, "#include <ppc/config/confpvm.h>\n\n");

  /*-----------------------------------------------------------------------*/
  {                                          /* task entry point prototype */
    int act_task;
    int act_param;
    
    fprintf(stub, "/* task entry points prototype */\n");
    fprintf(stub, "void ppc_main (\n");
    /* every tasks implemented by this module must have the same         */
    /* parameters. Here we use the first task implemented to             */
    /* determine the parameters.                                         */
    act_task = (int)modules[impl].tasks[0];
    for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
         act_param++)
    {
      char c_param_name [100];
      
      switch(tasks[act_task].parameters[act_param].type)
      {
	case OUTPUT_CHANNEL:
	  strcpy(c_param_name, "channel ");
	  break;
	case INPUT_CHANNEL:
	  strcpy(c_param_name, "channel ");
	  break;
	case INT:
	  strcpy(c_param_name, "int ");
	  break;
	case SHORT:
	  strcpy(c_param_name, "short ");
	  break;
	case UNSIGNED_SHORT:
	  strcpy(c_param_name, "unsigned short ");
	  break;
	case LONG:
	  strcpy(c_param_name, "long ");
	  break;
	case UNSIGNED_LONG:
	  strcpy(c_param_name, "unsigned long ");
	  break;
	case FLOAT:
	  strcpy(c_param_name, "float ");
	  break;
	case DOUBLE:
	  strcpy(c_param_name, "double ");
	  break;
	case CHAR:
	  strcpy(c_param_name, "char");
	  break;
	case OUTPUT_CHANNEL_ARRAY:
	  strcpy(c_param_name, "channel *");
	  break;
	case INPUT_CHANNEL_ARRAY:
	  strcpy(c_param_name, "channel *");
	  break;
	case INT_ARRAY:
	  strcpy(c_param_name, "int *");
	  break;
	case SHORT_ARRAY:
	  strcpy(c_param_name, "short *");
	  break;
	case UNSIGNED_SHORT_ARRAY:
	  strcpy(c_param_name, "unsigned *");
	  break;
	case LONG_ARRAY:
	  strcpy(c_param_name, "long *");
	  break;
	case UNSIGNED_LONG_ARRAY:
	  strcpy(c_param_name, "unsigned long *");
	  break;
	case FLOAT_ARRAY:
	  strcpy(c_param_name, "float *");
	  break;
	case DOUBLE_ARRAY:
	  strcpy(c_param_name, "double *");
	  break;
	case CHAR_ARRAY:
	  strcpy(c_param_name, "char *");
	  break;
	default:
	  fprintf(stderr, "ERROR: unknown task parameter type.\n");
	  exit(1);
      }
      fprintf(stub, "  %sparam_%d", c_param_name, act_param);
      if (act_param < (tasks[act_task].nb_parameters-1))
	fprintf(stub, ",");
      fprintf(stub, "\n");
    }
    fprintf(stub, ");\n\n");
  }

  /*-----------------------------------------------------------------------*/
                                  /* entry point and variables declaration */
  fprintf(stub, "/* PPC program entry point */\n");
  fprintf(stub, "void main (int argc, char *argv[])\n");
  fprintf(stub, "{\n");
  fprintf(stub, "  int my_nb; /* not PVM tid */\n");
  fprintf(stub, "  int main_tid;\n");
  fprintf(stub, "  int my_tid = pvm_mytid();\n");
  fprintf(stub, "  int the_channels [MAX_CHANNELS];\n");
  fprintf(stub, "  int tids [MAX_TASKS];\n\n");

  fprintf(stub, "  chan_init();\n");
  fprintf(stub, "  virtual_machine(FROM_HOST_CHANNEL_ID, ");
  fprintf(stub, "TO_HOST_CHANNEL_ID);\n\n");

  /*-----------------------------------------------------------------------*/
                                              /* main task additional test */
  /* The main task is responsible to start the other tasks. More than      */
  /* one task can belong to the same implementation as the main task.      */
  /* The main task is started as the program .run executable by the PPC    */
  /* script. We differenciate the main task and the normal tasks of the    */
  /* same implementation by the number of arguments they take when they    */
  /* are started. The main task takes none, and a normal task takes its    */
  /* task number.                                                          */
  /* We generate here the code that performs this test.                    */
  if (main_task)
  {
    fprintf(stub, "  /* test if main task of the parallel program */\n");
    fprintf(stub, "  if (argc == 1)\n");
    fprintf(stub, "  {\n");
    fprintf(stub, "    my_nb = check_main_task (%d);\n", impl);
    fprintf(stub, "    start_pvm_tasks(my_nb, \"%s\", tids, my_tid);\n",
      name);
    fprintf(stub, "    main_tid = my_nb;\n");
    fprintf(stub, "  }\n");
    fprintf(stub, "  else\n");
    fprintf(stub, "  {\n");
    strcpy(ident_str, "  ");
  }
  else
    strcpy(ident_str, "");

  /*-----------------------------------------------------------------------*/
                                                   /* get other tasks tids */
  fprintf(stub, "%s  /* get other tasks tids */\n", ident_str);
  fprintf(stub, "%s  pvm_recv(-1, -1);\n", ident_str);
  fprintf(stub, "%s  pvm_upkint(tids, MAX_TASKS, 1);\n", ident_str);
  fprintf(stub, "%s  sscanf(argv[1], \"%%d\", &my_nb);\n", ident_str);
  fprintf(stub, "%s  main_tid = pvm_parent();\n", ident_str);
  if (main_task)
    fprintf(stub, "  }\n");
  fprintf(stub, "\n");
  
  /*-----------------------------------------------------------------------*/
                                               /* open task's PVM channels */
  fprintf(stub, "  { /* open task's PVM channels */\n");
  fprintf(stub, "    int chan_idx;\n");
  fprintf(stub, "    int other_tid;\n");
  fprintf(stub, "    int n;\n\n");
  fprintf(stub, "    chan_idx = tasks[my_nb].channels[0];\n");
  fprintf(stub, "    for (n = 0; chan_idx != -1; n++)\n");
  fprintf(stub, "    {\n");
  fprintf(stub, "      if (channels[chan_idx].from == my_nb)\n");
  fprintf(stub, "        other_tid = tids[channels[chan_idx].to];\n");
  fprintf(stub, "      else\n");
  fprintf(stub, "        other_tid = tids[channels[chan_idx].from];\n");
  fprintf(stub, "      the_channels[chan_idx] = ");
  fprintf(stub, "new_pvm_channel(other_tid, chan_idx);\n");
  fprintf(stub, "      chan_idx = tasks[my_nb].channels[n+1];\n");
  fprintf(stub, "    }\n");
  fprintf(stub, "  }\n\n");

  /*-----------------------------------------------------------------------*/
  {                                              /* call PPC main function */
    int act_param;
    int act_task;
    
    fprintf(stub, "  { /* call PPC main function */\n");
    
    /* channel arrays cannot be simply given by their address find in      */
    /* the task parameter because these values reference the indice of     */
    /* the channel and not its value. Thus, we have to create some         */
    /* temporary channel array that contains the correct channel           */
    /* identifiers that will be passed to the task function.               */
    
				             /* declare the channel arrays */
    act_task = (int)modules[impl].tasks[0];
    for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
	  act_param++)
    {
      switch(tasks[act_task].parameters[act_param].type)
      {
	case OUTPUT_CHANNEL_ARRAY:
	case INPUT_CHANNEL_ARRAY:
	  fprintf(stub, "    channel chan_array_%d [%d];\n",
	    act_param, 
	    tasks[act_task].parameters[act_param].param.ptr.nitems);
	  break;
	default:
	  break;                                           /* do nothing */
      }
    }
					/* initialize the channel arrays */
    for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
	  act_param++)
    {
      switch(tasks[act_task].parameters[act_param].type)
      {
	case OUTPUT_CHANNEL_ARRAY:
	case INPUT_CHANNEL_ARRAY:
	  fprintf(stub, "    {\n");
	  fprintf(stub, "      int i;\n");
	  fprintf(stub, "      channel *act_ptr = tasks[my_nb].");
	  fprintf(stub, "parameters[%d].param.ptr.address;\n",
	    act_param);
	  fprintf(stub, "      for (i=0; i < %d; i++, act_ptr++)\n",
	    tasks[act_task].parameters[act_param].param.ptr.nitems);
	  fprintf(stub, "        chan_array_%d[i] = ", act_param);
	  fprintf(stub, "the_channels[*act_ptr];\n");
	  fprintf(stub, "    }\n\n");
	  break;
	default:
	  break;                                           /* do nothing */
      }
    }
    
    fprintf(stub, "    ppc_main(\n");
      for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
           act_param++)
      {
        char c_param_name [100];
	
        switch(tasks[act_task].parameters[act_param].type)
	{
	  case OUTPUT_CHANNEL:
	    if (tasks[act_task].parameters[act_param].param.value.channel_val ==
	        TO_HOST_CHANNEL_ID)
	      strcpy(c_param_name, "TO_HOST_CHANNEL_ID");
	    else
	      sprintf(c_param_name, 
	        "the_channels[tasks[my_nb].parameters[%d].param.value.channel_val]", 
		act_param);
	    break;
	  case INPUT_CHANNEL:
	    if (tasks[act_task].parameters[act_param].param.value.channel_val ==
	        FROM_HOST_CHANNEL_ID)
	      strcpy(c_param_name, "FROM_HOST_CHANNEL_ID");
	    else
	      sprintf(c_param_name, 
	        "the_channels[tasks[my_nb].parameters[%d].param.value.channel_val]", 
		act_param);
	    break;
	  case INT:
	    sprintf(c_param_name, 
	      "tasks[my_nb].parameters[%d].param.value.int_val",
	      act_param);
	    break;
	  case SHORT:
	    sprintf(c_param_name, 
	      "tasks[my_nb].parameters[%d].param.value.short_val",
	      act_param);
	    break;
	  case UNSIGNED_SHORT:
	    sprintf(c_param_name, 
	      "tasks[my_nb].parameters[%d].param.value.unsigned_short_val",
	      act_param);
	    break;
	  case LONG:
	    sprintf(c_param_name, 
	      "tasks[my_nb].parameters[%d].param.value.long_val",
	      act_param);
	    break;
	  case UNSIGNED_LONG:
	    sprintf(c_param_name, 
	      "tasks[my_nb].parameters[%d].param.value.unsigned_long_val",
	      act_param);
	    break;
	  case FLOAT:
	    sprintf(c_param_name, 
	      "tasks[my_nb].parameters[%d].param.value.float_val",
	      act_param);
	    break;
	  case DOUBLE:
	    sprintf(c_param_name, 
	      "tasks[my_nb].parameters[%d].param.value.double_val",
	      act_param);
	    break;
	  case CHAR:
	    sprintf(c_param_name, 
	      "tasks[my_nb].parameters[%d].param.value.char_val",
	      act_param);
	    break;
	  case OUTPUT_CHANNEL_ARRAY:
	    sprintf(c_param_name, 
	      "chan_array_%d",
	      act_param);
	    break;
	  case INPUT_CHANNEL_ARRAY:
	    sprintf(c_param_name, 
	      "chan_array_%d",
	      act_param);
	    break;
          case INT_ARRAY:
          case SHORT_ARRAY:
          case UNSIGNED_SHORT_ARRAY:
          case LONG_ARRAY:
          case UNSIGNED_LONG_ARRAY:
          case FLOAT_ARRAY:
          case DOUBLE_ARRAY:
          case CHAR_ARRAY:
	    sprintf(c_param_name, 
	      "tasks[my_nb].parameters[%d].param.ptr.address",
	      act_param);
	    break;
	}
        fprintf(stub, "      %s", c_param_name);
	if (act_param < (tasks[act_task].nb_parameters-1))
	  fprintf(stub, ",");
	fprintf(stub, "\n");
      }
      fprintf(stub, "    );\n");
    fprintf(stub, "  }\n\n");
  }

  /*-----------------------------------------------------------------------*/
                                                      /* test if main task */
  if (main_task)
  {
    fprintf(stub, "  /* test if main task of the parallel program */\n");
    fprintf(stub, "  if (argc == 1)\n");
    fprintf(stub, "  { /* wait for all tasks termination */\n");
    fprintf(stub, "    int to_terminate = 0;\n");
    fprintf(stub, "    int n;\n");
    fprintf(stub, "    int nb_tasks = 0;\n");
    fprintf(stub, "    int tmp;\n\n");
    fprintf(stub, "    for (n = 0; n < MAX_TASKS; n++)\n");
    fprintf(stub, "      if (!tasks[n].free && n != my_nb)\n");
    fprintf(stub, "        nb_tasks++;\n\n");
    fprintf(stub, "    for(to_terminate = 0; to_terminate < nb_tasks;");
    fprintf(stub, " to_terminate++)\n");
    fprintf(stub, "    {\n");
    fprintf(stub, "      pvm_recv(-1, -1);\n");
    fprintf(stub, "      pvm_upkint(&tmp, 1, 1);\n");
    fprintf(stub, "    }\n\n");
    fprintf(stub, "    for (n = 0; n < MAX_TASKS; n++)\n");
    fprintf(stub, "      if (!tasks[n].free && n != my_nb)\n");
    fprintf(stub, "      {\n");
    fprintf(stub, "        pvm_initsend(PvmDataDefault);\n");
    fprintf(stub, "        pvm_pkint(&tmp, 1, 1);\n");
    fprintf(stub, "        pvm_send(tids[n], 1);\n");
    fprintf(stub, "      }\n");
    fprintf(stub, "    pvm_exit();\n");
    fprintf(stub, "  }\n");
    fprintf(stub, "  else\n");
  }

  /*-----------------------------------------------------------------------*/
                                /* send terminate to main and wait for ack */
  fprintf(stub, "  { /* send terminate to main and wait for ack */\n");
  fprintf(stub, "    int tmp = my_nb;\n\n");
  fprintf(stub, "    pvm_initsend(PvmDataDefault);\n");
  fprintf(stub, "    pvm_pkint(&tmp, 1, 1);\n");
  fprintf(stub, "    pvm_send(main_tid, 1);\n\n");
  fprintf(stub, "    pvm_recv(main_tid, -1);\n");
  fprintf(stub, "    pvm_upkint(&tmp, 1, 1);\n");
  fprintf(stub, "  }\n\n");

  /*-----------------------------------------------------------------------*/
                                                        /* end of function */
  fprintf(stub, "  /* dispose memory used by the configuration */\n");
  fprintf(stub, "  free_config_data();\n");
  fprintf(stub, "  exit(0);\n");
  fprintf(stub, "}\n");

  fclose(stub);
}

/*-------------------------------------------------------------------------*/
/* generate_files:                                                         */
/* Generate the makefile for the PPC program in order to compile it        */
/* for the target machine. One version of this function must be            */
/* implemented for each target hardware.                                   */
/*-------------------------------------------------------------------------*/
void generate_files (char *name)
{
  char date_str [100];
  time_t act_time;
  char user_name [100];
  struct passwd *act_passwd;
  int main_implementation;/* implementation which implements the main task */

  time(&act_time);                           /* get string containing date */
  strcpy(date_str, ctime(&act_time));
  date_str[strlen(date_str)-1] = '\0';
  
  act_passwd = getpwuid(getuid());                /* get user's login name */
  strcpy(user_name, act_passwd->pw_name);

  {                                                /* generates main stubs */
    int n;
    int main_task;        /* true if actual module implement the main task */

    main_implementation = -1;
    for (n = 0; n < MAX_MODULES; n++)
      if (!modules[n].free)
      {
        main_task = check_main_task(n);
	if (main_task != -1)
	  main_implementation = n;
        generate_main_stub(name, user_name, date_str, n, main_task != -1);
      }
  }

  if (main_implementation == -1)
  {
    printf("CONFIGURATION ERROR :\n");
    printf("  No task are connected to the host by the to_host and\n");
    printf("  from_host channels.\n");
    printf("  (configuration aborted)\n");
    exit(1);
  }
  generate_makefile(name, user_name, date_str, main_implementation);
}

/*-------------------------------------------------------------------------*/
/* place_processor_on_host:                                                */
/* Place a virtual processor on a physical host given by its internet name */
/*-------------------------------------------------------------------------*/
void place_processor_on_host(processor proc, const char *host_name)
{
  if (!confpvm_init)
  {
    int act_proc;

    for (act_proc = 0; act_proc < MAX_PROCESSORS; act_proc++)
      hosts[act_proc][0] = '\0';
    confpvm_init = 1;
  }

  if (host_name == NULL)
    hosts[proc][0] = '\0';
  else
    strcpy(hosts[proc], host_name);

  if (proc > max_host_idx)
    max_host_idx = proc;
}

/*-------------------------------------------------------------------------*/
/* start_pvm_tasks:                                                        */
/* Called by the main PVM task which is responsible to start all other     */
/* tasks and to give them all information about the parallel program such  */
/* as the tasks's pids. Only one task of the parallel program must call    */
/* this function. This task is the one which takes the to_host and         */
/* from_host channels as parameters because it must be able to             */
/* communicate with the user by stdout and stdin.                          */
/* This functions uses the virtual machine configuration informations and  */
/* thus must be called after the function virtual_machine().               */
/* It returns the tid array containing the task ids of the started tasks.  */
/* my_task_nb and name must be set to the correct value when calling.      */
/* name corresponds to the name of the PPC program.                        */
/*-------------------------------------------------------------------------*/
void start_pvm_tasks (int my_task_nb, char *name, int tids [], int my_tid)
{
  int n;                                               /* counter variable */

  {                                                         /* start tasks */
    char module_name [200];
    int n;
    int spawn_stat;
    char *argv[] = {NULL, NULL};                  /* argument for the task */
    char n_str [10];                             /* string equivalent of n */
    char my_host_name [50];   /* name of host on which the prog is started */
    int where_flag;
    char act_host_name [50];              /* actual host to start the task */

    argv[0] = n_str;
    gethostname(my_host_name, 50);

    for (n = 0; n < MAX_TASKS; n++)
      if (!tasks[n].free && (n != my_task_nb))
      {
        sprintf(module_name, "%s_%s", 
	  name, modules[tasks[n].implemented].name);
	sprintf(n_str, "%d", n);
	
	if (tasks[n].on_processor == tasks[my_task_nb].on_processor)
	{
	  /* the current task being started is configured to run on the    */
	  /* same virtual processor as the one on which the main task      */
	  /* is running. This task must be executed on the current host    */
	  /* and the host name specified in the PVM hardware configuration */
	  /* is ignored.                                                   */
	  if (strcmp(hosts[tasks[n].on_processor], my_host_name))
	  {
	    printf("WARNING: \n");
	    printf("  Task %d is not running on the hardware \n", n);
	    printf("  configuration specified host but on the \n");
	    printf("  host from which the PPC program has been \n");
	    printf("  started\n");
	  }
	  strcpy(act_host_name, my_host_name);
	  where_flag = PvmTaskHost;
	}
	else if (!strlen(hosts[tasks[n].on_processor]))
	{                               /* automatic host selection by PVM */
	  act_host_name[0] = '\0';
	  where_flag = PvmTaskDefault;
	}
	else
	{                   /* start task on the programmer specified host */
	  strcpy(act_host_name, hosts[tasks[n].on_processor]);
	  where_flag = PvmTaskHost;
	}
	
        if ((spawn_stat = pvm_spawn(module_name, 
            argv, where_flag, act_host_name, 1, &tids[n]))
            < 0)
        {
          printf("error %d starting task on host %s.", spawn_stat, 
            act_host_name);
          pvm_exit();
        }
      }
  }
                                            
  tids[my_task_nb] = my_tid;

  for (n = 0; n < MAX_TASKS; n++)            /* send the tids to all tasks */
    if (!tasks[n].free && (n != my_task_nb))
    {
      pvm_initsend(PvmDataDefault);
      pvm_pkint(tids, MAX_TASKS, 1);
      pvm_send(tids[n], 1);
    }
}

/*-------------------------------------------------------------------------*/
/* print_hw_report:                                                        */
/* Print a hardware configuration report. Each hardware configuration      */
/* must implement a such function. It is called by the configuration main  */
/* function (confmain.c).                                                  */
/*-------------------------------------------------------------------------*/
void print_hw_report (void)
{
  int act_proc;

  printf("\nhardware configuration for PVM:\n");

  for (act_proc = 0; act_proc <= max_host_idx; act_proc++)
  {
    printf("  processor %3d placed on host ", act_proc);
    
    if (hosts[act_proc][0] == '\0')
      printf("(determined by PVM at run time)\n");
    else
      printf("%s\n", hosts[act_proc]);
  }
}
