/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* UNIX Hardware configuration implementation.                             */
/* The functions defined in this file must be called from the function     */
/* hardware_config() implemented by the user.                              */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 20.11.1993  Droux       Created                                         */
/* 09.02.1994  Droux       UNIX programs are always linked with server lib */
/* 25.06.1994  Droux       Special libraries for Solaris 2.x applications  */
/* 01.07.1994  Droux       Added print_hw_report()                         */
/* 08.07.1994  Droux       New structure of source directory               */
/* 17.08.1994  Droux       Support for custom hardware configuration file  */
/* 28.08.1994  Droux       Added "." prefix to architecture directory      */
/*-------------------------------------------------------------------------*/
/* filename: confunix.c                                                    */
/* Tested on :                                                             */
/*   NEXTSTEP 2.1, NeXT Computers                                          */
/*   NEXTSTEP 3.1, NeXT Computers                                          */
/*   NEXTSTEP 3.1, Intel Computers                                         */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*   SunOS Release 5.3 (Solaris 2), SPARC                                  */
/*   Linux 0.99.pl9, Intel Computers                                       */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#if defined(NEXT) || defined(NEXTINTEL)
#include <libc.h>
#endif
#include <syslog.h>
#include <time.h>
#include <stdio.h>
#include <pwd.h>
#include <ppc/csp/version.h>
#include <ppc/config/confunix.h>
#include <ppc/config/confpriv.h>
#include "confhard.h"
#include "confkey.h"

#define ARCH_DIR      ".unix/"
#define MAKEFILE_NAME ARCH_DIR "makefile"

/* Under UNIX, all tasks are linked together in a single executable. as    */
/* different tasks can contain the same object module, there can be        */
/* multiple instances of the same module linked in the same executable.    */
/* To avoid this duplication, we record a unique key for each              */
/* encountered module.                                                     */
static unsigned long modules_keys [MAX_MODULES*MAX_OBJECTS_PER_MODULE];
int nb_keys = 0;

/*-------------------------------------------------------------------------*/
/* generate_makefile:                                                LOCAL */
/* Generate the makefile for the current program.                          */
/*-------------------------------------------------------------------------*/
void generate_makefile (
  char *name, 
  char *user_name,
  char *date_str
)
{
  FILE *makefile;
  int n;
  int act_obj;
  int duplicated;                           /* 1 if a module is duplicated */
  int key_idx;                                      /* to browse keys list */
  unsigned long act_key;                            /* actual module's key */

  printf("generating UNIX makefile for \"%s\"\n", name);
  makefile = fopen(MAKEFILE_NAME, "w");
  fprintf(makefile, "\n");
  fprintf(makefile, " # This file has been generated by the PPC ");
  fprintf(makefile, "configurer\n");
  fprintf(makefile, " # PPC Version " PPC_VERSION "\n");
  fprintf(makefile, " # NEVER EDIT IT !\n");
  fprintf(makefile, " # Created on %s by %s\n", date_str, user_name);
  fprintf(makefile, " # PPC program name: \"%s\"\n", name);
  fprintf(makefile, " # Target architecture: UNIX\n");
  fprintf(makefile, "\n");
  fprintf(makefile, "CFLAGS=-I$(PPC_HOME)/include -D_$(ARCH) ");
  fprintf(makefile, "-D$(HOST_ARCH) -O $(OTHER_CFLAGS)\n");
  fprintf(makefile, "LFLAGS=-lconf -lconf$(ARCHLIB) -lcsp$(ARCHLIB) -lm ");
  fprintf(makefile, "-lserv$(ARCHLIB) \\\n");
  fprintf(makefile, "  -L$(PPC_HOME)/lib/$(HOST_ARCH) ");
  fprintf(makefile, "$(OTHERLIBS)\n");
  fprintf(makefile, "OBJ=.$(ARCH)/$(HOST_ARCH)/$(ARCH_FILE).o ");        
  fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/%s.o ", name);                      
  fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/_%s.o ", name); 
  
  for (n = 0; n < MAX_MODULES; n++)                  /* write object files */
  {
    if (!modules[n].free)
    {  
      for (act_obj = 0; act_obj < modules[n].nb_names; act_obj++)
      {
        act_key = confkey(modules[n].modules_names[act_obj]);
        duplicated = 0;
	for (key_idx = 0; key_idx < nb_keys; key_idx++)
	  duplicated = duplicated || (modules_keys[key_idx] == act_key);
	if (!duplicated)
	{
          fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/%s.o ", 
	    modules[n].modules_names[act_obj]);
	  modules_keys[nb_keys] = act_key;
	  nb_keys++;
	}
      }
    }
  }
  
  fprintf(makefile, "\n\n");                     /* write final executable */
  fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/%s.run: $(OBJ)\n", name); 
  fprintf(makefile, "\t$(CC) $(OBJ) -o $@ $(LFLAGS)\n\n");
  
  fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/_%s.o: .$(ARCH)/_%s.c\n", 
    name, name);
  fprintf(makefile, "\t$(CC) $(CFLAGS) -o $@ -c $?\n\n");
  
  nb_keys = 0;
  for (n = 0; n < MAX_MODULES; n++)  /* write dependencies for each module */
  {
    if (!modules[n].free)
    {
      for (act_obj = 0; act_obj < modules[n].nb_names; act_obj++)
      {
        act_key = confkey(modules[n].modules_names[act_obj]);
        duplicated = 0;
	for (key_idx = 0; key_idx < nb_keys; key_idx++)
	  duplicated = duplicated || (modules_keys[key_idx] == act_key);
	if (!duplicated)
	{
	  fprintf(makefile, ".$(ARCH)/$(HOST_ARCH)/%s.o: %s.c\n", 
	    modules[n].modules_names[act_obj],
	    modules[n].modules_names[act_obj]);
	  fprintf(makefile, "\t$(CC) $(CFLAGS) -Dppc_main=");
	  fprintf(makefile, "_%s_main ", modules[n].name);
	  fprintf(makefile, "-c $? -o $@\n\n");
	  modules_keys[nb_keys] = act_key;
	  nb_keys++;
	}
      }
    }
  }

  fclose(makefile);
}

/*-------------------------------------------------------------------------*/
/* generate_main:                                                    LOCAL */
/* Generate the main file of the current program.                          */
/* This main file will start all tasks and connect them together with      */
/* pipes.                                                                  */
/*-------------------------------------------------------------------------*/
void generate_main (
  char *name, 
  char *user_name,
  char *date_str
)
{
  FILE *mainfile;
  char mainfile_name [100];
  int n;
  int act_task;
  int act_param;
  
  printf("generating UNIX main file for \"%s\"\n", name);
  sprintf(mainfile_name, ARCH_DIR "_%s.c", name);
  mainfile = fopen(mainfile_name, "w");
  fprintf(mainfile, "\n/*\n");
  fprintf(mainfile, " * This file has been generated by the PPC ");
  fprintf(mainfile, "configurer\n");
  fprintf(mainfile, " * PPC Version " PPC_VERSION "\n");
  fprintf(mainfile, " * NEVER EDIT IT !\n");
  fprintf(mainfile, " * Created on %s by %s\n", date_str, user_name);
  fprintf(mainfile, " * PPC program name: \"%s\"\n", name);
  fprintf(mainfile, " * Target architecture: UNIX\n");
  fprintf(mainfile, " */\n\n");
  
  fprintf(mainfile, "#if defined(NEXT) || defined(NEXTINTEL)\n");
  fprintf(mainfile, "#include <libc.h>\n");
  fprintf(mainfile, "#endif\n\n");
  
  fprintf(mainfile, "#include <fcntl.h>\n");
  fprintf(mainfile, "#include <stdio.h>\n");
  fprintf(mainfile, "#include <sys/wait.h>\n\n");
  
  fprintf(mainfile, "#define channel conf_channel\n");
  fprintf(mainfile, "#include <ppc/config.h>\n");
  fprintf(mainfile, "#include <ppc/config/confpriv.h>\n");
  fprintf(mainfile, "#undef channel\n\n");
  
  fprintf(mainfile, "#include <ppc/csp.h>\n\n");

  fprintf(mainfile, "/* tasks entry points prototypes */\n");
  for (n = 0; n < MAX_MODULES; n++) 
  {
    if (!modules[n].free)
    {
      fprintf(mainfile, "void _%s_main (\n", modules[n].name);
      /* every tasks implemented by this module must have the same         */
      /* parameters. Here we use the first task implemented to             */
      /* determine the parameters.                                         */
      act_task = (int)modules[n].tasks[0];
      for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
           act_param++)
      {
        char c_param_name [100];
	
        switch(tasks[act_task].parameters[act_param].type)
	{
	  case OUTPUT_CHANNEL:
	    strcpy(c_param_name, "channel ");
	    break;
	  case INPUT_CHANNEL:
	    strcpy(c_param_name, "channel ");
	    break;
	  case INT:
	    strcpy(c_param_name, "int ");
	    break;
	  case SHORT:
	    strcpy(c_param_name, "short ");
	    break;
	  case UNSIGNED_SHORT:
	    strcpy(c_param_name, "unsigned short ");
	    break;
	  case LONG:
	    strcpy(c_param_name, "long ");
	    break;
	  case UNSIGNED_LONG:
	    strcpy(c_param_name, "unsigned long ");
	    break;
	  case FLOAT:
	    strcpy(c_param_name, "float ");
	    break;
	  case DOUBLE:
	    strcpy(c_param_name, "double ");
	    break;
	  case CHAR:
	    strcpy(c_param_name, "char");
	    break;
	  case OUTPUT_CHANNEL_ARRAY:
	    strcpy(c_param_name, "channel *");
	    break;
	  case INPUT_CHANNEL_ARRAY:
	    strcpy(c_param_name, "channel *");
	    break;
          case INT_ARRAY:
	    strcpy(c_param_name, "int *");
	    break;
          case SHORT_ARRAY:
	    strcpy(c_param_name, "short *");
	    break;
          case UNSIGNED_SHORT_ARRAY:
	    strcpy(c_param_name, "unsigned *");
	    break;
          case LONG_ARRAY:
	    strcpy(c_param_name, "long *");
	    break;
          case UNSIGNED_LONG_ARRAY:
	    strcpy(c_param_name, "unsigned long *");
	    break;
          case FLOAT_ARRAY:
	    strcpy(c_param_name, "float *");
	    break;
          case DOUBLE_ARRAY:
	    strcpy(c_param_name, "double *");
	    break;
          case CHAR_ARRAY:
	    strcpy(c_param_name, "char *");
	    break;
	  default:
	    fprintf(stderr, "ERROR: unknown task parameter type.\n");
	    exit(1);
	}
        fprintf(mainfile, "  %sparam_%d", c_param_name, act_param);
	if (act_param < (tasks[act_task].nb_parameters-1))
	  fprintf(mainfile, ",");
	fprintf(mainfile, "\n");
      }
      fprintf(mainfile, ");\n\n");
    }
  }
 
  fprintf(mainfile, "/* PPC program entry point */\n");
  fprintf(mainfile, "void main (int argc, char *argv[])\n");
  fprintf(mainfile, "{\n");
  fprintf(mainfile, "  int the_channels [MAX_CHANNELS];\n");
  fprintf(mainfile, "  char task_name [MAX_MODULE_NAME_LEN];\n");
  fprintf(mainfile, "  int n;\n");
  fprintf(mainfile, "  int fl;\n\n");

  fprintf(mainfile, "  virtual_machine(FROM_HOST_CHANNEL_ID, ");
  fprintf(mainfile, "TO_HOST_CHANNEL_ID);\n\n");

  fprintf(mainfile, "  /* open channels as pipes under UNIX */\n");
  fprintf(mainfile, "  for (n = 0; n < MAX_CHANNELS; n++)\n");
  fprintf(mainfile, "    if (!channels[n].free)\n");
  fprintf(mainfile, "      the_channels[n] = par_channel();\n\n");
    
  fprintf(mainfile, "  for (n = 0; n < MAX_TASKS; n++)\n");  
  fprintf(mainfile, "  {\n");  
  fprintf(mainfile, "    if (!tasks[n].free)\n");  
  fprintf(mainfile, "    {\n");  
  fprintf(mainfile, "      strcpy(task_name, modules[tasks[n].");  
  fprintf(mainfile, "implemented].name);\n\n"); 
  
  for (n = 0; n < MAX_MODULES; n++) 
  {
    if (!modules[n].free)
    {
      fprintf(mainfile, "      if (!strcmp(task_name, \"%s\"))\n", 
        modules[n].name);
      fprintf(mainfile, "      {\n");

      act_task = (int)modules[n].tasks[0];
      
      /* channel arrays cannot be simply given by their address find in    */
      /* the task parameter because these values reference the indice of   */
      /* the channel and not its value. Thus, we have to create some       */
      /* temporary channel array that contains the correct channel         */
      /* identifiers that will be passed to the task function.             */
      
                                             /* declare the channel arrays */
      for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
           act_param++)
      {
        switch(tasks[act_task].parameters[act_param].type)
	{
	  case OUTPUT_CHANNEL_ARRAY:
	  case INPUT_CHANNEL_ARRAY:
	    fprintf(mainfile, "        channel chan_array_%d [%d];\n",
	      act_param, 
	      tasks[act_task].parameters[act_param].param.ptr.nitems);
	    break;
	  default:
	    break;                                           /* do nothing */
	}
      }
                                          /* initialize the channel arrays */
      for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
           act_param++)
      {
        switch(tasks[act_task].parameters[act_param].type)
	{
	  case OUTPUT_CHANNEL_ARRAY:
	  case INPUT_CHANNEL_ARRAY:
	    fprintf(mainfile, "        {\n");
	    fprintf(mainfile, "          int i;\n");
	    fprintf(mainfile, "          channel *act_ptr = tasks[n].");
	    fprintf(mainfile, "parameters[%d].param.ptr.address;\n",
	      act_param);
	    fprintf(mainfile, "	  for (i=0; i < %d; i++, act_ptr++)\n",
	      tasks[act_task].parameters[act_param].param.ptr.nitems);
	    fprintf(mainfile, "	    chan_array_%d[i] = ", act_param);
	    fprintf(mainfile, "the_channels[*act_ptr];\n");
	    fprintf(mainfile, "        }\n\n");
	    break;
	  default:
	    break;                                           /* do nothing */
	}
      }
      
      fprintf(mainfile, "        if (!fork())\n");
      fprintf(mainfile, "        {\n");
      fprintf(mainfile, "          _%s_main (\n", modules[n].name);
      
      for (act_param = 0; act_param < tasks[act_task].nb_parameters; 
           act_param++)
      {
        char c_param_name [100];
	
        switch(tasks[act_task].parameters[act_param].type)
	{
	  case OUTPUT_CHANNEL:
	    if (tasks[act_task].parameters[act_param].param.value.channel_val ==
	        TO_HOST_CHANNEL_ID)
	      strcpy(c_param_name, "TO_HOST_CHANNEL_ID");
	    else
	      sprintf(c_param_name, 
	        "the_channels[tasks[n].parameters[%d].param.value.channel_val]", 
		act_param);
	    break;
	  case INPUT_CHANNEL:
	    if (tasks[act_task].parameters[act_param].param.value.channel_val ==
	        FROM_HOST_CHANNEL_ID)
	      strcpy(c_param_name, "FROM_HOST_CHANNEL_ID");
	    else
	      sprintf(c_param_name, 
	        "the_channels[tasks[n].parameters[%d].param.value.channel_val]", 
		act_param);
	    break;
	  case INT:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.value.int_val",
	      act_param);
	    break;
	  case SHORT:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.value.short_val",
	      act_param);
	    break;
	  case UNSIGNED_SHORT:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.value.unsigned_short_val",
	      act_param);
	    break;
	  case LONG:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.value.long_val",
	      act_param);
	    break;
	  case UNSIGNED_LONG:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.value.unsigned_long_val",
	      act_param);
	    break;
	  case FLOAT:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.value.float_val",
	      act_param);
	    break;
	  case DOUBLE:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.value.double_val",
	      act_param);
	    break;
	  case CHAR:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.value.char_val",
	      act_param);
	    break;
	  case OUTPUT_CHANNEL_ARRAY:
	    sprintf(c_param_name, 
	      "chan_array_%d",
	      act_param);
	    break;
	  case INPUT_CHANNEL_ARRAY:
	    sprintf(c_param_name, 
	      "chan_array_%d",
	      act_param);
	    break;
          case INT_ARRAY:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.ptr.address",
	      act_param);
	    break;
          case SHORT_ARRAY:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.ptr.address",
	      act_param);
	    break;
          case UNSIGNED_SHORT_ARRAY:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.ptr.address",
	      act_param);
	    break;
          case LONG_ARRAY:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.ptr.address",
	      act_param);
	    break;
          case UNSIGNED_LONG_ARRAY:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.ptr.address",
	      act_param);
	    break;
          case FLOAT_ARRAY:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.ptr.address",
	      act_param);
	    break;
          case DOUBLE_ARRAY:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.ptr.address",
	      act_param);
	    break;
          case CHAR_ARRAY:
	    sprintf(c_param_name, 
	      "tasks[n].parameters[%d].param.ptr.address",
	      act_param);
	    break;
	}
        fprintf(mainfile, "            %s", c_param_name);
	if (act_param < (tasks[act_task].nb_parameters-1))
	  fprintf(mainfile, ",");
	fprintf(mainfile, "\n");
      }
      fprintf(mainfile, "          );\n");
      fprintf(mainfile, "          exit(0);\n");
      fprintf(mainfile, "        }\n");
      fprintf(mainfile, "      }\n");
    }
  }
  fprintf(mainfile, "    }\n");
  fprintf(mainfile, "  }\n\n");
  
  fprintf(mainfile, "  /* wait the death of the tasks */\n");
  fprintf(mainfile, "  #ifdef SUN4SOL2\n");
  fprintf(mainfile, "    while(wait((int *)0) != -1);\n");
  fprintf(mainfile, "  #else\n");
  fprintf(mainfile, "    while(wait((union wait *)0) != -1);\n");
  fprintf(mainfile, "  #endif\n\n");
  
  fprintf(mainfile, "  /* close channels (pipes) */\n");
  fprintf(mainfile, "  for (n = 0; n < MAX_CHANNELS; n++)\n");
  fprintf(mainfile, "    if (!channels[n].free)\n");
  fprintf(mainfile, "      chan_close(the_channels[n]);\n\n");
  
  fprintf(mainfile, "  /* dispose memory used by the configuration */\n");
  fprintf(mainfile, "  free_config_data();\n");
  fprintf(mainfile, "}\n");
  
  fclose(mainfile);
}

/*-------------------------------------------------------------------------*/
/* generate_files:                                                         */
/* Generate the makefile for the PPC program in order to compile it        */
/* for the target machine. One version of this function must be            */
/* implemented for each target hardware.                                   */
/*-------------------------------------------------------------------------*/
void generate_files (char *name)
{
  char date_str [100];
  time_t act_time;
  char user_name [100];
  struct passwd *act_passwd;

  time(&act_time);                           /* get string containing date */
  strcpy(date_str, ctime(&act_time));
  date_str[strlen(date_str)-1] = '\0';
  
  act_passwd = getpwuid(getuid());                /* get user's login name */
  strcpy(user_name, act_passwd->pw_name);

  generate_makefile(name, user_name, date_str);
  generate_main(name, user_name, date_str);
}

/*-------------------------------------------------------------------------*/
/* print_hw_report:                                                        */
/* Print a hardware configuration report. Each hardware configuration      */
/* must implement a such function. It is called by the configuration main  */
/* function (confmain.c).                                                  */
/*-------------------------------------------------------------------------*/
void print_hw_report(void)
{
  printf("\nhardware configuration for UNIX:\n");
  printf("  all processors emulated on a single machine.\n");
}

