/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Client-Server Library.                                                  */
/* LoginWindow class implementation.                                       */
/* Subclass of Window, this class implements the PPC login window which    */
/* allows a user to select the host on which a server application must be  */
/* started.                                                                */
/* One instance of this class is created when the LoginPalette is          */
/* imported in a NEXTSTEP application. This instance must be then passed   */
/* as parameter to the connect method of the PPCServer class.              */
/* This class is never used directly by the programmer, but indirectly by  */
/* the PPCServer class.                                                    */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 09.07.1994  Droux       Created                                         */
/*-------------------------------------------------------------------------*/
/* filename: LoginWindow.m                                                 */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#import <syslog.h>
#import <pwd.h>
#import "LoginWindow.h"
#import "PPCHost.h"

#define DEFAULT_OWNER		"PPCLoginWindow"
#define DEFAULT_HOST_NAME	"DefaultHost"
#define DEFAULT_ARCH_NAME	"DefaultArchictecture"

@implementation LoginWindow

/*-------------------------------------------------------------------------*/
/* findHost:                                                         LOCAL */
/* Returns a pointer to the host object corresponding to the given host    */
/* name. If the name is not found, then returns nil.                       */
/*-------------------------------------------------------------------------*/
- findHost: (char *) name
{
  int actHost;
  
  for (actHost = 0; 
    actHost < nbHosts && strcmp(name, [hosts[actHost] getHostName]); 
    actHost++);
  
  if (actHost == nbHosts)
    return nil;
  return hosts[actHost];
}

/*-------------------------------------------------------------------------*/
/* freeHosts:                                                        LOCAL */
/* Free all hosts in the host table.                                       */
/*-------------------------------------------------------------------------*/
- freeHosts
{
  int actHost;
  
  for (actHost = 0; actHost < nbHosts; actHost++)
    [hosts[actHost] free];

  return self;
}

/*-------------------------------------------------------------------------*/
/* clearHostValues:                                                  LOCAL */
/* Clear the values of the optional login and optional password fields of  */
/* the hosts list group.                                                   */
/*-------------------------------------------------------------------------*/
- clearHostValues
{
  [selLoginFormCell setStringValue: ""];
  [selPathFormCell setStringValue: ""];
  [selPasswdSwitch setState: NO];
  return self;
}

/*-------------------------------------------------------------------------*/
/* clearSelectedHostValues:                                          LOCAL */
/* Clear the values of the selected host fields.                           */
/*-------------------------------------------------------------------------*/
- clearSelectedHostValues
{
  [serverFormCell setStringValue: ""];
  [archFormCell setStringValue: ""];
  [loginFormCell setStringValue: ""];
  [pathFormCell setStringValue: ""];
  [passwdTextField setEditable: NO];
  [passwdTextField setSelectable: NO];
  [[passwdTextField cell] setStringValue: ""];
  return self;
}

/*-------------------------------------------------------------------------*/
/* initUserName:                                                     LOCAL */
/* Init the instance variable userName with the actual user login name.    */
/*-------------------------------------------------------------------------*/
- initUserName
{
  struct passwd *act_passwd;
  
  act_passwd = getpwuid(getuid());
  strcpy(userName, act_passwd->pw_name);
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* readDefaults:                                                     LOCAL */
/* Read the default host and architecture from the user's database.        */
/* If the defaults are defined and still exist in the hosts list, then     */
/* initialize the instance variables defaultHost and defaultArchIdx and    */
/* return YES.                                                             */
/* If they are not defined, return NO.                                     */
/* If they are defined but the host does not exist in the hosts table      */
/* anymore, then clear the defaults and return NO.                         */
/*-------------------------------------------------------------------------*/
- (BOOL) readDefaults
{
  char *theHostName;
  char *theArchStr;
  
  theHostName = (char *)NXReadDefault(DEFAULT_OWNER, DEFAULT_HOST_NAME);
  theArchStr = (char *)NXReadDefault(DEFAULT_OWNER, DEFAULT_ARCH_NAME);
  
  if (theHostName == NULL)
  {                                    /* default host name is not defined */
    if (theArchStr != NULL)
      NXRemoveDefault(DEFAULT_OWNER, DEFAULT_ARCH_NAME);
    return NO;
  }
  
  if (theArchStr == NULL)
  {                                 /* default architecture is not defined */
    if (theHostName != NULL)
      NXRemoveDefault(DEFAULT_OWNER, DEFAULT_HOST_NAME);
    return NO;
  }
  
  if ((defaultHost = [self findHost: theHostName]) == nil)
  {               /* the default host does not exist anymore in hosts list */
    NXRemoveDefault(DEFAULT_OWNER, DEFAULT_HOST_NAME);
    NXRemoveDefault(DEFAULT_OWNER, DEFAULT_ARCH_NAME);
    return NO;
  }
  
  if ((defaultArchIdx = [defaultHost archIdxFromString: theArchStr]) == -1)
  {          /* the architecture for the default host is no more supported */
    NXRemoveDefault(DEFAULT_OWNER, DEFAULT_HOST_NAME);
    NXRemoveDefault(DEFAULT_OWNER, DEFAULT_ARCH_NAME);
    return NO;
  }
  
  return YES;                                          /* everything is OK */
}

/*-------------------------------------------------------------------------*/
/* initSelectionWithDefaults:                                        LOCAL */
/*-------------------------------------------------------------------------*/
- initSelectionWithDefaults
{
  [serverFormCell setStringValue: [defaultHost getHostName]];

  [archFormCell setStringValue: [defaultHost getArchString: defaultArchIdx]];

  if ([defaultHost loRequired: defaultArchIdx])
    [loginFormCell setStringValue: [defaultHost lo: defaultArchIdx]];
  else
    [loginFormCell setStringValue: userName];

  if ([defaultHost epRequired: defaultArchIdx])
    [pathFormCell setStringValue: [defaultHost ep: defaultArchIdx]];
  else
    [pathFormCell setStringValue: "(default)"];

  if ([defaultHost pwRequired: defaultArchIdx])
  {                         /* the user must be able to enter the password */
    [[passwdTextField cell] setStringValue: ""];
    [passwdTextField setEditable: YES];
  }
  else
  {
    [passwdTextField setEditable: NO];
    [passwdTextField setSelectable: NO];
  }
  
  [loginButton setEnabled: YES];                   /* login is now allowed */
  selectedHost = defaultHost;
  selectedArch = defaultArchIdx;
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* keyDown:                                                                */
/* keyDown event handler called when the user press a key. This method     */
/* perform user interaction corresponding to the pressed key if defined.   */
/*-------------------------------------------------------------------------*/
- keyDown: (NXEvent *)event
{
            /* login button has a key equivalent (normally the return key) */
  if (event->data.key.charCode == [loginButton keyEquivalent])
    [loginButton performClick: self];
    
  return [super keyDown: event];
}

/*-------------------------------------------------------------------------*/
/* init:                                                                   */
/* Misc instance initializations.                                          */
/*-------------------------------------------------------------------------*/
- init
{
  [super init];
  
  [hostsBrowser setTarget: self];
  [hostsBrowser setDelegate: self];

  [archBrowser setTarget: self];
  [archBrowser setDelegate: self];
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* free:                                                                   */
/* Misc instance deallocation.                                             */
/*-------------------------------------------------------------------------*/
- free
{
  [self freeHosts];
  return [super free];
}

/*-------------------------------------------------------------------------*/
/* initHosts:                                                              */
/* Read the hosts file and initialize the content of the login window.     */
/* This method must be called before showing the window.                   */
/*-------------------------------------------------------------------------*/
- initHosts
{
  if (hlist_read() < 0)
    return nil;               /* an error occured while reading hosts list */
  
  {                    /* initializes table containing the available hosts */
    int nbHostsInFile = hlist_nb_hosts();
    int currentHost;
    char *actName;
    architecture actArch;
    id actHost;
   
    [self freeHosts];
    nbHosts = 0;
     
    for (currentHost = 0; currentHost < nbHostsInFile; currentHost++)
    {
      actName = hlist_get_host_name_ptr(currentHost);

      if ((actHost = [self findHost: actName]) == nil)
      {                    /* the current host is not in the table, add it */
	hosts[nbHosts] = [[PPCHost alloc] init];         /* allocates host */
	[hosts[nbHosts] initHost: actName];               /* set host name */
	actHost = hosts[nbHosts];
	nbHosts++;                         /* update total number of hosts */
      }
      
      hlist_get_host_arch(currentHost, &actArch);
      if ([actHost addArch: actArch          /* add supported architecture */
	  loRequired: hlist_exec_lo_defined(currentHost)
	  lo: hlist_custom_name(currentHost)
	  pwRequired: hlist_pw_required(currentHost)
	  epRequired: hlist_exec_path_required(currentHost)
	  ep: hlist_exec_path(currentHost)] == nil)
      {                   /* an error occured during architecture addition */
        return nil;
      }
    }
  }
  
  [setHostButton setEnabled: NO];
  [loginButton setEnabled: NO];
  [self clearSelectedHostValues];
  [self clearHostValues];
  [hostsBrowser loadColumnZero];                  /* show hosts in browser */
  [archBrowser loadColumnZero];              /* init architectures browser */
  [self initUserName];
  
  if ([self readDefaults])
  {                                                /* defaults are defined */
    [setAsDefaultButton setEnabled: NO];
    [removeDefaultButton setEnabled: YES];
    [restoreDefaultButton setEnabled: YES];
    [self initSelectionWithDefaults];
  }
  else
  {                                                  /* no default defined */
    [setAsDefaultButton setEnabled: NO];
    [removeDefaultButton setEnabled: NO];
    [restoreDefaultButton setEnabled: NO];
  }
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* loginClicked:                                                           */
/* Message sent when the user click on the "Login" button. Will cause      */
/* the window to read its actual login contents and break the modal loop.  */
/*-------------------------------------------------------------------------*/
- loginClicked: sender
{
  canceled = NO;
  strcpy(selHostName, [selectedHost getHostName]);
  selHostArch = [selectedHost arch: selectedArch];
  strcpy(selPw, [[passwdTextField cell] stringValue]);
  [NXApp abortModal];                            /* break modal event loop */
  return self;
}

/*-------------------------------------------------------------------------*/
/* cancelClicked:                                                          */
/* Message sent when the user clicks on the "Cancel" button. Will cause    */
/* the window to break the modal loop.                                     */
/*-------------------------------------------------------------------------*/
- cancelClicked: sender
{
  canceled = YES;
  [NXApp abortModal];                            /* break modal event loop */
  return self;
}

/*-------------------------------------------------------------------------*/
/* passwdSwitchClicked:                                                    */
/* Message sent when the user clicks on the "Password Required" switch of  */
/* the hosts list.                                                         */
/*-------------------------------------------------------------------------*/
- passwdSwitchClicked: sender
{
  /* The user is not allowed to modify the hosts list. But if we disable   */
  /* the password switch, it is set in light gray, which is not very       */
  /* elegant. The solution here is to keep the switch enabled but when the */
  /* select it, it turns back to its old state and a beep is emmitted.     */
  [sender setState: ![sender state]];
  NXBeep();
  return self;
}

/*-------------------------------------------------------------------------*/
/* setHost:                                                                */
/* Message sent when the user clicks on the "Set Host" button.             */
/*-------------------------------------------------------------------------*/
- setHost: sender
{
  /* If the host is completely defined, copy the hosts properties from     */
  /* the hosts list to the selected host.                                  */
  
  selectedHost = [self findHost: 
    (char *) [[hostsBrowser selectedCell] stringValue]];
  selectedArch = [selectedHost archIdxFromString: 
    [[archBrowser selectedCell] stringValue]];
  
  [serverFormCell setStringValue: 
    [[hostsBrowser selectedCell] stringValue]];
  [archFormCell setStringValue:
    [[archBrowser selectedCell] stringValue]];
  [loginFormCell setStringValue:
    [selLoginFormCell stringValue]];
  [pathFormCell setStringValue:
    [selPathFormCell stringValue]];
    
  if ([selPasswdSwitch state])
  {                         /* the user must be able to enter the password */
    [[passwdTextField cell] setStringValue: ""];
    [passwdTextField setEditable: YES];
  }
  else
  {
    [passwdTextField setEditable: NO];
    [passwdTextField setSelectable: NO];
  }
  
  [loginButton setEnabled: YES];                   /* login is now allowed */
  return self;
}

/*-------------------------------------------------------------------------*/
/* setAsDefault:                                                           */
/* Message sent when the user clicks on the "Set as Default" button.       */
/*-------------------------------------------------------------------------*/
- setAsDefault: sender
{
                                              /* save defaults in database */
  NXWriteDefault(DEFAULT_OWNER, DEFAULT_HOST_NAME,
    [[hostsBrowser selectedCell] stringValue]);
  NXWriteDefault(DEFAULT_OWNER, DEFAULT_ARCH_NAME,
    [[archBrowser selectedCell] stringValue]);
                                      /* update instance variable defaults */
  defaultHost = [self findHost: 
    (char *)[[hostsBrowser selectedCell] stringValue]];
  defaultArchIdx = [defaultHost archIdxFromString: 
    [[archBrowser selectedCell] stringValue]];
    
  [removeDefaultButton setEnabled: YES];         /* operations now allowed */
  [restoreDefaultButton setEnabled: YES];

  [setHostButton performClick: self];       /* update selected host fields */
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* removeDefault:                                                          */
/* Message sent when the user clicks on the "Remove Default" button.       */
/*-------------------------------------------------------------------------*/
- removeDefault: sender
{
  NXRemoveDefault(DEFAULT_OWNER, DEFAULT_HOST_NAME);
  NXRemoveDefault(DEFAULT_OWNER, DEFAULT_ARCH_NAME);

  [removeDefaultButton setEnabled: NO];      /* operations no more allowed */
  [restoreDefaultButton setEnabled: NO];

  return self;
}

/*-------------------------------------------------------------------------*/
/* restoreDefault:                                                         */
/* Message sent when the user clicks on the "Restore Default" button.      */
/*-------------------------------------------------------------------------*/
- restoreDefault: sender
{
  [self initSelectionWithDefaults];
  return self;
}

/*-------------------------------------------------------------------------*/
/* hostsBrowserClick:                                                      */
/* Message sent when the user clicks on a cell of the hosts browser.       */
/*-------------------------------------------------------------------------*/
- hostsBrowserClick: sender
{
  [archBrowser loadColumnZero];              /* init architectures browser */
  [setHostButton setEnabled: NO];
  [setAsDefaultButton setEnabled: NO];
  [self clearHostValues];
  return self;
}

/*-------------------------------------------------------------------------*/
/* archBrowserClick:                                                       */
/* Message sent when the user clicks on a cell of the architectures        */
/* browser.                                                                */
/*-------------------------------------------------------------------------*/
- archBrowserClick: sender
{
  id host;
  int archIdx;
  
  /* If the user has clicked on the architecture, then the architectures   */
  /* have been initialized. If the architectures have been initialized,    */
  /* then this implies that a host has been selected. Thus, both the       */
  /* host name and the architecture are defined.                           */
  
  host = [self findHost: 
    (char *) [[hostsBrowser selectedCell] stringValue]];
  archIdx = [host archIdxFromString: 
    [[archBrowser selectedCell] stringValue]];
    
  /* If the user switch between different architectures of the same host   */
  /* rapidely, it is possible that the archBrowserClick: actionis called   */
  /* by the NXBrowser but no cell is selected. To protect the user against */
  /* random crashes, archIdxFromString: test for the validity of the       */
  /* given string and returns -1 if it was not valid. We test if for this  */
  /* condition and abort the process if it is the case.                    */
  if (archIdx == -1)
  {
    [self clearHostValues];
    return self;
  }
  
  if ([host loRequired: archIdx])
    [selLoginFormCell setStringValue: [host lo: archIdx]];
  else
    [selLoginFormCell setStringValue: userName];
  
  [selPasswdSwitch setState: [host pwRequired: archIdx]];
  
  if ([host epRequired: archIdx])
    [selPathFormCell setStringValue: [host ep: archIdx]];
  else
    [selPathFormCell setStringValue: "(default)"];
  
  [setHostButton setEnabled: YES];
  [setAsDefaultButton setEnabled: YES];

  return self;
}

/*-------------------------------------------------------------------------*/
/* browser:                                                                */
/* NXBrowser delegate method sent when one of the browsers needs the       */
/* contents of the cells.                                                  */
/*-------------------------------------------------------------------------*/
- (int) browser: sender
  fillMatrix: matrix
  inColumn: (int)column
{
  int nbElements;
  
  if (sender == hostsBrowser)
  {
    id currentCell;
    int currentHost;
  
    for (currentHost = 0; currentHost < nbHosts; currentHost++)
    {                             /* add current host to the hosts browser */
      [matrix addRow];
      currentCell = [matrix cellAt: currentHost: 0];
      [currentCell setStringValue: [hosts[currentHost] getHostName]];
      [currentCell setLeaf: YES];
    }
    
    nbElements = nbHosts;
  }
  else
  {
    id currentCell;
    id actHost;
    int currentArch;
    int nbArch;
    id selectedCell;
  
    selectedCell = [hostsBrowser selectedCell];
    if (selectedCell == nil)
      return 0;
      
    actHost = [self findHost: (char *) [selectedCell stringValue]];
    nbArch = [actHost nbArch];
    for (currentArch = 0; currentArch < nbArch; currentArch++)
    {
      [matrix addRow];
      currentCell = [matrix cellAt: currentArch: 0];
      [currentCell setStringValue: [actHost getArchString: currentArch]];
      [currentCell setLeaf: YES];
    }
    
    nbElements = nbArch;
  }
  
  return nbElements;
}

/*-------------------------------------------------------------------------*/
/* canceled:                                                               */
/* Return YES if the user has clicked on the "Cancel" button.              */
/*-------------------------------------------------------------------------*/
- (BOOL)canceled
{
  return canceled;
}

/*-------------------------------------------------------------------------*/
/* selHostName:                                                            */
/* Return the selected host name. Only valid after the user has clicked    */
/* on "Login".                                                             */
/*-------------------------------------------------------------------------*/
- (char *) selHostName
{
  return selHostName;
}

/*-------------------------------------------------------------------------*/
/* selHostArch:                                                            */
/* Return the selected host architecture. Only valid after the user has    */
/* clicked on "Login".                                                     */
/*-------------------------------------------------------------------------*/
- (architecture) selHostArch
{
  return selHostArch;
}

/*-------------------------------------------------------------------------*/
/* selPw:                                                                  */
/* Return the selected password. Only valid after the user has clicked on  */
/* "Login".                                                                */
/*-------------------------------------------------------------------------*/
- (char *) selPw
{
  return selPw;
}

@end
