/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Client-Server Library.                                                  */
/* PPCHost class implementation.                                           */
/* A PPCHost object correspond to a server host identified by its          */
/* internet address. One PPCHost can support more than one architecture.   */
/* (This is a difference compared to hlist host where a new host is        */
/* created for each different architecture.)                               */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 11.07.1994  Droux       Created                                         */
/*-------------------------------------------------------------------------*/
/* filename: PPCHost.m                                                     */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#import "PPCHost.h"
#import "PPCArch.h"

@implementation PPCHost

/*-------------------------------------------------------------------------*/
/* findArch:                                                         LOCAL */
/* Returns a pointer to the architecture object corresponding to the given */
/* architecture. If the architecture is not found, then returns nil.       */
/*-------------------------------------------------------------------------*/
- findArch: (architecture) archToFind
{
  int actArch;
  
  for (actArch = 0; 
    (actArch < nbArch) && (archToFind != [architectures[actArch] arch]); 
    actArch++);
  
  if (actArch == nbArch)
    return nil;

  return architectures[actArch];
}

/*-------------------------------------------------------------------------*/
/* freeArchs:                                                        LOCAL */
/* Free all architectures in the architecture table.                       */
/*-------------------------------------------------------------------------*/
- freeArchs
{
  int actArch;
  
  for (actArch = 0; actArch < nbArch; actArch++)
    [architectures[actArch] free];

  return self;
}

/*-------------------------------------------------------------------------*/
/* init:                                                                   */
/* Misc instance initializations.                                          */
/*-------------------------------------------------------------------------*/
- init
{
  [super init];
  nbArch = 0;
  return self;
}

/*-------------------------------------------------------------------------*/
/* free:                                                                   */
/* Misc instance deallocation.                                             */
/*-------------------------------------------------------------------------*/
- free
{
  [self freeArchs];
  return [super free];
}

/*-------------------------------------------------------------------------*/
/* initHost:                                                               */
/* Init the host object with the given internet name.                      */
/*-------------------------------------------------------------------------*/
- initHost: (char *) hostNameP
{
  hostName = hostNameP;
  return self;
}

/*-------------------------------------------------------------------------*/
/* addArch:                                                                */
/* Add the given architecture as supported by the host.                    */
/*-------------------------------------------------------------------------*/
- addArch: (architecture) arch                             /* architecture */
  loRequired: (BOOL) loRequiredP          /* YES if alternative login name */
  lo: (char *) loP                               /* alternative login name */
  pwRequired: (BOOL) pwRequiredP               /* YES if password required */
  epRequired: (BOOL) epRequiredP         /* YES if executable path defined */
  ep: (char *) epP;                                     /* executable path */
{
  id actArch;
  
  if ((actArch = [self findArch: arch]) == nil)
  {                          /* architecture not yet declared as supported */
    actArch = [[PPCArch alloc] init];
    architectures[nbArch] = actArch;
    nbArch++;
  }
  else
  {                      /* architecture already initialized, return error */
    ppc_errno = EPPC_LPANEL_MULTIHOST;
    return nil;
  }
  
  [actArch setArch: arch
    loRequired: loRequiredP
    lo: loP
    pwRequired: pwRequiredP
    epRequired: epRequiredP
    ep: epP];

  return self;
}

/*-------------------------------------------------------------------------*/
/* getHostName:                                                            */
/* Return the host internet name.                                          */
/*-------------------------------------------------------------------------*/
- (char *) getHostName
{
  return hostName;
}

/*-------------------------------------------------------------------------*/
/* nbArch:                                                                 */
/* Return the number of architectures supported by the host.               */
/*-------------------------------------------------------------------------*/
- (int) nbArch
{
  return nbArch;
}

/*-------------------------------------------------------------------------*/
/* arch:                                                                   */
/* Return the architecture of the specified supported architecture.        */
/*-------------------------------------------------------------------------*/
- (architecture) arch: (int) idx
{
  return [architectures[idx] arch];
}

/*-------------------------------------------------------------------------*/
/* getArchString:                                                          */
/* Return the name of the specified supported architecture.                */
/*-------------------------------------------------------------------------*/
- (char *) getArchString: (int) idx
{
  return [architectures[idx] getArchString];
}

/*-------------------------------------------------------------------------*/
/* archIdxFromString:                                                      */
/* Return the architecture identifier corresponding to the given           */
/* architecture name or -1 if not supported.                               */
/*-------------------------------------------------------------------------*/
- (int) archIdxFromString: (const char *)archName
{
  int actArch;
  
  if (archName == NULL)
    return -1;
    
  for (actArch = 0; 
    actArch < nbArch && 
    strcmp(archName, [architectures[actArch] getArchString]); 
    actArch++);
  
  if (actArch == nbArch)
    return -1;
    
  return actArch;
}

/*-------------------------------------------------------------------------*/
/* loRequired:                                                             */
/* Return YES if the specified architecture requires an alternative        */
/* login name.                                                             */
/*-------------------------------------------------------------------------*/
- (BOOL) loRequired: (int) idx
{
  return [architectures[idx] loRequired];
}

/*-------------------------------------------------------------------------*/
/* lo:                                                                     */
/* Return the alternative login name corresponding to the specified        */
/* architecture.                                                           */
/*-------------------------------------------------------------------------*/
- (char *) lo: (int) idx
{
  return [architectures[idx] lo];
}

/*-------------------------------------------------------------------------*/
/* pwRequired:                                                             */
/* Return YES if the specified architecture requires a password.           */
/*-------------------------------------------------------------------------*/
- (BOOL) pwRequired: (int) idx
{
  return [architectures[idx] pwRequired];
}

/*-------------------------------------------------------------------------*/
/* epRequired:                                                             */
/* Return YES if the specified architecture requires an optional           */
/* executable path.                                                        */
/*-------------------------------------------------------------------------*/
- (BOOL) epRequired: (int) idx
{
  return [architectures[idx] epRequired];
}

/*-------------------------------------------------------------------------*/
/* ep:                                                                     */
/* Return the executable path corresponding to the specified architecture. */
/*-------------------------------------------------------------------------*/
- (char *) ep: (int) idx
{
  return [architectures[idx] ep];
}

@end
