/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Client-Server Library.                                                  */
/* Server Class Interface.                                                 */
/* This class encapsulates a server into a single Objective-C class. The   */
/* client application views a server as an object to which it sends and    */
/* from which it receives message. In addition, the server class can       */
/* use a login window to allows the user of the application to select the  */
/* server host interactively.                                              */
/* The user of the class must allocate and initialize a PPCServer object   */
/* before connection.                                                      */
/*                                                                         */
/* The following methods are available:                                    */
/* connectToServer	Connect server                                     */
/* disconnectFromServer	Disconnect server                                  */
/* sendChannel		Send data to server                                */
/* receiveChannel	Receive data from server                           */
/* isConnected		Is server connected ?                              */
/* serverHost		Return the current server host name                */
/* serverArch		Return the current server host architecture        */
/* archString		Return a string corresponding to an architecture   */
/* lastServerName	Return last host name where connection failed      */
/* lastServerArch	Return last host architecture where conn. failed   */
/* lastErrorString	Return last error textual description              */
/* toServerChannel      Return channel to server                           */
/* fromServerChannel	Return channel from server                         */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 09.07.1994  Droux       Created                                         */
/*-------------------------------------------------------------------------*/
/* filename: PPCServer.h                                                   */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#import <appkit/appkit.h>
#import <ppc/cslib.h>
#import <rpc/rpc.h>

@interface PPCServer:Object
{
  BOOL isConnected;                          /* YES if server is connected */
  char hostName [MAX_HOST_NAME_LEN];                   /* actual host name */
  architecture hostArch;                       /* actual host architecture */
  int sessionId;                              /* CS-Lib session identifier */
  channel toServerChan;                           /* channel to PPC server */
  channel fromServerChan;                       /* channel from PPC server */
  BOOL handlerInstalled;          /* YES if a channel handler is installed */
}

/*-------------------------------------------------------------------------*/
/* init:                                                                   */
/* Initialize server instance.                                             */
/*-------------------------------------------------------------------------*/
- init;

/*-------------------------------------------------------------------------*/
/* connectToServer:                                                        */
/* Establish a connection with a server.                                   */
/* There are two mode of operation for this function depending on the      */
/* value of useLoginWindow:                                                */
/* 1. useLoginWindow = YES                                                 */
/*    Tell connect to use the object loginWindow to select the server      */
/*    host. In this case, loginWindow must be initialized to a             */
/*    LoginWindow object. This object is the login window which is created */
/*    when using the LoginPalette. The object which calls this method      */
/*    would have an outlet initialized to this window which it could       */
/*    pass as loginWindow. In this mode, the parameters serverHost,        */
/*    serverArch, and password are not used.                               */
/* 2. useLoginWindow = NO                                                  */
/*    Tell connect to use the parameters serverHost, serverArch and        */
/*    password to establish the connection with the server host as with    */
/*    the connect_to_server() function.                                    */
/* progName corresponds to the name of the server program to be started.   */
/* If errorPanel is set to YES, then an error panel will appears if an     */
/* error occured during the connection.                                    */
/*-------------------------------------------------------------------------*/
- connectToServer: (char *)progName                 /* server program name */
  loginWindow: loginW                               /* login window object */
  useLoginWindow: (BOOL)useLW          /* YES if login window must be used */
  serverHost: (char *)host                             /* server host name */
  serverArch: (architecture)arch                    /* server architecture */
  password: (char *)passwd                            /* optional password */
  errorPanel: (BOOL)errorP;                      /* use error panel or not */

/*-------------------------------------------------------------------------*/
/* disconnectFromServer:                                                   */
/* Disconnect server.                                                      */
/*-------------------------------------------------------------------------*/
- disconnectFromServer;

/*-------------------------------------------------------------------------*/
/* isConnected:                                                            */
/* Return YES if the server is connected, NO otherwise.                    */
/*-------------------------------------------------------------------------*/
- (BOOL) isConnected;

/*-------------------------------------------------------------------------*/
/* serverHost:                                                             */
/* Return a pointer to the string containing the name of the server host.  */
/* Return NULL if server is not connected.                                 */
/*-------------------------------------------------------------------------*/
- (char *) serverHost;

/*-------------------------------------------------------------------------*/
/* serverArch:                                                             */
/* Return the architecture of the host on which server is running.         */
/* Possible values are: UNIX, PVM, or PACTC.                               */
/* Return UNKNOWN if server is not connected.                              */
/*-------------------------------------------------------------------------*/
- (architecture) serverArch;

/*-------------------------------------------------------------------------*/
/* archString:                                                             */
/* Return a pointer to a string containing a textual description of the    */
/* given architecture.                                                     */
/*-------------------------------------------------------------------------*/
- (char *) archString: (architecture) arch;

/*-------------------------------------------------------------------------*/
/* lastServerName:                                                         */
/* Return a pointer to a string containing the name of the host on which   */
/* the server has been tried to be started. This method is usefull when    */
/* a connection failed.                                                    */
/*-------------------------------------------------------------------------*/
- (char *) lastServerName;

/*-------------------------------------------------------------------------*/
/* lastServerArch:                                                         */
/* Return the architecture of the host on which the server has been tried  */
/* to be started. This method is usefull when a connection failed.         */
/*-------------------------------------------------------------------------*/
- (architecture) lastServerArch;

/*-------------------------------------------------------------------------*/
/* lastErrorString:                                                        */
/* Return a pointer to a string containing the description of the error    */
/* which occurs during the last connection attempt if it failed.           */
/*-------------------------------------------------------------------------*/
- (char *) lastErrorString;

/*-------------------------------------------------------------------------*/
/* sendChannel:                                                            */
/* Send the given data to the server.                                      */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value>                             */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    [server sendChannel: INT, a, FLOAT_ARRAY, 100, b, END];              */
/*-------------------------------------------------------------------------*/
- sendChannel: (int)firstParam, ...;

/*-------------------------------------------------------------------------*/
/* receiveChannel:                                                         */
/* Get the given data from the server.                                     */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value address>                     */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    [server receiveChannel: INT, &a, FLOAT_ARRAY, 100, b, END];          */
/*-------------------------------------------------------------------------*/
- receiveChannel: (int)firstParam, ...;

/*-------------------------------------------------------------------------*/
/* toServerChannel:                                                        */
/* Return the channel identifier to the server.                            */
/*-------------------------------------------------------------------------*/
- (channel) toServerChannel;

/*-------------------------------------------------------------------------*/
/* fromServerChannel:                                                      */
/* Return the channel identifier from the server.                          */
/*-------------------------------------------------------------------------*/
- (channel) fromServerChannel;

/*-------------------------------------------------------------------------*/
/* installChannelHandler:                                                  */
/* Install the given channel handler for the server.                       */
/* The handler must be declared as follows:                                */
/*    - channelHandler: server;                                            */
/* The handler is installed as follows:                                    */
/*    [server installChannelHandler: @selector(channelHandler:)];          */
/* Returns nil if a handler is already installed.                          */
/*-------------------------------------------------------------------------*/
- installChannelHandler: sender
  handler: (SEL)handlerP;

/*-------------------------------------------------------------------------*/
/* removeChannelHandler:                                                   */
/* Remove the actual channel handler if defined.                           */
/*-------------------------------------------------------------------------*/
- removeChannelHandler;

@end
