/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Client-Server Library.                                                  */
/* Server Class Implementation.                                            */
/* This class encapsulates a server into a single Objective-C class. The   */
/* client application views a server as an object to which it sends and    */
/* from which it receives message. In addition, the server class can       */
/* use a login window to allows the user of the application to select the  */
/* server host interactively.                                              */
/* The user of the class must allocate and initialize a PPCServer object   */
/* before connection.                                                      */
/*                                                                         */
/* The following methods are available:                                    */
/* connectToServer	Connect server                                     */
/* disconnectFromServer	Disconnect server                                  */
/* sendChannel		Send data to server                                */
/* receiveChannel	Receive data from server                           */
/* isConnected		Is server connected ?                              */
/* serverHost		Return the current server host name                */
/* serverArch		Return the current server host architecture        */
/* archString		Return a string corresponding to an architecture   */
/* lastServerName	Return last host name where connection failed      */
/* lastServerArch	Return last host architecture where conn. failed   */
/* lastErrorString	Return last error textual description              */
/* toServerChannel      Return channel to server                           */
/* fromServerChannel	Return channel from server                         */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 09.07.1994  Droux       Created                                         */
/*-------------------------------------------------------------------------*/
/* filename: PPCServer.m                                                   */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#import <syslog.h>
#import "PPCServer.h"
#import "LoginWindow.h"
#include "pclichan.h"

@implementation PPCServer

/*-------------------------------------------------------------------------*/
/* The following variables are used to communicate channel handler         */
/* parameters to the callHandler function. As this function is not an      */
/* Objective-C method, it is impossible for it to access instance          */
/* variables.                                                              */
/*-------------------------------------------------------------------------*/
SEL chanHandler;                              /* installed channel handler */
id handlerTarget;                           /* handler message destination */
id mySelf;                                               /* equals to self */

/*-------------------------------------------------------------------------*/
/* callHandler:                                                            */
/* This C function is called by the Display Poscript system when data      */
/* arrives on the file descriptor associated with the server receive       */
/* channel if a channel handler has been installed. It calls calls the     */
/* installed channel handler.                                              */
/*-------------------------------------------------------------------------*/
void callHandler (int fd, void *data)
{
  [handlerTarget perform: chanHandler with: mySelf];
}

/*-------------------------------------------------------------------------*/
/* init:                                                                   */
/* Initialize server instance.                                             */
/*-------------------------------------------------------------------------*/
- init
{
  [super init];
  isConnected = NO;
  return self;
}

/*-------------------------------------------------------------------------*/
/* connectToServer:                                                        */
/* Establish a connection with a server.                                   */
/* There are two mode of operation for this function depending on the      */
/* value of useLoginWindow:                                                */
/* 1. useLoginWindow = YES                                                 */
/*    Tell connect to use the object loginWindow to select the server      */
/*    host. In this case, loginWindow must be initialized to a             */
/*    LoginWindow object. This object is the login window which is created */
/*    when using the LoginPalette. The object which calls this method      */
/*    would have an outlet initialized to this window which it could       */
/*    pass as loginWindow. In this mode, the parameters serverHost,        */
/*    serverArch, and password are not used.                               */
/* 2. useLoginWindow = NO                                                  */
/*    Tell connect to use the parameters serverHost, serverArch and        */
/*    password to establish the connection with the server host as with    */
/*    the connect_to_server() function.                                    */
/* progName corresponds to the name of the server program to be started.   */
/* If errorPanel is set to YES, then an error panel will appears if an     */
/* error occured during the connection.                                    */
/*-------------------------------------------------------------------------*/
- connectToServer: (char *)progName                 /* server program name */
  loginWindow: loginW                               /* login window object */
  useLoginWindow: (BOOL)useLW          /* YES if login window must be used */
  serverHost: (char *)host                             /* server host name */
  serverArch: (architecture)arch                    /* server architecture */
  password: (char *)passwd                            /* optional password */
  errorPanel: (BOOL)errorP;                      /* use error panel or not */
{
  char localPw [100];
  
  if (isConnected)
    [self disconnectFromServer];
  
  if (useLW)
  {                                 /* the login window has been requested */
    if ([loginW initHosts] == nil)
    {                     /* an error occured while parsing the hosts list */
      if (errorP)
        NXRunAlertPanel("PPC Hosts List Error",
	  ppc_errlist[ppc_errno],
	  NULL, NULL, NULL);
      return nil;         
    }
    
    /* Enter in modal event loop for login window. The next message will   */
    /* cause loginWindow to show up. It will return only when the "Cancel" */
    /* or "Login" buttons of the window will be pressed. This allows the   */
    /* application to wait until the user select the server host to be     */
    /* used.                                                               */
    [NXApp runModalFor: loginW];                 /* enter modal event loop */
    [loginW orderOut: self];                /* hide window from the screen */
    
    if ([loginW canceled])
    {
      ppc_errno = EPPC_LPANEL_CANCELED;
      if (errorP)
        NXRunAlertPanel("PPC Connection Error",
	  ppc_errlist[ppc_errno],
	  NULL, NULL, NULL);
      return nil;
    }
    
    strcpy(hostName, [loginW selHostName]);
    hostArch = [loginW selHostArch];
    strcpy(localPw, [loginW selPw]);
  }
  else
  {                              /* do not use login window but parameters */
    strcpy(hostName, host);
    hostArch = arch;
    strcpy(localPw, passwd);
  }
  
  sessionId = connect_to_server(            /* start and connect to server */
    &toServerChan,
    &fromServerChan,
    progName,
    hostName,
    hostArch,
    localPw);
    
  if (sessionId == -1)
  {                                                   /* connection failed */
    if (errorP)
    {
      NXRunAlertPanel("PPC Connection Error",
	"Connection to %s (%s) failed:\n%s",
	NULL, NULL, NULL,
	last_server_host_name(),
	arch_string(last_server_arch()),
	ppc_errlist[ppc_errno]);
    }
    return nil;
  }
  
  isConnected = YES;                          /* update instance variables */
  handlerInstalled = NO;
  return self;
}

/*-------------------------------------------------------------------------*/
/* disconnectFromServer:                                                   */
/* Disconnect server.                                                      */
/*-------------------------------------------------------------------------*/
- disconnectFromServer
{
  if (!isConnected)
    return self;
  if (handlerInstalled)
    [self removeChannelHandler];
  disconnect_from_server(sessionId);    
  isConnected = NO;
  return self;
}

/*-------------------------------------------------------------------------*/
/* isConnected:                                                            */
/* Return YES if the server is connected, NO otherwise.                    */
/*-------------------------------------------------------------------------*/
- (BOOL) isConnected
{
  return isConnected;
}

/*-------------------------------------------------------------------------*/
/* serverHost:                                                             */
/* Return a pointer to the string containing the name of the server host.  */
/* Return NULL if server is not connected.                                 */
/*-------------------------------------------------------------------------*/
- (char *) serverHost
{
  if (!isConnected)
    return NULL;
  return hostName;
}

/*-------------------------------------------------------------------------*/
/* serverArch:                                                             */
/* Return the architecture of the host on which server is running.         */
/* Possible values are: UNIX, PVM, or PACTC.                               */
/* Return UNKNOWN if server is not connected.                              */
/*-------------------------------------------------------------------------*/
- (architecture) serverArch
{
  if (!isConnected)
    return UNKNOWN;
  return hostArch;
}

/*-------------------------------------------------------------------------*/
/* archString:                                                             */
/* Return a pointer to a string containing a textual description of the    */
/* given architecture.                                                     */
/*-------------------------------------------------------------------------*/
- (char *) archString: (architecture) arch
{
  return arch_string(arch);
}

/*-------------------------------------------------------------------------*/
/* lastServerName:                                                         */
/* Return a pointer to a string containing the name of the host on which   */
/* the server has been tried to be started. This method is usefull when    */
/* a connection failed.                                                    */
/*-------------------------------------------------------------------------*/
- (char *) lastServerName
{
  if (isConnected)
    return hostName;
  return last_server_host_name();
}

/*-------------------------------------------------------------------------*/
/* lastServerArch:                                                         */
/* Return the architecture of the host on which the server has been tried  */
/* to be started. This method is usefull when a connection failed.         */
/*-------------------------------------------------------------------------*/
- (architecture) lastServerArch
{
  if (isConnected)
    return hostArch;
  return last_server_arch();
}

/*-------------------------------------------------------------------------*/
/* lastErrorString:                                                        */
/* Return a pointer to a string containing the description of the error    */
/* which occurs during the last connection attempt if it failed.           */
/*-------------------------------------------------------------------------*/
- (char *) lastErrorString
{
  return ppc_errlist[ppc_errno];
}

/*-------------------------------------------------------------------------*/
/* sendChannel:                                                            */
/* Send the given data to the server.                                      */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value>                             */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    [server sendChannel: INT, a, FLOAT_ARRAY, 100, b, END];              */
/*-------------------------------------------------------------------------*/
- sendChannel: (int)firstParam, ...
{
  #define SEND_INIT \
    va_start(pvar, firstParam); \
    act_type = firstParam; \
    chan_id = toServerChan;
  #include "clichan.send"
  return self;
}

/*-------------------------------------------------------------------------*/
/* receiveChannel:                                                         */
/* Get the given data from the server.                                     */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value address>                     */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    [server receiveChannel: INT, &a, FLOAT_ARRAY, 100, b, END];          */
/*-------------------------------------------------------------------------*/
- receiveChannel: (int)firstParam, ...
{
  #define RECEIVE_INIT \
    va_start(pvar, firstParam); \
    act_type = firstParam; \
    chan_id = fromServerChan;
  #include "clichan.receive"
  return self;
}

/*-------------------------------------------------------------------------*/
/* toServerChannel:                                                        */
/* Return the channel identifier to the server.                            */
/*-------------------------------------------------------------------------*/
- (channel) toServerChannel
{
  return toServerChan;
}

/*-------------------------------------------------------------------------*/
/* fromServerChannel:                                                      */
/* Return the channel identifier from the server.                          */
/*-------------------------------------------------------------------------*/
- (channel) fromServerChannel
{
  return fromServerChan;
}

/*-------------------------------------------------------------------------*/
/* installChannelHandler:                                                  */
/* Install the given channel handler for the server.                       */
/* The handler must be declared as follows:                                */
/*    - channelHandler: server;                                            */
/* The handler is installed as follows:                                    */
/*    [server installChannelHandler: @selector(channelHandler:)];          */
/* Returns nil if a handler is already installed.                          */
/*-------------------------------------------------------------------------*/
- installChannelHandler: sender
  handler: (SEL)handlerP
{
  if (!isConnected)
    return self;
  if (handlerInstalled)
    return nil;
    
  chanHandler = handlerP;
  handlerTarget = sender;
  mySelf = self;

  DPSAddFD(
    channel_socket(fromServerChan), 
    callHandler, 
    NULL,
    NX_BASETHRESHOLD);
    
  handlerInstalled = YES;
  return self;
}
/*-------------------------------------------------------------------------*/
/* removeChannelHandler:                                                   */
/* Remove the actual channel handler if defined.                           */
/*-------------------------------------------------------------------------*/
- removeChannelHandler
{
  if (!handlerInstalled)
    return self;
    
  DPSRemoveFD(channel_socket(fromServerChan));
    
  handlerInstalled = NO;
  return self;
}

@end
