/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Client-Server Library.                                                  */
/* Client channels implementation. The client channels are provided        */
/* to allows non-PPC custom application to communicate with a PPC          */
/* program using PPC-like channels.                                        */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 08.02.1994  Droux       Created                                         */
/* 17.02.1994  Droux       va_arg(char|short|float) bug fixed              */
/* 22.06.1994  Droux       Added channel_socket()                          */
/* 08.07.1994  Droux       New structure of source directory               */
/* 12.07.1994  Droux       clichan splitted into three parts.              */
/* 13.07.1994  Droux       Added libc.h inclusion under NEXTSTEP           */
/* 30.09.1994  Droux       Suppressed syslog.h for Solaris 2 compatibility */
/*-------------------------------------------------------------------------*/
/* filename: clichan.c                                                     */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*   SunOS Release 5.3 (Solaris 2), SPARC                                  */
/*   Linux 0.99.pl9, Intel Computers                                       */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define clichan_impl

#if defined(NEXT) || defined(NEXTINTEL)
  #include <libc.h>
#endif
#include <varargs.h>
#include <stdio.h>
#include <rpc/rpc.h>
#include <ppc/cslib.h>
#include <ppc/cslib/error.h>
#include "pclichan.h"

typedef enum
{              /* specifies the direction of the channel when creating XDR */
  INPUT,
  OUTPUT
} direction;

cli_chan_descr cli_chans [MAX_CS_CHANS];                /* client channels */

static int cli_chan_init = 0;                  /* module initialized ? 1:0 */

/*-------------------------------------------------------------------------*/
/* cli_chan_init_it:                                                 LOCAL */
/* Init the client channels module.                                        */
/*-------------------------------------------------------------------------*/
void cli_chan_init_it (void)
{
  int c;
  
  for(c = 0; c <  MAX_CS_CHANS; c++)
    cli_chans[c].free = 1;
  cli_chan_init = 1;
}

/*-------------------------------------------------------------------------*/
/* init_cli_chan:                                                    LOCAL */
/* Init a new client channel.                                              */
/* Returns the channel identifer or -1.                                    */
/*-------------------------------------------------------------------------*/
channel init_cli_chan (
  int socket,
  direction dir
)
{
  enum xdr_op x_op;                                       /* XDR direction */
  int c;                                                     /* channel id */
  char stream_mode [2];                    /* mode when opening new stream */
                                /* look for a free  slot in channels table */
  for (c = 0; c < MAX_CS_CHANS && !cli_chans[c].free; c++);
  if (c == MAX_CS_CHANS)
  {
    ppc_errno = EPPC_CHAN_OUTOF;
    return -1;
  }
  
  if (dir == INPUT)     /* determine XDR direction and stream opening mode */
  {
    x_op = XDR_DECODE;
    strcpy(stream_mode, "r");
  }
  else
  {
    x_op = XDR_ENCODE;
    strcpy(stream_mode, "w");
  }
  
  cli_chans[c].free = 0;                              /* mark slot as busy */
  cli_chans[c].socket = socket;
  cli_chans[c].stream = fdopen(socket, stream_mode);
  
  xdrstdio_create(            /* create XDR based client-server connection */
    &(cli_chans[c].xdrs),                                    /* XDR handle */
    cli_chans[c].stream,                                      /* IO stream */
    x_op);

  return c;
}

/*-------------------------------------------------------------------------*/
/* new_client_channel:                                                     */
/* Create two new channels and initialize them according to the given      */
/* sockets. The new channels ids are returned in from_server and           */
/* to_server. Returns -1 if an error occured, 0 otherwise.                 */
/*-------------------------------------------------------------------------*/
int new_client_channel(
  int fs_socket, 
  int ts_socket, 
  channel *from_server, 
  channel *to_server
)
{
  if (!cli_chan_init) cli_chan_init_it();
  
  if ((*from_server = init_cli_chan(fs_socket, INPUT)) == -1)
    return -1;
  if ((*to_server = init_cli_chan(ts_socket, OUTPUT)) == -1)
    return -1;
    
  return 0;
}

/*-------------------------------------------------------------------------*/
/* close_client_channels:                                                  */
/* Close the given client channels.                                        */
/*-------------------------------------------------------------------------*/
void close_client_channels(
  channel from_server,
  channel to_server
)
{
  cli_chans[from_server].free = 1;
  cli_chans[to_server].free = 1;
  fclose(cli_chans[from_server].stream);
  fclose(cli_chans[to_server].stream);
  close(cli_chans[from_server].socket);
  close(cli_chans[to_server].socket);
  xdr_destroy(&(cli_chans[from_server].xdrs));
  xdr_destroy(&(cli_chans[to_server].xdrs));
}

/*-------------------------------------------------------------------------*/
/* send_channel:                                                           */
/* Send the given data on the given channel.                               */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value>                             */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    send_channel(c, INT, a, FLOAT_ARRAY, 100, b, END);                   */
/*-------------------------------------------------------------------------*/
void send_channel (va_alist)
va_dcl
{
  #define SEND_INIT \
    va_start(pvar); \
    chan_id = va_arg(pvar, channel); \
    act_type = va_arg(pvar, int);
  #include "clichan.send"
}

/*-------------------------------------------------------------------------*/
/* receive_channel:                                                        */
/* Get the given data from the specified channel.                          */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value address>                     */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    receive_channel(c, INT, &a, FLOAT_ARRAY, 100, b, END);               */
/*-------------------------------------------------------------------------*/
void receive_channel (va_alist)
va_dcl
{
  #define RECEIVE_INIT \
    va_start(pvar); \
    chan_id = va_arg(pvar, channel); \
    act_type = va_arg(pvar, int);
  #include "clichan.receive"
}

/*-------------------------------------------------------------------------*/
/* channel_socket:                                                         */
/* Return the socket associated with the given channel.                    */
/*-------------------------------------------------------------------------*/
int channel_socket (
  channel chan
)
{
  return cli_chans[chan].socket;
}
