/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Client-Server Library.                                                  */
/* Hosts file list management. Every manipulation of the hosts list must   */
/* be done via this module.                                                */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 08.02.1994  Droux       Created                                         */
/* 08.07.1994  Droux       New structure of source directory               */
/* 09.07.1994  Droux       Added MAX_HOST_NAME_LEN                         */
/* 10.07.1994  Droux       hlist_read() returns error if hlist not present */
/* 10.07.1994  Droux       hlist_read() can read hlist more than 1 time    */
/* 11.07.1994  Droux       Added MAX_LO_LEN and MAX_EP_LEN                 */
/* 11.07.1994  Droux       Added hlist_nb_hosts()                          */
/* 11.07.1994  Droux       nb_arch in .c replaced by NB_ARCH in .h         */
/* 12.07.1994  Droux       hlist_read() returns error if no host in hlist  */
/* 13.07.1994  Droux       Added libc.h inclusion under NEXTSTEP           */
/*-------------------------------------------------------------------------*/
/* filename: hlist.c                                                       */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*   SunOS Release 5.3 (Solaris 2), SPARC                                  */
/*   Linux 0.99.pl9, Intel Computers                                       */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#if defined(NEXT) || defined(NEXTINTEL)
  #include <libc.h>
#endif
#include <stdio.h>
#include <ppc/csp/types.h>
#include <ppc/cslib/error.h>
#include <ppc/cslib/hlist.h>

#define HOSTS_FILE_NAME "ppchosts"                 /* hosts list file name */
#define MAX_LINE_LEN 200             /* max characters per hosts list line */

typedef struct
{
  char host_name[MAX_HOST_NAME_LEN];                          /* host name */
  architecture host_arch;                             /* host architecture */
  char lo[MAX_LO_LEN];                            /* alternative user name */
  int pw;                           /* 1 if password required, 0 otherwise */
  char ep[MAX_EP_LEN];                                  /* executable path */
} host_descr;

static int init = 0;                            /* 1 if module initialized */
static int nb_entries;                  /* number of entries in hosts list */
host_descr hosts [MAX_HOSTS];          /* hosts as recorded in hosts liste */
static char *priv_arch_string[] =   /* arch strings as stored in host file */
{                                        /* eg. arch_string[UNIX] = "unix" */
  "unix",   /* indices in arch_string must correspond to architecture type */
  "pvm",
  "pactc",
  "unknown"
};

/*-------------------------------------------------------------------------*/
/* arch_string:                                                            */
/* Return a pointer to the string corresponding to the given architecture. */
/* For example, arch_string(UNIX) will return a pointer to a string        */
/* containing "UNIX".                                                      */
/*-------------------------------------------------------------------------*/
char *arch_string(architecture arch)
{
  return priv_arch_string[arch];
}

/*-------------------------------------------------------------------------*/
/* blank_char:                                                       LOCAL */
/* Return 1 if the given char is a blank char, 0 if it isn't.              */
/*-------------------------------------------------------------------------*/
int blank_char (char c)
{
  return (c == '\t') || (c == ' ');
}

/*-------------------------------------------------------------------------*/
/* get_next_token:                                                   LOCAL */
/* Get the next token from the given string. Returns 1 if end of line      */
/* reached.                                                                */
/*-------------------------------------------------------------------------*/
int get_next_token(
  int *c,             /* counter, must be set to 0 before starting parsing */
  char str[],                                               /* line string */
  char token[]                                           /* returned token */
)
{
  int tk_len;                                              /* token length */
  
  while(blank_char(str[*c]))                                /* skip blanks */
    (*c)++;
    
  if (str[*c] == '\n' || str[*c] == '\0')
    return 1;                                       /* end of line reached */

  for (tk_len = 0; !blank_char(str[*c]) && str[*c] != '\0'; (*c)++, tk_len++)
    token[tk_len] = str[*c];
  
  token[tk_len] = '\0';                                /* terminate string */
  return 0;
}

/*-------------------------------------------------------------------------*/
/* get_next_line:                                                    LOCAL */
/* Get the next line from the given file. Returns 1 if the end of line     */
/* is reached.                                                             */
/*-------------------------------------------------------------------------*/
int get_next_line(
  FILE *file,
  char line[],
  int first_line                  /* true if first line of look-ahead read */
)
{
  int idx;
  static int act_char;
  
  if (!first_line)
  {
    if (act_char == EOF)
      return 1;
  }
  
  act_char = fgetc(file);
  for (idx = 0; act_char != EOF && act_char != '\n'; idx++)
  {
    line[idx] = (char)act_char;
    act_char = fgetc(file);
  }
  line[idx] = '\0';
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* read_hosts_list:                                                        */
/*                                                                         */
/* Read the hosts list file. This function is automatically called during  */
/* the first host interrogation.                                           */
/*                                                                         */
/* Return a negative value for unsuccessful call and stores an error code  */
/* in errno, otherwise returns zero. Possible errors are:                  */
/*                                                                         */
/*   [EPPC_HLIST_UNKARCH]	unknown architecture specifier in hosts    */
/*				list.                                      */
/*   [EPPC_HLIST_LOEQ] 		missing '=' for \"lo\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_LOVAL]		missing value for \"lo\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_EPEQ]		missing '=' for \"ep\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_EPVAL]		missing value for \"ep\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_UNKOPT]	unknown option in hosts list.              */
/*   [EPPC_HLIST_ITMEXC]	too many items in hosts list line.         */
/*   [EPPC_HLIST_PARTDEF]	host only partially defined in hosts list. */
/*                                                                         */
/*-------------------------------------------------------------------------*/
int hlist_read ()
{
  char act_line [MAX_LINE_LEN];                  /* actual line being read */
  char file_name [50];                                  /* hosts file name */
  FILE *hl;                                             /* hosts list file */
  int eofile;                                  /* 1 if end of file reached */
  int c;                           /* actual char in the line being parsed */
  int eoline;                                  /* 1 if end of line reached */
  char token[MAX_LINE_LEN];                         /* actual word in line */
  int nb_tokens;                 /* number of tokens parsed in actual line */
  int act_arch;                        /* actual architecture being tested */
  char act_option[50];                       /* actual option being parsed */
  char *option_value;                               /* actual option value */
  int host_ok;               /* 1 if host name has been completely defined */
  
  init = 1;
  nb_entries = 0;
  
  sprintf(file_name, "%s/" HOSTS_FILE_NAME, getenv("HOME"));
  if ((hl = fopen(file_name, "r")) == NULL)
  {                                             /* try hidden file version */
    sprintf(file_name, "%s/." HOSTS_FILE_NAME, getenv("HOME"));
    if ((hl = fopen(file_name, "r")) == NULL)
    {                                              /* hosts list not found */
      ppc_errno = EPPC_HLIST_NOHLIST;
      return -1;
    }
  }
   
  eofile = get_next_line(hl, act_line, 1);  /* parse file using look-ahead */
  while(!eofile)                    
  {
    hosts[nb_entries].pw = 0;            /* initializes current host entry */
    strcpy(hosts[nb_entries].lo, "");
    strcpy(hosts[nb_entries].ep, "");
    host_ok = 0;
    
    c = 0;                    /* parse actual line using a look-ahead read */
    eoline = get_next_token(&c, act_line, token);
    nb_tokens = 0;
    while (!eoline)
    {
      if (token[0] == '#')
        eoline = 1;                     /* end of line detected on comment */
      else
      {
        switch(nb_tokens)
	{
	  case 0:               /* the current token must be the host name */
	    strcpy(hosts[nb_entries].host_name, token);
	    break;
	    
	  case 1:       /* the current token must be the host architecture */
	    for (act_arch = 0; 
	         act_arch < NB_ARCH && 
		 strcmp(token, priv_arch_string[act_arch]);
		 act_arch++);
	    if (act_arch == NB_ARCH)
	    {              /* unknown architecture specifier in hosts list */
	      ppc_errno = EPPC_HLIST_UNKARCH; 
	      return -1; 
	    }
	    hosts[nb_entries].host_arch = act_arch;
	    host_ok = 1;
	    break;
	    
	  case 2:                       /* we have here one of the options */
	  case 3:
	  case 4:
	    strcpy(act_option, token);
	    act_option[2] = '\0';    /* each option is two characters wide */
	    option_value = &token[3];
	    
	    if (!strcmp(act_option, "lo"))
	    {                                     /* alternative user name */
              if (token[2] != '=')
	      {            /* missing '=' for \"lo\" option in hosts list. */
	        ppc_errno = EPPC_HLIST_LOEQ;
		return -1;
	      }
              strcpy(hosts[nb_entries].lo, option_value);
	      if (!strlen(hosts[nb_entries].lo))
	      {          /* missing value for \"lo\" option in hosts list. */
	        ppc_errno = EPPC_HLIST_LOVAL;
		return -1;
	      }
	    }
	    else if (!strcmp(act_option, "pw"))
	    {                                        /* mandatory password */
	      hosts[nb_entries].pw = 1;
	    }
	    else if (!strcmp(act_option, "ep"))
	    {                           /* alternative program search path */
              if (token[2] != '=')
	      {            /* missing '=' for \"ep\" option in hosts list. */
	        ppc_errno = EPPC_HLIST_EPEQ;
		return -1;
	      }
              strcpy(hosts[nb_entries].ep, option_value);
	      if (!strlen(hosts[nb_entries].ep))
	      {          /* missing value for \"ep\" option in hosts list. */
	        ppc_errno = EPPC_HLIST_EPVAL;
		return -1;
	      }
	    }
	    else
	    {                             /* unknown option in hosts list. */
	      ppc_errno = EPPC_HLIST_UNKOPT;
	      return -1;
	    }
	    break;
	    
	  default:
	  {                          /* too many items in hosts list line. */
	    ppc_errno = EPPC_HLIST_ITMEXC;
	    return -1;
	  }
	}
	nb_tokens++;
        eoline = get_next_token(&c, act_line, token);
      }
    }    
    
    if (nb_tokens > 0)
    {                          /* something was defined on the actual line */
      if (!host_ok)
      {                      /* host only partially defined in hosts list. */
        ppc_errno = EPPC_HLIST_PARTDEF;
	return -1;
      }
      #if DEBUG
      printf("> host %d registered:\n", nb_entries);
      printf("  host_name = %s\n", hosts[nb_entries].host_name);
      printf("  host_arch = %s\n", 
        priv_arch_string[hosts[nb_entries].host_arch]);
      printf("  lo        = ");
      if (strlen(hosts[nb_entries].lo))
        printf("%s\n", hosts[nb_entries].lo);
      else
        printf("(none)\n");
      printf("  pw        = ");
      if (hosts[nb_entries].pw)
        printf("YES\n");
      else
        printf("NO\n");
      printf("  ep        = ");
      if (strlen(hosts[nb_entries].ep))
        printf("%s\n", hosts[nb_entries].ep);
      else
        printf("(none)\n");
      #endif
      nb_entries++;
    }
    eofile = get_next_line(hl, act_line, 0);/* parse file using look-ahead */
  }
  
  fclose(hl);
  
  if (!nb_entries)
  {                                      /* no hosts defined in hosts list */
    ppc_errno = EPPC_HLIST_NOHOSTDEF;
    return -1;
  }
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* hlist_find_host_name:                                                   */
/*                                                                         */
/* Finds a host of the given architecture and returns its id.              */
/*                                                                         */
/* Return a negative value for unsuccessful call and stores an error code  */
/* in errno. Possible errors are:                                          */
/*                                                                         */
/*   [EPPC_HLIST_UNKARCH]	unknown architecture specifier in hosts    */
/*				list.                                      */
/*   [EPPC_HLIST_LOEQ] 		missing '=' for \"lo\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_LOVAL]		missing value for \"lo\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_EPEQ]		missing '=' for \"ep\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_EPVAL]		missing value for \"ep\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_UNKOPT]	unknown option in hosts list.              */
/*   [EPPC_HLIST_ITMEXC]	too many items in hosts list line.         */
/*   [EPPC_HLIST_PARTDEF]	host only partially defined in hosts list. */
/*   [EPPC_HLIST_ARCHNFND]	specified architecture not found in hosts  */
/*                              list.                                      */
/*                                                                         */
/*-------------------------------------------------------------------------*/
int hlist_find_host_name (architecture arch)
{
  int idx;                                    /* host index to be returned */
  
  if (!init)
  {                                                     /* read hosts list */
    if (hlist_read() == -1)
    {                     /* an error occured while reading the hosts list */
      return -1;
    }
  }
  
  for (idx = 0; idx < nb_entries && hosts[idx].host_arch != arch; idx++);
  
  if (idx == nb_entries)
  {                     /* specified architecture not found in hosts list. */
    ppc_errno = EPPC_HLIST_ARCHNFND;
    return -1;
  }
  
  return idx;
}

/*-------------------------------------------------------------------------*/
/* hlist_find_host_arch:                                                   */
/*                                                                         */
/* Finds a host of the given name and returns its id.                      */
/*                                                                         */
/* Return a negative value for unsuccessful call and stores an error code  */
/* in errno. Possible errors are:                                          */
/*                                                                         */
/*   [EPPC_HLIST_UNKARCH]	unknown architecture specifier in hosts    */
/*				list.                                      */
/*   [EPPC_HLIST_LOEQ] 		missing '=' for \"lo\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_LOVAL]		missing value for \"lo\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_EPEQ]		missing '=' for \"ep\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_EPVAL]		missing value for \"ep\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_UNKOPT]	unknown option in hosts list.              */
/*   [EPPC_HLIST_ITMEXC]	too many items in hosts list line.         */
/*   [EPPC_HLIST_PARTDEF]	host only partially defined in hosts list. */
/*   [EPPC_HLIST_HOSTNFND]	specified host name not found in hosts     */
/*                              list.                                      */
/*                                                                         */
/*-------------------------------------------------------------------------*/
int hlist_find_host_arch (char *host_name)
{
  int idx;                                    /* host index to be returned */
  
  if (!init)
  {                                                     /* read hosts list */
    if (hlist_read() == -1)
    {                     /* an error occured while reading the hosts list */
      return -1;
    }
  }
  
  for (idx = 0; 
       idx < nb_entries && strcmp(hosts[idx].host_name, host_name); 
       idx++);
  
  if (idx == nb_entries)
  {                        /* specified host name not found in hosts list. */
    ppc_errno = EPPC_HLIST_HOSTNFND;
    return -1;
  }
  
  return idx;
}

/*-------------------------------------------------------------------------*/
/* hlist_find_host_name_arch:                                              */
/*                                                                         */
/* Finds a host of the given name AND architecture and returns its id.     */
/*                                                                         */
/* Return a negative value for unsuccessful call and stores an error code  */
/* in errno. Possible errors are:                                          */
/*                                                                         */
/*   [EPPC_HLIST_UNKARCH]	unknown architecture specifier in hosts    */
/*				list.                                      */
/*   [EPPC_HLIST_LOEQ] 		missing '=' for \"lo\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_LOVAL]		missing value for \"lo\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_EPEQ]		missing '=' for \"ep\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_EPVAL]	missing value for \"ep\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_UNKOPT]	unknown option in hosts list.              */
/*   [EPPC_HLIST_ITMEXC]	too many items in hosts list line.         */
/*   [EPPC_HLIST_PARTDEF]	host only partially defined in hosts list. */
/*   [EPPC_HLIST_HOSTARCHNFND]	specified host name with given             */
/*                              architecture not found in hosts list.      */
/*                                                                         */
/*-------------------------------------------------------------------------*/
int hlist_find_host_name_arch (char *host_name, architecture arch)
{
  int idx;                                    /* host index to be returned */
  
  if (!init)
  {                                                     /* read hosts list */
    if (hlist_read() == -1)
    {                     /* an error occured while reading the hosts list */
      return -1;
    }
  }
  
  for (idx = 0; 
       (idx < nb_entries) 
       && ((strcmp(hosts[idx].host_name, host_name))
       || (hosts[idx].host_arch != arch)); 
       idx++);
  
  if (idx == nb_entries)
  {/* specified host name with given architecture not found in hosts list. */
    ppc_errno = EPPC_HLIST_HOSTARCHNFND;
    return -1;
  }
  
  return idx;
}

/*-------------------------------------------------------------------------*/
/* hlist_get_host_name:                                                    */
/*                                                                         */
/* Copy in the string str the host name corresponding to the given         */
/* host id.                                                                */
/*                                                                         */
/* Return a negative value for unsuccessful call and stores an error code  */
/* in errno, otherwise returns zero. Possible errors are:                  */
/*                                                                         */
/*   [EPPC_HLIST_UNKARCH]	unknown architecture specifier in hosts    */
/*				list.                                      */
/*   [EPPC_HLIST_LOEQ] 	missing '=' for \"lo\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_LOVAL]	missing value for \"lo\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_EPEQ]		missing '=' for \"ep\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_EPVAL]	missing value for \"ep\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_UNKOPT]	unknown option in hosts list.              */
/*   [EPPC_HLIST_ITMEXC]	too many items in hosts list line.         */
/*   [EPPC_HLIST_PARTDEF]	host only partially defined in hosts list. */
/*   [EPPC_HLIST_BADIDNA]	bad host id given when inquiring host      */
/*				name, internal error.                      */
/*                                                                         */
/*-------------------------------------------------------------------------*/
int hlist_get_host_name (
  int id, 
  char str[]
)
{
  if (!init)
  {                                                     /* read hosts list */
    if (hlist_read() == -1)
    {                     /* an error occured while reading the hosts list */
      return -1;
    }
  }
  
  if (id < 0 || id >= nb_entries)
  {         /* bad host id given when inquiring host name, internal error. */
    ppc_errno = EPPC_HLIST_BADIDNA;
    return -1;
  }
  
  strcpy(str, hosts[id].host_name);
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* hlist_get_host_name_ptr:                                                */
/* Returns a pointer to the host name corresponding to the given host id,  */
/* NULL if an error occured. Similar to hlist_get_host_name with only a    */
/* different interface.                                                    */
/*-------------------------------------------------------------------------*/
char *hlist_get_host_name_ptr (
  int id
)
{
  if (!init)
  {                                                     /* read hosts list */
    if (hlist_read() == -1)
    {                     /* an error occured while reading the hosts list */
      return NULL;
    }
  }
  
  if (id < 0 || id >= nb_entries)
  {         /* bad host id given when inquiring host name, internal error. */
    return NULL;
  }
  
  return hosts[id].host_name;
}

/*-------------------------------------------------------------------------*/
/* hlist_get_host_arch:                                                    */
/*                                                                         */
/* Copy in arch the architecture corresponding to the given host.          */
/*                                                                         */
/* Return a negative value for unsuccessful call and stores an error code  */
/* in errno, otherwise returns zero. Possible errors are:                  */
/*                                                                         */
/*   [EPPC_HLIST_UNKARCH]	unknown architecture specifier in hosts    */
/*				list.                                      */
/*   [EPPC_HLIST_LOEQ] 		missing '=' for \"lo\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_LOVAL]		missing value for \"lo\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_EPEQ]		missing '=' for \"ep\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_EPVAL]		missing value for \"ep\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_UNKOPT]	unknown option in hosts list.              */
/*   [EPPC_HLIST_ITMEXC]	too many items in hosts list line.         */
/*   [EPPC_HLIST_PARTDEF]	host only partially defined in hosts list. */
/*   [EPPC_HLIST_BADIDAR]	bad host id given when inquiring host      */
/*				architecture, internal error.              */
/*                                                                         */
/*-------------------------------------------------------------------------*/
int hlist_get_host_arch (int id, architecture *arch)
{
  if (!init)
  {                                                     /* read hosts list */
    if (hlist_read() == -1)
    {                     /* an error occured while reading the hosts list */
      return -1;
    }
  }
  
  if (id < 0 || id >= nb_entries)
  { /* bad host id given when inquiring host architecture, internal error. */
    ppc_errno = EPPC_HLIST_BADIDAR;
    return -1;
  }
  
  *arch = hosts[id].host_arch;
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* hlist_pw_required:                                                      */
/*                                                                         */
/* Returns 1 if the given host requires a password when starting a server, */
/* 0 if it does not, and -1 if an error occured.                           */
/*                                                                         */
/* Possible errors are:                                                    */
/*   [EPPC_HLIST_UNKARCH]	unknown architecture specifier in hosts    */
/*				list.                                      */
/*   [EPPC_HLIST_LOEQ] 		missing '=' for \"lo\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_LOVAL]		missing value for \"lo\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_EPEQ]		missing '=' for \"ep\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_EPVAL]		missing value for \"ep\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_UNKOPT]	unknown option in hosts list.              */
/*   [EPPC_HLIST_ITMEXC]	too many items in hosts list line.         */
/*   [EPPC_HLIST_PARTDEF]	host only partially defined in hosts list. */
/*   [EPPC_HLIST_BADIDPW]      bad host id given when inquiring host      */
/*				password necessity, internal error.        */
/*-------------------------------------------------------------------------*/
int hlist_pw_required (
  int id                                                        /* host id */
)
{
  if (!init)
  {                                                     /* read hosts list */
    if (hlist_read() == -1)
    {                     /* an error occured while reading the hosts list */
      return -1;
    }
  }
  
  if (id < 0 || id >= nb_entries)
  {           /* bad host id given when inquiring host password necessity. */
    ppc_errno = EPPC_HLIST_BADIDPW;
    return -1;
  }
  
  return hosts[id].pw;
}

/*-------------------------------------------------------------------------*/
/* hlist_exec_path_required:                                               */
/*                                                                         */
/* Returns 1 if a custom executable file has been defined for the given    */
/* host, 0 if not, and -1 if an error occured.                             */
/*                                                                         */
/* Possible errors are:                                                    */
/*                                                                         */
/*   [EPPC_HLIST_UNKARCH]	unknown architecture specifier in hosts    */
/*				list.                                      */
/*   [EPPC_HLIST_LOEQ] 		missing '=' for \"lo\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_LOVAL]		missing value for \"lo\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_EPEQ]		missing '=' for \"ep\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_EPVAL]		missing value for \"ep\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_UNKOPT]	unknown option in hosts list.              */
/*   [EPPC_HLIST_ITMEXC]	too many items in hosts list line.         */
/*   [EPPC_HLIST_PARTDEF]	host only partially defined in hosts list. */
/*   [EPPC_HLIST_BADIDPR]      bad host id given when inquiring for       */
/*				custom exec path presence.                 */
/*-------------------------------------------------------------------------*/
int hlist_exec_path_required (
  int id
)
{
  if (!init)
  {                                                     /* read hosts list */
    if (hlist_read() == -1)
    {                     /* an error occured while reading the hosts list */
      return -1;
    }
  }
  
  if (id < 0 || id >= nb_entries)
  {     /* bad host id given when inquiring for custom exec path presence. */
    ppc_errno = EPPC_HLIST_BADIDPR;
    return -1;
  }
  
  return (strlen(hosts[id].ep) != 0);
}

/*-------------------------------------------------------------------------*/
/* hlist_exec_path:                                                        */
/*                                                                         */
/* Returns a pointer to a string containing the custom executable          */
/* path for the given host. Returns NULL if not defined or an error        */
/* occured.                                                                */
/*-------------------------------------------------------------------------*/
char *hlist_exec_path (
  int id
)
{
  if (!init)
  {                                                     /* read hosts list */
    if (hlist_read() == -1)
    {                     /* an error occured while reading the hosts list */
      return NULL;
    }
  }
  
  if (id < 0 || id >= nb_entries)
  {        /* bad host id given when inquiring for custom exec path value. */
    return NULL;
  }
  
  return hosts[id].ep;
}

/*-------------------------------------------------------------------------*/
/* hlist_exec_lo_defined:                                                  */
/*                                                                         */
/* Returns 1 if a custom login name has been defined for the given         */
/* host, 0 if not, and -1 if an error occured.                             */
/*                                                                         */
/* Possible errors are:                                                    */
/*                                                                         */
/*   [EPPC_HLIST_UNKARCH]	unknown architecture specifier in hosts    */
/*				list.                                      */
/*   [EPPC_HLIST_LOEQ] 		missing '=' for \"lo\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_LOVAL]		missing value for \"lo\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_EPEQ]		missing '=' for \"ep\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_EPVAL]		missing value for \"ep\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_UNKOPT]	unknown option in hosts list.              */
/*   [EPPC_HLIST_ITMEXC]	too many items in hosts list line.         */
/*   [EPPC_HLIST_PARTDEF]	host only partially defined in hosts list. */
/*   [EPPC_HLIST_BADIDLI]      bad host id given when inquiring for       */
/*				custom exec path presence.                 */
/*-------------------------------------------------------------------------*/
int hlist_exec_lo_defined (
  int id
)
{
  if (!init)
  {                                                     /* read hosts list */
    if (hlist_read() == -1)
    {                     /* an error occured while reading the hosts list */
      return -1;
    }
  }
  
  if (id < 0 || id >= nb_entries)
  {     /* bad host id given when inquiring for custom user name presence. */
    ppc_errno = EPPC_HLIST_BADIDLI;
    return -1;
  }
  
  return (strlen(hosts[id].lo) != 0);
}

/*-------------------------------------------------------------------------*/
/* hlist_custom_name:                                                      */
/*                                                                         */
/* Returns a pointer to a string containing the custom user name           */
/* for the given host. Returns NULL if not defined or an error             */
/* occured.                                                                */
/*-------------------------------------------------------------------------*/
char *hlist_custom_name (
  int id
)
{
  if (!init)
  {                                                     /* read hosts list */
    if (hlist_read() == -1)
    {                     /* an error occured while reading the hosts list */
      return NULL;
    }
  }
  
  if (id < 0 || id >= nb_entries)
  {        /* bad host id given when inquiring for custom exec path value. */
    return NULL;
  }
  
  return hosts[id].lo;
}

/*-------------------------------------------------------------------------*/
/* hlist_nb_hosts:                                                         */
/* Returns the number of hosts declared in the hosts list.                 */
/*-------------------------------------------------------------------------*/
int hlist_nb_hosts (void)
{
  if (!init)
  {                                                     /* read hosts list */
    if (hlist_read() == -1)
    {                     /* an error occured while reading the hosts list */
      return -1;
    }
  }
  
  return nb_entries;
}
