/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Client-Server Library.                                                  */
/* Implementation of the functions allowing the asynchronous processing    */
/* of messages by clients.                                                 */
/* The Functions implemented here are declared in the header file          */
/* cliasync.h                                                              */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 22.06.1994  Droux       Created                                         */
/* 13.07.1994  Droux       Added libc.h inclusion under NEXTSTEP           */
/*-------------------------------------------------------------------------*/
/* filename: cliasync.c                                                    */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*   SunOS Release 5.3 (Solaris 2), SPARC                                  */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#ifdef LINUX
  /* Unfortunately, the latest version of Linux at this date (June 24th    */
  /* 1994) does not support the asychronous mode of sockets. It is thus    */
  /* impossible to use this module under this OS.                          */
  #warning The asynchronous message processing of CS-Lib is not implemented\
  under Linux due to serious limitations of this operating system.
#else
#ifdef SUN4SOL2
  #include <stropts.h>          /* stream options modified under Solaris 2 */
#endif
#if defined(NEXT) || defined(NEXTINTEL)
  #include <libc.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <ppc/cslib.h>
#include "pclichan.h"
#include "pcliconn.h"

/* Under Solaris 2, the SIGIO mechanism is replaced by SIGPOLL. We define  */
/* here the signal which must be used under the current operating system.  */
#ifdef SUN4SOL2
  #define SIGNAL_NR SIGPOLL
#else
  #define SIGNAL_NR SIGIO
#endif

/* The signal management functions have been modified under Solaris 2. We  */
/* define here some macro whose content is selected depending upon the OS  */
/* actually in use.                                                        */
#ifdef SUN4SOL2
  #define DECLARE_SIGNAL
  #define DISABLE_SIGNAL(x) 	sighold(x)
  #define RESTORE_SIGNAL(x) 	sigrelse(x)
  #define SIGNAL(sig, disp)	sigset(sig, disp)
  #define IGNORE_SIGNAL(x)	sigignore(x);
#else
  #define DECLARE_SIGNAL	int old_mask;
  #define DISABLE_SIGNAL(x)	old_mask = sigblock(sigmask(x))
  #define RESTORE_SIGNAL(x)	sigsetmask(old_mask)
  #define SIGNAL(sig, disp)	signal(sig, disp)
  #define IGNORE_SIGNAL(x)	signal(x, SIG_IGN)
#endif

typedef struct /* structure describing the properties of a channel handler */
{
  int free;                                           /* 1 if slot is free */
  int socket;                       /* channel from which the data arrives */
  void (*handler)(int, channel);                           /* user handler */
  channel chan;                                   /* corresponding channel */
  int session;                                 /* corresponding session id */
} handler_type;

static handler_type ppc_handlers [MAX_CHAN_HANDLERS];     /* user handlers */
static int nb_handlers;            /* number of handlers defined until now */
static int cliasync_init = 0;      /* 1 if module initialized, 0 otherwise */

static fd_set global_readfds;          /* sockets to be tested by select() */
static int max_readfd;                /* highest read fd for select() call */
static struct timeval select_timeout;              /* timeout for select() */

/*-------------------------------------------------------------------------*/
/* cliasync_initialize:                                              LOCAL */
/* Initialize the module.                                                  */
/*-------------------------------------------------------------------------*/
void cliasync_initialize (void)
{
  int act_handle;
  
  for (act_handle = 0; act_handle < MAX_CHAN_HANDLERS; act_handle++)
    ppc_handlers[act_handle].free = 1;
  
  nb_handlers = 0;
  
  FD_ZERO(&global_readfds);         /* initialize values for select() call */
  max_readfd = 0;
  select_timeout.tv_sec = 0;
  select_timeout.tv_usec = 0;
  
  cliasync_init = 1;
}

/*-------------------------------------------------------------------------*/
/* ppc_channel_handler:                                              LOCAL */
/* Called automatically when some data is ready on one of the channels     */
/* for which a handler has been installed. The function determines from    */
/* which server the message is arriving and call the corresponding user    */
/* defined channel handler.                                                */
/*-------------------------------------------------------------------------*/
void ppc_channel_handler (
  int sig_nr                           /* signal number, must be SIGNAL_NR */
)
{
  static fd_set readfds;
  static int nb_ready;
  static int last_selected = -1;
  static int id;                                      /* handle identifier */

  readfds = global_readfds;
  
  do 
  {
    /* It is possible that multiple messages may cause a single signal     */
    /* to occur. In this case, we have to perform a loop here while        */
    /* there is something to read on the specified channels. This is       */
    /* also a way of speeding up the calculation a bit if the delays       */
    /* between messages is very short.                                     */
    
    nb_ready = select(max_readfd+1, &readfds, NULL, NULL, &select_timeout);
    
    if (nb_ready < 0)                             /* this is a fatal error */
    {
      syslog(LOG_ERR, "fatal error during PPC asynchronous mode select()");
      exit(1);
    }
    
    if (nb_ready == 0) 
    {                            /* the data does not comes from a channel */
      return;
    }
			   /* determine from where the data actually comes */
    for (last_selected = (last_selected + 1) % (max_readfd + 1);
	  !FD_ISSET(last_selected, &readfds);
	  last_selected = (last_selected + 1) % (max_readfd + 1));
	  
    for (id = 0;         /* determine the slot corresponding to the socket */
	  id < MAX_CHAN_HANDLERS && 
	  (ppc_handlers[id].free || 
	  (ppc_handlers[id].socket!=last_selected)); 
	  id++);
  
    ppc_handlers[id].handler(
      ppc_handlers[id].session,
      ppc_handlers[id].chan);
      
    nb_ready--;
  } while (nb_ready > 0);
}

/*-------------------------------------------------------------------------*/
/* init_for_signal:                                                        */
/* Initialize the specified socket such as a signal (SIGIO or SIGPOLL for  */
/* Solaris) must be send to the process when it receives data.             */
/* Returns -1 if an error occured, 1 otherwise.                            */
/*-------------------------------------------------------------------------*/
int init_for_signal 
(
  int socket                     /* socket on which the data should arrive */
)
{
  #ifdef SUN4SOL2
  {
    ioctl(socket, I_SETSIG, S_INPUT | S_HIPRI);
    
    return 1;
  }
  #else
  {
    int socket_flags;  /* socket flags to be modified to asynchronous mode */
    
    if ((socket_flags = fcntl(socket, F_GETFL, 0)) == -1)
    {
      ppc_errno = EPPC_ASYNC_GETFL;
      return -1;
    }
    
    socket_flags |= FASYNC;                      /* enable asynchronous IO */
  
    if (fcntl(socket, F_SETFL, socket_flags) < 0)        
    {
      ppc_errno = EPPC_ASYNC_FASYNC;
      return -1;
    }
    
    if (fcntl(socket, F_SETOWN, getpid()) < 0)                /* set owner */
    {
      ppc_errno = EPPC_ASYNC_SETOWN;
      return -1;
    }
  
    return 1;
  }
  #endif
}

/*-------------------------------------------------------------------------*/
/* disable_for_signal:                                                     */
/* Restore the state of the socket as it was before applying the function  */
/* init_for_signal on it.                                                  */
/*-------------------------------------------------------------------------*/
void disable_for_signal
(
  int socket                                          /* socket to restore */
)
{
  #ifdef SUN4SOL2
  {
    ioctl(socket, 
      I_SETSIG, ioctl(socket, I_GETSIG, 0) & ~ (S_INPUT | S_HIPRI));
  }
  #else
  {
    int socket_flags;  /* socket flags to be modified to asynchronous mode */

    socket_flags = fcntl(socket, F_GETFL, 0);     /* enable synchronous IO */
    socket_flags &= ~FASYNC;
    fcntl(socket, F_SETFL, socket_flags);        
  }
  #endif
}

/*-------------------------------------------------------------------------*/
/* install_channel_handler:                                                */
/* Install the given function as the handler for the given session.        */
/* When the server program corresponding to the specified session id send  */
/* a message to the client, the function handler() will be automatically   */
/* called. When called, the function will receive two values: the session  */
/* id of the host which just send a message, and the channel from which    */
/* the data can be read.                                                   */
/*-------------------------------------------------------------------------*/
int install_channel_handler (
  int session_id,
  void (*handler)(int, channel)
)
{
  int socket;         /* socket corresponding to the session input channel */
  channel chan;              /* input channel corresponding to the session */
  int id;                                               /* handler slot id */
  DECLARE_SIGNAL;

  DISABLE_SIGNAL(SIGNAL_NR);
  
  if (!cliasync_init) cliasync_initialize();
  
  if (nb_handlers == MAX_CHAN_HANDLERS)
  {
    ppc_errno = EPPC_ASYNC_TOOMANY;
    RESTORE_SIGNAL(SIGNAL_NR);
    return -1;
  }
  
  chan = session_read_channel(session_id);    /* get corresponding channel */
  if (chan == -1)
  {
    ppc_errno = EPPC_ASYNC_BADSE;
    RESTORE_SIGNAL(SIGNAL_NR);
    return -1;
  }
  
  socket = channel_socket(chan);    /* get corresponding socket descriptor */
  
  if (init_for_signal(socket) == -1)
  {
    RESTORE_SIGNAL(SIGNAL_NR);
    return -1;
  }
  
  if ((int)SIGNAL(SIGNAL_NR, ppc_channel_handler) == -1) /* signal catcher */
  {
    ppc_errno = EPPC_ASYNC_SIG;
    RESTORE_SIGNAL(SIGNAL_NR);
    return -1;
  }
                                                   /* look for a free slot */
  for (id = 0; (id < MAX_CHAN_HANDLERS) && (!ppc_handlers[id].free); id++);
  
  ppc_handlers[id].free = 0;                        /* update handler slot */
  ppc_handlers[id].socket = socket;
  ppc_handlers[id].handler = handler;
  ppc_handlers[id].session = session_id;
  ppc_handlers[id].chan = chan;
  
  FD_SET(socket, &global_readfds);   /* put socket in set to be select()ed */
  
  if (socket > max_readfd)         /* update highest read fd for select() */
    max_readfd = socket;

  nb_handlers++;                        /* update total number of handlers */
  
  RESTORE_SIGNAL(SIGNAL_NR);

  return 1;
}

/*-------------------------------------------------------------------------*/
/* remove_channel_handler:                                                 */
/* Removes the channel handler installed for the given session.            */
/*-------------------------------------------------------------------------*/
void remove_channel_handler (
  int session_id
)
{
  int id;                               /* session index in handlers table */
  int socket;                       /* TCP socket corresponding to session */
  DECLARE_SIGNAL;

  DISABLE_SIGNAL(SIGNAL_NR);              /* temporarily disable SIGNAL_NR */
  
  for (id = 0;          /* determine the slot corresponding to the session */
       id < MAX_CHAN_HANDLERS && 
       (ppc_handlers[id].free || (ppc_handlers[id].session != session_id)); 
       id++);
  
  if (id == MAX_CHAN_HANDLERS)
  {                                                   /* session not found */
    return;                                           
    RESTORE_SIGNAL(SIGNAL_NR);
  }
  
  ppc_handlers[id].free = 1;
  nb_handlers--;
    
  if (!nb_handlers)
    IGNORE_SIGNAL(SIGNAL_NR);                          /* ignore signal IO */

  socket = ppc_handlers[id].socket;
  
  disable_for_signal(socket);
  
  FD_CLR(socket, &global_readfds);           /* remove descriptor from set */
  
  max_readfd = 0;              /* update highest read fd for socket() call */
  for (id = 0; id < MAX_CHAN_HANDLERS; id++)
    if (!ppc_handlers[id].free && ppc_handlers[id].socket > max_readfd)
      max_readfd = ppc_handlers[id].socket;

  RESTORE_SIGNAL(SIGNAL_NR);
}
#endif
