/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Client-Server Library.                                                  */
/* This module implements the client-side connection functions of the      */
/* PPC client-server library.                                              */
/* The Functions implemented here are declared in the header file cslib.h  */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 07.02.1994  Droux       Created                                         */
/* 17.06.1994  Droux       Check for NEXTSTEP and Solartis perm. denied    */
/* 17.06.1994  Droux       Check for unexpected message from server        */
/* 19.06.1994  Droux       Password checking implemented                   */
/* 21.06.1994  Droux       Test rsh termination by means of a timeout      */
/* 22.06.1994  Droux       Added session_read_channel()                    */
/* 30.06.1994  Droux       More reliable connection under Linux            */
/* 08.07.1994  Droux       New structure of source directory               */
/* 13.07.1994  Droux       Forware version checking msg with pw option     */
/*-------------------------------------------------------------------------*/
/* filename: cliconn.c                                                     */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*   SunOS Release 5.3 (Solaris 2), SPARC                                  */
/*   Linux 0.99.pl9, Intel                                                 */
/*   Linux 1.0.8, Intel                                                    */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pwd.h>
#include <signal.h>
#include <ppc/cslib.h>
#include <ppc/cslib/hlist.h>
#include <ppc/cslib/error.h>
#include "pclichan.h"
#include "ion.h"
#include "hfinder.h"
#include "csversion.h"

#if defined(NEXT) || defined(NEXTINTEL)
  #include <libc.h>
#endif
#if defined(SUN4SOL2) || defined(LINUX)
  #include <unistd.h>
#endif
                    /* possible error codes received during server startup */
#define CONN_ERROR_EXEC '9'               /* cannot start remote "starter" */
#define CONN_ERROR_BADP '8'                    /* incorrect password given */
#define CONN_ERROR_RSHF '7'                       /* rsh utility not found */
#define CONN_ERROR_PATH '1'        /* executable directory does not exists */
#define CONN_ERROR_PROG '2'                 /* specified program not found */
#define CONN_ERROR_COMP '3' /* program not compiled for given architecture */
#define CONN_ERROR_HOST '5'     /* program not compiled to run on the host */
#define CONN_ERROR_PVMN '4'       /* PVM is not running on the server host */

static int cslib_init = 0;                      /* 1 if module initialized */

typedef struct                                           /* session_descr: */
{                         /* contains status of a client-server connection */
  int free;                    /* free tag, 1 if element free, 0 otherwise */
  int host_id;                            /* hlist host id for the session */
  int exec_pid;                                        /* exec process pid */
  channel from_server;                              /* channel from server */
  channel to_server;                                  /* channel to server */
} session_descr;

static session_descr sessions[MAX_NB_SESSIONS];         /* active sessions */

/* The two following variables are used by the functions                   */
/* last_server_host_name() and last_server_arch() to allows the client     */
/* program to get the host name and architecture of the server on which    */
/* the program program has been started, or failed to start.               */
/* They can be used by the client when the session startup failed.         */
static char last_server_host_name_value [50];
static architecture last_server_arch_value;

static int to_be_killed;          /* if the rsh process does not terminate */

/*-------------------------------------------------------------------------*/
/* init_cslib_module:                                                LOCAL */
/* Initializes the cslib module. Called only one time automatically.       */
/*-------------------------------------------------------------------------*/
void init_cslib_module (void)
{
  int c;
  
  for (c = 0; c < MAX_NB_SESSIONS; c++)
    sessions[c].free = 1;
  strcpy(last_server_host_name_value, "(none)");
  last_server_arch_value = UNKNOWN;
  cslib_init = 1;
}

/*-------------------------------------------------------------------------*/
/* killer:                                                           LOCAL */
/* It is possible that the remote shell process never terminates for some  */
/* obscure reason. If this is the case, we have to kill it by sending a    */
/* SIGKILL signal. This function is called automatically if the rsh        */
/* process does not terminate after a short time determined by my_wait().  */
/* The process id is communicated by the function my_wait() by means of    */
/* the global variable to_be_killed.                                       */
/*-------------------------------------------------------------------------*/
void killer (int sig)
{
  kill(to_be_killed, SIGKILL);
}

/*-------------------------------------------------------------------------*/
/* my_wait:                                                          LOCAL */
/* My version of the wait function. It is implemented either using the     */
/* BSD wait4() functions by default, or POSIX 1003.1 waitpid() under       */
/* Solaris 2.                                                              */
/* The function blocks until the given process is terminated.              */
/*-------------------------------------------------------------------------*/
void my_wait (
  int pid
)
{
  to_be_killed = pid;          /* set timeout stuff if rsh won't terminate */
  signal(SIGALRM, killer);
  alarm(2);
  
  #ifdef SUN4SOL2                                             /* Solaris 2 */
    waitpid(pid, (int *)NULL, 0);
  #else                                 /* NEXTSTEP, Solaris 1, Linux, ... */
    wait4(pid, (union wait *)NULL, 0, (struct rusage *)NULL);
  #endif
  
  alarm(0);
}

/*-------------------------------------------------------------------------*/
/* start_server:                                                     LOCAL */
/*                                                                         */
/* Starts the given server program on the given host. Returns two file     */
/* descriptors connected to the stdin and stdout streams of the started    */
/* program.                                                                */
/* Returns -1 if an error occured, 0 otherwise.                            */
/*                                                                         */
/* Possible errors are:                                                    */
/*                                                                         */
/*   [EPPC_CONN_PIPECRE]	error creating pipes for connection with   */
/*                              remote server program.                     */
/*   [EPPC_CONN_FORK]           error forking new process to start remote  */
/*                              server program.                            */
/*   [EPPC_CONN_EXEC]           cannot execute the \"rsh\" program. Must   */
/*                              be in /usr/bin or /usr/ucb.                */
/*   [EPPC_CONN_RSHF]           rsh program not found. Must be placed in   */
/*                              /usr/bin or /usr/ucb.                      */
/*   [EPPC_CONN_PWNI]           remote execution with password not yet     */
/*                              implemented.                               */
/*   [EPPC_CONN_PATH]		executable directory does not exists on    */
/*				server host.                               */
/*   [EPPC_CONN_PROG]		specified program not found on server host.*/
/*   [EPPC_CONN_COMP]		specified program not compiled for the     */
/*				given architecture                         */
/*   [EPPC_CONN_HOST]		specified program not compiled for the     */
/*				given host.                                */
/*   [EPPC_CONN_PERM]		permission denied to execute a program on  */
/*				the server host.                           */
/*-------------------------------------------------------------------------*/
int start_server(
  int host_id,         /* host on which the server program must be started */
  char *prog_name,                                  /* server program name */
  char *password,                                  /* facultative password */
  int *s_from_server,                     /* stream from the server stdout */
  int *s_to_server,                          /* stream to the server stdin */
  int *exec_pid_ptr                                    /* exec process pid */
)
{
  int from_server_pipe [2];      /* pipes to communicate with exec process */
  int to_server_pipe[2];                                       /* ... idem */
  int exec_pid;                          /* exec process pid. rsh or rexec */
  
  if (hlist_pw_required(host_id) && (password == NULL))
  {            /* test if password is required and given if it is the case */
    ppc_errno = EPPC_CONN_PARE;
    return -1;
  }
  
  if (pipe(from_server_pipe) == -1)
  {     /* error creating pipes for connection with remote server program. */
    ppc_errno = EPPC_CONN_PIPECRE;
    return -1;
  }
  
  if (pipe(to_server_pipe) == -1)
  {     /* error creating pipes for connection with remote server program. */
    close(from_server_pipe[0]);
    close(to_server_pipe[1]);
    ppc_errno = EPPC_CONN_PIPECRE;
    return -1;
  }
  
  if ((exec_pid = fork()) == -1)
  {           /* error forking new process to start remote server program. */
    close(from_server_pipe[0]);
    close(from_server_pipe[1]);
    close(to_server_pipe[0]);
    close(to_server_pipe[1]);
    ppc_errno = EPPC_CONN_FORK;
    return -1;
  }
  
  if (!exec_pid)
  {                                           /* this is the child process */
    close(0);                            /* close stdin before redirection */
    close(1);                           /* close stdout before redirection */
    close(2);                           /* close stderr before redirection */
    
    close(to_server_pipe[1]);              /* not used by the exec process */
    close(from_server_pipe[0]);                                /* ... idem */
    
    dup2(to_server_pipe[0], 0);                       /* stdin redirection */
    dup2(from_server_pipe[1], 1);                    /* stdout redirection */
    dup2(from_server_pipe[1], 2);                    /* stderr redirection */
    
    if (!hlist_pw_required(host_id))
    {              /* password not necessary, start remote server with rsh */
      char *argv[] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
      char command1[] = "/usr/ucb/rsh";     /* one of the two possibilties */
      char command2[] = "/usr/bin/rsh";
      char *command;
      char user_option[] = "-l";     /* alternative loggin name rsh option */
      char remote_command[] = "starter";
      architecture arch;                       /* server host architecture */
      struct stat stat_buf;              /* to check if rsh program exists */
      int lo_offset;         /* offset to add to argv index if lo required */
      
      /* Under some operating systems, the rsh command is found under      */
      /* /usr/ucb/rsh (NEXTSTEP, Solaris 1). Under some other OS's, it     */
      /* is however in /usr/bin/rsh (Linux, Solaris 2). We check here      */
      /* where the rsh program is located and modify the command pointer   */
      /* according to the obtained result. An error is returned if it      */
      /* is not found in one of the two paths.                             */
      if (stat(command1, &stat_buf) != -1)
        command = command1;
      else
        if (stat(command2, &stat_buf) != -1)
	  command = command2;
	else
	{ /* rsh program not found. Must be placed in /usr/bin or /usr/ucb */
	  printf("%c\n", CONN_ERROR_RSHF);
	  exit(1);
	}
	
      hlist_get_host_arch(host_id, &arch);          
      argv[0] = command;                            /* construct argv list */
      argv[1] = hlist_get_host_name_ptr(host_id);

      /* If an alternative loggin name has been defined for the server     */
      /* host, then the option "-l user" must be added to the list of      */
      /* rsh arguments. Here we test this condition and set the            */
      /* appropriate offset to assign values of further argv values.       */
      if (hlist_exec_lo_defined(host_id))
      {
        argv[2] = user_option;
        argv[3] = hlist_custom_name(host_id);
	lo_offset = 2;
      }
      else
        lo_offset = 0;
        
      argv[2+lo_offset] = remote_command;
      argv[3+lo_offset] = arch_string(arch);
      argv[4+lo_offset] = prog_name;

      if (hlist_exec_path_required(host_id))
        argv[5+lo_offset] = hlist_exec_path(host_id);
            
      if (execv(argv[0], argv) == -1)
      {                        /* failed to start "starter" on remote host */
        /* we are not in the user process here. Thus, it is not possible   */
	/* to simply return an error code to it. Instead, we do this       */
	/* indirectly by sending a little string to the user process which */
	/* will analyze it. If "starter" can be started correcly, then it  */
	/* will also send its status code to the user process.             */
	printf("%c\n", CONN_ERROR_EXEC);
	exit(1);
      }
    }
    else
    {                /* password necessary, start remote server with rexec */
      struct servent *serv_info;      /* exec service info to get TCP port */
      int serv_socket;             /* socket to the started remote process */
      char command [200];      /* string containing the command to execute */
      architecture arch;                            /* server architecture */
      char *host;                  /* host on which the program is started */
      char user_name [100];
      
      strcpy(command, "starter "); /* construct the remote command to exec */
      hlist_get_host_arch(host_id, &arch);            /* host architecture */
      strcat(command, arch_string(arch));
      strcat(command, " ");
      strcat(command, prog_name);                          /* program name */
      if (hlist_exec_path_required(host_id))           /* path if required */
      {
        strcat(command, " ");
        strcat(command, hlist_exec_path(host_id));
      }
      
      serv_info = getservbyname("exec", "tcp");   /* get exec service info */
      host = hlist_get_host_name_ptr(host_id);

      if (hlist_exec_lo_defined(host_id))
        strcpy(user_name, hlist_custom_name(host_id));
      else
      {
        struct passwd *act_passwd;
        act_passwd = getpwuid(getuid());          /* get user's login name */
        strcpy(user_name, act_passwd->pw_name);
      }

      serv_socket = rexec(                         /* start remote program */
        &host, 
        serv_info->s_port,
        user_name,
        password,
        command,
        NULL);
	
      if (serv_socket == -1)
      {                        /* failed to start "starter" on remote host */
        /* we are not in the user process here. Thus, it is not possible   */
	/* to simply return an error code to it. Instead, we do this       */
	/* indirectly by sending a little string to the user process which */
	/* will analyze it. If "starter" can be started correcly, then it  */
	/* will also send its status code to the user process.             */
	printf("%c\n", CONN_ERROR_EXEC);
	exit(1);
      }

      {                         /* forward starter status and port numbers */
        char starter_status [2];
	char port_str [PORT_NAME_LEN];
	char version_str [CSLIB_VERSION_LEN];
	char check_result [COMPATIBILITY_LENGTH];
       
        read(serv_socket, starter_status, 2);
	write(1, starter_status, 2);
	
	if (starter_status[0] == 'L')
	  exit(1);
    
        readn(serv_socket, version_str, CSLIB_VERSION_LEN);
	writen(1, version_str, CSLIB_VERSION_LEN);
	
	readn(0, check_result, COMPATIBILITY_LENGTH);
	writen(serv_socket, check_result, COMPATIBILITY_LENGTH);

        readn(serv_socket, port_str, PORT_NAME_LEN);
	writen(1, port_str, PORT_NAME_LEN);
      
        readn(serv_socket, port_str, PORT_NAME_LEN);
	writen(1, port_str, PORT_NAME_LEN);
      }
      	
      exit(1);
   }
  }

  close(to_server_pipe[0]);               /* only used by the exec process */
  close(from_server_pipe[1]);                                  /* ... idem */
  
  {                                          /* get starter status message */
    char msg [30];                                 /* the message received */
    
    readn(from_server_pipe[0], msg, 2);
    msg[2] = '\0';                             /* remove trailing new-line */
    switch (msg[0])
    {
      case CONN_ERROR_RSHF:
        my_wait(exec_pid);
	ppc_errno = EPPC_CONN_RSHF;
	return -1;
      case CONN_ERROR_EXEC:        /* the starter itself cannot be started */
        my_wait(exec_pid);       /* wait for the death of the exec process */
	ppc_errno = EPPC_CONN_EXEC;
	return -1;
      case CONN_ERROR_PATH:        /* executable directory does not exists */
        my_wait(exec_pid);
	ppc_errno = EPPC_CONN_PATH;       
	return -1;
      case CONN_ERROR_PROG:                 /* specified program not found */
        my_wait(exec_pid);
	ppc_errno = EPPC_CONN_PROG;
	return -1;
      case CONN_ERROR_COMP: /* program not compiled for given architecture */
        my_wait(exec_pid);
	ppc_errno = EPPC_CONN_COMP;
	return -1;
      case CONN_ERROR_HOST:     /* program not compiled to run on the host */
        my_wait(exec_pid);
	ppc_errno = EPPC_CONN_HOST;
	return -1;
      case CONN_ERROR_BADP:                          /* bad password given */
        my_wait(exec_pid);
	ppc_errno = EPPC_CONN_BADP;
	return -1;
      case CONN_ERROR_PVMN:       /* PVM is not running on the server host */
        my_wait(exec_pid);
        ppc_errno = EPPC_CONN_PVMN;
        return -1;
      case 'p':                                       /* permission denied */
        /* We have received the first two letters of a "permission denied" */
	/* error message from rsh. The remaining of the message is read,   */
	/* and an error is returned to the caller.                         */
        read(from_server_pipe[0], msg, 16);
        my_wait(exec_pid);
	ppc_errno = EPPC_CONN_PERM;
	return -1;
      case 'P':                                       /* permission denied */
        /* We have received the first two letters of a "Permission         */
	/* denied." error message from rsh. The remaining of the message   */
	/* is read, and an error is returned to the caller.                */
        read(from_server_pipe[0], msg, 17);
        my_wait(exec_pid);
	ppc_errno = EPPC_CONN_PERM;
	return -1;
      case 'L':                                       /* permission denied */
        /* We have received the first two letters of a "Login incorrect."  */
	/* error message from rsh. The remaining of the message is read,   */
	/* and an error is returned to the caller.                         */
        read(from_server_pipe[0], msg, 15);
        my_wait(exec_pid);
	ppc_errno = EPPC_CONN_PERM;
	return -1;
      case '0':                                  /* no error. Champagne !! */
        break;
      default:                                 /* unknown message received */
      {
	char *host;
	
        host = hlist_get_host_name_ptr(host_id);
        if ((msg[0] == host[0]) && (msg[1] == host[1]))
	{     /* the two first characters received correspond to the name */
	      /* of the server host. We received probably a message of    */
	      /* the form: "<host>: unknown host" generated by rexec().   */
          my_wait(exec_pid);
	  ppc_errno = EPPC_CONN_UNHO;
	  return -1;	
	}
      }
    }
  }
  
  *s_from_server = from_server_pipe[0];  /* update values before returning */
  *s_to_server = to_server_pipe[1];
  *exec_pid_ptr = exec_pid;
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* my_bzero:                                                         LOCAL */
/* Fill the given buffer with the specified number of 0 bytes.             */
/*-------------------------------------------------------------------------*/
void my_bzero (void *buffer, int length)
{
  char *b = (char *)buffer;
  int curr_char;
  
  for (curr_char = 0; curr_char < length; curr_char++)
    b[curr_char] = 0;
}

/*--------------------------------------------------------------------------*/
/* copy_struct:                                                       LOCAL */
/* copy the <len> first bytes of <source> to <dest>.                        */
/*--------------------------------------------------------------------------*/
void copy_struct (void *source, void *dest, int len)
{
  char *s = (char *)source;
  char *d = (char *)dest;

  while (len)
  {
    *d = *s;
    d++;
    s++;
    len--;
  }
}

/*-------------------------------------------------------------------------*/
/* create_client_socket:                                             LOCAL */
/* Creates a TCP socket based connection between the client and the        */
/* server. Returns a file descriptor corresponding to the socket or -1     */
/* if an error occured.                                                    */
/*-------------------------------------------------------------------------*/
int create_client_socket (
  char *host_name,
  int from_server
)
{
  int socket_fd;
  struct hostent *foreign_addr;               /* address of a given server */
  struct sockaddr_in sock_addr;                          /* socket address */
  int connect_retries = 10;                          /* retries connection */
  int servlen;
  int connect_status;
  int port_num;
  char port_name[PORT_NAME_LEN];
                                            /* get port number from server */
  readn(from_server, port_name, PORT_NAME_LEN);
  if (sscanf(port_name,"%d",&port_num) != 1)
  {             /* error connecting to server: unexpected message received */
    ppc_errno = EPPC_CONN_SCANF;
    return -1;
  }

  servlen = sizeof(struct sockaddr_in);
  if ((socket_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) /* create sockets */
  {              /* error creating TCP socket connection with server host. */
    ppc_errno = EPPC_CONN_SOCK;
    return -1;
  }
  
  if ((foreign_addr = gethostbyname(host_name)) == 0)    /* get the server */
  {              /* error creating TCP socket connection with server host. */
    ppc_errno = EPPC_CONN_SOCK;
    return -1;
  }

  my_bzero((char *)&sock_addr, servlen);
  copy_struct(foreign_addr->h_addr,&sock_addr.sin_addr, 
    foreign_addr->h_length);
  sock_addr.sin_family = AF_INET;
  sock_addr.sin_port = htons(port_num);

  /* A strange behaviour has been observed here under Linux: sometimes,    */
  /* the server cannot call the listen() and the accept() functions        */
  /* before the client calls the connect() function. If this is the        */
  /* case, the two entities cannot be connected. To ensure that the        */
  /* server will enter in accept() before the client enters in connect(),  */
  /* we add here a little delay of a few 10th of seconds.                  */
  #ifdef LINUX
    usleep(200000);
  #endif

  do {                                            /* connect to the server */
    connect_status = connect(socket_fd,
      (struct sockaddr *)&sock_addr, servlen);  
    connect_retries--;
  } while((connect_status < 0) && (connect_retries));
  
  if (!connect_retries)                       /* test if connection failed */
  {              /* error creating TCP socket connection with server host. */
    ppc_errno = EPPC_CONN_SOCK;
    return -1;
  }

  return socket_fd;
}

/*-------------------------------------------------------------------------*/
/* connect_to_server:                                                      */
/* Start and connect to the server specified by prog_name, host_name,      */
/* arch, and evt. password and returns the two channels to_server and      */
/* from_server.                                                            */
/* Returns the session identifier if the server has been correctly         */
/* started, or -1 if an error occured.                                     */
/*                                                                         */
/* Possible errors:                                                        */
/*                                                                         */
/*   [EPPC_HLIST_UNKARCH]	unknown architecture specifier in hosts    */
/*				list.                                      */
/*   [EPPC_HLIST_LOEQ] 		missing '=' for \"lo\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_LOVAL]		missing value for \"lo\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_EPEQ]		missing '=' for \"ep\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_EPVAL]		missing value for \"ep\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_UNKOPT]	unknown option in hosts list.              */
/*   [EPPC_HLIST_ITMEXC]	too many items in hosts list line.         */
/*   [EPPC_HLIST_PARTDEF]	host only partially defined in hosts list. */
/*   [EPPC_HLIST_ARCHNFND]	specified architecture not found in hosts  */
/*                              list.                                      */
/*   [EPPC_HLIST_HOSTNFND]	specified host name not found in hosts     */
/*                              list.                                      */
/*   [EPPC_HLIST_HOSTARCHNFND]	specified host name with given             */
/*                              architecture not found in hosts list.      */
/*   [EPPC_HLIST_BADIDAR]	bad host id given when inquiring host      */
/*				architecture, internal error.              */
/*   [EPPC_HFIND_NODEF]		at least host name or architecture must be */
/*				specified to find host.                    */
/*   [EPPC_CONN_OOFSLOTS]      	too many client-server sessions activated  */
/*                              at the same time.                          */
/*   [EPPC_CONN_PIPECRE]	error creating pipes for connection with   */
/*                              remote server program.                     */
/*   [EPPC_CONN_FORK]          	error forking new process to start remote  */
/*                              server program.                            */
/*   [EPPC_CONN_EXEC]         	cannot execute the \"rsh\" program. Must   */
/*                              be in /usr/bin or /usr/ucb.                */
/*   [EPPC_CONN_RSHF]          	rsh program not found. Must be placed in   */
/*                              /usr/bin or /usr/ucb.                      */
/*   [EPPC_CONN_PWNI]          	remote execution with password not yet     */
/*                              implemented.                               */
/*   [EPPC_CONN_PATH]		executable directory does not exists on    */
/*				server host.                               */
/*   [EPPC_CONN_PROG]		specified program not found on server host.*/
/*   [EPPC_CONN_COMP]		specified program not compiled for the     */
/*				given architecture                         */
/*   [EPPC_CONN_HOST]		specified program not compiled for the     */
/*				given host.                                */
/*   [EPPC_CONN_PERM]		permission denied to execute a program on  */
/*				the server host.                           */
/*   [EPPC_CONN_SOCK]		error creating TCP socket connection with  */
/*				server host. Internal error.               */
/*-------------------------------------------------------------------------*/
int connect_to_server (
  channel *to_server,                             /* channel to the server */
  channel *from_server,                         /* channel from the server */
  char *prog_name,                                  /* server program name */
  char *host_name,                                     /* server host name */
  architecture arch,                                /* server architecture */
  char *password                                   /* facultative password */
)
{
  int act_host;                                /* actually started host id */
  int act_session;                        /* new session id to be returned */
  int exec_pid;                                        /* exec process pid */
  int fs_fd;                    /* from server stdout pipe file descriptor */
  int ts_fd;                       /* to server stdin pipe file descriptor */
  int fs_socket;                      /* socket connection from the server */
  int ts_socket;                        /* socket connection to the server */
  
  if (!cslib_init) init_cslib_module();        /* init module if necessary */
  
  for (act_session = 0;                                  /* find free slot */
    act_session < MAX_NB_SESSIONS && !sessions[act_session].free;
    act_session++);
    
  if (act_session == MAX_NB_SESSIONS)
  {         /* too many client-server sessions activated at the same time. */
    ppc_errno = EPPC_CONN_OOFSLOTS;
    return -1;
  }

  /* --------------------------------------------------------------------- */
  /* The current client-server startup is implemented as follows:          */
  /* (S = server, C = client)                                              */
  /* 1. C: Find the host according to the given name and architecture      */
  /* 2. C: Start the server on the server host with rexec if pw entry      */
  /*       found in hosts file, rsh otherwise.                             */
  /*       From this action will result two streams connected to the       */
  /*       stdin and stdout of the server                                  */
  /* 3. S: The "server starter" is started by the client.                  */
  /* 4. S: "starter" checks the parameters given and if the specified      */
  /*       server program exists. Return a status message to the client.   */
  /* 5. C: Get status message from server.                                 */
  /* 6. S: Starts the server program.                                      */
  /*       server PPC program.                                             */
  /* 7. S: Creates two sockets and send the corresponding ports to the     */
  /*       client.                                                         */
  /* 8. S: Listen the socket                                               */
  /* 9. C: Get the ports from the server and connect to the socket.        */
  /* --------------------------------------------------------------------- */
  
           /* find a host corresponding to the given name and architecture */
  if ((act_host = find_host(host_name, arch)) == -1)
    return -1;
                    /* update global host name and arch for error handling */
  strcpy(last_server_host_name_value, hlist_get_host_name_ptr(act_host));
  hlist_get_host_arch(act_host, &last_server_arch_value);
  
                    /* start the requested server on the server host found */
  if (start_server(act_host, prog_name, password, 
                   &fs_fd, &ts_fd, &exec_pid) == -1)
    return -1;
                                           /* check protocol compatibility */
  if (client_check_version(fs_fd, ts_fd) == -1)
  {         /* client and server versions of the protocol are incompatible */
    ppc_errno = EPPC_CONN_COMPAT;
    return -1;
  }
                                  /* connects to the server by TCP sockets */
  if ((fs_socket = 
    create_client_socket(hlist_get_host_name_ptr(act_host), fs_fd)) == -1)
    return -1;
  if ((ts_socket = 
    create_client_socket(hlist_get_host_name_ptr(act_host), fs_fd)) == -1)
    return -1;
                                        /* create the new session channels */
  if (new_client_channel(fs_socket, ts_socket, from_server, to_server) == -1)
    return -1;
  
  sessions[act_session].free = 0;    /* execution correct, update sessions */
  sessions[act_session].host_id = act_host;
  sessions[act_session].exec_pid = exec_pid;
  sessions[act_session].from_server = *from_server;
  sessions[act_session].to_server = *to_server;
    
  return act_session;
}

/*-------------------------------------------------------------------------*/
/* disconnect_from_server:                                                 */
/* Disconnect from the given server.                                       */
/* This function returns only when the corresponding server is terminated. */
/*-------------------------------------------------------------------------*/
void disconnect_from_server(
  int session_id
)
{
  if (!cslib_init) init_cslib_module();        /* init module if necessary */

  if (session_id < 0 || session_id >= MAX_NB_SESSIONS)
    return;
  if (sessions[session_id].free)
    return;
    
  my_wait(sessions[session_id].exec_pid);
  
  close_client_channels(
    sessions[session_id].from_server, sessions[session_id].to_server);
}

/*-------------------------------------------------------------------------*/
/* server_host_name:                                                       */
/* Returns a pointer to a string containing the name of the server         */
/* corresponding to the given session identifier.                          */
/* Returns NULL if session_id does not correspond to a valid session.      */
/*-------------------------------------------------------------------------*/
char *server_host_name(
  int session_id
)
{
  if (!cslib_init) init_cslib_module();        /* init module if necessary */
  
  if (session_id < 0 || session_id >= MAX_NB_SESSIONS)
    return NULL;
  if (sessions[session_id].free)
    return NULL;
  
  return hlist_get_host_name_ptr(sessions[session_id].host_id);
}

/*-------------------------------------------------------------------------*/
/* server_arch:                                                            */
/* Returns the architecture of the server corresponding to the given       */
/* session identifier.                                                     */
/* Returns UNKNOWN if session_id does not correspond to a valid session.   */
/*-------------------------------------------------------------------------*/
architecture server_arch(
  int session_id
)
{
  architecture arch;                        /* architecture to be returned */

  if (!cslib_init) init_cslib_module();        /* init module if necessary */
  
  if (session_id < 0 || session_id >= MAX_NB_SESSIONS)
    return UNKNOWN;
  if (sessions[session_id].free)
    return UNKNOWN;

  if (hlist_get_host_arch(sessions[session_id].host_id, &arch) == -1)
    return UNKNOWN;
    
  return arch;
}

/*-------------------------------------------------------------------------*/
/* last_server_host_name:                                                  */
/* Returns a pointer to a string containing the name of the last server    */
/* for which a connection has been tried.                                  */
/* Can be used for error management.                                       */
/*-------------------------------------------------------------------------*/
char *last_server_host_name(void)
{
  if (!cslib_init) init_cslib_module();        /* init module if necessary */

  return last_server_host_name_value;
}

/*-------------------------------------------------------------------------*/
/* last_server_arch:                                                       */
/* Returns the architecture of the last server for which a connection has  */
/* been tried or established.                                              */
/* Can be used for error management.                                       */
/*-------------------------------------------------------------------------*/
architecture last_server_arch(void)
{
  if (!cslib_init) init_cslib_module();        /* init module if necessary */

  return last_server_arch_value;
}

/*-------------------------------------------------------------------------*/
/* session_read_channel:                                                   */
/* Returns the channel from which the client receives data from the        */
/* specified server.                                                       */
/* Returns -1 if the given session id is not correct.                      */
/*-------------------------------------------------------------------------*/
channel session_read_channel (
  int session_id
)
{
  if (!cslib_init) init_cslib_module();        /* init module if necessary */

  if (session_id < 0 || session_id >= MAX_NB_SESSIONS)
    return -1;
  if (sessions[session_id].free)
    return -1;
    
  return sessions[session_id].from_server;
}
