/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Client-Server Library.                                                  */
/* Host finder. This module implement functions used to select the host    */
/* on which the server application must be started.                        */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 08.02.1994  Droux       Created                                         */
/* 08.07.1994  Droux       New structure of source directory               */
/*-------------------------------------------------------------------------*/
/* filename: hfinder.c                                                     */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*   SunOS Release 5.3 (Solaris 2), SPARC                                  */
/*   Linux 0.99.pl9, Intel Computers                                       */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <ppc/cslib/error.h>
#include <ppc/cslib/hlist.h>
#include "hfinder.h"

/*-------------------------------------------------------------------------*/
/* find_host:                                                              */
/*                                                                         */
/* Return the host corresponding to the given host name and architecture.  */
/* host_name and arch are facultative. If either host_name or arch are set */
/* to respectively NULL and UNKNOWN, then the missing one is determined    */
/* from the hosts file.                                                    */
/* Returns the host id if server found, -1 othewise.                       */
/*                                                                         */
/* Possible errors are:                                                    */
/*                                                                         */
/*   [EPPC_HLIST_UNKARCH]	unknown architecture specifier in hosts    */
/*				list.                                      */
/*   [EPPC_HLIST_LOEQ] 		missing '=' for \"lo\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_LOVAL]		missing value for \"lo\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_EPEQ]		missing '=' for \"ep\" option in hosts     */
/*				list.                                      */
/*   [EPPC_HLIST_EPVAL]		missing value for \"ep\" option in hosts   */
/*				list.                                      */
/*   [EPPC_HLIST_UNKOPT]	unknown option in hosts list.              */
/*   [EPPC_HLIST_ITMEXC]	too many items in hosts list line.         */
/*   [EPPC_HLIST_PARTDEF]	host only partially defined in hosts list. */
/*   [EPPC_HLIST_ARCHNFND]	specified architecture not found in hosts  */
/*                              list.                                      */
/*   [EPPC_HLIST_HOSTNFND]	specified host name not found in hosts     */
/*                              list.                                      */
/*   [EPPC_HLIST_HOSTARCHNFND]	specified host name with given             */
/*                              architecture not found in hosts list.      */
/*   [EPPC_HLIST_BADIDAR]	bad host id given when inquiring host      */
/*				architecture, internal error.              */
/*   [EPPC_HFIND_NODEF]		at least host name or architecture must be */
/*				specified to find host.                    */
/*                                                                         */
/*-------------------------------------------------------------------------*/
int find_host(
  char *host_name,                           /* desired host (facultative) */
  architecture arch                  /* desired architecture (facultative) */
)
{
  int host_idx;                            /* host index from hlist module */
  
  if ((host_name == NULL) && (arch == UNKNOWN))
  {  /* at least host name or architecture must be specified to find host. */
    ppc_errno = EPPC_HFIND_NODEF;
    return -1;                     /* at least one of both must be defined */
  }
  
  if (host_name == NULL)
  { /* if the host name is missing, then look for a host of the given type */
    if ((host_idx = hlist_find_host_name(arch)) == -1)
      return -1;
  } 
  else if (arch == UNKNOWN)
  {  /* if architecture is unknown, then look for the architecture of host */ 
    if ((host_idx = hlist_find_host_arch(host_name)) == -1)
      return -1;
  } else
  {          /* both name and architecture are defined, check if it exists */
    if ((host_idx = hlist_find_host_name_arch(host_name, arch)) == -1)
      return -1;
  }

  /* At this point, we are sure that the hosts list contains a host that   */
  /* fullfits the requirements.                                            */
  return host_idx;
}
