/*
 *
 *  @(#)b004unix.c  1.7 92/12/31    (c) PACT
 *
 *  PSERVER - PACT host server
 *
 *  Copyright (c) 1991, 1992 PACT
 *  All Rights Reserved
 *
 *  USV/386/B004 interface module.
 *
 */

#include    <stdio.h>                   /* standard I/O */
#include    <fcntl.h>                   /* file control */
#include    <errno.h>                   /* error numbers */
#include    <sys/trp.h>                 /* driver interface */
#include    "linkio.h"                  /* link interface definition */


/*
 *
 *      Boolean type.
 *
 */

typedef int bool_t;
#undef  FALSE
#define FALSE   0
#undef  TRUE
#define TRUE    1


/*
 *
 *      The fact that a link is currently active link is recorded.
 *
 */

#define FREELINK    (-1)
static  link_t  activelink = FREELINK;


/*
 *
 *      Default base port address is 0x150, as per Inmos B004 specs.
 *
 */

#define C012_DEFAULT_BASE   0x150


/*
 *
 *      Include defs for driver, and provide a nice ioctl() to work with it.
 *
 */

static  bool_t  my_ioctl (fd, x, ptrp)
int             fd;
int             x;
struct trpio_t  *ptrp;
{
    if (ioctl (fd, x, ptrp) == -1) {
        perror ("ioctl /dev/trp");
        return TRUE;
     }
    return FALSE;
}


/*
 *
 *      void WaitLink (bool_t wfl)
 *
 *  EXTENSION.  Indicates whether we want to wait for board to become free
 *  instead of fail if busy.
 *
 */

static  bool_t  waitforlink = FALSE;

void    WaitLink (wfl)
bool_t  wfl;
{
    waitforlink = wfl;
}


/*
 *
 *      link_t  OpenLink (char *link)
 *
 *  Open link.  If link is NULL or empty, the default link is used.
 *  Otherwise, link should be a hex number indicating the base address
 *  of the port to use.  Returns negative number if an error occurs,
 *  and a positive linkid if the link was successfully opened.
 *
 */

#if __STDC__
link_t  OpenLink (char *link)
#else
link_t  OpenLink (link)
char    *link;
#endif
{
    int oflags;
    int trpfd;

    /* If a link is already active, we can't open another.              */
    if (activelink != FREELINK)
        return ER_LINK_CANT;

    oflags = O_RDWR | O_SYNC | O_NDELAY;

    /* Open device.                                                     */
    do {
        int busy = FALSE;
        if (link && link[0] && strcmp (link, "any")) {
            trpfd = open (link, oflags);
            if (trpfd < 0 && errno == EBUSY)
                busy = TRUE;
        }
        else {
            int i;
            for (i=0, trpfd=-1; i<16 && trpfd < 0; i++) {
                char    try[16];
                sprintf (try, "/dev/trp%02d", i);
                trpfd = open (try, oflags);
                if (trpfd < 0) {
                    if (errno == EBUSY)
                         busy = TRUE;
                }
                else
                    if (link && ! strcmp (link, "any"))
                        fprintf (stderr, "opened link %s\n", try);
            }
        }
        if (trpfd < 0)
            if (busy && waitforlink)
                sleep (1);
            else
                return ER_LINK_CANT;
    } while (trpfd < 0);

    /* Flag link as active and return `link descriptor'.                */
    activelink = trpfd;
    return activelink;
}


/*
 *
 *      int CloseLink (link_t linkid)
 *
 *  Close link.  Must be open to be closed.  If any error occurs (e.g. link
 *  is not open), a negative number is returned.  Otherwise, 1 is returned.
 *
 */

#if __STDC__
int     CloseLink (link_t linkid)
#else
int     CloseLink (linkid)
link_t  linkid;
#endif
{
    /* Check if link is actually open.                                  */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Close device.                                                    */
    (void) close (linkid);

    /* Free link.                                                       */
    activelink = FREELINK;

    /* Return 1 to indicate success.                                    */
    return 1;
}


/*
 *
 *      int ReadLink (link_t linkid, char *buf, unsigned len, int timout)
 *
 *  Read len bytes from specified link into buffer.  Returns number of
 *  bytes read (before timeout), or a negative number if an error occurs
 *  (wrong link or 0-size transfer).  The timout parameter is currently
 *  ignored - a crude timeout mechanism with a max loop count is provided.
 *
 */

#if __STDC__
int     ReadLink (link_t linkid, char *buf, unsigned len, int timout)
#else
int     ReadLink (linkid, buf, len, timout)
link_t       linkid;
char        *buf;
unsigned    len;
int         timout;
#endif
{
    /* If linkid is not the active link, fail.                          */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Zero-size transfers can't be done.                               */
    if (len < 1)
        return ER_LINK_CANT;

    /* Read from transputer link and return # bytes done.               */
    return read (linkid, buf, len);
}


/*
 *
 *      int WriteLink (link_t linkid, char *buf, unsigned len, int timout)
 *
 *  Write len bytes from specified link into buffer.  Returns number of
 *  bytes written (before timeout), or a negative number if an error occurs
 *  (wrong link or 0-size transfer).  The timout parameter is currently
 *  ignored - a crude timeout mechanism with a max loop count is provided.
 *
 */

#if __STDC__
int     WriteLink (link_t linkid, char *buf, unsigned len, int timout)
#else
int     WriteLink (linkid, buf, len, timout)
link_t      linkid;
char        *buf;
unsigned    len;
int         timout;
#endif
{
    /* If linkid is not the active link, fail.                          */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Zero-size transfers can't be done.                               */
    if (len < 1)
        return ER_LINK_CANT;

    /* Write to transputer link and return # bytes done.                */
    return write (linkid, buf, len);
}


/*
 *
 *      int ResetLink (link_t linkid)
 *
 *  Reset transputer.  Returns 1 if successful, and a negative number to
 *  indicate an error.
 *
 */

#if __STDC__
int     ResetLink (link_t linkid)
#else
int     ResetLink (linkid)
link_t   linkid;
#endif
{
    /* If specified link is not active return error status.             */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Tell driver to reset the transputer.  Return error or success.   */
    return my_ioctl (linkid, T_RESET, NULL) ? ER_LINK_SOFT : 1;
}


/*
 *
 *      int AnalyseLink (link_t linkid)
 *
 *  Analyse transputer.  Returns 1 if successful, and a negative number to
 *  indicate an error.
 *
 */

#if __STDC__
int     AnalyseLink (link_t linkid)
#else
int     AnalyseLink (linkid)
link_t  linkid;
#endif
{
    /* If specified link is not active return error status.             */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Tell driver to reset the transputer.  Return error or success.   */
    return my_ioctl (linkid, T_ANAL, NULL) ? ER_LINK_SOFT : 1;
}


/*
 *
 *      int TestError (link_t linkid)
 *
 *  Returns state of ErrorFlag.
 *
 */

#if __STDC__
int     TestError (link_t linkid)
#else
int     TestError (linkid)
link_t  linkid;
#endif
{
    struct trpio_t  trp;

    /* If specified link is not active return error status.             */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Ask driver for error status.  Return error or status.            */
    return my_ioctl (linkid, T_TSTER, &trp) ? ER_LINK_SOFT : trp.timeout;
}


/*
 *
 *      int TestRead (link_t linkid)
 *
 *  Returns TRUE if transputer is ready to receive data.
 *
 */

#if __STDC__
int     TestRead (link_t linkid)
#else
int     TestRead (linkid)
link_t  linkid;
#endif
{
    struct trpio_t  trp;

    /* If specified link is not active return error status.             */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Ask driver for read status.  Return error or status.             */
    return my_ioctl (linkid, T_TSTRD, &trp) ? ER_LINK_SOFT : !trp.timeout;
}


/*
 *
 *      int TestWrite (link_t linkid)
 *
 *  Returns TRUE if transputer is ready to receive data.
 *
 */

#if __STDC__
int     TestWrite (link_t linkid)
#else
int     TestWrite (linkid)
link_t  linkid;
#endif
{
    struct trpio_t  trp;

    /* If specified link is not active return error status.             */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Ask driver for write status.  Return error or status.            */
    return my_ioctl (linkid, T_TSTWR, &trp) ? ER_LINK_SOFT : !trp.timeout;
}

