/*
 *
 *  @(#)getopt.c    1.2 92/12/20    (c) PACT
 *
 *  Copyright (c) 1991, 1992 PACT
 *  All Rights Reserved
 *
 *  Generic UNIX-style command-line option parser. Is compatible with the
 *  UNIX System V.3 function with the same name.
 *
 */


#include    <stdio.h>


/*
 *
 *      int getopt (int argc, char **argv, char *optstring)
 *
 *      extern  char    *optarg;
 *      extern  int     optind, opterr;
 *
 *  The getopt() function returns the next option letter in argv that matches
 *  a letter in optstring. It supports all the rules of the UNIX command
 *  syntax standard (see intro(1)).
 *  optstring must contain the option letters the command using getopt will
 *  recognize; if a letter is followed by a colon, the option is expected to
 *  have an argument, or group of arguments, which must be separated from it
 *  by white space.
 *  When all options have been processed (i.e., up to the first non-option
 *  argument), getopt() returns -1. The special option '--' may be used to
 *  delimit the end of the options; when it is encountered, -1 will be
 *  returned, and '--' will be skipped.
 *
 *  optarg is set to point to the start of the option-argument on return from
 *  getopt().
 *
 *  getopt() places in optind the argv index of the next argument to be
 *  processed. optind is external and is initialized to 1 before the first
 *  call to getopt().
 *
 *  getopt() prints an error message on standard error and returns a question
 *  mark (?) when it encounters an option letter not included in optstring or
 *  no option argument after an option that expects one. This error message
 *  may be disabled by setting opterr to 0.
 *
 *  Rules:
 *      1.  Command names must be between two and nine characters.
 *      2.  Command names must include lowercase letters and digits only.
 *      3.  Option names must be a single character in length.
 *      4.  All options must be delimited by the - character.
 *     *5.  Options with no arguments may be grouped behind one delimiter.
 *     *6.  The first option-argument following an option must be preceded
 *          by white space.
 *      7.  Option arguments cannot be optional.
 *      8.  Groups of option arguments following an option must be separated
 *          by commas or separated by whitespace and quoted.
 *      9.  All options must precede operands on the command line.
 *      10. The characters -- may be used to delimit the end of the options.
 *      11. The order of the options relative to one another should not matter.
 *      12. The order of operands may matter and position-related interpretations
 *          should be determined on a command-specific basis.
 *      13. The - character preceded and followed by white space should be used
 *          only to mean standard input.
 *
 *   Note: in this implementation, rules 5 and 6 may be violated (like on
 *      UNIX System V.3).
 *
 *  See also: getopt(3C) in the UNIX Programmer's Reference Manual
 *            getopts(1), intro(1) in the UNIX User's Reference Manual
 *
 *
 */

char    *optarg = NULL;                 /* option argument */
int     optind = 1;                     /* next argv entry to do */
int     opterr = 1;                     /* 0 if no error message to be generated */


#ifdef LINUX
int my_getopt (argc, argv, optstring)
#else
int my_getopt (argc, argv, optstring)
#endif
int     argc;
char    **argv;
char    *optstring;
{
    static  int index = 0;              /* if !0, index in argv entry: next option not processed */

    int opt;                            /* option found */
    if (index != 0 && argv[optind][index] == '\0')  /* end of group reached */
        optind++, index=0;                          /* look at next argv entry */
    if (index == 0 && argv[optind][0] == '-') {     /* more options */
        if (argv[optind][1] == '-' && argv[optind][2] == '\0') {    /* '--' -> */           optind++;                   /* skip arg */
            optind++;                   /* skip */
            return -1;                  /* and return -1 */
        }
        else if (argv[optind][1] == '\0')   /* '-' -> */
            return -1;                      /* not an option but stdin */
        index = 1;                      /* now parsing group of options: first option after delimiter */
    }
    if (index != 0) {                   /* parsing group of options? */
        extern  char    *strchr();      /* finds character in string */
        char    *s;
        opt = argv[optind][index++];    /* option */
        s = strchr (optstring, opt);    /* find option in optstring */
        if (s == NULL) {                /* illegal option? */
            fprintf (stderr, "%s: illegal option -- %c\n", argv[0], opt);
            return '?';                 /* return '?' */
        }
        if (*(s+1) == ':') {        /* option argument expected? */
#ifndef STRICT  /* Rule 6 violation */
            if (argv[optind][index] != '\0') {  /* option arg immediately follows option? */
                optarg = &argv[optind][index];  /* this is the argument */
                optind++, index=0;      /* end of this group */
            }
            else
#endif
                 {
                optind++, index=0;      /* end of this group */
                if (optind >= argc) {   /* no args left? */
                    fprintf (stderr, "%s: option requires an argument -- %c\n", argv[0], opt);
                    return '?';
                }
                optarg = argv[optind];  /* this is the argument */
                optind++;               /* next option, index is still 0 */
            }
        }
    }
    else
        opt = -1;                       /* not an option: return -1 */
    return opt;                         /* return option */
}

