/*
 *
 *  @(#)volvox.c    1.1 93/02/04    (c) PACT
 *
 *  PSERVER - PACT host server
 *
 *  Copyright (c) 1991-1993 PACT
 *  All Rights Reserved
 *
 *  Archipel Volvox transputer board interface module.
 *
 */

#include    <stdio.h>                   /* standard I/O */
#include    <fcntl.h>                   /* file control */
#include    <errno.h>                   /* errno */
#include    <sys/types.h>               /* types */
#include    <sys/param.h>               /* for HZ */
#include    <sys/transputer.h>          /* Volvox driver interface */
#include    "linkio.h"                  /* link interface definition */


/*
 *
 *  Currently active link and its name.
 *
 */

#define FREELINK    (-1)
static  link_t      activelink = FREELINK;
static  char        linkname[80] = "";


/*
 *
 *  Default link.
 *
 */

#define DEFAULTLINK     "/dev/vxv0"


/*
 *
 *  Remember current timeout, so it can be changed when necessary.
 *
 */

static int      curtimout = -1;


/*
 *
 *      link_t  OpenLink (char *link)
 *
 *  Open link.  If link is NULL or empty, the default link is used.
 *  Otherwise, link should be the name of the device to use.  Returns
 *  negative number if an error occurs, and a positive linkid if the
 *  link was successfully opened.
 *
 */

#if __STDC__
link_t  OpenLink (char *link)
#else
link_t  OpenLink (link)
char *link;
#endif
{
    /* If a link is already active, we can't open another.              */
    if (activelink != FREELINK)
        return ER_LINK_CANT;

    /* Use default device if no link selected.                          */
    if (!link || *link == '\0')
        link = DEFAULTLINK;

    /* Try to open device.                                              */
    activelink = open (link, O_RDWR);

    /* Error if not able to open link.                                  */
    if (activelink < 0)
        return ER_LINK_CANT;

    /* Return linkid to indicate success.                               */
    strcpy (linkname, link);
    return activelink;
}


/*
 *
 *      int CloseLink (link_t linkid)
 *
 *  Close link.  Must be open to be closed.  If any error occurs (e.g. link
 *  is not open), a negative number is returned.  Otherwise, 1 is returned.
 *
 */

#if __STDC__
int     CloseLink (link_t linkid)
#else
int     CloseLink (linkid)
link_t  linkid;
#endif
{
    /* Check if link is actually open.                                  */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Close device.                                                    */
    (void) close (linkid);

    /* Free link.                                                       */
    activelink = FREELINK;

    /* Return 1 to indicate success.                                    */
    return 1;
}


/*
 *
 *      int ReadLink (link_t linkid, char *buf, unsigned len, int timout)
 *
 *  Read len bytes from specified link into buffer.  Returns number of
 *  bytes read (before timeout), or a negative number if an error occurs
 *  (wrong link or 0-size transfer).  The timout parameter specifies
 *  the max number of tenths of seconds the transfer may take.
 *
 */

#if  __STDC__
int     ReadLink (link_t linkid, char *buf, unsigned len, int timout)
#else
int     ReadLink (linkid, buf, len, timout)
link_t      linkid;
char        *buf;
unsigned    len;
int         timout;
#endif
{
    int res;

    /* If linkid is not the active link, fail.                          */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Zero-size transfers can't be done.                               */
    if (len < 1)
        return ER_LINK_CANT;

    /* Change timeout if necessary.                                     */
    timout = (timout * HZ) / 10;
    if (timout && timout != curtimout) {
        (void) ioctl (linkid, VXV_W_TOR, &timout);
        (void) ioctl (linkid, VXV_W_TOW, &timout);
        curtimout = timout;
    }

    /* Read from transputer link and return # bytes done.               */
    res = read (linkid, buf, len);
    if (res == -1)
        return (errno == EIO) ? 0 : ER_LINK_SOFT;
    else
        return res;
}


/*
 *
 *      int WriteLink (link_t linkid, char *buf, unsigned len, int timout)
 *
 *  Write len bytes from specified link into buffer.  Returns number of
 *  bytes written (before timeout), or a negative number if an error occurs
 *  (wrong link or 0-size transfer).  The timout parameter specifies
 *  the max number of tenths of seconds the transfer may take.
 *
 */

#if __STDC__
int     WriteLink (link_t linkid, char *buf, unsigned len, int timout)
#else
int     WriteLink (linkid, buf, len, timout)
link_t      linkid;
char        *buf;
unsigned    len;
int         timout;
#endif
{
    int res;

    /* If linkid is not the active link, fail.                          */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Zero-size transfers can't be done.                               */
    if (len < 1)
         return ER_LINK_CANT;

    /* Change timeout if necessary.                                     */
    timout = (timout * HZ) / 10;
    if (timout && timout != curtimout) {
        (void) ioctl (linkid, VXV_W_TOR, &timout);
        (void) ioctl (linkid, VXV_W_TOW, &timout);
        curtimout = timout;
    }

    /* Write to transputer link and return # bytes done.                */
    res = write (linkid, buf, len);
    if (res == -1)
        return (errno == EIO) ? 0 : ER_LINK_SOFT;
    else
        return res;
}


/*
 *
 *      int ResetLink (link_t linkid)
 *
 *  Reset transputer.  Returns 1 if successful, and a negative number to
 *  indicate an error.
 *
 */

#if __STDC__
int     ResetLink (link_t linkid)
#else
int     ResetLink (linkid)
link_t  linkid;
#endif
{
    int mask;

    /* If specified link is not active return error status.             */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Tell driver to reset board.                                      */
    mask = 0x10;
    (void) ioctl (linkid, VXV_RESET, &mask);

    /* Return success.                                                  */
    return 1;
}


/*
 *
 *      int AnalyseLink (link_t linkid)
 *
 *  Analyse transputer.  Returns 1 if successful, and a negative number to
 *  indicate an error.
 *
 */

#if __STDC__
int     AnalyseLink (link_t linkid)
#else
int     AnalyseLink (linkid)
link_t  linkid;
#endif
{
     int mask;

    /* If specified link is not active return error status.             */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Tell driver to analyse board.                                    */
    mask = 0x10;
    (void) ioctl (linkid, VXV_ANALYSE, &mask);

    /* Return success.                                                  */
    return 1;
}


/*
 *
 *      int TestError (link_t linkid)
 *
 *  Returns state of ErrorFlag.
 *
 */

#if __STDC__
int     TestError (link_t linkid)
#else
int     TestError (linkid)
link_t  linkid;
#endif
{
    int status;

    /* If specified link is not active return error status.             */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Ask driver for error status.  Return it.                         */
    (void) ioctl (linkid, VXV_STAT_ERR, &status);
    return status & 0x10;
}


/*
 *
 *      int TestRead (link_t linkid)
 *
 *  Returns TRUE if transputer is ready to receive data.
 *
 */

#if __STDC__
int     TestRead (link_t linkid)
#else
int     TestRead (linkid)
link_t  linkid;
#endif
{
    int status;

    /* If specified link is not active return error status.             */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Ask driver for read status.  Return it.                          */
    (void) ioctl (linkid, VXV_STAT_READ, &status);
    return status & 1;
}


/*
 *
 *      int TestWrite (link_t linkid)
 *
 *  Returns TRUE if transputer is ready to receive data.
 *
 */

#if __STDC__
int     TestWrite (link_t linkid)
#else
int     TestWrite (linkid)
link_t  linkid;
#endif
{
    int status;

    /* If specified link is not active return error status.             */
    if (linkid != activelink)
        return ER_LINK_BAD;

    /* Ask driver for write status.  Return it.                         */
    (void) ioctl (linkid, VXV_STAT_WRITE, &status);
    return status & 1;
}

