/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Client-Server Library.                                                  */
/* Server channels.                                                        */
/* This module implements the server channels. Server channels are used    */
/* to send and receive data respectively to and from a client application. */
/* These functions are not directly used by a PPC program but are called   */
/* from the channel implementation of the supported parallel               */
/* architectures.                                                          */
/* This module is only used by UNIX and PVM PPC programs. For PACT-C       */
/* programs, see chanservpact.c.                                           */
/* Note that there are only two possible channels, one for input, the      */
/* other for output.                                                       */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 16.02.1994  Droux       Created                                         */
/* 08.07.1994  Droux       New structure of source directory               */
/*-------------------------------------------------------------------------*/
/* filename: chanservunix.c                                                */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <stdio.h>
#include <syslog.h>
#include <rpc/rpc.h>
#include <ppc/csp/types.h>
#include "chanservunix.h"
#include "ion.h"

typedef enum
{              /* specifies the direction of the channel when creating XDR */
  INPUT,
  OUTPUT
} direction;

typedef struct         /* contains the characteristics of a client channel */
{
  int free;                                           /* 1 if slot is free */
  XDR xdrs;              /* client-server connection is implemented by XDR */
  int socket;         /* TCP socket on which data must be send or received */
  FILE *stream;                           /* stream associated with socket */
} serv_chan_descr;

/* A PPC server program can be connected to only one client application.   */
/* Thus, only two channels are necessary, one from the client, and the     */
/* other to it.                                                            */
serv_chan_descr serv_chans [2];                         /* server channels */

/*-------------------------------------------------------------------------*/
/* init_serv_chan:                                                   LOCAL */
/* Init a new server channel.                                              */
/*-------------------------------------------------------------------------*/
void init_serv_chan (
  channel c,
  int socket,
  direction dir
)
{
  enum xdr_op x_op;                                       /* XDR direction */
  char stream_mode [2];                    /* mode when opening new stream */
  
  if (dir == INPUT)     /* determine XDR direction and stream opening mode */
  {
    x_op = XDR_DECODE;
    strcpy(stream_mode, "r");
  }
  else
  {
    x_op = XDR_ENCODE;
    strcpy(stream_mode, "w");
  }
  
  serv_chans[c].socket = socket;
  serv_chans[c].stream = fdopen(socket, stream_mode);
  
  xdrstdio_create(            /* create XDR based client-server connection */
    &(serv_chans[c].xdrs),                                   /* XDR handle */
    serv_chans[c].stream,                                     /* IO stream */
    x_op);
}

/*-------------------------------------------------------------------------*/
/* servunix_init_channels:                                                 */
/* Initializes the two channels from and to the client application with    */
/* the given values.                                                       */
/* This function is called by the chanunix module.                         */
/*-------------------------------------------------------------------------*/
void servunix_init_channels(
  int from_client,
  int to_client
)
{
  init_serv_chan(FROM_CLIENT, from_client, INPUT);
  init_serv_chan(TO_CLIENT, to_client, OUTPUT);
}

/*-------------------------------------------------------------------------*/
/* servunix_chan_write:                                                    */
/* Send the given data to the client.                                      */
/* <nitems> corresponds to the number of items of type <type> to be        */
/* transmitted.                                                            */
/*-------------------------------------------------------------------------*/
int servunix_chan_write (
  channel id, 
  void *vbuf, 
  int nitems, 
  int type
)
{
  static XDR *xdrs;
  xdrs = &(serv_chans[id].xdrs);
  
  switch(type)
  {
    case INT:
      xdr_int(xdrs, (int *)vbuf);
      break;
    case SHORT:
      xdr_short(xdrs, (short *)vbuf);
      break;
    case UNSIGNED_SHORT:
      xdr_u_short(xdrs, (unsigned short *)vbuf);
      break;
    case LONG:
      xdr_long(xdrs, (long *)vbuf);
      break;
    case UNSIGNED_LONG:
      xdr_u_long(xdrs, (unsigned long *)vbuf);
      break;
    case FLOAT:
      xdr_float(xdrs, (float *)vbuf);
      break;
    case DOUBLE:
      xdr_double(xdrs, (double *)vbuf);
      break;
    case CHAR:
      xdr_char(xdrs, (char *)vbuf);
      break;
    case INT_ARRAY:
      xdr_vector(xdrs, (char *)vbuf, nitems, 
	type_size[type], xdr_int);
      break;
    case SHORT_ARRAY:
      xdr_vector(xdrs, (char *)vbuf, nitems, 
	type_size[type], xdr_short);
      break;
    case UNSIGNED_SHORT_ARRAY:
      xdr_vector(xdrs, (char *)vbuf, nitems, 
	type_size[type], xdr_u_short);
      break;
    case LONG_ARRAY:
      xdr_vector(xdrs, (char *)vbuf, nitems, 
	type_size[type], xdr_long);
      break;
    case UNSIGNED_LONG_ARRAY:
      xdr_vector(xdrs, (char *)vbuf, nitems, 
	type_size[type], xdr_u_long);
      break;
    case FLOAT_ARRAY:
      xdr_vector(xdrs, (char *)vbuf, nitems, 
	type_size[type], xdr_float);
      break;
    case DOUBLE_ARRAY:
      xdr_vector(xdrs, (char *)vbuf, nitems, 
	type_size[type], xdr_double);
      break;
    case CHAR_ARRAY:
      xdr_vector(xdrs, (char *)vbuf, nitems, 
	type_size[type], xdr_char);
      break;
  }
  if (id == TO_CLIENT)
    fflush(serv_chans[id].stream);
  return 0;
}

/*-------------------------------------------------------------------------*/
/* unix_start_write:                                                       */
/* Starts a channel transmission of data. Must be called before any call   */
/* to unix_chan_write for each message transmission.                       */
/* Does nothing under here. Implemented for compatibility with the PVM     */
/* lower-half module.                                                      */
/*-------------------------------------------------------------------------*/
void servunix_start_write (void)
{
}

/*-------------------------------------------------------------------------*/
/* servunix_end_write:                                                     */
/* Terminates a channel transmission of data. Must be called after all     */
/* calls to unix_chan_write for each message transmission.                 */
/* Does nothing here. Implemented for compatibility with the PVM           */
/* lower-half module.                                                      */
/*-------------------------------------------------------------------------*/
void servunix_end_write (
  channel id
)
{
}

/*-------------------------------------------------------------------------*/
/* servunix_start_read:                                                    */
/* Starts a channel reception of data. Must be called before any call      */
/* to unix_chan_read for each message reception.                           */
/* Does nothing here. Implemented for compatibility with the PVM           */
/* lower-half module.                                                      */
/*-------------------------------------------------------------------------*/
void servunix_start_read (
  channel id
)
{
}
