/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Client-Server Library.                                                  */
/* The functions implemented by this module are used by a PACT-C server    */
/* program to send and receive typed data respectively to and from a       */
/* remote client through the extended pserver.                             */
/* These functions follow the XDR specifications, i.e. the format of the   */
/* data received or sent respectively from and to the client are           */
/* converted from and to the XDR format to and from the transputer format. */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 30.08.1994  Droux       Created                                         */
/*-------------------------------------------------------------------------*/
/* filename: chservpact.c                                                  */
/* Tested on :                                                             */
/*   PACT-C v93.1, T800, T805                                              */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include "chservpact.h"
#include "pactcserv.h"

#define BUF_LEN 990

#define SWAP_32(x, y) {src = (char *)x; \
  dest = &(((char *)y)[3]); \
  *(dest--) = *(src++); \
  *(dest--) = *(src++); \
  *(dest--) = *(src++); \
  *dest = *src;}

#define SWAP_64(x, y) { \
  SWAP_32(x, y); \
  SWAP_32(&(((char *)x)[4]), &(((char *)y)[4])); \
  }

#define WRITEN(fd, ptr, cnt) for (nleft = cnt, out_buffer = ptr; nleft > 0;) \
  { \
    ntransmitted = write(fd, out_buffer, nleft); \
    nleft -= ntransmitted; \
    out_buffer += ntransmitted; \
  }
  
#define READN(fd, ptr, cnt) {\
  for (nleft = cnt, in_buffer = ptr; nleft > 0;) \
  { \
    ntransmitted = read(fd, in_buffer, nleft); \
    nleft -= ntransmitted; \
    in_buffer += ntransmitted; \
  }}

static char out_tmp_32 [4];
static char out_tmp_64 [8];
static char in_tmp_32 [4];
static char in_tmp_64 [8];
static char *src;
static char *dest;
static int nleft;
static int ntransmitted;
static char *in_buffer;
static char *out_buffer;
static long out_tmp_long;
static long in_tmp_long;
static char input_buffer [BUF_LEN];
static char output_buffer [BUF_LEN];

/*-------------------------------------------------------------------------*/
/* scalar output                                                           */
/*-------------------------------------------------------------------------*/

void cslib_xdr_send_int (int *n)
{
  SWAP_32(n, out_tmp_32);
  WRITEN(SERVER_TO_CLIENT_FD, out_tmp_32, 4);
}

void cslib_xdr_send_short (short *n)
{
  out_tmp_long = (long)*n;
  SWAP_32(&out_tmp_long, out_tmp_32);
  WRITEN(SERVER_TO_CLIENT_FD, out_tmp_32, 4);
}

void cslib_xdr_send_float (float *n)
{
  SWAP_32(n, out_tmp_32);
  WRITEN(SERVER_TO_CLIENT_FD, out_tmp_32, 4);
}

void cslib_xdr_send_double (double *n)
{
  SWAP_64(n, out_tmp_64);
  WRITEN(SERVER_TO_CLIENT_FD, out_tmp_64, 8);
}

void cslib_xdr_send_char (char *n)
{
  out_tmp_long = (long)*n;
  SWAP_32(&out_tmp_long, out_tmp_32);
  WRITEN(SERVER_TO_CLIENT_FD, out_tmp_32, 4);
}

/*-------------------------------------------------------------------------*/
/* vector output                                                           */
/*-------------------------------------------------------------------------*/

void cslib_xdr_send_int_vector (int *n, int l)
{
  int left;
  int *data;
  int idx;
  int limit = BUF_LEN/4;
  int *out_buf;
  
  for (left = l, data = n; left > 0;)
  {
    out_buf = (int *)output_buffer;
    for (idx = 0; idx < left && idx < limit; idx++)
    {
      SWAP_32(data, (char *)out_buf);
      data++;
      out_buf++;
    }
    WRITEN(SERVER_TO_CLIENT_FD, (char *)output_buffer, idx*4);
    left -= idx;
  }
}

void cslib_xdr_send_short_vector (short *n, int l)
{
  int left;
  short *data;
  int idx;
  int limit = BUF_LEN/4;
  int *out_buf;
  
  for (left = l, data = n; left > 0;)
  {
    out_buf = (int *)output_buffer;
    for (idx = 0; idx < left && idx < limit; idx++)
    {
      out_tmp_long = (long)*data;
      SWAP_32(&out_tmp_long, (char *)out_buf);
      data++;
      out_buf++;
    }
    WRITEN(SERVER_TO_CLIENT_FD, (char *)output_buffer, idx*4);
    left -= idx;
  }
}

void cslib_xdr_send_float_vector (float *n, int l)
{
  int left;
  float *data;
  int idx;
  int limit = BUF_LEN/4;
  int *out_buf;
  
  for (left = l, data = n; left > 0;)
  {
    out_buf = (int *)output_buffer;
    for (idx = 0; idx < left && idx < limit; idx++)
    {
      SWAP_32(data, (char *)out_buf);
      data++;
      out_buf++;
    }
    WRITEN(SERVER_TO_CLIENT_FD, (char *)output_buffer, idx*4);
    left -= idx;
  }
}

void cslib_xdr_send_double_vector (double *n, int l)
{
  int left;
  double *data;
  int idx;
  int limit = BUF_LEN/8;
  double *out_buf;
  
  for (left = l, data = n; left > 0;)
  {
    out_buf = (double *)output_buffer;
    for (idx = 0; idx < left && idx < limit; idx++)
    {
      SWAP_64(data, (char *)out_buf);
      data++;
      out_buf++;
    }
    WRITEN(SERVER_TO_CLIENT_FD, (char *)output_buffer, idx*8);
    left -= idx;
  }
}

void cslib_xdr_send_char_vector (char *n, int l)
{
  int left;
  char *data;
  int idx;
  int limit = BUF_LEN/4;
  int *out_buf;
  int tmp_int;
  
  for (left = l, data = n; left > 0;)
  {
    out_buf = (int *)output_buffer;
    for (idx = 0; idx < left && idx < limit; idx++)
    {
      tmp_int = (int)*data;
      SWAP_32(&tmp_int, (char *)out_buf);
      data++;
      out_buf++;
    }
    WRITEN(SERVER_TO_CLIENT_FD, (char *)output_buffer, idx*4);
    left -= idx;
  }
}

/*-------------------------------------------------------------------------*/
/* scalar input                                                            */
/*-------------------------------------------------------------------------*/

void cslib_xdr_get_int (int *n)
{
  READN(CLIENT_TO_SERVER_FD, in_tmp_32, 4);
  SWAP_32(in_tmp_32, n);
}

void cslib_xdr_get_short (short *n)
{
  READN(CLIENT_TO_SERVER_FD, in_tmp_32, 4);
  SWAP_32(in_tmp_32, &in_tmp_long);
  *n = (short)in_tmp_long;
}

void cslib_xdr_get_float (float *n)
{
  READN(CLIENT_TO_SERVER_FD, in_tmp_32, 4);
  SWAP_32(in_tmp_32, n);
}

void cslib_xdr_get_double (double *n)
{
  READN(CLIENT_TO_SERVER_FD, in_tmp_64, 8);
  SWAP_64(in_tmp_64, n);
}

void cslib_xdr_get_char (char *n)
{
  READN(CLIENT_TO_SERVER_FD, in_tmp_32, 4);
  SWAP_32(in_tmp_32, &in_tmp_long);
  *n = (char)in_tmp_long;
}

/*-------------------------------------------------------------------------*/
/* vector input                                                            */
/*-------------------------------------------------------------------------*/

void cslib_xdr_get_int_vector (int *n, int l)
{
  int left;
  int *data;
  int idx;
  int limit = BUF_LEN/4;
  int *in_buf;
  int to_read;
  
  for (left = l, data = n; left > 0;)
  {
    to_read = left > limit ? limit : left;
    in_buf = (int *)input_buffer;
    READN(CLIENT_TO_SERVER_FD, (char *)input_buffer, to_read*4);
    for (idx = 0; idx < to_read; idx++)
    {
      SWAP_32((char *)in_buf, data);
      data++;
      in_buf++;
    }
    left -= to_read;
  }
}

void cslib_xdr_get_short_vector (short *n, int l)
{
  int left;
  short *data;
  int idx;
  int limit = BUF_LEN/4;
  int *in_buf;
  int to_read;
  
  for (left = l, data = n; left > 0;)
  {
    to_read = left > limit ? limit : left;
    in_buf = (int *)input_buffer;
    READN(CLIENT_TO_SERVER_FD, (char *)input_buffer, to_read*4);
    for (idx = 0; idx < to_read; idx++)
    {
      SWAP_32((char *)in_buf, &in_tmp_long);
      *data = (short)in_tmp_long;
      data++;
      in_buf++;
    }
    left -= to_read;
  }
}

void cslib_xdr_get_float_vector (float *n, int l)
{
  int left;
  float *data;
  int idx;
  int limit = BUF_LEN/4;
  int *in_buf;
  int to_read;
  
  for (left = l, data = n; left > 0;)
  {
    to_read = left > limit ? limit : left;
    in_buf = (int *)input_buffer;
    READN(CLIENT_TO_SERVER_FD, (char *)input_buffer, to_read*4);
    for (idx = 0; idx < to_read; idx++)
    {
      SWAP_32((char *)in_buf, data);
      data++;
      in_buf++;
    }
    left -= to_read;
  }
}

void cslib_xdr_get_double_vector (double *n, int l)
{
  int left;
  double *data;
  int idx;
  int limit = BUF_LEN/8;
  int *in_buf;
  int to_read;
  
  for (left = l, data = n; left > 0;)
  {
    to_read = left > limit ? limit : left;
    in_buf = (int *)input_buffer;
    READN(CLIENT_TO_SERVER_FD, (char *)input_buffer, to_read*8);
    for (idx = 0; idx < to_read; idx++)
    {
      SWAP_64((char *)in_buf, data);
      data++;
      in_buf++;
    }
    left -= to_read;
  }
}

void cslib_xdr_get_char_vector (char *n, int l)
{
  int left;
  char *data;
  int idx;
  int limit = BUF_LEN/4;
  int *in_buf;
  int to_read;
  int tmp_int;
  
  for (left = l, data = n; left > 0;)
  {
    to_read = left > limit ? limit : left;
    in_buf = (int *)input_buffer;
    READN(CLIENT_TO_SERVER_FD, (char *)input_buffer, to_read*4);
    for (idx = 0; idx < to_read; idx++)
    {
      SWAP_32((char *)in_buf, &tmp_int);
      *data = (char)tmp_int;
      data++;
      in_buf++;
    }
    left -= to_read;
  }
}

