/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Alternate construction PVM implementation.                              */
/* This source file contains the implementation of the functions defined   */
/* in the header file alt.h.                                               */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 15.01.1994  Droux       Created                                         */
/* 08.07.1994  Droux       New structure of source directory               */
/*-------------------------------------------------------------------------*/
/* filename: altpvm.c                                                      */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <sys/types.h>
#include <sys/time.h>
#include <syslog.h>
#include <signal.h>
#ifdef SUN4
#include <stddef.h>
#else
#include <stdlib.h>
#endif
#include <ppc/csp/chan.h>
#include <ppc/csp/alt.h>
#include <ppc/csp/chanpvmuh.h>

#define MAX_PAUSE 50                                 /* max pause of .5 ms */
#define INIT_PAUSE 1                              /* initial pause of 1 us */

typedef enum {                           /* alternate construct categories */
  NONE,
  PVM_ONLY,                    /* alternate construct on PVM channels only */
  UNIX_ONLY,                  /* alternate construct on UNIX channels only */
  PVM_AND_UNIX             /* alternate construct on PVM and UNIX channels */
} mix_categories;

static int alt_tids [MAX_ALT_CHANS];     /* tids of channels to be watched */
static int alt_tags [MAX_ALT_CHANS];     /* tags of channels to be watched */

/*-------------------------------------------------------------------------*/
/* alt_init:                                                               */
/* Initializes the alternate construct.                                    */
/* This functions is not called by the user directly but it is generated   */
/* by the macro alt_begin().                                               */
/*-------------------------------------------------------------------------*/
void alt_init (alt_status *act_alt)
{
  act_alt->nb_chans = act_alt->the_timeout = 0;
}

/*-------------------------------------------------------------------------*/
/* unix_only_alt_wait:                                                     */
/* alt_wait called when only UNIX channels have been declared.             */
/*-------------------------------------------------------------------------*/
void unix_only_alt_wait (alt_status *act_alt, channel *winner)
{
  static last_returned = 0;
  fd_set readfds;
  fd_set writefds;
  fd_set exceptfds;
  struct timeval timeout;
  void *timeout_ptr;
  int nfound;
  int n;
  int max_fd;
  
  FD_ZERO(&readfds);                                    /* clear all flags */
  FD_ZERO(&writefds);
  FD_ZERO(&exceptfds);
  
  if (act_alt->the_timeout)                  /* determine timeout to apply */
  {
    timeout.tv_sec = 0;
    timeout.tv_usec = act_alt->the_timeout;
    timeout_ptr = &timeout;
  }
  else
    timeout_ptr = NULL;
  
  max_fd = 0;
  for (n = 0; n < act_alt->nb_chans; n++)    /* add channels to be checked */
  {
    int act_fd = unix_read_fd(act_alt->chans[n]);
    FD_SET(act_fd, &readfds);
    if (act_fd > max_fd)
      max_fd = act_fd;
  }
     
  nfound = select(max_fd + 1,                /* wait for something to come */
    &readfds, &writefds, &exceptfds, 
    timeout_ptr);
    
  if (!nfound)
    *winner = TIMEOUT;
  else
  {
    int idx;
    for (n = 0; n < act_alt->nb_chans; n++)
    {
      idx = (n + last_returned) % act_alt->nb_chans;
      if (FD_ISSET(unix_read_fd(act_alt->chans[idx]), &readfds))
      {
        last_returned = idx + 1;
	*winner = act_alt->chans[idx];
	break;
      }
    }
  }
}

/*-------------------------------------------------------------------------*/
/* actual_time:                                                            */
/* Returns the number of microseconds elapsed since midnight (0 hour),     */
/* January 1, 1970.                                                        */
/*-------------------------------------------------------------------------*/
long actual_time ()
{
  struct timeval tp;
  struct timezone tzp;
  
  gettimeofday(&tp, &tzp);
  return tp.tv_sec*1000000 + tp.tv_usec;
}

/*-------------------------------------------------------------------------*/
/* pvm_only_alt_wait:                                                      */
/* alt_wait called when only PVM channels have been declared.              */
/*-------------------------------------------------------------------------*/
void pvm_only_alt_wait (alt_status *act_alt, channel *winner)
{
  static int inc = 1;
  static last_returned = 0;              /* last winner index returned + 1 */
  int c;                                         /* counter for misc usage */
  unsigned delay;         /* pause between two PVM polling in microseconds */
  long end_time;
  
  for (c = 0; c < act_alt->nb_chans; c++)
  {                                          /* PVM channels tags and tids */
    pvm_tid_and_tag(
      lh_chan_id(act_alt->chans[c]), 
      &alt_tids[c], &alt_tags[c]);
  }
  
  delay = INIT_PAUSE;
  end_time = actual_time() + act_alt->the_timeout;
  do
  {
    for (c = last_returned % act_alt->nb_chans; c < act_alt->nb_chans; c++)
    {
      if (pvm_nrecv(alt_tids[c], alt_tags[c]) > 0)
      {                            /* a message is present on this channel */
        *winner = act_alt->chans[c];           /* update winner channel id */
	pvm_disable_recv(lh_chan_id(act_alt->chans[c]));   /* disable recv */
   	            /* update last_returned to simulate non-predeterminism */
	last_returned = c + inc;
	inc++;
	return;                                        /* return to caller */
      }
    }
    
    usleep(delay);       /* sleep during a short time between two pollings */
    if (delay < MAX_PAUSE)
      delay *= 2;
  } while ((actual_time() <= end_time) || (act_alt->the_timeout == 0));
  
  *winner = TIMEOUT;           /* there is nothing to be read on a channel */
}

/*-------------------------------------------------------------------------*/
/* pvm_and_unix_alt_wait:                                                  */
/* alt_wait called when PVM and UNIX channels have been declared.          */
/*-------------------------------------------------------------------------*/
void pvm_and_unix_alt_wait (
  alt_status *act_alt, 
  channel *winner,
  lh_types *chan_types)
{
  static int inc = 1;
  static last_returned = 0;              /* last winner index returned + 1 */
  int c;                                         /* counter for misc usage */
  unsigned delay;         /* pause between two PVM polling in microseconds */
  fd_set readfds;                         /* file descriptors for select() */
  fd_set writefds;
  fd_set exceptfds;
  struct timeval timeout;                     /* zero timeout for select() */
  long end_time;
  int unix_pipe;
  int found;
  
  FD_ZERO(&readfds);                                    /* clear all flags */
  FD_ZERO(&writefds);
  FD_ZERO(&exceptfds);

  timeout.tv_sec = 0;                           /* set timeout for polling */
  timeout.tv_usec = 0;

  for (c = 0; c < act_alt->nb_chans; c++)
  {
    if (chan_types[c] == PVM_CHAN)
    {                                        /* PVM channels tags and tids */
      pvm_tid_and_tag(
	lh_chan_id(act_alt->chans[c]), 
	&alt_tids[c], &alt_tags[c]);
    }
  }
  
  delay = INIT_PAUSE;
  end_time = actual_time() + act_alt->the_timeout;
  do
  {
    for (c = last_returned % act_alt->nb_chans; c < act_alt->nb_chans; c++)
    {
      if (chan_types[c] == PVM_CHAN)
      {                                            /* poll the PVM channel */
        found = pvm_nrecv(alt_tids[c], alt_tags[c]);
      }
      else
      {                                           /* poll the UNIX channel */
        FD_ZERO(&readfds);
        unix_pipe = unix_read_fd(lh_chan_id(act_alt->chans[c]));
        FD_SET(unix_pipe, &readfds);
        found = select(unix_pipe+1, 
	  &readfds, &writefds, &exceptfds, &timeout);
      }
      
      if (found)
      {                            /* a message is present on this channel */
        *winner = act_alt->chans[c];           /* update winner channel id */
        if (chan_types[c] == PVM_CHAN)
	  pvm_disable_recv(lh_chan_id(act_alt->chans[c])); /* disable recv */
   	            /* update last_returned to simulate non-predeterminism */
	last_returned = c + inc;
	inc++;
	return;                                        /* return to caller */
      }
    }
    
    usleep(delay);     
    if (delay < MAX_PAUSE)
      delay *= 2;
  } while ((actual_time() <= end_time) || (act_alt->the_timeout == 0));
  
  *winner = TIMEOUT;           /* there is nothing to be read on a channel */
}

/*-------------------------------------------------------------------------*/
/* alt_wait:                                                               */
/* Activate the alternate construct. Waits for data comming from one of the*/
/* channels given in the specified alt status variable. Exists if some     */
/* data is arrived or if the timeout is reached. Returns TIMEOUT in this   */
/* case or the winner channel id otherwise.                                */
/*-------------------------------------------------------------------------*/
void alt_wait (alt_status *act_alt, channel *winner)
{
  lh_types chan_types [MAX_ALT_CHANS];         /* lower-half channel types */
  int c;                                         /* counter for misc usage */
  mix_categories mix_category;                   /* channels of each types */
  
  /* determine the mix category, i.e. the type of channels that have been  */
  /* declared to be watched during the alternate construct. If every       */
  /* declared channels are of the same type, an optimized selection        */
  /* algorithm is executed according to the channel type.                  */
  mix_category = NONE;
  for (c = 0; c < act_alt->nb_chans; c++)      /* fill channel types array */
  {
    chan_types[c] = lh_chan_type(act_alt->chans[c]);
    switch(mix_category)
    {
      case NONE:
        if (chan_types[c] == PVM_CHAN)
	  mix_category = PVM_ONLY;
	else
	  mix_category = UNIX_ONLY;
        break;
      case PVM_ONLY:
        if (chan_types[c] == PVM_CHAN)
	  mix_category = PVM_ONLY;
	else
	  mix_category = PVM_AND_UNIX;
        break;
      case UNIX_ONLY:
        if (chan_types[c] == PVM_CHAN)
	  mix_category = PVM_AND_UNIX;
	else
	  mix_category = UNIX_ONLY;
        break;
      case PVM_AND_UNIX:
	mix_category = PVM_AND_UNIX;
        break;
      default:
        break;
    }
  }
       /* call the alternate selection function according to the mix types */
  switch(mix_category)
  {
    case NONE:
      break;
    case PVM_ONLY:
      pvm_only_alt_wait(act_alt, winner);
      break;
    case UNIX_ONLY:
      unix_only_alt_wait(act_alt, winner);
      break;
    case PVM_AND_UNIX:
      pvm_and_unix_alt_wait(act_alt, winner, chan_types);
      break;
    default:
      break;
  }
}

/*-------------------------------------------------------------------------*/
/* alt_add_channel:                                                        */
/* Add the given channel to the list of channels to be listen.             */
/*-------------------------------------------------------------------------*/
void alt_add_channel (alt_status *act_alt, channel c)
{
  act_alt->chans[act_alt->nb_chans++] = c;
}

/*-------------------------------------------------------------------------*/
/* alt_add_timeout:                                                        */
/* Specifies the timeout of the alternate construct. If this function is   */
/* not called, then the alternate constrct terminates only if something is */
/* available on one of the specified channels.                             */
/*-------------------------------------------------------------------------*/
void alt_add_timeout (alt_status *act_alt, long t)
{
  act_alt->the_timeout = t;
}
