/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* PACT-C channels implementation.                                         */
/* This file implements the functions that are defined in the header file  */
/* chan.h and chanpriv.h.                                                  */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 20.01.1994  Droux       Created                                         */
/* 17.02.1994  Droux       va_arg(char|short|float) bug fixed              */
/* 15.07.1994  Droux       New structure of source directory               */
/*-------------------------------------------------------------------------*/
/* filename: chanpact.c                                                    */
/* Tested on :                                                             */
/*   PACT-C 93.1, T800, T805                                               */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define chan_impl

#include <stdarg.h>
#include <par.h>
#include <sem.h>
#include <process.h>
#include <limits.h>
#include <ppc/csp/chanpriv.h>
#include <ppc/csp/chanpact.h>
#include "chanpactc.h"
#include "pactcserv.h"
#include "chservpact.h"

#define MAX_PACTC_CHANS 100              /* max # of channels for the task */
#define SYNC_CHAR 'X'

typedef enum        /* identifies whether a channel is a PACT-C channel or */
{                                                      /* a CS-Lib channel */
  PACTC_CHAN_TYPE,
  CSLIB_CHAN_TYPE
} chan_type_enum;

/* PACT-C 93.1 does not support the alternate construct on virtual         */
/* channels, but PPC does. In order to avoid this PACT-C limitation,       */
/* a PPC channel is made by two PACT-C channels: a synchronization and     */
/* a data channels.                                                        */

typedef struct
{
  chan_type_enum type;
  void |* data_channel;
  char |* sync_channel;
} pactc_chan_type;

static int nb_pactc_channels = 0;           /* actual # of PACT-C channels */
static pactc_chan_type pactc_channels [MAX_PACTC_CHANS];  /* PACT channels */
static sem_t prot_sem = SEMINIT(1);   /* nb_pactc_channels mutex semaphore */

/*-------------------------------------------------------------------------*/
/* chan_init:                                                              */
/* Initializes the channels implementation. This function is never called  */
/* directly by the programmer.                                             */
/*-------------------------------------------------------------------------*/
void chan_init (void)
{
}

/*-------------------------------------------------------------------------*/
/* par_channel:                                                            */
/* Creates a new channel to communicate between two parallel threads.      */
/*-------------------------------------------------------------------------*/
channel par_channel (void)
{
  int chan_idx;                            /* channel index to be returned */
  pactc_chan_type *ptr;
  
  semp(&prot_sem);                                    /* acquire semaphore */
  chan_idx = nb_pactc_channels;
  ptr = &(pactc_channels[chan_idx]);
  ptr->data_channel = challoc(1);
  ptr->sync_channel = (char |*)ptr->data_channel;
  ptr->type = PACTC_CHAN_TYPE;
  nb_pactc_channels++;
  semv(&prot_sem);                                    /* release semaphore */
  
  return chan_idx;
}

/*-------------------------------------------------------------------------*/
/* chan_close:                                                             */
/* Close the given channel.                                                */
/*-------------------------------------------------------------------------*/
void chan_close (channel id)
{ 
  channel dumb = id;                     /* just to avoid annoying warning */
}

/*-------------------------------------------------------------------------*/
/* send_channel:                                                           */
/* Send the given data on the given channel.                               */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value>                             */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    send_channel(c, INT, a, FLOAT_ARRAY, 100, b, END);                   */
/*-------------------------------------------------------------------------*/
void send_channel (channel id, ...)
{
  va_list pvar;
  void |*data_chan;
  int nitems;
  int act_type;                                        /* actual item type */

  va_start(pvar, id);
  
  if (pactc_channels[id].type == PACTC_CHAN_TYPE)
  {                                                      /* PACT-C channel */
    |*pactc_channels[id].sync_channel = SYNC_CHAR;
    data_chan = pactc_channels[id].data_channel;
    
    act_type = va_arg(pvar, int);            /* get first type, look-ahead */
    while(act_type != END)
    {
      switch(act_type)
      {
	case INT:
	  {
	    int val = va_arg(pvar, int);
	    chwrite(data_chan, &val, type_size[act_type]);
	    break;
	  }
	case SHORT:
	  {
	    short val = va_arg(pvar, int);
	    chwrite(data_chan, &val, type_size[act_type]);
	    break;
	  }
	case UNSIGNED_SHORT:
	  {
	    unsigned short val = va_arg(pvar, unsigned int);
	    chwrite(data_chan, &val, type_size[act_type]);
	    break;
	  }
	case LONG:
	  {
	    long val = va_arg(pvar, long);
	    chwrite(data_chan, &val, type_size[act_type]);
	    break;
	  }
	case UNSIGNED_LONG:
	  {
	    unsigned long val = va_arg(pvar, unsigned long);
	    chwrite(data_chan, &val, type_size[act_type]);
	    break;
	  }
	case FLOAT:
	  {
	    float val = va_arg(pvar, double);
	    chwrite(data_chan, &val, type_size[act_type]);
	    break;
	  }
	case DOUBLE:
	  {
	    double val = va_arg(pvar, double);
	    chwrite(data_chan, &val, type_size[act_type]);
	    break;
	  }
	case CHAR:
	  {
	    char val = va_arg(pvar, int);
	    chwrite(data_chan, &val, type_size[act_type]);
	    break;
	  }
	case INT_ARRAY:
	case SHORT_ARRAY:
	case UNSIGNED_SHORT_ARRAY:
	case LONG_ARRAY:
	case UNSIGNED_LONG_ARRAY:
	case FLOAT_ARRAY:
	case DOUBLE_ARRAY:
	case CHAR_ARRAY:
	  {
	    nitems = va_arg(pvar, int);
	    chwrite(data_chan, va_arg(pvar, void *), 
	      nitems*type_size[act_type]);
	    break; 
	  }
	default: 
	  exit(1);
      }
      act_type = va_arg(pvar, int);
    }
  }
  else
  {                                                      /* CS-Lib channel */
    act_type = va_arg(pvar, int);            /* get first type, look-ahead */
    while(act_type != END)
    {
      switch(act_type)
      {
	case INT:
	{
	  int val = va_arg(pvar, int);
	  cslib_xdr_send_int(&val);
	  break;
	}
	case SHORT:
	{
	  short val = va_arg(pvar, int);
	  cslib_xdr_send_short(&val);
	  break;
	}
	case UNSIGNED_SHORT:
	{
	  unsigned short val = va_arg(pvar, unsigned int);
	  cslib_xdr_send_u_short(&val);
	  break;
	}
	case LONG:
	{
	  long val = va_arg(pvar, long);
	  cslib_xdr_send_long(&val);
	  break;
	}
	case UNSIGNED_LONG:
	{
	  unsigned long val = va_arg(pvar, unsigned long);
	  cslib_xdr_send_u_long(&val);
	  break;
	}
	case FLOAT:
	{
	  float val = va_arg(pvar, double);
	  cslib_xdr_send_float(&val);
	  break;
	}
	case DOUBLE:
	{
	  double val = va_arg(pvar, double);
	  cslib_xdr_send_double(&val);
	  break;
	}
	case CHAR:
	{
	  char val = va_arg(pvar, int);
	  cslib_xdr_send_char(&val);
	  break;
	}
	case INT_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_send_int_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case SHORT_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_send_short_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case UNSIGNED_SHORT_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_send_u_short_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case LONG_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_send_long_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case UNSIGNED_LONG_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_send_u_long_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case FLOAT_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_send_float_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case DOUBLE_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_send_double_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case CHAR_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_send_char_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	default: 
	  exit(1);
      }
      act_type = va_arg(pvar, int);
    }
  }
  va_end(pvar);
}

/*-------------------------------------------------------------------------*/
/* receive_channel:                                                        */
/* Get the given data from the specified channel.                          */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value address>                     */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    receive_channel(c, INT, &a, FLOAT_ARRAY, 100, b, END);               */
/*-------------------------------------------------------------------------*/
void receive_channel (channel id, ...)
{
  va_list pvar;
  void |*data_chan;
  int act_type;                                        /* actual item type */
  int nitems;
  char sync_char;
  
  va_start(pvar, id);

  if (pactc_channels[id].type == PACTC_CHAN_TYPE)
  {                                                      /* PACT-C channel */
    sync_char = |*pactc_channels[id].sync_channel;
    data_chan = pactc_channels[id].data_channel;
  
    act_type = va_arg(pvar, int);
    while(act_type != END)
    {
      switch(act_type)
      {
	case INT:
	case SHORT:
	case UNSIGNED_SHORT:
	case LONG:
	case UNSIGNED_LONG:
	case FLOAT:
	case DOUBLE:
	case CHAR:
	  nitems = 1;
	  break;
	case INT_ARRAY:
	case SHORT_ARRAY:
	case UNSIGNED_SHORT_ARRAY:
	case LONG_ARRAY:
	case UNSIGNED_LONG_ARRAY:
	case FLOAT_ARRAY:
	case DOUBLE_ARRAY:
	case CHAR_ARRAY:
	  nitems = va_arg(pvar, int);
	  break;
	default: 
	  exit(1);
      }
      chread(data_chan, va_arg(pvar, void *), nitems*type_size[act_type]);
      act_type = va_arg(pvar, int);
    }
  }
  else
  {                                                      /* CS-Lib channel */
    act_type = va_arg(pvar, int);            /* get first type, look-ahead */
    while(act_type != END)
    {
      switch(act_type)
      {
	case INT:
	{
	  int *val = va_arg(pvar, int *);
	  cslib_xdr_get_int(val);
	  break;
	}
	case SHORT:
	{
	  short *val = va_arg(pvar, short *);
	  cslib_xdr_get_short(val);
	  break;
	}
	case UNSIGNED_SHORT:
	{
	  unsigned short *val = va_arg(pvar, unsigned short *);
	  cslib_xdr_get_u_short(val);
	  break;
	}
	case LONG:
	{
	  long *val = va_arg(pvar, long *);
	  cslib_xdr_get_long(val);
	  break;
	}
	case UNSIGNED_LONG:
	{
	  unsigned long *val = va_arg(pvar, unsigned long *);
	  cslib_xdr_get_u_long(val);
	  break;
	}
	case FLOAT:
	{
	  float *val = va_arg(pvar, float *);
	  cslib_xdr_get_float(val);
	  break;
	}
	case DOUBLE:
	{
	  double *val = va_arg(pvar, double *);
	  cslib_xdr_get_double(val);
	  break;
	}
	case CHAR:
	{
	  char *val = va_arg(pvar, char *);
	  cslib_xdr_get_char(val);
	  break;
	}
	case INT_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_get_int_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case SHORT_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_get_short_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case UNSIGNED_SHORT_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_get_u_short_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case LONG_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_get_long_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case UNSIGNED_LONG_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_get_u_long_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case FLOAT_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_get_float_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case DOUBLE_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_get_double_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	case CHAR_ARRAY:
	{
	  nitems = va_arg(pvar, int);
	  cslib_xdr_get_char_vector(va_arg(pvar, void *), nitems);
	  break;
	}
	default: 
	  exit(1);
      }
      act_type = va_arg(pvar, int);
    }
  }
  va_end(pvar);
}

/*-------------------------------------------------------------------------*/
/* synchronizer:                                                           */
/* The synchronizer process allows the alternate construct on virtual      */
/* channels. Before sending the data from a task, a single byte is sent    */
/* via the synchronizer channel.                                           */
/*-------------------------------------------------------------------------*/
void synchronizer (
  Process *p,
  char |* from_sink,
  char |* to_dest
)
{
  char sync_char;
  Process *tmp = p;                                       /* avoid warning */
  
  for (;;)
  {
    sync_char = |*from_sink;
    |*to_dest = sync_char;
  }
}

/*-------------------------------------------------------------------------*/
/* open_pactc_channel:                                                     */
/* Open a new PACT-C channel and initialize it with the value given        */
/* as parameter. Returns the channel identifier.                           */
/* This function is called during the startup phase to map a PACT-C        */
/* channel on a PPC channel.                                               */
/*-------------------------------------------------------------------------*/
channel open_pactc_channel (
  void |* data_channel,                                    /* data channel */
  char |* sync_channel,                          /* sychronization channel */
  direction_type direction,                     /* input or output channel */
  connection_status_type connected                     /* connected or not */
)
{
  if (connected == NOT_CONNECTED)
    return 0;        /* channel exist but is not connected to another task */
  
  pactc_channels[nb_pactc_channels].type = PACTC_CHAN_TYPE;
  pactc_channels[nb_pactc_channels].data_channel = data_channel;
  
  if (direction == CHAN_INPUT)
  {                                          /* start synchronizer process */
    Process *sync_proc;
    char |* new_chan;
    
    new_chan = (char |*)challoc(1);
    pactc_channels[nb_pactc_channels].sync_channel = new_chan;
    
    sync_proc = ProcAlloc(synchronizer, 0, 2, 
      sync_channel, new_chan);
    ProcRun(sync_proc);
  }
  else
    pactc_channels[nb_pactc_channels].sync_channel = sync_channel;
    
  nb_pactc_channels++;
  return nb_pactc_channels-1;
}

/*-------------------------------------------------------------------------*/
/* open_pactc_channels_array:                                              */
/* Opens an array of new PACT-C channel and initialize it with the values  */
/* given as parameter. The channels ids are stored in the array of         */
/* channels given as parameter. Returns this array.                        */
/*-------------------------------------------------------------------------*/
channel *open_pactc_channels_array (
  void |* data_channels[],   /* PACT-C data channels, as given at start-up */
  char |* sync_channels[],            /* idem but synchronization channels */
  channel ids[],                        /* where to store the PPC channels */
  int nitems,                                        /* number of channels */
  direction_type direction,                     /* input or output channel */
  connection_status_type connected[]                   /* connected or not */
)
{
  int c;                                        /* counter, actual channel */
  
  for (c = 0; c < nitems; c++)
    ids[c] = open_pactc_channel(
      data_channels[c], sync_channels[c], direction, connected[c]);
  
  return ids;
}

/*-------------------------------------------------------------------------*/
/* get_sync_pactc_chan:                                                    */
/* Return the synchronization PACT-C channel corresponding to the          */
/* specified PPC channel.                                                  */
/* This function is used by the alternate construct.                       */
/*-------------------------------------------------------------------------*/
char |*get_sync_pactc_chan (channel id)
{
  return pactc_channels[id].sync_channel;
}

/*-------------------------------------------------------------------------*/
/* new_cslib_channel:                                                      */
/* Create two new CS-Lib channels.                                         */
/*-------------------------------------------------------------------------*/
void new_cslib_channel (
  channel *to_client,
  channel *from_client
)
{
  int chan_idx;                            /* channel index to be returned */

  semp(&prot_sem);                                    /* acquire semaphore */
  
  chan_idx = nb_pactc_channels;
  pactc_channels[chan_idx].type = CSLIB_CHAN_TYPE;
  pactc_channels[chan_idx+1].type = CSLIB_CHAN_TYPE;
  nb_pactc_channels += 2;
  
  semv(&prot_sem);                                    /* release semaphore */

  *to_client = chan_idx;
  *from_client = chan_idx+1;
}
