/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Lower-Half PVM channels implementation.                                 */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 07.01.1994  Droux       Created                                         */
/* 08.07.1994  Droux       New structure of source directory               */
/*-------------------------------------------------------------------------*/
/* filename: chanpvmlh.c                                                   */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#if defined(NEXT) || defined(NEXTINTEL)
#include <libc.h>
#endif
#include <fcntl.h>
#include <stdio.h>
#include <syslog.h>
#include <pvm3.h>
#include <ppc/csp/chanpvmlh.h>

#define MAX_LH_PVM_CHANNELS 500             /* max lower-half PVM channels */
#define STRIDE 1            /* stride for PVM packets management functions */

typedef struct     /* structure containing the lower-half pvm channels ... */
{                                                        /* ... properties */
  int free;                                              /* 1 if free slot */
  int other_tid;                                  /* correspondant task id */
  int tag;                                  /* PVM tag to identify channel */
} lh_pvm_channel_descr;

lh_pvm_channel_descr lh_pvm_channels [MAX_LH_PVM_CHANNELS]; /* lh channels */
static int disable_recv_flag = 0;                         /* 1 if disabled */
static int disable_recv_tid;                       /* disabled channel tid */
static int disable_recv_tag;                       /* disabled channel tag */

/*-------------------------------------------------------------------------*/
/* pvm_chan_init:                                                          */
/* Initializes the lower-half PVM channel implementation.                  */
/* This function is never directly called by the programmer but when the   */
/* parallel program is started.                                            */
/*-------------------------------------------------------------------------*/
void pvm_chan_init (void)
{
  int c;                                                        /* counter */
  
  for (c = 0; c < MAX_LH_PVM_CHANNELS; c++)
    lh_pvm_channels[c].free = 1;                     /* all slots are free */
}

/*-------------------------------------------------------------------------*/
/* lh_new_pvm_channel:                                                     */
/* Declares a new lower-half PVM channel.                                  */
/*-------------------------------------------------------------------------*/
channel lh_new_pvm_channel (int tid, int tag)
{
  int c;                                                        /* counter */
  int chan_id;                                    /* upper-half channel id */
  
  for (c = 0; (c < MAX_LH_PVM_CHANNELS) && (!lh_pvm_channels[c].free); c++);
  chan_id = c;
  
  lh_pvm_channels[chan_id].free = 0;
  lh_pvm_channels[chan_id].other_tid = tid;
  lh_pvm_channels[chan_id].tag = tag;
  
  return chan_id;
}

/*-------------------------------------------------------------------------*/
/* pvm_chan_write:                                                         */
/* Write the given data items to the specified channel.                    */
/* <nitems> corresponds to the number of items of type <type> to be        */
/* transmitted.                                                            */
/*-------------------------------------------------------------------------*/
int pvm_chan_write (channel id, void *vbuf, int nitems, int type)
{
  switch(type)
  {
    case INT:
    case INT_ARRAY:
      pvm_pkint(vbuf, nitems, STRIDE);
      break;
    case SHORT:
    case UNSIGNED_SHORT:
    case SHORT_ARRAY:
    case UNSIGNED_SHORT_ARRAY:
      pvm_pkshort(vbuf, nitems, STRIDE);
      break;
    case LONG:
    case UNSIGNED_LONG:
    case LONG_ARRAY:
    case UNSIGNED_LONG_ARRAY:
      pvm_pklong(vbuf, nitems, STRIDE);
      break;
    case FLOAT:
    case FLOAT_ARRAY:
      pvm_pkfloat(vbuf, nitems, STRIDE);
      break;
    case DOUBLE:
    case DOUBLE_ARRAY:
      pvm_pkdouble(vbuf, nitems, STRIDE);
      break;
    case CHAR:
    case CHAR_ARRAY:
      pvm_pkbyte(vbuf, nitems, STRIDE);
      break;
  }
  return 0;
}

/*-------------------------------------------------------------------------*/
/* pvm_chan_read:                                                          */
/* Read the given data items from the specified channel.                   */
/* <nitems> corresponds to the number of items of type <type> to be        */
/* received.                                                               */
/*-------------------------------------------------------------------------*/
int pvm_chan_read (channel id, void *vbuf, int nitems, int type)
{
  switch(type)
  {
    case INT:
    case INT_ARRAY:
      pvm_upkint(vbuf, nitems, STRIDE);
      break;
    case SHORT:
    case UNSIGNED_SHORT:
    case SHORT_ARRAY:
    case UNSIGNED_SHORT_ARRAY:
      pvm_upkshort(vbuf, nitems, STRIDE);
      break;
    case LONG:
    case UNSIGNED_LONG:
    case LONG_ARRAY:
    case UNSIGNED_LONG_ARRAY:
      pvm_upklong(vbuf, nitems, STRIDE);
      break;
    case FLOAT:
    case FLOAT_ARRAY:
      pvm_upkfloat(vbuf, nitems, STRIDE);
      break;
    case DOUBLE:
    case DOUBLE_ARRAY:
      pvm_upkdouble(vbuf, nitems, STRIDE);
      break;
    case CHAR:
    case CHAR_ARRAY:
      pvm_upkbyte(vbuf, nitems, STRIDE);
      break;
  }
  return 0;
}

/*-------------------------------------------------------------------------*/
/* pvm_start_write:                                                        */
/* Starts a channel transmission of data. Must be called before any call   */
/* to pvm_chan_write for each message transmission.                        */
/*-------------------------------------------------------------------------*/
void pvm_start_write (void)
{
  pvm_initsend(PvmDataDefault);
}

/*-------------------------------------------------------------------------*/
/* pvm_end_write:                                                          */
/* Terminates a channel transmission of data. Must be called after all     */
/* calls to pvm_chan_write for each message transmission.                  */
/*-------------------------------------------------------------------------*/
void pvm_end_write (channel id)
{
  pvm_send(lh_pvm_channels[id].other_tid, lh_pvm_channels[id].tag);
}

/*-------------------------------------------------------------------------*/
/* pvm_start_read:                                                         */
/* Starts a channel reception of data. Must be called before any call      */
/* to pvm_chan_read for each message reception.                            */
/*-------------------------------------------------------------------------*/
void pvm_start_read (channel id)
{
  if (disable_recv_flag)
    if ((disable_recv_tid == lh_pvm_channels[id].other_tid) &&
        (disable_recv_tag == lh_pvm_channels[id].tag))
    {
      disable_recv_flag = 0;
      return;                     /* channel disabled by ALT, does nothing */
    }
    
  pvm_recv(lh_pvm_channels[id].other_tid, lh_pvm_channels[id].tag);
}

/*-------------------------------------------------------------------------*/
/* pvm_tid_and_tag:                                                        */
/* Returns the tid and tag corresponding to the given lower-half channel.  */
/* This function is used by the PVM alternate construct.                   */
/*-------------------------------------------------------------------------*/
void pvm_tid_and_tag (channel id, int *tid, int *tag)
{
  *tid = lh_pvm_channels[id].other_tid;
  *tag = lh_pvm_channels[id].tag;
}

/*-------------------------------------------------------------------------*/
/* pvm_disable_recv:                                                       */
/* Calling this function will disable the next call to pvm_start_read on   */
/* the specified channel. This function is used by the PVM alternate       */
/* construct in which the equivalent of pvm_start_read is called when data */
/* is available on the channel.                                            */
/* Warning: if two calls to pvm_disable_recv are made before calling       */
/* pvm_start_read, the effect of the first call is not taken in account.   */
/*-------------------------------------------------------------------------*/
void pvm_disable_recv (channel id)
{
  disable_recv_flag = 1;
  disable_recv_tid = lh_pvm_channels[id].other_tid;
  disable_recv_tag = lh_pvm_channels[id].tag;
}
