/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Upper-Half PVM channels implementation.                                 */
/* This file implements the functions that are defined in the header file  */
/* chan.h and chanpriv.h.                                                  */
/* It uses the Lower-Half UNIX channels implementation for communication   */
/* between threads of the same task and the Lower-Half PVM channels        */
/* implementation for communication between threads of different tasks.    */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 07.01.1994  Droux       Created                                         */
/* 17.02.1994  Droux       va_arg(char|short|float) bug fixed              */
/* 08.07.1994  Droux       New structure of source directory               */
/*-------------------------------------------------------------------------*/
/* filename: chanpvmuh.c                                                   */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define chan_impl

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <varargs.h>
#include <syslog.h>
#include <ppc/csp/chanpriv.h>
#include <ppc/csp/chanunixlh.h>
#include <ppc/csp/chanpvmlh.h>
#include <ppc/csp/chanpvmuh.h>
#include "chanservunix.h"

#define MAX_UH_PVM_CHANNELS 1000            /* max upper-half PVM channels */

typedef struct     /* structure containing the upper-half PVM channels ... */
{                                                       /* ... properties. */
  int free;                                              /* 1 if free slot */
  lh_types lh_type;                             /* lower-half channel type */
  channel lh_id;                                  /* lower-half channel id */
} uh_pvm_channel_descr;

uh_pvm_channel_descr uh_pvm_channels [MAX_UH_PVM_CHANNELS]; /* uh channels */

/*-------------------------------------------------------------------------*/
/* chan_init:                                                              */
/* Initializes the channels implementation. This function is never called  */
/* directly by the programmer.                                             */
/*-------------------------------------------------------------------------*/
void chan_init (void)
{
  int c;                                                        /* counter */
  
  pvm_chan_init();                /* initializes the PVM lower-half module */
  for (c = 0; c < MAX_UH_PVM_CHANNELS; c++)
    uh_pvm_channels[c].free = 1;                     /* all slots are free */
}

/*-------------------------------------------------------------------------*/
/* par_channel:                                                            */
/* Creates a new channel to communicate between two parallel threads.      */
/*-------------------------------------------------------------------------*/
channel par_channel (void)
{
  int c;                                                        /* counter */
  int chan_id;                                    /* upper-half channel id */
  
                                 /* a par_channel is always a UNIX channel */
  for (c = 0; (c < MAX_UH_PVM_CHANNELS) && (!uh_pvm_channels[c].free); c++);
  chan_id = c;
  
  uh_pvm_channels[chan_id].free = 0;
  uh_pvm_channels[chan_id].lh_type = UNIX_CHAN;
  uh_pvm_channels[chan_id].lh_id = unix_par_channel ();
  
  return chan_id;
}

/*-------------------------------------------------------------------------*/
/* chan_close:                                                             */
/* Close the given channel.                                                */
/*-------------------------------------------------------------------------*/
void chan_close (channel id)
{
                           /* only UNIX channels can be closed explicitely */
  if (uh_pvm_channels[id].lh_type == UNIX_CHAN)
    unix_chan_close(uh_pvm_channels[id].lh_id);
}

/*-------------------------------------------------------------------------*/
/* send_channel:                                                           */
/* Send the given data on the given channel.                               */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value>                             */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    send_channel(c, INT, a, FLOAT_ARRAY, 100, b, END);                   */
/*-------------------------------------------------------------------------*/
void send_channel (va_alist)
va_dcl
{
  va_list pvar;                  /* needed for variable arguements parsing */
  int act_type;                                        /* actual item type */
  int chan_id;                                    /* lower-half channel id */
  int uh_chan_id;                                 /* upper-half channel id */
  void *address;
  int len;
  int nitems;
  int (* writef) (channel, void *, int, int);    /* write function pointer */
  void (* start_write)(void);              /* start_write function address */
  void (* end_write)(channel);               /* end_write function address */
  
  va_start(pvar);
  uh_chan_id = va_arg(pvar, channel);
  chan_id = uh_pvm_channels[uh_chan_id].lh_id;
  
  switch (uh_pvm_channels[uh_chan_id].lh_type)
  {
    case UNIX_CHAN:
      writef = (int (*)(channel, void *, int, int)) unix_chan_write;
      start_write = (void (*)(void))unix_start_write;
      end_write = (void (*)(channel))unix_end_write;
      break;
    case PVM_CHAN:
      writef = (int (*)(channel, void *, int, int)) pvm_chan_write;
      start_write = (void (*)(void))pvm_start_write;
      end_write = (void (*)(channel))pvm_end_write;
      break;
    case CSLIB_CHAN:
      writef = (int (*)(channel, void *, int, int)) servunix_chan_write;
      start_write = (void (*)(void))servunix_start_write;
      end_write = (void (*)(channel))servunix_end_write;
      break;
  }
  
  start_write();
  act_type = va_arg(pvar, int);
  while(act_type != END)
  {
    switch(act_type)
    {
      case INT:
        {
	  int val = va_arg(pvar, int);
	  writef(chan_id, &val, 1, act_type);
	  break;
	}
      case SHORT:
        {
	  short val = va_arg(pvar, int);
	  writef(chan_id, &val, 1, act_type);
	  break;
	}
      case UNSIGNED_SHORT:
        {
	  unsigned short val = va_arg(pvar, unsigned int);
	  writef(chan_id, &val, 1, act_type);
	  break;
	}
      case LONG:
        {
	  long val = va_arg(pvar, long);
	  writef(chan_id, &val, 1, act_type);
	  break;
	}
      case UNSIGNED_LONG:
        {
	  unsigned long val = va_arg(pvar, unsigned long);
	  writef(chan_id, &val, 1, act_type);
	  break;
	}
      case FLOAT:
        {
	  float val = va_arg(pvar, double);
	  writef(chan_id, &val, 1, act_type);
	  break;
	}
      case DOUBLE:
        {
	  double val = va_arg(pvar, double);
	  writef(chan_id, &val, 1, act_type);
	  break;
	}
      case CHAR:
        {
	  char val = va_arg(pvar, int);
	  writef(chan_id, &val, 1, act_type);
	  break;
	}
      case INT_ARRAY:
      case SHORT_ARRAY:
      case UNSIGNED_SHORT_ARRAY:
      case LONG_ARRAY:
      case UNSIGNED_LONG_ARRAY:
      case FLOAT_ARRAY:
      case DOUBLE_ARRAY:
      case CHAR_ARRAY:
	nitems = va_arg(pvar, int);
	writef(chan_id, va_arg(pvar, void *), nitems, act_type);
	break; 
      default: 
        fprintf(stderr, "error: unknown type identifier\n");
	exit(1);
    }
    act_type = va_arg(pvar, int);
  }
  end_write(chan_id);
}

/*-------------------------------------------------------------------------*/
/* receive_channel:                                                        */
/* Get the given data from the specified channel.                          */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value address>                     */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    receive_channel(c, INT, &a, FLOAT_ARRAY, 100, b, END);               */
/*-------------------------------------------------------------------------*/
void receive_channel (va_alist)
va_dcl
{
  va_list pvar;                  /* needed for variable arguements parsing */
  int act_type;                                        /* actual item type */
  int chan_id;                                    /* lower-half channel id */
  int uh_chan_id;                                 /* upper-half channel id */
  int nitems;
  void *address;
  int (* readf) (channel, void *, int, int);      /* read function pointer */
  void (* start_read)(channel);             /* start_read function address */
  
  va_start(pvar);
  uh_chan_id = va_arg(pvar, channel);
  chan_id = uh_pvm_channels[uh_chan_id].lh_id;
  
  switch (uh_pvm_channels[uh_chan_id].lh_type)
  {
    case UNIX_CHAN:
      readf = (int (*)(channel, void *, int, int)) unix_chan_read;
      start_read = (void (*)(channel))unix_start_read;
      break;
    case PVM_CHAN:
      readf = (int (*)(channel, void *, int, int)) pvm_chan_read;
      start_read = (void (*)(channel))pvm_start_read;
      break;
    case CSLIB_CHAN:
      readf = (int (*)(channel, void *, int, int)) servunix_chan_read;
      start_read = (void (*)(channel))servunix_start_read;
      break;
  }

  start_read(chan_id);
  act_type = va_arg(pvar, int);
  while(act_type != END)
  {
    switch(act_type)
    {
      case INT:
      case SHORT:
      case UNSIGNED_SHORT:
      case LONG:
      case UNSIGNED_LONG:
      case FLOAT:
      case DOUBLE:
      case CHAR:
        nitems = 1;
	break;
      case INT_ARRAY:
      case SHORT_ARRAY:
      case UNSIGNED_SHORT_ARRAY:
      case LONG_ARRAY:
      case UNSIGNED_LONG_ARRAY:
      case FLOAT_ARRAY:
      case DOUBLE_ARRAY:
      case CHAR_ARRAY:
        nitems = va_arg(pvar, int);
	break;
      default: 
        fprintf(stderr, "error: unknown type identifier\n");
	exit(1);
    }
    readf(chan_id, va_arg(pvar, void *), nitems, act_type);
    act_type = va_arg(pvar, int);
  }
}

/*-------------------------------------------------------------------------*/
/* new_pvm_channel:                                                        */
/* Declares a new PVM channel.                                             */
/*-------------------------------------------------------------------------*/
channel new_pvm_channel (int tid, int tag)
{
  int c;                                                        /* counter */
  int chan_id;                                    /* upper-half channel id */
  
  for (c = 0; (c < MAX_UH_PVM_CHANNELS) && (!uh_pvm_channels[c].free); c++);
  chan_id = c;
  
  uh_pvm_channels[chan_id].free = 0;
  uh_pvm_channels[chan_id].lh_type = PVM_CHAN;
  uh_pvm_channels[chan_id].lh_id = lh_new_pvm_channel(tid, tag);
  
  return chan_id;
}

/*-------------------------------------------------------------------------*/
/* lh_chan_type:                                                           */
/* Returns the lower-half channel type corresponding to the given channel. */
/*-------------------------------------------------------------------------*/
lh_types lh_chan_type (channel id)
{
  return uh_pvm_channels[id].lh_type;
}

/*-------------------------------------------------------------------------*/
/* lh_chan_id:                                                             */
/* Returns the lower-half channel id corresponding to the given channel.   */
/*-------------------------------------------------------------------------*/
channel lh_chan_id (channel id)
{
  return uh_pvm_channels[id].lh_id;
}

/*-------------------------------------------------------------------------*/
/* new_cslib_channel:                                                      */
/* Return the two channels from and to the client application and          */
/* initialize them with the given values.                                  */
/* This function is called by the unixconn module.                         */
/*-------------------------------------------------------------------------*/
void new_cslib_channel(
  int from_client,
  int to_client,
  channel *from_client_chan,
  channel *to_client_chan
)
{
  int c;                                                        /* counter */
  int chan_id;                                    /* upper-half channel id */

  servunix_init_channels(from_client, to_client);
  
  for (c = 0; (c < MAX_UH_PVM_CHANNELS) && (!uh_pvm_channels[c].free); c++);
  *from_client_chan = c;
  
  uh_pvm_channels[*from_client_chan].free = 0;
  uh_pvm_channels[*from_client_chan].lh_type = CSLIB_CHAN;
  uh_pvm_channels[*from_client_chan].lh_id = FROM_CLIENT;
  
  for (c = 0; (c < MAX_UH_PVM_CHANNELS) && (!uh_pvm_channels[c].free); c++);
  *to_client_chan = c;
  
  uh_pvm_channels[*to_client_chan].free = 0;
  uh_pvm_channels[*to_client_chan].lh_type = CSLIB_CHAN;
  uh_pvm_channels[*to_client_chan].lh_id = TO_CLIENT;
}

