/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Lower-Half UNIX channels implementation.                                */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 25.11.1993  Droux       Created                                         */
/* 08.07.1994  Droux       New structure of source directory               */
/*-------------------------------------------------------------------------*/
/* filename: chanunixlh.c                                                  */
/* Tested on :                                                             */
/*   NEXTSTEP 2.1, NeXT Computers                                          */
/*   NEXTSTEP 3.1, NeXT Computers                                          */
/*   NEXTSTEP 3.1, Intel Computers                                         */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*   SunOS Release 5.3 (Solaris 2), SPARC                                  */
/*   Linux 0.99.pl9, Intel Computers                                       */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/


#if defined(NEXT) || defined(NEXTINTEL)
#include <libc.h>
#endif
#include <fcntl.h>
#include <stdio.h>
#include <ppc/csp/chanunixlh.h>
#include <ppc/csp/types.h>
#include "ion.h"

#define PIPE_IN  0                     /* indices for accessing UNIX pipes */
#define PIPE_OUT 1

typedef struct 
{
  int free;                                           /* == 1 if free slot */
  int in;                         /* file descriptor on which we must read */
  int out;                       /* file descriptor on which we must write */
  int creator_pid;          /* PID of the process that created the channel */
} unix_channel_descr;

unix_channel_descr unix_channels [MAX_UNIX_CHANNELS];          /* channels */
int cspunix_init = 0;                           /* 1 if module initialized */

/*-------------------------------------------------------------------------*/
/* unix_par_channel:                                                       */
/* Creates a new channel to communicate between two parallel threads.      */
/* UNIX version.                                                           */
/*-------------------------------------------------------------------------*/
channel unix_par_channel (void)
{
  int the_pipe [2];
  int n;
  int chan_id;
  int fl;
  
  if (!cspunix_init)
  {                                              /* module not initialized */
    for (n = 0; n < MAX_UNIX_CHANNELS; n++)
      unix_channels[n].free = 1;
    cspunix_init = 1;
  }
                             /* look for a free slot in the channels array */
  for (n = 0; (n < MAX_UNIX_CHANNELS) && (!unix_channels[n].free); n++);
  chan_id = n;
  
  pipe(the_pipe);                                         /* create a pipe */
  
  #if !defined(LINUX) && !defined(SUN4SOL2)
    fl = fcntl(the_pipe[PIPE_IN], F_GETFL, 0);
    fl |= FSYNC;
    fcntl(the_pipe[PIPE_IN], F_SETFL, fl);
    
    fl = fcntl(the_pipe[PIPE_OUT], F_GETFL, 0);
    fl |= FSYNC;
    fcntl(the_pipe[PIPE_OUT], F_SETFL, fl);
  #endif
  
  unix_channels[chan_id].in = the_pipe[PIPE_IN];
  unix_channels[chan_id].out = the_pipe[PIPE_OUT];
  unix_channels[chan_id].creator_pid = getpid();
  unix_channels[chan_id].free = 0;
  
  return chan_id;
}

/*-------------------------------------------------------------------------*/
/* unix_chan_close:                                                        */
/* Close the given channel.                                                */
/* UNIX version.                                                           */
/*-------------------------------------------------------------------------*/
void unix_chan_close (channel id)
{
  close(unix_channels[id].in);
  close(unix_channels[id].in);
  unix_channels[id].free = 1;
}

/*-------------------------------------------------------------------------*/
/* unix_chan_write:                                                        */
/* Write the given data items to the specified channel.                    */
/* <nitems> corresponds to the number of items of type <type> to be        */
/* transmitted.                                                            */
/*-------------------------------------------------------------------------*/
int unix_chan_write (channel id, void *vbuf, int nitems, int type)
{
  return writen(unix_channels[id].out, (char *)vbuf, 
    nitems*type_size[type]);
}

/*-------------------------------------------------------------------------*/
/* unix_chan_read:                                                         */
/* Read the given data items from the specified channel.                   */
/* <nitems> corresponds to the number of items of type <type> to be        */
/* received.                                                               */
/*-------------------------------------------------------------------------*/
int unix_chan_read (channel id, void *vbuf, int nitems, int type)
{
  return readn(unix_channels[id].in, (char *)vbuf, 
    nitems*type_size[type]);
}

/*-------------------------------------------------------------------------*/
/* unix_start_write:                                                       */
/* Starts a channel transmission of data. Must be called before any call   */
/* to unix_chan_write for each message transmission.                       */
/* Does nothing under UNIX. Implemented for compatibility with the PVM     */
/* lower-half module.                                                      */
/*-------------------------------------------------------------------------*/
void unix_start_write (void)
{
}

/*-------------------------------------------------------------------------*/
/* unix_end_write:                                                         */
/* Terminates a channel transmission of data. Must be called after all     */
/* calls to unix_chan_write for each message transmission.                 */
/* Does nothing under UNIX. Implemented for compatibility with the PVM     */
/* lower-half module.                                                      */
/*-------------------------------------------------------------------------*/
void unix_end_write (channel id)
{
}

/*-------------------------------------------------------------------------*/
/* unix_start_read:                                                        */
/* Starts a channel reception of data. Must be called before any call      */
/* to unix_chan_read for each message reception.                           */
/* Does nothing under UNIX. Implemented for compatibility with the PVM     */
/* lower-half module.                                                      */
/*-------------------------------------------------------------------------*/
void unix_start_read (channel id)
{
}

/*-------------------------------------------------------------------------*/
/* unix_read_fd:                                                           */
/* Returns the read file descriptor of the given channel. This function    */
/* is particularly usefull in the alt construct.                           */
/*-------------------------------------------------------------------------*/
int unix_read_fd (channel c)
{
  return unix_channels[c].in;
}
