/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Upper-Half UNIX channels implementation.                                */
/* This file implements the functions that are defined in the header file  */
/* chan.h and chanpriv.h. It must be used for the UNIX PPC simulation.     */
/* It uses exclusively the Lower-Half UNIX channels.                       */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 25.11.1993  Droux       Created                                         */
/* 16.02.1994  Droux       Adapted for client-server library               */
/* 17.02.1994  Droux       va_arg(char|short|float) bug fixed              */
/* 08.07.1994  Droux       New structure of source directory               */
/*-------------------------------------------------------------------------*/
/* filename: chanunixuh.c                                                  */
/* Tested on :                                                             */
/*   NEXTSTEP 2.1, NeXT Computers                                          */
/*   NEXTSTEP 3.1, NeXT Computers                                          */
/*   NEXTSTEP 3.1, Intel Computers                                         */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*   SunOS Release 5.3 (Solaris 2), SPARC                                  */
/*   Linux 0.99.pl9, Intel Computers                                       */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define chan_impl

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <varargs.h>
#include <ppc/csp/chanpriv.h>
#include <ppc/csp/chanunixlh.h>
#include "chanunixuh.h"
#include "chanservunix.h"

/* Two kind of channels are accessed through this module: UNIX channels    */
/* and CS-Lib channels. When an operation on a channel is requested by the */
/* PPC program, this module must be able to determinate if this channel    */
/* is a UNIX or a CS-Lib channel. In order to allows this distinction,     */
/* we define the CS-Lib channel ids as constant beyond the lower-half      */
/* UNIX channel possible values. This is possible since only two           */
/* two CS-Lib channels can be opened by a server program.                  */
#define FROM_CLIENT_CHAN_ID MAX_UNIX_CHANNELS
#define TO_CLIENT_CHAN_ID   (FROM_CLIENT_CHAN_ID+1)

/*-------------------------------------------------------------------------*/
/* chan_init:                                                              */
/* Initializes the channels implementation. This function is never called  */
/* directly by the programmer.                                             */
/*-------------------------------------------------------------------------*/
void chan_init (void)
{
  /* under UNIX, only the lower-half UNIX implementation is used. This     */
  /* implementation initializes itself automatically during the first      */
  /* call to the function unix_par_channel(). An explicit configuration    */
  /* is thus not required.                                                 */
}

/*-------------------------------------------------------------------------*/
/* new_cslib_channel:                                                      */
/* Return the two channels from and to the client application and          */
/* initialize them with the given values.                                  */
/* This function is called by the unixconn module.                         */
/*-------------------------------------------------------------------------*/
void new_cslib_channel(
  int from_client,
  int to_client,
  channel *from_client_chan,
  channel *to_client_chan
)
{
  servunix_init_channels(from_client, to_client);
  *from_client_chan = FROM_CLIENT_CHAN_ID;
  *to_client_chan = TO_CLIENT_CHAN_ID;
}

/*-------------------------------------------------------------------------*/
/* par_channel:                                                            */
/* Creates a new channel to communicate between two parallel threads.      */
/*-------------------------------------------------------------------------*/
channel par_channel (void)
{
  return unix_par_channel ();
}

/*-------------------------------------------------------------------------*/
/* chan_close:                                                             */
/* Close the given channel.                                                */
/*-------------------------------------------------------------------------*/
void chan_close (channel id)
{
  unix_chan_close(id);
}

/*-------------------------------------------------------------------------*/
/* send_channel:                                                           */
/* Send the given data on the given channel.                               */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value>                             */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    send_channel(c, INT, a, FLOAT_ARRAY, 100, b, END);                   */
/*-------------------------------------------------------------------------*/
void send_channel (va_alist)
va_dcl
{
  va_list pvar;                  /* needed for variable arguements parsing */
  int act_type;                                        /* actual item type */
  int chan_id;                                               /* channel id */
  int nitems;
  static int (* writef) (channel, void *, int, int);

  va_start(pvar);
  chan_id = va_arg(pvar, channel);
  
  if (chan_id == TO_CLIENT_CHAN_ID)
  {
    chan_id = TO_CLIENT;
    writef = (int (*)(channel, void *, int, int)) servunix_chan_write;
  }
  else
    writef = (int (*)(channel, void *, int, int)) unix_chan_write;
  
  act_type = va_arg(pvar, int);
  while(act_type != END)
  {
    switch(act_type)
    {
      case INT:
        {
	  int val = va_arg(pvar, int);
	  writef(chan_id, &val, 1, INT);
	  break;
	}
      case SHORT:
        {
	  short val = va_arg(pvar, int);
	  writef(chan_id, &val, 1, SHORT);
	  break;
	}
      case UNSIGNED_SHORT:
        {
	  unsigned short val = va_arg(pvar, unsigned int);
	  writef(chan_id, &val, 1, UNSIGNED_SHORT);
	  break;
	}
      case LONG:
        {
	  long val = va_arg(pvar, long);
	  writef(chan_id, &val, 1, LONG);
	  break;
	}
      case UNSIGNED_LONG:
        {
	  unsigned long val = va_arg(pvar, unsigned long);
	  writef(chan_id, &val, 1, UNSIGNED_LONG);
	  break;
	}
      case FLOAT:
        {
	  float val = va_arg(pvar, double);
	  writef(chan_id, &val, 1, FLOAT);
	  break;
	}
      case DOUBLE:
        {
	  double val = va_arg(pvar, double);
	  writef(chan_id, &val, 1, DOUBLE);
	  break;
	}
      case CHAR:
        {
	  char val = va_arg(pvar, int);
	  writef(chan_id, &val, 1, CHAR);
	  break;
	}
      case INT_ARRAY:
      case SHORT_ARRAY:
      case UNSIGNED_SHORT_ARRAY:
      case LONG_ARRAY:
      case UNSIGNED_LONG_ARRAY:
      case FLOAT_ARRAY:
      case DOUBLE_ARRAY:
      case CHAR_ARRAY:
	nitems = va_arg(pvar, int);
	writef(chan_id, va_arg(pvar, void *), nitems, act_type);
	break; 
      default: 
        fprintf(stderr, "error: unknown type identifier\n");
	exit(1);
    }
    act_type = va_arg(pvar, int);
  }
}

/*-------------------------------------------------------------------------*/
/* receive_channel:                                                        */
/* Get the given data from the specified channel.                          */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value address>                     */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    send_channel(c, INT, &a, FLOAT_ARRAY, 100, b, END);                  */
/*-------------------------------------------------------------------------*/
void receive_channel (va_alist)
va_dcl
{
  va_list pvar;                  /* needed for variable arguements parsing */
  int act_type;                                        /* actual item type */
  int chan_id;                                               /* channel id */
  int nitems;
  static int (* readf) (channel, void *, int, int);      
  
  va_start(pvar);
  chan_id = va_arg(pvar, channel);
  
  if (chan_id == FROM_CLIENT_CHAN_ID)
  {
    chan_id = FROM_CLIENT;
    readf = (int (*)(channel, void *, int, int)) servunix_chan_read;
  }
  else
    readf = (int (*)(channel, void *, int, int)) unix_chan_read;

  act_type = va_arg(pvar, int);
  while(act_type != END)
  {
    switch(act_type)
    {
      case INT:
      case SHORT:
      case UNSIGNED_SHORT:
      case LONG:
      case UNSIGNED_LONG:
      case FLOAT:
      case DOUBLE:
      case CHAR:
        nitems = 1;
	break;
      case INT_ARRAY:
      case SHORT_ARRAY:
      case UNSIGNED_SHORT_ARRAY:
      case LONG_ARRAY:
      case UNSIGNED_LONG_ARRAY:
      case FLOAT_ARRAY:
      case DOUBLE_ARRAY:
      case CHAR_ARRAY:
        nitems = va_arg(pvar, int);
	break;
      default: 
        fprintf(stderr, "error: unknown type identifier\n");
	exit(1);
    }
    readf(chan_id, va_arg(pvar, void *), nitems, act_type);
    act_type = va_arg(pvar, int);
  }
}
