/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* PACT-C cleaner.                                                         */
/* It is possible, for some reasons I didn't find out, that a PACT-C       */
/* program cannot be started twice. During the second run, the program     */
/* blocks. To avoid this problem, we execute a little program, called      */
/* cleaner, and implemented here, which allows the same parallel program   */
/* to be executed twice without problem.                                   */
/* This program is an adaptation of the PACT Parallel C worm               */
/* demonstration copyright (c) 1991, 1992 PACT                             */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 19.08.1994  Droux       Created                                         */
/*-------------------------------------------------------------------------*/
/* filename: cleaner.c                                                     */
/* Tested on :                                                             */
/*  PACT-C 93.1, T800, T805                                                */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <par.h>
#include <node.h>
#include <asm.h>

static float VERSION = 1.0;
static char REV = 'd';

#pragma PACT stackcheck off
#pragma PACT stacknew 200

typedef int bool_t;
#define FALSE 0
#define TRUE 1
typedef int link_t;

typedef struct neighbour 
{
  int id;
  char dev;
  char booted;
  char error;
  char noise;
} neighbour_t;

#define DEV_NC      0
#define DEV_HOST    1
#define DEV_PROC    2

#define ERR_OK      0
#define ERR_CONN    1
#define ERR_BOOT    2
#define ERR_RUN     3

#define NOFLINKS    4

typedef struct proc 
{
  int id;     
  int bt;     
  int depth;  
  int type;   
  char rev;    
  int cpuspeed;   
  int linkspeed;  
  neighbour_t neighbour[NOFLINKS];  
} proc_t;

#define NOPROC      (-1)          
#define ROOTPROC    1             

#define POKE 0
#define PEEK 1
#define BOOT 2

typedef char tag_t;

#define HEARTBEAT   1     
#define BRANCHDONE  2     
#define PROCESSOR   3     

static proc_t thisproc;      
static proc_t *procs = NULL; 
static int maxid = 0;     
static int nextid = 0;    

static int cycletime (
  int count, 
  int *address
)
{
  int     x[2];
  int     starttime;

  x[0] = 0;
  x[1] = count / 17;

  starttime = ldtimer();

  __asm ("        align               ");
  __asm (".lp:                        ");
  __asm ("",      *address            );
  __asm ("",      &x                  );
  __asm ("        ldc     .elp - .lp  ");
  __asm ("        lend                ");
  __asm (".elp:                       ");
  
  return ldtimer() - starttime;
}

static void endcycletime (void) {} 

static void attrib (void)
{
  int local;            
  int extmemstart;      

  thisproc.type = _Node.devtype;
  thisproc.rev = (_Node.devid > 0 ? 'a' + _Node.devid%10 : ' ');

  extmemstart = _Node.devtype < 400 ? _Node.devtype == 212 ? 0x8800
                                                           : 0x9000
                                : _Node.devtype <= 414 ? 0x80000800
                                                      : 0x80001000;

  if ((int) endcycletime < extmemstart && (int) &local < extmemstart) 
  {
    int cycles, usec;
    cycles = 10000;
    stpri (0);
    usec = cycletime (cycles, (int *) 0x80000000);
    stpri (1);
    thisproc.cpuspeed = (cycles + usec/2) / usec;

    stpri (0);
    usec = cycletime (cycles, (int *) extmemstart);
    stpri (1);
  }
  else
    thisproc.cpuspeed = 0;

  if (_Node.id == ROOTPROC) 
  {
    int usec;
    int len;
    int replen;

    len = 512-2-1-2-1;
    |* (short |*) _Node.toserver = 512-2; 
    |* (char  |*) _Node.toserver = 32;    
    |* (short |*) _Node.toserver = len;   
    |* (char  |*) _Node.toserver = '\0';  
    stpri (0);
    usec = ldtimer ();
    chwrite (_Node.toserver, (void *) 0x80000000, len);
    usec = ldtimer () - usec;
    stpri (1);

    replen = |* (short |*) _Node.fromserver;
    while (replen-- > 0)
      (void) |* (char |*) _Node.fromserver;

    thisproc.linkspeed = (1000 * len + usec/2) / usec;
  }
}


static char *readworm (
  const char *prog, 
  int *psize
)
{
  #define  DIRSEPC     '/'
  #define  DEFTPATH    "/pcc/bin"
  #define  DEFEXT      ".run"

  char    *worm = NULL;
  FILE    *f = NULL;
  char    *tpath = NULL,
          *paths,
          *path,
          *name;

  if (prog[0] == '\0' || prog[0] == '?')
    prog = "cleaner";

  tpath = getenv ("TPATH");
  if (! tpath)
    tpath = "";
  paths = malloc (strlen(".")+1 + strlen(tpath)+1 + 
          strlen(DEFTPATH) + 1);
  if (! paths) 
  {
    fprintf (stderr, "can't allocate memory for executable search path\n");
    return NULL;
  }
  strcat (strcat (strcat (strcpy (paths, ".;"), tpath), ";"), DEFTPATH);

  for (path=strtok(paths,";"); path && !f; path=strtok(NULL,";")) 
  {
    name = malloc (strlen(path)+ 1 + strlen(prog) + strlen (DEFEXT) + 1);
    if (! name) 
    {
      fprintf (stderr, "can't allocate memory for executable search path");
      break;
    }
    if (path[strlen(path)-1] == DIRSEPC)
      path[strlen(path)-1] = '\0';
    if (!strcmp (path, "."))
      strcpy (name, prog);
    else
      sprintf (name, "%s%c%s", path, DIRSEPC, prog);
    strcat (name, DEFEXT);
    if (! (f = fopen (name, "rb"))) 
    {
      name[strlen(name)-strlen(DEFEXT)] = '\0';
      if (! (f = fopen (name, "rb")))
        free (name);
    }
  }

  free (paths);

  if (! f)
    fprintf (stderr, "can't find worm executable \"%s\" anywhere\n", prog);
  else 
  {
    fseek (f, 0, SEEK_END);
    *psize = ftell (f);
    rewind (f);

    if (! (worm = malloc (*psize)))
      fprintf (stderr, "can't allocate memory for worm image\n");
    else 
    {
      if (fread (worm, 1, *psize, f) != *psize) 
      {
        fprintf (stderr, "error reading worm image \"%s\"\n", prog);
        free (worm);
        worm = NULL;
      }
      (void) fclose (f);
      free (name);
    }
  }

  return worm;
}


static void probelink (link_t l)
{
  void |*out = LINKxOUT (l),  
       |*in  = LINKxIN  (l);  
  char peek = PEEK;           
  char poke = POKE;           
  int addr = 0x80000100;                                  
  int wrval = 0xAA55AA55;    
  int rdval = 0;             
  int ticks = CLOCKS_PER_SEC/100;

  if (chtwrite (out, &poke,  sizeof(poke),  ticks) || 
      chtwrite (out, &addr,  sizeof(addr),  ticks) || 
      chtwrite (out, &wrval, sizeof(wrval), ticks) || 
      chtwrite (out, &peek,  sizeof(peek),  ticks) || 
      chtwrite (out, &addr,  sizeof(addr),  ticks) || 
      chtwrite (in,  &rdval, sizeof(rdval), ticks))   
    thisproc.neighbour[l].dev = DEV_NC;
  else 
  {
    if (rdval == wrval) 
    {
      thisproc.neighbour[l].dev    = DEV_PROC;
      thisproc.neighbour[l].booted = FALSE;
    }
    else if (rdval == ~wrval) {
            int confirm;
            thisproc.neighbour[l].dev    = DEV_PROC;
            thisproc.neighbour[l].booted = TRUE;

            if (chtread (in, &thisproc.neighbour[l].id, 
                sizeof(thisproc.neighbour[l].id), ticks) ||
                chtread (in, &confirm, sizeof(confirm), ticks) ||
                confirm != wrval) {
                thisproc.neighbour[l].error = ERR_CONN;
                thisproc.neighbour[l].id    = NOPROC;
            }
        }

        else
            thisproc.neighbour[l].dev = DEV_NC;
    }
}

static  int bootlink (link_t l, int id, int wormsize, char *worm)
{
    void    |*out = LINKxOUT (l),     
            |*in  = LINKxIN  (l);
    char    req[8];       
    char    ans[4] = { 2, 0, 1, 0 }; 
    int     ticks = CLOCKS_PER_SEC;  
    int     err = ERR_OK;            

    if (chtwrite (out, worm, wormsize, ticks))
        err = ERR_BOOT;

    else if ((chtread (in, req, sizeof(req), ticks) || req[2] != 42) ||
             (chtwrite (out, ans, sizeof(ans), ticks)) ||
             (chtread (in, req, sizeof(req), ticks) || 
             (req[2] != 40 && req[2] != 167)) ||
             (chtwrite (out, ans, sizeof(ans), ticks)))
        err = ERR_RUN;

    else if (chtwrite (out, &id,       sizeof(id),       ticks) ||
             chtwrite (out, &wormsize, sizeof(wormsize), ticks) ||
             chtwrite (out, worm,      wormsize,         ticks) ||
             chtread  (in,  req,       1,                ticks))
        err = ERR_RUN;

    return err;
}

static  void    info (proc_t proc)
{
    if (_Node.id == ROOTPROC) {
        printf ("%4d\r", proc.id);

        if (proc.id > maxid) {
            procs = (proc_t *) realloc (procs, (proc.id+1) * sizeof (proc_t));
            if (! procs) {
                fprintf (stderr, "out of memory (T # %d)\n", proc.id);
                exit (2);
            }
            while (++maxid < proc.id) 
                procs[maxid].id = NOPROC;
        }
        procs[proc.id] = proc;
    }

    else {
        |* (tag_t  |*) _Node.toserver = PROCESSOR;
        |* (proc_t |*) _Node.toserver = proc;
    }
}

static  void    bootbranch (link_t l, int wormsize, char *worm)
{
    void    |*in = LINKxIN (l);     
     tag_t   tag;                   

    if (_Node.id != ROOTPROC)
        |* (tag_t |*) _Node.toserver = HEARTBEAT;

    if (thisproc.neighbour[l].error = bootlink (l, nextid, wormsize, worm))
        return;

    do {
        proc_t  child;
        int     i;

        if (chtread (in, &tag, sizeof(tag), CLOCKS_PER_SEC/2)) {

            if (_Node.id != ROOTPROC)
                |* (tag_t |*) _Node.toserver = HEARTBEAT;

            if (chtread (in, &tag, sizeof(tag), CLOCKS_PER_SEC/2)) {
                thisproc.neighbour[l].error = ERR_RUN;
                return;
            }
        }

        switch (tag) {

        case BRANCHDONE:
            break;

        case HEARTBEAT:
            if (_Node.id != ROOTPROC)
                |* (tag_t |*) _Node.toserver = HEARTBEAT;
            break;

        case PROCESSOR:

            if (chtread (in, &child, sizeof(child), CLOCKS_PER_SEC)) {
                thisproc.neighbour[l].error = ERR_RUN;
                return;
            }

            child.depth++;
            if (child.id+1 > nextid)
                nextid = child.id + 1;

            for (i=0; i<NOFLINKS; i++) {

                if (child.neighbour[i].dev == DEV_HOST) {
                    child.neighbour[i].dev = DEV_PROC;
                    child.neighbour[i].id  = _Node.id;
                }

                if (child.neighbour[i].dev == DEV_PROC && 
                    child.neighbour[i].id == _Node.id) {
                    thisproc.neighbour[l].dev = DEV_PROC;
                    thisproc.neighbour[l].id  = child.id;
                }
            }

            info (child);
            break;
        }
    } while (tag != BRANCHDONE);
}

static  bool_t  gatekeeper (link_t l, int |*stopchan)
{
    void    |*out = LINKxOUT (l),     
            |*in = LINKxIN (l);
    int     quit = FALSE;             
    int     err = FALSE;               
    int     ticks = CLOCKS_PER_SEC/100;
    char    scratch[1+sizeof(int)];    
    int     x;                         

    while (! quit) {
        alt {

        guard stopchan:
            (void) |*stopchan;
            quit = TRUE;
            break;

        guard in:
            err |= chtread (in, scratch, sizeof(scratch), ticks) ||
                    chtread (in, &x, sizeof(x), ticks) ||
                   chtread (in, scratch, sizeof(scratch), ticks) ||
                   (x = ~x, chtwrite (out, &x, sizeof(x), ticks)) ||
                   chtwrite (out, &thisproc.id, sizeof(thisproc.id), ticks) ||
                   (x = ~x, chtwrite (out, &x, sizeof(x), ticks));
            break;
        }
    }

    return err;
}

static  void    usage (char *progname)
{
    fprintf (stderr, "Usage: %s [ -o file ]\n", progname);

    exit (2);
}

static  int report (FILE *f)
{
    int     id = 1;
    int     err = 0;
    int     i;
    link_t  l;

    fprintf (f, "worm %2.1f%c\n", VERSION, REV);
    fprintf (f, "   # Part rate Mb Bt [ ");
    for (l=0; l<NOFLINKS; l++)
        fprintf (f, " Link%d ", l);
    fprintf (f, "]\n");

    for (i=1; i<=maxid; i++)
        if (procs[i].id != NOPROC)
            procs[i].id = id++;

    for (i=1; i<=maxid; i++) {
        if (procs[i].id != NOPROC) {

            fprintf (f, "%4d T%3d%c-", procs[i].id-1, procs[i].type,
                                                            procs[i].rev);
            if (procs[i].cpuspeed)
                fprintf (f, "%02d", procs[i].cpuspeed);
            else
                fprintf (f, "??");
            fprintf (f, " %3.2lf %d [ ", (double)procs[i].linkspeed/1000,
                                                            procs[i].bt);

            for (l=0; l<NOFLINKS; l++)
                if (procs[i].neighbour[l].error == ERR_OK) {
                    if (procs[i].neighbour[l].dev == DEV_PROC) {
                        int nbid = procs[i].neighbour[l].id;
                        if (procs[nbid].id == NOPROC)
                            fprintf (f, " *BOOT");
                        else {
                            int ll;
                            fprintf (f, "%4d:", procs[nbid].id-1);
                            for (ll=0; ll<NOFLINKS; ll++)
                                if (procs[nbid].neighbour[ll].id == procs[i].id)
                                    break;
                            if (ll == NOFLINKS)
                                fprintf (f, "?");
                            else
                                fprintf (f, "%d", ll);
                        }
                    }

                    else if (l >= 2 && procs[i].type == 400)
                        fprintf (f, "      ");
                    else
                        fprintf (f, procs[i].neighbour[l].dev == DEV_NC
                                                        ? "   ..." : "  HOST");

                    if (procs[i].neighbour[l].noise) {
                        fprintf (f, "*");
                        err |= 1;  
                    }
                    else
                        fprintf (f, " ");
                }
                else {
                    fprintf (f, procs[i].neighbour[l].error == ERR_CONN ? " *CONN*" :
                                procs[i].neighbour[l].error == ERR_BOOT ? " *BOOT*" :
                                                                          "  *RUN*");
                    err |= 2;
                }
            fprintf (f, "]\n");
        }
    }

    return err;
}

int main (int argc, char *argv[])
{
    int     bl = _Node.fromserver-LINK0IN; 
    FILE    *outfile = NULL;           
    char    *worm;                     
    int     wormsize = 0;              
    int     i, j, l;                   

    if (argv[0][0]) {
        extern  int     getopt();  
        extern  char    *optarg;   
        extern  int     optind;    
        bool_t  errflg = FALSE;    

        while ((i = getopt (argc, argv, "o:?")) != -1)
            switch (i) {
            case 'o':
                outfile = fopen (optarg, "w");
                if (! outfile) {
                    fprintf (stderr, "unable to open '%s' for output\n", optarg);
                    exit (2);
                }
                break;
            case '?':
                errflg = TRUE;
            }

        if (errflg)
            usage (argv[0]);

        worm = readworm (argv[0], &wormsize);
    }

    else {
        _Node.id = |* (int |*) _Node.fromserver;
        wormsize = |* (int |*) _Node.fromserver;
        worm = malloc (wormsize);
        if (worm) {
            int usec;
            stpri (0);
             usec = ldtimer ();
            chread (_Node.fromserver, worm, wormsize);
            usec = ldtimer () - usec;
            stpri (0);
            thisproc.linkspeed = (1000 * wormsize + usec/2) / usec;
        }
        else
            while (wormsize-- > 0)
                (void) |* (char |*) _Node.fromserver;
        |* (char |*) _Node.toserver = 0; 
    }

    nextid = _Node.id + 1;

    thisproc.id   = _Node.id;
    thisproc.bt   = bl;
    thisproc.depth= 0;
    for (l=0; l<NOFLINKS; l++) {
        thisproc.neighbour[l].id = NOPROC;
        thisproc.neighbour[l].dev = DEV_NC;
    }
    thisproc.neighbour[bl].dev = DEV_HOST;

    attrib ();

    if (worm) {

        for (l=0; l<NOFLINKS; l++) {
            if (thisproc.neighbour[l].dev != DEV_HOST)
                par {
                    int |stop[NOFLINKS];
                    par (j=0; j<NOFLINKS; j++) {
                        if (l != j && bl != j)
                            if (gatekeeper (j, &stop[j]))
                                thisproc.neighbour[j].noise = TRUE;
                    }

                    {
                        probelink (l);
                        for (i=0; i<NOFLINKS; i++)
                            if (i != l && i != bl)
                                |stop[i] = TRUE;
                    }
                }

            if (thisproc.neighbour[l].dev == DEV_PROC && !thisproc.neighbour[l].booted) {
                par {
                    int |stop[NOFLINKS];

                    par (j=0; j<NOFLINKS; j++) {
                        if (l != j && bl != j)
                            if (gatekeeper (j, &stop[j]))
                                thisproc.neighbour[j].noise = TRUE;
                     }

                    {
                        bootbranch (l, wormsize, worm);
                        for (i=0; i<NOFLINKS; i++)
                            if (i != l && i != bl)
                                |stop[i] = TRUE;
                    }
                }
            }
        }
    }

    info (thisproc);

    if (_Node.id == ROOTPROC) {
        int stat;
        if (outfile)
            (void) fclose (outfile);
        exit (stat);
    }

    else {
        |* (tag_t |*) _Node.toserver = BRANCHDONE;
        par (i=0; i<NOFLINKS; i++) {
            int |dummy;
            gatekeeper (i, &dummy);
        }
    }
}
