/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* PPC command line utility.                                               */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 26.08.1994  Droux       Created                                         */
/*-------------------------------------------------------------------------*/
/* filename: ppccmd.c                                                      */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   Linux 1.0.8, Intel                                                    */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#if defined(NEXT) || defined(NEXTINTEL)
  #include <libc.h>
#endif
#if defined(LINUX)
  #include <unistd.h>
  #include <fcntl.h>
  #include <sys/wait.h>
#endif
#if defined(SUN4) || defined(SUN4SOL2)
  #include <fcntl.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#if defined(SUN4) || defined(SUN4SOL2)
  #include <dirent.h>
  #define direct dirent
#else
  #include <sys/dir.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>

#if defined(NEXT) || defined(NEXTINTEL)
  #define MAKE_NAME "/bin/make"
#elif defined(LINUX)
  #define MAKE_NAME "/usr/bin/make"
#elif defined(SUN4SOL2)
  #define MAKE_NAME "/usr/ccs/bin/make"
#else
  #define MAKE_NAME "/bin/make"
#endif

#if defined(SUN4)
  extern char *sys_errlist[];
  #define strerror(x) sys_errlist[x]
#endif

#define PROG_NAME "ppc"                                /* PPC utility name */
#define ALLOWED_ACTIONS "cmrdivh"                   /* allowed key actions */
#define MAX_FILE_NAME_LEN 100                  /* maximum file name length */
#define PPC_HOME_NAME "PPC_HOME"     /* PPC home environment variable name */
#define MAX_ARCH_LEN 100          /* maximum architecture specifier length */
#define DEFAULT_MODE 07755                   /* default file creation mode */
#define PVM_HOME_NAME "PVM_ROOT"     /* PVM home environment variable name */
#define EXECV_FAILED 1234                     /* exit code if execv failed */
#define PVM_TMP_DIR "/tmp"
#define PVMD_FILE "pvmd"
#define PACTC_HOME_NAME "PCC"

typedef enum                              /* possible target architectures */
{
  UNIX_ARCH,
  PVM_ARCH,
  PACTC_ARCH,
  NB_ARCH
} arch_enum;

char *arch_names[] =                 /* strings corresponding to arch_enum */
{
  "unix",
  "pvm",
  "pactc"
};

char *output_arch_names[] =                   /* idem but for verbose mode */
{
  "UNIX Simulation",
  "PVM",
  "Transputer Networks"
};

typedef enum                  /* architectures officially supported by PPC */
{
  NEXT_HOST,
  NEXTINTEL_HOST,
  LINUX_HOST,
  SUN4_HOST,
  SUN4SOL2_HOST
} human_arch;

char *human_arch_str[] =            /* strings corresponding to human_arch */
{
  "NEXTSTEP for NeXT computers",
  "NEXTSTEP for Intel computers",
  "Linux for Intel computers",
  "SunOS 4 for SPARC",
  "Solaris 2 for SPARC"
};

struct dir_entry
{
  char name [MAX_FILE_NAME_LEN];
  struct dir_entry *next;
};

extern int optind;                                     /* used by getopt() */
extern char *optarg;

char ppc_home [MAX_FILE_NAME_LEN];                   /* PPC home directory */
char install_directory [MAX_FILE_NAME_LEN];      /* installation directory */
char hardware_file [MAX_FILE_NAME_LEN];     /* hardware configuration file */
char vmc_file [MAX_FILE_NAME_LEN];   /* virtual machine configuration file */
arch_enum arch;                                     /* target architecture */
struct dir_entry *files_list = NULL;  /* list containing current dir files */
int verbose = 0;                                      /* 1 if verbose mode */
char host_arch [100];                         /* current host architecture */
char pvm_home [MAX_FILE_NAME_LEN];                   /* PVM home directory */
char pactc_home [MAX_FILE_NAME_LEN];     /* PACT-C compiler home directory */

/*-------------------------------------------------------------------------*/
/* print_usage:                                                            */
/* Print usage and exit.                                                   */
/*-------------------------------------------------------------------------*/
void print_usage (char *msg)
{
  if (msg != NULL)
    printf("%s: %s\n", PROG_NAME, msg);
  printf("%s: usage: {%s} -[afhiv]\n", 
    PROG_NAME, ALLOWED_ACTIONS);
  printf("%s: type '%s h' for help\n", PROG_NAME, PROG_NAME);
  exit(1);
}

/*-------------------------------------------------------------------------*/
/* print_title:                                                            */
/* Print program title.                                                    */
/*-------------------------------------------------------------------------*/
void print_title (void)
{
  printf("PPC command line utility. Version 0.9.0, September 1994\n");
  printf("Copyright (c) 1993, 1994 Engineering College of Biel-Bienne.\n");
}

/*-------------------------------------------------------------------------*/
/* print_help:                                                             */
/* Print help and exit.                                                    */
/*-------------------------------------------------------------------------*/
void print_help (void)
{
  print_title();
  printf("Usage: ppc key [options]\n");
  printf("key must be a combination of:\n");
  printf("  c         Configure program. Examine options -a, -f and -h.\n");
  printf("  m         Make program. Examine options -a, -f, and -h.\n");
  printf("  r         Run program. Examine option -a and -f.\n");
  printf("  d         Delete hardware dependent files. Examine option -a.\n");
  printf("  i         Install server program. Examine options -a, -f, and -i.\n");
  printf("  v         Print configuration report. Examine options -a, -f, and -h.\n");
  printf("  h         Print help.\n");
  printf("options:\n");
  printf("  -a arch   Specify target architecture. Default is unix. Allowed\n");
  printf("            architectures are unix, pvm, and pactc.\n");
  printf("  -f file   Specify virtual machine configuration file. Default is the\n");
  printf("            first .vmc file found in current directory.\n");
  printf("  -h file   Specify hardware configuration file. Default is <arch>.c.\n");
  printf("  -i path   Specify alternate server executable directory. Default is\n");
  printf("            $PPC_HOME/exec.\n");
  printf("  -v        Run the utility in verbose mode.\n");
  exit(2);
}

/*-------------------------------------------------------------------------*/
/* host_arch_human:                                                        */
/* For the given host architecture, return a human readable architecture   */
/* description, or the original architecture if unknown.                   */
/*-------------------------------------------------------------------------*/
char *host_arch_human (char *arch)
{
  if (!strcmp(arch, "NEXT"))
    return human_arch_str[NEXT_HOST];
  if (!strcmp(arch, "NEXTINTEL"))
    return human_arch_str[NEXTINTEL_HOST];
  if (!strcmp(arch, "LINUX"))
    return human_arch_str[LINUX_HOST];
  if (!strcmp(arch, "SUN4"))
    return human_arch_str[SUN4_HOST];
  if (!strcmp(arch, "SUN4SOL2"))
    return human_arch_str[SUN4SOL2_HOST];
  return arch;
}

/*-------------------------------------------------------------------------*/
/* check_key:                                                              */
/* Check if the given key is valid. If not, prints an error message,       */
/* the program usage and exit.                                             */
/*-------------------------------------------------------------------------*/
void check_key (char *key)
{
  int c;
  char err_msg [100];
  
  for (c = 0; c < strlen(key); c++)
    if (strchr(ALLOWED_ACTIONS, key[c]) == NULL)
    {
      sprintf(err_msg, "%c: unknown action", key[c]);
      print_usage(err_msg);
    }
}

/*-------------------------------------------------------------------------*/
/* get_ext:                                                                */
/* Return in ext the extension of the given file name. ext is set to an    */
/* empty string if the file has no extension.                              */
/*-------------------------------------------------------------------------*/
void get_ext (
  char *name,
  char *ext
)
{
  int c;
  int i;
  int name_len = strlen(name);
  
  for (c = name_len-1, i = 0; c && (name[c] != '.'); c--, i++);
  if (c)
  {
    ext[i] = '\0';
    for (c = 0, i--; i >= 0; i--, c++)
      ext[c] = name[name_len-1-i];
  }
  else
    ext[0] = '\0';
}

/*-------------------------------------------------------------------------*/
/* process_options:                                                        */
/* Parse the specified options and update the corresponding global         */
/* variables.                                                              */
/* If an error occures, then print an error message, the program usage,    */
/* and exit.                                                               */
/*-------------------------------------------------------------------------*/
void process_options (int argc, char **argv)
{
  int option;                            /* current option being processed */

  while((option = getopt(argc-1, &argv[1], "a:i:h:f:v")) != EOF)
  {
    switch(option)
    {
      case 'a':                            /* optional target architecture */
      {                                    
        int correct = 0;
        int act_arch;
	
	if (optarg == NULL)
          print_usage("option -a must be followed by unix, pvm, or pactc");
        for (act_arch = 0; act_arch < NB_ARCH && !correct; act_arch++)
	  correct = !strcmp(optarg, arch_names[act_arch]);
	if (!correct)
	  print_usage("specified architecture incorrect (must be unix, pvm, or pactc)");
	arch = act_arch-1;
        break;
      }
      case 'i':                         /* optional installation directory */
      {                                
        struct stat path_stat;

        if (optarg == NULL)
          print_usage("option -i must be followed by a path");
        if(stat(optarg, &path_stat) == -1)
	  print_usage("specified installation path does not exist or is not accessible");
        if (!(path_stat.st_mode & S_IFDIR))
	  print_usage("specified installation path is not a directory");
        strcpy(install_directory, optarg);
        break;
      }
      case 'h':               /* optional hardware configuration file name */
      {
        struct stat file_stat;
	char ext [100];
	
        if (optarg == NULL)
          print_usage("option -h must be followed by a file name");
        if(stat(optarg, &file_stat) == -1)
	{                           /* <file> does not exist, try <file>.c */
	  char file_plus_ext [MAX_FILE_NAME_LEN];
	  sprintf(file_plus_ext, "%s.c", optarg);
          if(stat(file_plus_ext, &file_stat) == -1)
	    print_usage("specified hardware configuration file does not exist");
	}
        strcpy(hardware_file, optarg);
	get_ext(hardware_file, ext);    /* suppress .c extension if exists */
	if (!strcmp(ext, "c"))
	  hardware_file[strlen(hardware_file)-strlen(ext)-1] = '\0';
	break;
      }
      case 'f':        /* optional virtual machine configuration file name */
      {
        struct stat file_stat;
	char ext [100];
	
        if (optarg == NULL)
          print_usage("option -f must be followed by a file name");
        strcpy(vmc_file, optarg);
        if(stat(optarg, &file_stat) == -1)
	{                         /* <file> does not exist, try <file>.vmc */
	  char file_plus_ext [MAX_FILE_NAME_LEN];
	  sprintf(file_plus_ext, "%s.vmc", optarg);
          if(stat(file_plus_ext, &file_stat) == -1)
	    print_usage("specified virtual machine configuration file does not exist");
	}
        strcpy(vmc_file, optarg);
	get_ext(vmc_file, ext);       /* suppress .vmc extension if exists */
	if (!strcmp(ext, "vmc"))
	  vmc_file[strlen(vmc_file)-strlen(ext)-1] = '\0';
	break;
      }
      case 'v':                                  /* toggle to verbose mode */
        verbose = 1;
	break;
      default:
        print_usage(NULL);
	break;
    }
  }
}

/*-------------------------------------------------------------------------*/
/* found_hardware_file:                                                    */
/* Found the name of the hardware configuration file corresponding to the  */
/* current architecture in the current directory.                          */
/*-------------------------------------------------------------------------*/
void found_hardware_file (void)
{
  struct dir_entry *act_entry = files_list;
  int found = 0;
  char file_name [MAX_FILE_NAME_LEN];
  
  sprintf(file_name, "%s.c", arch_names[arch]);
  while (act_entry != NULL && !found)
  {
    found = !strcmp(act_entry->name, file_name);
    act_entry = act_entry->next;
  }
  
  if (!found)
  {                       /* default hardware configuration file not found */
    printf("ppc: default hardware configuration file %s not found\n",
      file_name);
    print_usage(NULL);
  }
  
  strcpy(hardware_file, arch_names[arch]);
}

/*-------------------------------------------------------------------------*/
/* found_vmc_file:                                                         */
/* Found the name of the virtual configuration file in the current         */
/* directory.                                                              */
/*-------------------------------------------------------------------------*/
void found_vmc_file (void)
{
  struct dir_entry *act_entry = files_list;
  char ext [100];
  int found = 0;
  
  while (act_entry != NULL && !found)
  {
    get_ext(act_entry->name, ext);
    if (!(found = !strcmp(ext, "vmc")))
      act_entry = act_entry->next;
  }

  if (!found)
  {                                 /* no virtual configuration file found */
    printf("ppc: no virtual machine configuration file found in current directory\n");
    print_usage(NULL);
  }
  
  strcpy(vmc_file, act_entry->name);
  vmc_file[strlen(vmc_file)-strlen(ext)-1] = '\0';
}

/*-------------------------------------------------------------------------*/
/* get_dir_files:                                                          */
/* Initialize the list files_list with the files found in the current      */
/* directory.                                                              */
/*-------------------------------------------------------------------------*/
void get_dir_files (void)
{
  DIR *dir;
  struct direct *dirEnt;
  struct dir_entry *act_entry = files_list;
  struct dir_entry *new_entry;
  
  dir = opendir(".");
  for (dirEnt = readdir(dir); dirEnt != NULL; dirEnt = readdir(dir))
  {
    if (!strcmp(dirEnt->d_name, ".") || !strcmp(dirEnt->d_name, ".."))
      continue;
    new_entry = malloc(sizeof(struct dir_entry));
    if (files_list == NULL)
      files_list = new_entry;
    else
      act_entry->next = new_entry;
    act_entry = new_entry;
    strcpy(act_entry->name, dirEnt->d_name);
  }
}

/*-------------------------------------------------------------------------*/
/* is_pvm_running:                                                         */
/* Return 1 if PVM is running on the current host, 0 otherwise.            */
/*-------------------------------------------------------------------------*/
int is_pvm_running (void)
{
  DIR *dir;
  struct direct *dir_ent;
  char *point_ptr;
  int idx;
  
  if ((dir = opendir(PVM_TMP_DIR)) == NULL)
    return 0;

  for (dir_ent = readdir(dir); 
       (dir_ent != NULL); 
       dir_ent = readdir(dir))
  {
    point_ptr = dir_ent->d_name + strlen(dir_ent->d_name) - 1;
    idx = strlen(dir_ent->d_name);
    while(*point_ptr != '.' && idx--)
      point_ptr--;
    *point_ptr = '\0';
    if (!strcmp(dir_ent->d_name, PVMD_FILE))
      return 1;
  }

  return 0;
}

/*-------------------------------------------------------------------------*/
/* run_cmd:                                                                */
/* Execute the specified command.                                          */
/*-------------------------------------------------------------------------*/
void run_cmd (
  char *path, 
  char *argv[]
)
{
  if (verbose)
  {
    int c;
    for (c = 0; argv[c] != NULL; c++)
      printf("%s ", argv[c]);
    printf("\n");
  }
  
  if (!fork())
  {                                                       /* child process */
    execv(path, argv);
    exit(EXECV_FAILED);
  }
  wait(NULL);
}

/*-------------------------------------------------------------------------*/
/* make_configurer:                                                        */
/* Make program configurer. Called by configure() and view_config().       */
/*-------------------------------------------------------------------------*/
void make_configurer (void)
{
  struct stat dir_stat;
  char dir_name [100];
  char host_dir_name [100];
  static char other_libs [MAX_FILE_NAME_LEN];
  static char makefile_name [MAX_FILE_NAME_LEN];

  sprintf(dir_name, ".%s", arch_names[arch]);
  if(stat(dir_name, &dir_stat) == -1)
  {                    /* architecture directory does not exist, create it */
    if (verbose) 
      printf("Creating architecture directory\n");
    
    if (mkdir(dir_name, DEFAULT_MODE) == -1)
    {
      printf("ppc: error creating architecture directory: %s\n",
        strerror(errno));
      exit(1);
    }
  }
  
  sprintf(host_dir_name, "%s/%s", dir_name, HOST_ARCH);
  if (stat(host_dir_name, &dir_stat) == -1)
  {                            /* host directory does not exist, create it */
    if (verbose)
      printf("Creating host directory\n");

    if (mkdir(host_dir_name, DEFAULT_MODE) == -1)
    {
      printf("ppc: error creating host directory: %s\n",
        strerror(errno));
      exit(1);
    }
  }

  if (verbose)
    printf("Making Configurer\n");
  
  other_libs[0] = '\0';                  /* some makefile options defaults */
  sprintf(makefile_name, "%s/bin/config.mak", ppc_home);

  switch(arch)
  {
    case UNIX_ARCH:
      break;
    case PVM_ARCH:
      sprintf(other_libs, "-L%s/lib/%s -lpvm3", pvm_home, HOST_ARCH);
      break;
    case PACTC_ARCH:
      sprintf(makefile_name, "%s/bin/config.pactc.mak", ppc_home);
      break;
    default:
      break;
  }

  {
    char *argv[] = 
      {NULL, "-f", NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
    char arch_option [100];
    char arch_lib [100];
    char host_arch [100];
    char app_name [100];
    char other_libs_opt [100];
    char arch_file_opt [100];

    argv[0] = MAKE_NAME;
    argv[2] = makefile_name;
    sprintf(arch_option, "ARCH=%s", arch_names[arch]);
    argv[3] = arch_option;
    sprintf(arch_lib, "ARCHLIB=%s", arch_names[arch]);
    argv[4] = arch_lib;
    sprintf(host_arch, "HOST_ARCH=%s", HOST_ARCH);
    argv[5] = host_arch;
    sprintf(app_name, "APP_NAME=%s", vmc_file);
    argv[6] = app_name;
    sprintf(other_libs_opt, "OTHERLIB=%s", other_libs);
    argv[7] = other_libs_opt;
    sprintf(arch_file_opt, "ARCH_FILE=%s", hardware_file);
    argv[8] = arch_file_opt;
    run_cmd(MAKE_NAME, argv);
  }
}

/*-------------------------------------------------------------------------*/
/* configure:                                                              */
/* Configure program.                                                      */
/*-------------------------------------------------------------------------*/
void configure (void)
{
  if (verbose)
    printf("Configuring %s for %s\n", vmc_file, output_arch_names[arch]);
  
  make_configurer();
  
  if (verbose)
    printf("Executing configurer\n");
    
  {
    char *argv[] = {NULL, NULL};
    char configurer_name [100];
    
    sprintf(configurer_name, ".%s/%s/%s.con", 
      arch_names[arch], HOST_ARCH, vmc_file);
    argv[0] = configurer_name;
    run_cmd(configurer_name, argv);
  }
  
  if (verbose)
    printf("\n");
}

/*-------------------------------------------------------------------------*/
/* make:                                                                   */
/* Make program.                                                           */
/*-------------------------------------------------------------------------*/
void make (void)
{
  static char other_libs [100];
  static char makefile_name [100];

  if (verbose)
    printf("Making %s for %s\n", vmc_file, output_arch_names[arch]);
  
  sprintf(makefile_name, ".%s/makefile", arch_names[arch]);
  
  #ifdef SUN4SOL2
    strcpy(other_libs, "-lnsl -lsocket");
  #else
    other_libs[0] = '\0';
  #endif

  {
    char *argv[] = 
      {NULL, "-f", NULL, NULL, NULL, NULL, NULL, NULL, NULL};
    char arch_option [100];
    char arch_lib [100];
    char host_arch [100];
    char other_libs_opt [100];
    char arch_file_opt [100];

    argv[0] = MAKE_NAME;
    argv[2] = makefile_name;
    sprintf(arch_option, "ARCH=%s", arch_names[arch]);
    argv[3] = arch_option;
    sprintf(arch_lib, "ARCHLIB=%s", arch_names[arch]);
    argv[4] = arch_lib;
    sprintf(host_arch, "HOST_ARCH=%s", HOST_ARCH);
    argv[5] = host_arch;
    sprintf(other_libs_opt, "OTHERLIBS=%s", other_libs);
    argv[6] = other_libs_opt;
    sprintf(arch_file_opt, "ARCH_FILE=%s", hardware_file);
    argv[7] = arch_file_opt;
    run_cmd(MAKE_NAME, argv);
  }

  if (verbose)
    printf("\n");
}

/*-------------------------------------------------------------------------*/
/* run:                                                                    */
/* Run program.                                                            */
/*-------------------------------------------------------------------------*/
void run (void)
{
  if (verbose)
    printf("Executing %s for %s\n", vmc_file, output_arch_names[arch]);
    
  switch(arch)
  {
    case UNIX_ARCH:
    {
      char *argv[] = {NULL, NULL};
      char prog_name [100];
      
      sprintf(prog_name, ".%s/%s/%s.run",
        arch_names[arch], HOST_ARCH, vmc_file);
      argv[0] = prog_name;
      
      run_cmd(prog_name, argv);
      break;
    }
    case PVM_ARCH:
    {
      char *argv[] = {NULL, NULL};
      char prog_name [100];

      if (!is_pvm_running())
      {
        printf("%s: PVM is not running on the current host\n",
	  PROG_NAME);
	exit(1);
      }
      
      sprintf(prog_name, ".%s/%s/%s.run",
        arch_names[arch], HOST_ARCH, vmc_file);
      argv[0] = prog_name;
      
      run_cmd(prog_name, argv);
      break;
    }
    case PACTC_ARCH:
    {
      int null_file = open("/dev/null", O_RDONLY, 0);
      
      if (null_file == -1)
      {
        printf("%s: error opening /dev/null: %s\n",
	  PROG_NAME, strerror(errno));
      }
      
      if (!fork())
      {
        char *argv[] = {NULL, NULL, NULL};
	char prog_name [MAX_FILE_NAME_LEN];
	char cleaner_name [MAX_FILE_NAME_LEN];
	
	sprintf(prog_name, "%s/bin/pserver", pactc_home);
	argv[0] = prog_name;
	sprintf(cleaner_name, "%s/bin/cleaner.run", ppc_home);
	argv[1] = cleaner_name;
	
	if (verbose)
	  printf("%s %s\n", prog_name, cleaner_name);
	
        close(1);
	dup2(null_file, 1);
	execv(prog_name, argv);
      }      
      wait(NULL);
      {
	char *argv[] = {NULL, NULL, NULL};
	char prog_name [MAX_FILE_NAME_LEN];
	char run_name [100];
	
	sprintf(prog_name, "%s/bin/pserver", pactc_home);
	argv[0] = prog_name;
	sprintf(run_name, ".%s/PACTC/%s.run",
	  arch_names[arch], vmc_file);
	argv[1] = run_name;
	
	run_cmd(prog_name, argv);
      }
      break;
    }
    default:
      break;
  }
    
  if (verbose)
    printf("\n");
}

/*-------------------------------------------------------------------------*/
/* delete:                                                                 */
/* Delete program.                                                         */
/*-------------------------------------------------------------------------*/
void delete (void)
{
  char makefile_name [MAX_FILE_NAME_LEN];

  if (verbose)
    printf("Deleting %s files of %s\n", output_arch_names[arch], vmc_file);

  sprintf(makefile_name, "%s/bin/config.mak", ppc_home);
  {
    char *argv[] = 
      {NULL, "-f", NULL, "delete", NULL, NULL, NULL, NULL, NULL};
    char dir_to_delete [100];
    char app_name_opt [100];
    char arch_file_opt [100];
    char host_arch_opt [100];

    argv[0] = MAKE_NAME;
    argv[2] = makefile_name;
    sprintf(dir_to_delete, "DIR_TO_DELETE=.%s", arch_names[arch]);
    argv[4] = dir_to_delete;
    sprintf(app_name_opt, "APP_NAME=%s", vmc_file);
    argv[5] = app_name_opt;
    sprintf(arch_file_opt, "ARCH_FILE=%s", arch_names[arch]);
    argv[6] = arch_file_opt;
    sprintf(host_arch_opt, "HOST_ARCH=%s", HOST_ARCH);
    argv[7] = host_arch_opt;
    run_cmd(MAKE_NAME, argv);
  }

  if (verbose)
    printf("\n");
}

/*-------------------------------------------------------------------------*/
/* create_directory:                                                       */
/* Create the specified directory, print error message if the operation    */
/* failed.                                                                 */
/*-------------------------------------------------------------------------*/
void create_directory (
  char *dir
)
{
  struct stat f_stat;

  if (stat(dir, &f_stat) == -1)
  {                                 /* directory does not exist, create it */
    if (verbose)
      printf("Creating directory %s\n", dir);
    mkdir(dir, DEFAULT_MODE);
    if (stat(dir, &f_stat) == -1)
    {
      printf("%s: error creating directory %s\n", PROG_NAME, dir);
      exit(1);
    }
  }
}

/*-------------------------------------------------------------------------*/
/* install:                                                                */
/* Install program.                                                        */
/*-------------------------------------------------------------------------*/
void install (void)
{
  struct stat f_stat;
  char exec_name [100];
  char dir_name [MAX_FILE_NAME_LEN];
  char makefile_name [MAX_FILE_NAME_LEN];
  char arch_dir [100];
  
  if (verbose)
    printf("Installing %s server for %s in %s\n", vmc_file, 
      output_arch_names[arch], install_directory);
  
  if (arch == PACTC_ARCH)
    strcpy(arch_dir, "PACTC");
  else
    strcpy(arch_dir, HOST_ARCH);
  
  sprintf(exec_name, ".%s/%s/%s.run", arch_names[arch], arch_dir, vmc_file);
  if (stat(exec_name, &f_stat) == -1)
  {
    printf("%s: program \"%s\" has not been compiled for %s\n", 
      PROG_NAME, vmc_file, output_arch_names[arch]);
    exit(1);
  }
  
  create_directory(install_directory);
  sprintf(dir_name, "%s/%s", install_directory, vmc_file);
  create_directory(dir_name);
  sprintf(dir_name, "%s/%s/%s", install_directory, vmc_file, 
    arch_names[arch]);
  create_directory(dir_name);
  sprintf(dir_name, "%s/%s/%s/%s", install_directory, vmc_file, 
    arch_names[arch], HOST_ARCH);
  create_directory(dir_name);
  
  if (arch == PACTC_ARCH)
    sprintf(makefile_name, "%s/bin/config.pactc.mak", ppc_home);
  else
    sprintf(makefile_name, "%s/bin/config.mak", ppc_home);
  
  {
    char *argv[] = 
      {NULL, "-f", NULL, "install", NULL, NULL, NULL, NULL, NULL};
    char arch_option [100];
    char host_arch [100];
    char app_name [100];
    char install_dir_opt [MAX_FILE_NAME_LEN];

    argv[0] = MAKE_NAME;
    argv[2] = makefile_name;
    sprintf(arch_option, "ARCH=%s", arch_names[arch]);
    argv[4] = arch_option;
    sprintf(host_arch, "HOST_ARCH=%s", HOST_ARCH);
    argv[5] = host_arch;
    sprintf(app_name, "APP_NAME=%s", vmc_file);
    argv[6] = app_name;
    sprintf(install_dir_opt, "INSTALLDIR=%s", dir_name);
    argv[7] = install_dir_opt;
    run_cmd(MAKE_NAME, argv);
  }

  if (verbose)
    printf("\n");
}

/*-------------------------------------------------------------------------*/
/* view_config:                                                            */
/* View program configuration.                                             */
/*-------------------------------------------------------------------------*/
void view_config (void)
{
  if (verbose)
    printf("\"%s\" configuration for %s:\n", 
      vmc_file, output_arch_names[arch]);

  make_configurer();

  if (verbose)
    printf("Executing configurer\n");
    
  {
    char *argv[] = {NULL, "v", NULL};
    char configurer_name [100];
    
    sprintf(configurer_name, ".%s/%s/%s.con", 
      arch_names[arch], HOST_ARCH, vmc_file);
    argv[0] = configurer_name;
    run_cmd(configurer_name, argv);
  }
  
  if (verbose)
    printf("\n");
}

/*-------------------------------------------------------------------------*/
/* process_key:                                                            */
/* Process the actions found in key.                                       */
/*-------------------------------------------------------------------------*/
void process_key (
  char *key
)
{
  char action_idx;
  
  for (action_idx = 0; action_idx < strlen(key); action_idx++)
  {
    switch(key[action_idx])
    {
      case 'c': configure();   break;
      case 'm': make();        break;
      case 'r': run();         break;
      case 'd': delete();      break;
      case 'i': install();     break;
      case 'v': view_config(); break;
      default:                 break;
    }
  }
}

/*-------------------------------------------------------------------------*/
/* main:                                                                   */
/* PPC utility entry point.                                                */
/*-------------------------------------------------------------------------*/
void main (int argc, char **argv)
{
  char *ppc_home_ptr;
  
  if (argc == 1)                            /* check if a key is specified */
    print_usage("missing key");
    
  if (strchr(argv[1], 'h'))           /* check if help action is specified */
    print_help();
    
  check_key(argv[1]);      /* check if key is combination of valid actions */

  if ((ppc_home_ptr = getenv(PPC_HOME_NAME)) == NULL)
  {
    printf("%s: %s environment variable not defined\n",
     PROG_NAME, PPC_HOME_NAME);
    exit(4);
  }
  strcpy(ppc_home, ppc_home_ptr);                   /* initialize PPC home */
  
  get_dir_files();     /* get names of the files found in the current dir. */

  sprintf(install_directory, "%s/exec", ppc_home);         /* set defaults */
  strcpy(vmc_file, "");
  strcpy(hardware_file, "");
  arch = UNIX_ARCH;

  process_options(argc, argv);
  
  if(!strlen(vmc_file))
    if (strchr(argv[1], 'c') || strchr(argv[1], 'm') || 
        strchr(argv[1], 'r') || strchr(argv[1], 'i') ||
	strchr(argv[1], 'v'))
      found_vmc_file();                       /* vmc file name is required */
  
  if(!strlen(hardware_file))
    if (strchr(argv[1], 'c') || strchr(argv[1], 'm') || 
	strchr(argv[1], 'v'))
      found_hardware_file();             /* hardware file name is required */
  
  switch(arch)  /* special operations depending on the target architecture */
  {   
    case PVM_ARCH:
    {                                            /* get PVM home directory */
      char *pvm_home_ptr;
      
      if ((pvm_home_ptr = getenv(PVM_HOME_NAME)) == NULL)
      {
	printf("%s: %s environment variable not defined\n",
	  PROG_NAME, PVM_HOME_NAME);
	exit(4);
      }
      strcpy(pvm_home, pvm_home_ptr);               /* initialize PVM home */
      break;
    }
    case PACTC_ARCH:
    {                                         /* get PACT-C home directory */
      char *pactc_home_ptr;
      
      if ((pactc_home_ptr = getenv(PACTC_HOME_NAME)) == NULL)
      {
	printf("%s: %s environment variable not defined\n",
	  PROG_NAME, PACTC_HOME_NAME);
	exit(4);
      }
      strcpy(pactc_home, pactc_home_ptr);        /* initialize PACT-C home */
      break;
    }
    default:
      break;
  }
  
  if (verbose)
  {
    print_title();
    printf("\n");
    printf("target architecture                : %s\n", 
      output_arch_names[arch]);
    if (strlen(vmc_file))
      printf("virtual machine configuration file : %s.vmc\n", vmc_file);
    if (strlen(hardware_file))
      printf("hardware configuration file        : %s.c\n", hardware_file);
    if (strchr(argv[1], 'i'))
      printf("server installation directory      : %s\n", install_directory);
    printf("host computer architecture         : %s\n", 
      host_arch_human(HOST_ARCH));
    printf("\n");
  }
  
  process_key(argv[1]);                          /* process actions in key */

  exit(0);
}
